//
// Element class implementation file
//
// Revision -
//   $Revision: 1.5 $  $Date: 2002/01/31 15:05:57 $  $Author: wash $
//
 
#include <cstring>

#include "Element.hh"
#include "Jacobian.hh"

namespace FemLib {

Element::Element(const char *ishape_name) :
    shape_name(0),intpts_name(0)
{
    Configure(ishape_name) ;
}


Element::~Element()
{
    if (shape_name != 0) delete [] shape_name ;
}

#include <cstdio>

double Element::Area(int face_num,
                     const TVector<Vec3D> &coords,
                     int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetSurfaceShapeDefaultPts(face_num) ;

    // get a map for the nodes of the requested face.  The
    // first element in the map is the number of nodes on the
    // surface.

    const int *face_map = shape->GetFaceMap(face_num) ;
    ShapeFunc::SurfaceShape surf_shape = shape->GetFaceShape(face_num) ;

    // gather the coordinates for this face

    TVector<Vec3D> lcoords(face_map[0]) ;
    for (int i=0 ; i<face_map[0] ; ++i) {
        lcoords[i] = coords[face_map[i+1]] ;
    }

    // Get the shape functions

    IntegrationPts::SurfaceShape intg_shape =
             surf_shape == ShapeFunc::TRIANGLE ?
             IntegrationPts::TRIANGLE :
             IntegrationPts::QUADRILATERAL ;

    // get the integration coordinates and weights

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts =
               intpts->GetSurfaceCoords(intg_shape,int_pts) ;
    const TVector<double> &iwts =
               intpts->GetSurfaceWeights(intg_shape,int_pts) ;

    // do the integration

    double area = 0.0 ;
    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the shape function derivatives and
        // compute the jacobian, and it's determinate

        TMatrix<double> deriv =
                   shape->SurfaceShapeDeriv(surf_shape,ipts[i]) ;
        SurfaceJacobian jacob(deriv,lcoords) ;
        double jac_det = jacob.Determinate() ;
        jac_det = fabs(jac_det) ;

        // update the vector

        area += jac_det * iwts[i] ;
    }
    return(area) ;
}

bool Element::CheckForValidElement(
        const TVector<Vec3D> &coords,
        int int_pts) const
{
    double s_det = 0.0 ;

    // get the integration coordinates 

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts = intpts->GetVolumeCoords(int_pts) ;

    // check to make sure that the jacobian has a consistent sign

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the derivatives of the shape functions,
        // compute the jacobian, and it's determinate

        TMatrix<double> deriv = shape->VolumeShapeDeriv(ipts[i]) ;
        Jacobian jacob(deriv,coords) ;
        double jac_det = jacob.Determinate() ;

        // check the sign on the jacobian so that inverted element
        // numbering still works

        if (i == 0) s_det = jac_det ;
        if (jac_det*s_det <= 0.0) return(false) ;
    }
    return(true) ;
}


void Element::Configure(const char *ishape_name)
{
    shape = ShapeFuncFactory::Instance()->
                        CreateShapeFunc(ishape_name) ;

    if (shape_name != 0) delete [] shape_name ;
    shape_name = new char[strlen(ishape_name)+1] ;
    strcpy(shape_name,ishape_name) ;
}

TMatrix<double> Element::DeformationStiffMtx(
        const TVector<Vec3D> &coords,
        const TMatrix<double> &material_mat,
        int iint_pts) const
{
    return(DeformationStiffMtxHelp(coords,&material_mat,true,iint_pts)) ;
}

TMatrix<double> Element::DeformationStiffMtx(
        const TVector<Vec3D> &coords,
        const TMatrix<double> *material_mats,
        int iint_pts) const
{
    return(DeformationStiffMtxHelp(coords,material_mats,false,iint_pts)) ;
}

TMatrix<double> Element::DeformationStiffMtxHelp(
        const TVector<Vec3D> &coords,
        const TMatrix<double> *material_mats,
        bool uniform_material_mat,
        int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetDerivDefaultPts() ;
    int num_shape = shape->NumVolumeShape() ;
    TMatrix<double> Kmtx(num_shape*3,num_shape*3) ;

    // get the integration coordinates and weights

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts = intpts->GetVolumeCoords(int_pts) ;
    const TVector<double> &iwts = intpts->GetVolumeWeights(int_pts) ;

    // do the integration loop

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the derivatives of the shape functions,
        // compute the jacobian, it's determinate, and inverse

        TMatrix<double> deriv = shape->VolumeShapeDeriv(ipts[i]) ;
        Jacobian jacob(deriv,coords) ;
        double jac_det = jacob.Determinate() ;
        TMatrix<double> jac_inv = jacob.Inverse(jac_det) ;
        jac_det = fabs(jac_det) ;

        // transform the shape function derivatives to Cartesian coords

        deriv *= jac_inv ;

        // update the matrix

        if (uniform_material_mat)
            StiffUpd(num_shape,jac_det*iwts[i],
                     deriv,material_mats[0],Kmtx) ;
        else
            StiffUpd(num_shape,jac_det*iwts[i],
                     deriv,material_mats[i],Kmtx) ;
    }
    return(Kmtx) ;
}


#define MAX_ITS 25
#define NAT_TOL 0.001

bool Element::FindNaturalCoords(const Vec3D &cartesian,
                                const TVector<Vec3D> &node_coords,
                                Vec3D &natural) const
{
    Vec3D ncoord ;
    int i, num_its = 0 ;
    int num_shape = shape->NumVolumeShape() ;

    // set the natural coordinates to a known bad value

    natural[0] = natural[1] = natural[2] = 1e26 ;

    // loop trying to converge

    while (num_its < MAX_ITS) {
        ++num_its ;

        TVector<double> shapef = shape->VolumeShapeFunc(ncoord) ;
        TMatrix<double> deriv = shape->VolumeShapeDeriv(ncoord) ;

        // interpolate to find the Cartesion coordinat at this point

        Vec3D guess(0.0,0.0,0.0) ;
        for (i=0 ; i<num_shape ; ++i) {
            guess += shapef[i] * node_coords[i] ;
        }

        // compute the jacobian and its inverse */

        Jacobian jacob(deriv,node_coords) ;
        double jac_det = jacob.Determinate() ;
        TMatrix<double> jac_inv = jacob.Inverse(jac_det) ;
        jac_det = fabs(jac_det) ;

        // update our guessed natural coordinates based on a Newton step

        Vec3D delta = guess - cartesian ;
        Vec3D update ;
        for (i=0 ; i<3 ; ++i) {
            update[i] = jac_inv(0,i)*delta[0] +
                        jac_inv(1,i)*delta[1] +
                        jac_inv(2,i)*delta[2] ;
        }
        Vec3D new_nc = ncoord - update ; 

        // check for convergance */

        if ((fabs(new_nc[0]-ncoord[0]) > NAT_TOL) ||
            (fabs(new_nc[1]-ncoord[1]) > NAT_TOL) ||
            (fabs(new_nc[2]-ncoord[2]) > NAT_TOL)) {
            ncoord = new_nc ;
        } else {
            natural = new_nc ;
            break ;
        }
    }

    // now check to see if this coordinate is inside the element

    return(shape->InsideElement(natural)) ;
}


const TVector<Vec3D> &Element::GetVolumeIntegrationPts(
        int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetDerivDefaultPts() ;
    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts = intpts->GetVolumeCoords(int_pts) ;
    return(ipts) ;
}


void Element::InitialStressStiffUpd(
        TMatrix<double> &stiff,
        const TVector<Vec3D> &coords,
        const TVector<StressVec> &stress,
        int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetDerivDefaultPts() ;
    int num_shape = shape->NumVolumeShape() ;
    assert((stiff.Rows() == num_shape*3) &&
           (stiff.Cols() == num_shape*3)) ;

    // get the integration coordinates and weights

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts = intpts->GetVolumeCoords(int_pts) ;
    const TVector<double> &iwts = intpts->GetVolumeWeights(int_pts) ;

    // do the integration loop

    TMatrix<double> sg(3,num_shape) ;

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the derivatives of the shape functions,
        // compute the jacobian, it's determinate, and inverse

        TMatrix<double> deriv = shape->VolumeShapeDeriv(ipts[i]) ;
        Jacobian jacob(deriv,coords) ;
        double jac_det = jacob.Determinate() ;
        TMatrix<double> jac_inv = jacob.Inverse(jac_det) ;
        jac_det = fabs(jac_det) ;

        // transform the shape function derivatives to Cartesian coords

        deriv *= jac_inv ;

        // update the matrix

        for (int j=0 ; j<num_shape ; ++j) {
            sg(0,j) = stress[i][0]*deriv(0,j) + stress[i][3]*deriv(1,j) +
                      stress[i][5]*deriv(2,j) ;
            sg(1,j) = stress[i][3]*deriv(0,j) + stress[i][1]*deriv(1,j) +
                      stress[i][4]*deriv(2,j) ;
            sg(2,j) = stress[i][5]*deriv(0,j) + stress[i][4]*deriv(1,j) +
                      stress[i][2]*deriv(2,j) ;
        }
        for (int j=0 ; j<num_shape ; ++j) {
            for (int k=0 ; k<num_shape ; ++k) {
                double term = (deriv(0,j)*sg(0,k) + deriv(1,j)*sg(1,k) +
                               deriv(2,j)*sg(2,k)) * jac_det * iwts[i] ;
                stiff(j*3,k*3) += term ;
                stiff(j*3+1,k*3+1) += term ;
                stiff(j*3+2,k*3+2) += term ;
            }
        }
    }
}


TVector<Vec3D> Element::IntegrateNormalSurfaceScalar(
        int face_num,
        const TVector<Vec3D> &coords,
        double value,
        int iint_pts) const
{
    return(IntegrateNormalSurfaceScalarHelp1(face_num,coords,
                                        &value,true,iint_pts)) ;
}

TVector<Vec3D> Element::IntegrateNormalSurfaceScalar(
        int face_num,
        const TVector<Vec3D> &coords,
        double *nodal_value,
        int iint_pts) const
{
    return(IntegrateNormalSurfaceScalarHelp1(face_num,coords,
                               nodal_value,false,iint_pts)) ;
}

TVector<Vec3D> Element::IntegrateNormalSurfaceScalarHelp1(
        int face_num,
        const TVector<Vec3D> &coords,
        double *scalar,
        bool uniform_scalar,
        int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetSurfaceShapeDefaultPts(face_num) ;

    // get a map for the nodes of the requested face.  The
    // first element in the map is the number of nodes on the
    // surface.

    const int *face_map = shape->GetFaceMap(face_num) ;
    ShapeFunc::SurfaceShape surf_shape = shape->GetFaceShape(face_num) ;

    // gather the coordinates for this face

    TVector<Vec3D> lcoords(face_map[0]) ;
    TVector<double> lscalar(face_map[0]) ;
    for (int i=0 ; i<face_map[0] ; ++i) {
        lcoords[i] = coords[face_map[i+1]] ;
        lscalar[i] = uniform_scalar ? scalar[0] : scalar[face_map[i+1]] ;
    }

    // do the integration

    TVector<Vec3D> vect = IntegrateNormalSurfaceScalarHelp2(
          surf_shape,lcoords,lscalar,int_pts) ;

    // scatter the results into a vector for the whole element

    int num_shape = shape->NumVolumeShape() ;
    TVector<Vec3D> Fvect(num_shape) ;
    for (int j=0 ; j<face_map[0] ; j++ ) {
        Fvect[face_map[j+1]] = vect[j] ;
    }
    return(Fvect) ;
}

TVector<Vec3D> Element::IntegrateNormalSurfaceScalarHelp2(
        const ShapeFunc::SurfaceShape surface_shape,
        const TVector<Vec3D> &coords,
        const TVector<double> &scalar,
        int int_pts) const
{
    IntegrationPts::SurfaceShape intg_shape =
             surface_shape == ShapeFunc::TRIANGLE ?
             IntegrationPts::TRIANGLE :
             IntegrationPts::QUADRILATERAL ;

    int num_shape = shape->NumSurfaceShape(surface_shape) ;
    TVector<Vec3D> Fvec(num_shape) ;

    // get the integration coordinates and weights

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts =
               intpts->GetSurfaceCoords(intg_shape,int_pts) ;
    const TVector<double> &iwts =
               intpts->GetSurfaceWeights(intg_shape,int_pts) ;

    // do the integration loop

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the shape functions and derivatives and
        // compute the jacobian, and it's determinate

        TVector<double> shapef =
                   shape->SurfaceShapeFunc(surface_shape,ipts[i]) ;
        TMatrix<double> deriv =
                   shape->SurfaceShapeDeriv(surface_shape,ipts[i]) ;
        SurfaceJacobian jacob(deriv,coords) ;

        // the normal to the surface is stored in the third row
        // of the jacobian, but it is not normalized

        Vec3D normal(jacob(2,0),jacob(2,1),jacob(2,2)) ;

        // update the vector

        double lvalue = 0 ;
        for (int j=0 ; j<num_shape ; ++j) {
            lvalue += shapef[j] * scalar[j] ;
        }
        for (int j=0 ; j<num_shape ; ++j) {
            Fvec(j) += shapef[j] * iwts[i] * lvalue * normal ;
        }
    }
    return(Fvec) ;
}

TVector<double> Element::IntegrateSurfaceScalar(
        int face_num,
        const TVector<Vec3D> &coords,
        double scalar,
        int int_pts) const
{
    return(IntegrateSurfaceScalarHelp1(face_num,coords,&scalar,true,int_pts)) ;
}


TVector<double> Element::IntegrateSurfaceScalar(
        int face_num,
        const TVector<Vec3D> &coords,
        double *nodal_scalars,
        int int_pts) const
{
    return(IntegrateSurfaceScalarHelp1(face_num,coords,nodal_scalars,false,int_pts)) ;
}

TVector<double> Element::IntegrateSurfaceScalarHelp1(
        int face_num,
        const TVector<Vec3D> &coords,
        double *scalar,
        bool uniform_scalar,
        int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetSurfaceShapeDefaultPts(face_num) ;

    // get a map for the nodes of the requested face.  The
    // first element in the map is the number of nodes on the
    // surface.

    const int *face_map = shape->GetFaceMap(face_num) ;
    ShapeFunc::SurfaceShape surf_shape = shape->GetFaceShape(face_num) ;

    // gather the coordinates for this face

    TVector<Vec3D> lcoords(face_map[0]) ;
    TVector<double> lscalar(face_map[0]) ;
    for (int i=0 ; i<face_map[0] ; ++i) {
        lcoords[i] = coords[face_map[i+1]] ;
        lscalar[i] = uniform_scalar ? scalar[0] : scalar[face_map[i+1]] ;
    }

    // do the integration

    TVector<double> vect = IntegrateSurfaceScalarHelp2(
          surf_shape,lcoords,lscalar,int_pts) ;

    // scatter the results into a vector for the whole element

    int num_shape = shape->NumVolumeShape() ;
    TVector<double> Fvect(num_shape) ;
    for (int j=0 ; j<face_map[0] ; j++ ) {
        Fvect[face_map[j+1]] = vect[j] ;
    }
    return(Fvect) ;
}


TVector<double> Element::IntegrateSurfaceScalarHelp2(
        const ShapeFunc::SurfaceShape surface_shape,
        const TVector<Vec3D> &coords,
        const TVector<double> &scalar,
        int int_pts) const
{
    IntegrationPts::SurfaceShape intg_shape =
             surface_shape == ShapeFunc::TRIANGLE ?
             IntegrationPts::TRIANGLE :
             IntegrationPts::QUADRILATERAL ;

    int num_shape = shape->NumSurfaceShape(surface_shape) ;
    TVector<double> Fvec(num_shape) ;

    // get the integration coordinates and weights

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts =
               intpts->GetSurfaceCoords(intg_shape,int_pts) ;
    const TVector<double> &iwts =
               intpts->GetSurfaceWeights(intg_shape,int_pts) ;

    // do the integration loop

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the shape functions and derivatives and
        // compute the jacobian, and it's determinate

        TVector<double> shapef =
                   shape->SurfaceShapeFunc(surface_shape,ipts[i]) ;
        TMatrix<double> deriv =
                   shape->SurfaceShapeDeriv(surface_shape,ipts[i]) ;
        SurfaceJacobian jacob(deriv,coords) ;
        double jac_det = jacob.Determinate() ;
        jac_det = fabs(jac_det) ;

        // update the vector

        double lvalue = 0.0 ;
        for (int j=0 ; j<num_shape ; ++j) {
            lvalue += shapef[j] * scalar[j] ;
        }
        double factor = lvalue * jac_det * iwts[i] ;

        for (int j=0 ; j<num_shape ; ++j) {
            Fvec(j) += shapef[j] * factor ;
        }
    }
    return(Fvec) ;
}


TVector<Vec3D> Element::IntegrateSurfaceVector(
        int face_num,
        const TVector<Vec3D> &coords,
        const Vec3D &value,
        int iint_pts) const
{
    TVector<Vec3D> tmp(1) ;
    tmp(0) = value ;
    return(IntegrateSurfaceVectorHelp1(face_num,coords,tmp,true,iint_pts)) ;
}

TVector<Vec3D> Element::IntegrateSurfaceVector(
        int face_num,
        const TVector<Vec3D> &coords,
        const TVector<Vec3D> &nodal_value,
        int iint_pts) const
{
    return(IntegrateSurfaceVectorHelp1(face_num,coords,nodal_value,false,iint_pts)) ;
}

TVector<Vec3D> Element::IntegrateSurfaceVectorHelp1(
        int face_num,
        const TVector<Vec3D> &coords,
        const TVector<Vec3D> &value,
        bool uniform_scalar,
        int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetSurfaceShapeDefaultPts(face_num) ;

    // get a map for the nodes of the requested face.  The
    // first element in the map is the number of nodes on the
    // surface.

    const int *face_map = shape->GetFaceMap(face_num) ;
    ShapeFunc::SurfaceShape surf_shape = shape->GetFaceShape(face_num) ;

    // gather the coordinates for this face

    TVector<Vec3D> lcoords(face_map[0]) ;
    TVector<Vec3D> lvector(face_map[0]) ;
    for (int i=0 ; i<face_map[0] ; ++i) {
        lcoords[i] = coords[face_map[i+1]] ;
        lvector[i] = uniform_scalar ? value[0] : value[face_map[i+1]] ;
    }

    // do the integration

    TVector<Vec3D> vect = IntegrateSurfaceVectorHelp2(
          surf_shape,lcoords,lvector,int_pts) ;

    // scatter the results into a vector for the whole element

    int num_shape = shape->NumVolumeShape() ;
    TVector<Vec3D> Fvect(num_shape) ;
    for (int j=0 ; j<face_map[0] ; j++ ) {
        Fvect[face_map[j+1]] = vect[j] ;
    }
    return(Fvect) ;
}

TVector<Vec3D> Element::IntegrateSurfaceVectorHelp2(
        const ShapeFunc::SurfaceShape surface_shape,
        const TVector<Vec3D> &coords,
        const TVector<Vec3D> &values,
        int int_pts) const
{
    IntegrationPts::SurfaceShape intg_shape =
             surface_shape == ShapeFunc::TRIANGLE ?
             IntegrationPts::TRIANGLE :
             IntegrationPts::QUADRILATERAL ;

    int num_shape = shape->NumSurfaceShape(surface_shape) ;
    TVector<Vec3D> Fvec(num_shape) ;

    // get the integration coordinates and weights

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts =
               intpts->GetSurfaceCoords(intg_shape,int_pts) ;
    const TVector<double> &iwts =
               intpts->GetSurfaceWeights(intg_shape,int_pts) ;

    // do the integration loop

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the shape functions and derivatives and
        // compute the jacobian, and it's determinate

        TVector<double> shapef =
                   shape->SurfaceShapeFunc(surface_shape,ipts[i]) ;
        TMatrix<double> deriv =
                   shape->SurfaceShapeDeriv(surface_shape,ipts[i]) ;
        SurfaceJacobian jacob(deriv,coords) ;
        double jac_det = jacob.Determinate() ;
        jac_det = fabs(jac_det) ;

        // update the vector

        double factor = jac_det * iwts[i] ;

        Vec3D lvalue ;
        for (int j=0 ; j<num_shape ; ++j) {
            lvalue += shapef[j] * values[j] ;
        }
        for (int j=0 ; j<num_shape ; ++j) {
//            for (int k=0 ; k<3 ; ++k) {
//                Fvec(j)[k] += shapef[j] * lvalue[k] * factor ;
            Fvec(j) += shapef[j] * lvalue * factor ;
//            }
        }
    }
    return(Fvec) ;
}

TVector<double> Element::IntegrateVolume(
        const TVector<Vec3D> &coords,
        double scalar,
        int iint_pts) const
{
    return(IntegrateVolumeScalarHelp(coords,&scalar,true,iint_pts)) ;
}


TVector<double> Element::IntegrateVolume(
        const TVector<Vec3D> &coords,
        double *scalar,
        int iint_pts) const
{
    return(IntegrateVolumeScalarHelp(coords,scalar,false,iint_pts)) ;
}


TVector<double> Element::IntegrateVolumeScalarHelp(
        const TVector<Vec3D> &coords,
        double *scalar,
        bool uniform_scalar,
        int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetDerivDefaultPts() ;
    int num_shape = shape->NumVolumeShape() ;
    TVector<double> Fvec(num_shape) ;

    // get the integration coordinates and weights

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts = intpts->GetVolumeCoords(int_pts) ;
    const TVector<double> &iwts = intpts->GetVolumeWeights(int_pts) ;

    // do the integration loop

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the shape functions and derivatives and
        // compute the jacobian, and it's determinate

        TVector<double> shapef = shape->VolumeShapeFunc(ipts[i]) ;
        TMatrix<double> deriv = shape->VolumeShapeDeriv(ipts[i]) ;
        Jacobian jacob(deriv,coords) ;
        double jac_det = jacob.Determinate() ;
        jac_det = fabs(jac_det) ;

        // update the vector

        double factor = uniform_scalar ? 
                        scalar[0] * jac_det * iwts[i] :
                        scalar[i] * jac_det * iwts[i] ;

        for (int j=0 ; j<num_shape ; ++j) {
            Fvec(j) += shapef[j] * factor ;
        }
    }
    return(Fvec) ;
}

TVector<Vec3D> Element::IntegrateVolume(
        const TVector<Vec3D> &coords,
        const StressVec &stress,
        int iint_pts) const
{
    TVector<StressVec> tmp(1) ;
    tmp(0) = stress ;
    return(IntegrateVolumeStressHelp(coords,tmp,true,iint_pts)) ;
}

TVector<Vec3D> Element::IntegrateVolume(
        const TVector<Vec3D> &coords,
        const TVector<StressVec> &stress,
        int iint_pts) const
{
    return(IntegrateVolumeStressHelp(coords,stress,false,iint_pts)) ;
}

TVector<Vec3D> Element::IntegrateVolumeStressHelp(
        const TVector<Vec3D> &coords,
        const TVector<StressVec> &stress,
        bool uniform_stress,
        int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetDerivDefaultPts() ;
    int num_shape = shape->NumVolumeShape() ;
    TVector<Vec3D> Fvec(num_shape) ;

    // get the integration coordinates and weights

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts = intpts->GetVolumeCoords(int_pts) ;
    const TVector<double> &iwts = intpts->GetVolumeWeights(int_pts) ;

    // do the integration loop

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the shape functions and derivatives and
        // compute the jacobian, and it's determinate

        TMatrix<double> deriv = shape->VolumeShapeDeriv(ipts[i]) ;
        Jacobian jacob(deriv,coords) ;
        double jac_det = jacob.Determinate() ;
        TMatrix<double> jac_inv = jacob.Inverse(jac_det) ;
        jac_det = fabs(jac_det) ;

        // transform the shape function derivatives to Cartesian coords

        deriv *= jac_inv ;
 
        double factor = jac_det * iwts[i] ;

        if (uniform_stress) {
            for (int j=0 ; j<num_shape ; ++j) {
                Fvec[j][0] += (deriv(0,j)*stress[0][0] +
                               deriv(1,j)*stress[0][3] +
                               deriv(2,j)*stress[0][5]) * factor ;
                Fvec[j][1] += (deriv(1,j)*stress[0][1] +
                               deriv(0,j)*stress[0][3] +
                               deriv(2,j)*stress[0][4]) * factor ;
                Fvec[j][2] += (deriv(2,j)*stress[0][2] +
                               deriv(1,j)*stress[0][4] +
                               deriv(0,j)*stress[0][5]) * factor ;
            }
        } else {
            for (int j=0 ; j<num_shape ; ++j) {
                Fvec[j][0] += (deriv(0,j)*stress[i][0] +
                               deriv(1,j)*stress[i][3] +
                               deriv(2,j)*stress[i][5]) * factor ;
                Fvec[j][1] += (deriv(1,j)*stress[i][1] +
                               deriv(0,j)*stress[i][3] +
                               deriv(2,j)*stress[i][4]) * factor ;
                Fvec[j][2] += (deriv(2,j)*stress[i][2] +
                               deriv(1,j)*stress[i][4] +
                               deriv(0,j)*stress[i][5]) * factor ;
            }
        }
    }
    return(Fvec) ;
}


TVector<Vec3D> Element::InternalPointForces(
        const TVector<Vec3D> &nat_coords,
        const TVector<Vec3D> &values) const
{
    assert(nat_coords.Rows() == values.Rows()) ;

    int num_shape = shape->NumVolumeShape() ;
    TVector<Vec3D> Fvec(num_shape) ;

    // loop through all the input points

    for (int i=0 ; i<nat_coords.Rows() ; ++i) {

        // compute the shape functions and contributions
        // to the force fector at this point

        TVector<double> shapef = shape->VolumeShapeFunc(nat_coords[i]) ;
        for (int j=0 ; j<num_shape ; ++j) {
            Fvec[j] += shapef[j] * values[i] ;
        }
    }
    return(Fvec) ;
}


double Element::InterpolateSurfaceScalar(
        int face_num,
        const Vec3D &natural_coords,
        const TVector<double> &node_values) const
{
    // get a map for the nodes of the requested face.  The
    // first element in the map is the number of nodes on the
    // surface.

    const int *face_map = shape->GetFaceMap(face_num) ;
    ShapeFunc::SurfaceShape surf_shape = shape->GetFaceShape(face_num) ;

    // gather the coordinates for this face

    TVector<double> lnode_values(face_map[0]) ;
    for (int i=0 ; i<face_map[0] ; ++i) {
        lnode_values[i] = node_values[face_map[i+1]] ;
    }
    return(InterpolateSurfaceScalarHelp(
                surf_shape,natural_coords,lnode_values)) ;
}

double Element::InterpolateSurfaceScalar(
        int face_num,
        const Vec3D &natural_coords,
        const double *node_values) const
{
    // get a map for the nodes of the requested face.  The
    // first element in the map is the number of nodes on the
    // surface.

    const int *face_map = shape->GetFaceMap(face_num) ;
    ShapeFunc::SurfaceShape surf_shape = shape->GetFaceShape(face_num) ;

    // gather the coordinates for this face

    TVector<double> lnode_values(face_map[0]) ;
    for (int i=0 ; i<face_map[0] ; ++i) {
        lnode_values[i] = node_values[face_map[i+1]] ;
    }
    return(InterpolateSurfaceScalarHelp(
                surf_shape,natural_coords,lnode_values)) ;
}

double Element::InterpolateSurfaceScalarHelp(
        const ShapeFunc::SurfaceShape surface_shape,
        const Vec3D &natural_coords,
        const TVector<double> &node_values) const
{
    int num_shape = shape->NumSurfaceShape(surface_shape) ;
    TVector<double> shapef =
        shape->SurfaceShapeFunc(surface_shape,natural_coords) ;
    assert(shapef.Rows() == node_values.Rows()) ;

    double value = 0.0 ;
    for (int i=0 ; i<num_shape ; ++i) {
        value += shapef[i] * node_values[i] ;
    }
    return(value) ;
}


TMatrix<double> Element::ScalarFieldDerivMtx(
        const TVector<Vec3D> &coords,
        const TMatrix<double> &material_mat,
        int iint_pts) const
{
    return(ScalarFieldDerivMtxHelp(coords,&material_mat,true,iint_pts)) ;
}

TMatrix<double> Element::ScalarFieldDerivMtx(
        const TVector<Vec3D> &coords,
        const TMatrix<double> *material_mats,
        int iint_pts) const
{
    return(ScalarFieldDerivMtxHelp(coords,material_mats,false,iint_pts)) ;
}

TMatrix<double> Element::ScalarFieldDerivMtxHelp(
        const TVector<Vec3D> &coords,
        const TMatrix<double> *material_mats,
        bool uniform_material_mat,
        int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetDerivDefaultPts() ;
    int num_shape = shape->NumVolumeShape() ;
    TMatrix<double> Kmtx(num_shape,num_shape) ;

    // get the integration coordinates and weights

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts = intpts->GetVolumeCoords(int_pts) ;
    const TVector<double> &iwts = intpts->GetVolumeWeights(int_pts) ;

    // do the integration loop

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the derivatives of the shape functions,
        // compute the jacobian, it's determinate, and inverse

        TMatrix<double> deriv = shape->VolumeShapeDeriv(ipts[i]) ;
        Jacobian jacob(deriv,coords) ;
        double jac_det = jacob.Determinate() ;
        TMatrix<double> jac_inv = jacob.Inverse(jac_det) ;
        jac_det = fabs(jac_det) ;

        // transform the shape function derivatives to Cartesian coords

        deriv *= jac_inv ; 

        // update the matrix

        if (uniform_material_mat)
            Kmtx.TripleProd(deriv,material_mats[0],jac_det*iwts[i]) ;
        else
            Kmtx.TripleProd(deriv,material_mats[i],jac_det*iwts[i]) ;
    }
    return(Kmtx) ;
}


TMatrix<double> Element::ScalarFieldShapeMtx(
        const TVector<Vec3D> &coords,
        double material_prop,
        int iint_pts) const
{
    return(ScalarFieldShapeMtxHelp(coords,&material_prop,true,iint_pts)) ;
}

TMatrix<double> Element::ScalarFieldShapeMtx(
        const TVector<Vec3D> &coords,
        double *material_props,
        int iint_pts) const
{
    return(ScalarFieldShapeMtxHelp(coords,material_props,false,iint_pts)) ;
}

TMatrix<double> Element::ScalarFieldShapeMtxHelp(
        const TVector<Vec3D> &coords,
        double *material_props,
        bool uniform_material_props,
        int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetShapeDefaultPts() ;
    int num_shape = shape->NumVolumeShape() ;
    TMatrix<double> Cmtx(num_shape,num_shape) ;

    // get the integration coordinates and weights

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts = intpts->GetVolumeCoords(int_pts) ;
    const TVector<double> &iwts = intpts->GetVolumeWeights(int_pts) ;

    // do the integration loop

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the shape functions and derivatives and
        // compute the jacobian, and it's determinate

        TVector<double> shapef = shape->VolumeShapeFunc(ipts[i]) ;
        TMatrix<double> deriv = shape->VolumeShapeDeriv(ipts[i]) ;
        Jacobian jacob(deriv,coords) ;
        double jac_det = jacob.Determinate() ;
        jac_det = fabs(jac_det) ;

        // update the matrix

        double factor = uniform_material_props ? 
                        material_props[0] * jac_det * iwts[i] :
                        material_props[i] * jac_det * iwts[i] ;

        for (int j=0 ; j<num_shape ; ++j) {
            for (int k=0 ; k<num_shape ; ++k) {
                Cmtx(j,k) += shapef[j] * shapef[k] * factor ;
            }
        }
    }
    return(Cmtx) ;
}


TMatrix<double> Element::StiffMtxCheck(
        const TVector<Vec3D> &coords,
        const TMatrix<double> &material_mat,
        int iint_pts) const
{
    return(StiffMtxCheckHelp(coords,&material_mat,true,iint_pts)) ;
}

TMatrix<double> Element::StiffMtxCheck(
        const TVector<Vec3D> &coords,
        const TMatrix<double> *material_mats,
        int iint_pts) const
{
    return(StiffMtxCheckHelp(coords,material_mats,false,iint_pts)) ;
}

TMatrix<double> Element::StiffMtxCheckHelp(
        const TVector<Vec3D> &coords,
        const TMatrix<double> *material_mats,
        bool uniform_material_mat,
        int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetDerivDefaultPts() ;
    int num_shape = shape->NumVolumeShape() ;
    TMatrix<double> Kmtx(num_shape*3,num_shape*3) ;

    // get the integration coordinates and weights

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts = intpts->GetVolumeCoords(int_pts) ;
    const TVector<double> &iwts = intpts->GetVolumeWeights(int_pts) ;

    // do the integration loop

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the derivatives of the shape functions,
        // compute the jacobian, it's determinate, and inverse

        TMatrix<double> deriv = shape->VolumeShapeDeriv(ipts[i]) ;
        Jacobian jacob(deriv,coords) ;
        double jac_det = jacob.Determinate() ;
        TMatrix<double> jac_inv = jacob.Inverse(jac_det) ;
        jac_det = fabs(jac_det) ;

        // transform the shape function derivatives to Cartesian coords

        deriv *= jac_inv ;

        // update the matrix

        if (uniform_material_mat)
            StiffUpdChk(num_shape,jac_det*iwts[i],
                        deriv,material_mats[0],Kmtx) ;
        else
            StiffUpdChk(num_shape,jac_det*iwts[i],
                        deriv,material_mats[i],Kmtx) ;
    }
    return(Kmtx) ;
}



double Element::Volume(const TVector<Vec3D> &coords,
                       int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetDerivDefaultPts() ;
    double volume = 0.0 ;

    // get the integration coordinates and weights

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts = intpts->GetVolumeCoords(int_pts) ;
    const TVector<double> &iwts = intpts->GetVolumeWeights(int_pts) ;

    // do the integration loop

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the derivatives of the shape functions,
        // compute the jacobian and it's determinate

        TMatrix<double> deriv = shape->VolumeShapeDeriv(ipts[i]) ;
        Jacobian jacob(deriv,coords) ;
        double jac_det = jacob.Determinate() ;
        jac_det = fabs(jac_det) ;

        volume += jac_det * iwts[i] ;
    }
    return(volume) ;
}




double Element::InterpolateVolume(
        const Vec3D &natural_coords,
        const TVector<double> &node_values) const
{
    TVector<double> shapef = shape->VolumeShapeFunc(natural_coords) ;
    assert(shapef.Rows() == node_values.Rows()) ;

    double value = 0.0 ;
    for (int i=0 ; i<shapef.Rows() ; ++i) {
        value += shapef[i] * node_values[i] ;
    }
    return(value) ;
}

double Element::InterpolateVolume(
        const Vec3D &natural_coords,
        const double *node_values) const
{
    TVector<double> shapef = shape->VolumeShapeFunc(natural_coords) ;

    double value = 0.0 ;
    for (int i=0 ; i<shapef.Rows() ; ++i) {
        value += shapef[i] * node_values[i] ;
    }
    return(value) ;
}

Vec3D Element::InterpolateVolume(
        const Vec3D &natural_coords,
        const TVector<Vec3D> &node_values) const
{
    TVector<double> shapef = shape->VolumeShapeFunc(natural_coords) ;
    assert(shapef.Rows() == node_values.Rows()) ;

    Vec3D value(0.0,0.0,0.0) ;
    for (int i=0 ; i<shapef.Rows() ; ++i) {
        value += shapef[i] * node_values[i] ;
    }
    return(value) ;
}




void Element::StiffUpd(int num_shape,double factor,
                       const TMatrix<double> &d,
                       const TMatrix<double> &C,
                       TMatrix<double> &k) const
{
    TMatrix<double> dc(3*num_shape,6) ;

    for (int i=0 ; i<num_shape ; ++i) {
        for (int j=0 ; j<6 ; ++j) {
            dc(i*3  ,j) = (d(0,i)*C(0,j)+d(1,i)*C(3,j)+d(2,i)*C(5,j)) * factor ;
            dc(i*3+1,j) = (d(1,i)*C(1,j)+d(0,i)*C(3,j)+d(2,i)*C(4,j)) * factor ;
            dc(i*3+2,j) = (d(2,i)*C(2,j)+d(1,i)*C(4,j)+d(0,i)*C(5,j)) * factor ;
        }
    }

    for (int i=0 ; i<num_shape*3 ; i+=3) {
        for (int j=0 ; j<num_shape ; ++j) {
            k(i,j*3)   += dc(i,0)*d(0,j)+dc(i,3)*d(1,j)+dc(i,5)*d(2,j) ;
            k(i,j*3+1) += dc(i,1)*d(1,j)+dc(i,3)*d(0,j)+dc(i,4)*d(2,j) ;
            k(i,j*3+2) += dc(i,2)*d(2,j)+dc(i,4)*d(1,j)+dc(i,5)*d(0,j) ;

            k(i+1,j*3)   += dc(i+1,0)*d(0,j)+dc(i+1,3)*d(1,j)+dc(i+1,5)*d(2,j) ;
            k(i+1,j*3+1) += dc(i+1,1)*d(1,j)+dc(i+1,3)*d(0,j)+dc(i+1,4)*d(2,j) ;
            k(i+1,j*3+2) += dc(i+1,2)*d(2,j)+dc(i+1,4)*d(1,j)+dc(i+1,5)*d(0,j) ;

            k(i+2,j*3)   += dc(i+2,0)*d(0,j)+dc(i+2,3)*d(1,j)+dc(i+2,5)*d(2,j) ;
            k(i+2,j*3+1) += dc(i+2,1)*d(1,j)+dc(i+2,3)*d(0,j)+dc(i+2,4)*d(2,j) ;
            k(i+2,j*3+2) += dc(i+2,2)*d(2,j)+dc(i+2,4)*d(1,j)+dc(i+2,5)*d(0,j) ;
        }
    }
}
 


void Element::StiffUpdChk(int num_shape,double factor,
                       const TMatrix<double> &d,
                       const TMatrix<double> &C,
                       TMatrix<double> &k) const
{
    TMatrix<double> b(6,3*num_shape) ;
    TMatrix<double> bt(3*num_shape,6) ;

    for (int i=0 ; i<num_shape ; ++i) {
        b(0,i*3) = d(0,i) ;
        b(3,i*3) = d(1,i) ;
        b(5,i*3) = d(2,i) ;
        b(1,i*3+1) = d(1,i) ;
        b(3,i*3+1) = d(0,i) ;
        b(4,i*3+1) = d(2,i) ;
        b(2,i*3+2) = d(2,i) ;
        b(4,i*3+2) = d(1,i) ;
        b(5,i*3+2) = d(0,i) ;
        bt(i*3,0) = d(0,i) ;
        bt(i*3,3) = d(1,i) ;
        bt(i*3,5) = d(2,i) ;
        bt(i*3+1,1) = d(1,i) ;
        bt(i*3+1,3) = d(0,i) ;
        bt(i*3+1,4) = d(2,i) ;
        bt(i*3+2,2) = d(2,i) ;
        bt(i*3+2,4) = d(1,i) ;
        bt(i*3+2,5) = d(0,i) ;
    }

    TMatrix<double> bc = bt * C ;
    TMatrix<double> bcb = bc * b ;

    k.TripleProd(b,C,factor) ;
}
 
TVector<StrainVec> Element::Strains(
        const TVector<Vec3D> &coords,
        const TVector<Vec3D> &nodal_disp,
        int iint_pts) const
{
    int int_pts = iint_pts > 0 ? iint_pts :
                  shape->GetDerivDefaultPts() ;
    int num_shape = shape->NumVolumeShape() ;
    TVector<StrainVec> strain(int_pts) ;

    // get the integration coordinates

    const IntegrationPts *intpts = shape->GetIntegrationPtsObject() ;
    const TVector<Vec3D>  &ipts = intpts->GetVolumeCoords(int_pts) ;

    // compute the strains at these points

    for (int i=0; i<ipts.Rows() ; ++i) {

        // get the derivatives of the shape functions,
        // compute the jacobian, it's determinate, and inverse

        TMatrix<double> deriv = shape->VolumeShapeDeriv(ipts[i]) ;
        Jacobian jacob(deriv,coords) ;
        double jac_det = jacob.Determinate() ;
        TMatrix<double> jac_inv = jacob.Inverse(jac_det) ;
        jac_det = fabs(jac_det) ;

        // transform the shape function derivatives to Cartesian coords

        deriv *= jac_inv ;

        // compute the strains

        for (int j=0 ; j<num_shape ; ++j) {
            strain[i][0] += deriv(0,j) * nodal_disp[j].x() ;
            strain[i][1] += deriv(1,j) * nodal_disp[j].y() ;
            strain[i][2] += deriv(2,j) * nodal_disp[j].z() ;
            strain[i][3] += deriv(0,j) * nodal_disp[j].y() +
                            deriv(1,j) * nodal_disp[j].x() ;
            strain[i][4] += deriv(1,j) * nodal_disp[j].z() +
                            deriv(2,j) * nodal_disp[j].y() ;
            strain[i][5] += deriv(2,j) * nodal_disp[j].x() +
                            deriv(1,j) * nodal_disp[j].z() ;
        }
    }
    return(strain) ;
}


// ---------------------------------------------------------------
//  ElementFactory implementation

ElementFactory *ElementFactory::Instance()
{
    if (_instance == 0) {
        _instance = new ElementFactory ;
    }
    return _instance ;
}

ElementFactory *ElementFactory::_instance = 0 ;

int ElementFactory::alloc_quantum = 10 ;

Element *ElementFactory::CreateElement(const char *shape_name)
{
    // check to see if we already have an element of this type

    // THIS IS A DUMB IMPLEMENTATION, IT NEEDS TO BE UPDATED
    // WITH A HASH TABLE

    for (int i=0 ; i<num_in_cache ; ++i) {
        if (strncmp(shape_name,element_cache[i].shape_name,40) == 0)
            return(element_cache[i].element) ;
    }

    // if we get here we need to create a new element.  First make
    // sure we have room in the cache

    if (num_in_cache >= num_allocated) {
        ElementCacheItem *tmp =
            new ElementCacheItem[num_allocated+alloc_quantum] ;
        if (num_allocated > 0) {
            for (int j=0 ; j<num_allocated ; ++j)
                tmp[j] = element_cache[j] ;
            delete [] element_cache ;
        }
        element_cache = tmp ;
    }

    // create the element

    strncpy(element_cache[num_in_cache].shape_name,shape_name,40) ;
    element_cache[num_in_cache].element = new Element(shape_name) ;
    ++num_in_cache ;
    return(element_cache[num_in_cache-1].element) ;
}

} // namespace
