//
// Element class header file
//
// Revision -
//   $Revision: 1.6 $  $Date: 2002/01/31 15:05:57 $  $Author: wash $
//
 
#ifndef Element_h
#define Element_h

#include "Vec3D.hh"
#include "ColTensor.hh"
#include "TVector.hh"
#include "TMatrix.hh"
#include "ShapeFunc.hh"
#include "IntegrationPts.hh"
#include "Jacobian.hh"

namespace FemLib {

class Element {
    public:

//        enum SurfaceShape {TRIANGLE, QUADRILATERAL} ;

        Element() : shape(0),shape_name(0) {} ;

        Element(const char *shape_name) ;

        ~Element() ;

        double Area(int face_num,
                    const TVector<Vec3D> &coords,
                    int iint_pts = 0) const ;

        // computes the area of an element's face

        bool CheckForValidElement(
                             const TVector<Vec3D> &coords,
                             int iint_pts = 0) const ;

        // checks to see if an element's geometry is valid, that is,
        // the jacobian derivative at all Gauss points is != 0 and all
        // have the same size

        void Configure(const char *shape_name) ;

        // configures an element to formulate matricies using the type
        // of shape functions defined in the shape_name (should be a
        // string such as 'Tet2IShapeFunc'

        TMatrix<double> DeformationStiffMtx(
                             const TVector<Vec3D> &coords,
                             const TMatrix<double> &material_mat,
                             int iint_pts = 0) const ;

        // formulates an elasticity stiffness matrix assuming that the
        // constitutive matrix is uniform over the element

        TMatrix<double> DeformationStiffMtx(
                             const TVector<Vec3D> &coords,
                             const TMatrix<double> *material_mats,
                             int iint_pts = 0) const ;

        // formulates an elasticity stiffness matrix with a constitutive
        // matrix given at each Gauss point

        bool FindNaturalCoords(const Vec3D &cartesian,
                               const TVector<Vec3D> &node_coords,
                               Vec3D &natural) const ;

        // finds the natural coordinates in an element that
        // correspond to the given Cartesian coordinates.  Returns
        // false if the point is not in the element.

        int GetDerivDefaultPts() const { return(shape->GetDerivDefaultPts()) ; } ;

        // returns the default number of Gauss points used for integrating
        // derivative based matricies (e.g., stiffness or conduction matrices)

        int GetShapeDefaultPts() const { return(shape->GetShapeDefaultPts()) ; } ;

        // returns the default number of Gauss points used for integrating
        // shape function based matrices (e.g., mass or storage matrices)

        int GetSurfDefaultPts(int face) const
            { return(shape->GetSurfaceShapeDefaultPts(face)) ; } ;

        // returns the default number of Gauss points used for surface
        // integrations.

        const TVector<Vec3D> &GetVolumeIntegrationPts(int int_pts = 0) const ;

        // returns a reference to the Gauss point coordinates used
        // for volume integrations

        void InitialStressStiffUpd(
                             TMatrix<double> &stiff,
                             const TVector<Vec3D> &coords,
                             const TVector<StressVec> &stress,
                             int iint_pts = 0) const ;

        // updates an element stiffness matrix with a geometrical
        // nonlinear contribution due to the current stress state
        // (the so-called "initial stiffness matrix").

        TVector<Vec3D> IntegrateNormalSurfaceScalar(
                             int face_num,
                             const TVector<Vec3D> &coords,
                             double scalar,
                             int iint_pts = 0) const ;

        // integrates a scalar pressure over one surface of an element.
        // The pressure is assumed to be uniform over the surface.  The
        // returned vector has a length equal to the number of shape
        // functions, some of the values being zero.

        TVector<Vec3D> IntegrateNormalSurfaceScalar(
                             int face_num,
                             const TVector<Vec3D> &coords,
                             double *nodal_scalars,
                             int iint_pts = 0) const ;

        // integrates a scalar pressure over one surface of an element.
        // Values of pressure are given for each node point.  The nodal
        // scalar vector should contain values for all nodes, even
        // those not on the face See above for return values.

        TVector<double> IntegrateSurfaceScalar(
                             int face_num,
                             const TVector<Vec3D> &coords,
                             double scalar,
                             int iint_pts = 0) const ;

        // integrates a scalar over one surface of an element.  The scalar
        // is assumed to be uniform over the surface.  The returned vector
        // has a length equal to the number of shape functions, some of
        // the values being zero.

        TVector<double> IntegrateSurfaceScalar(
                             int face_num,
                             const TVector<Vec3D> &coords,
                             double *nodal_scalars,
                             int iint_pts = 0) const ;

        // integrates a scalar over one surface of an element.  Values of
        // the scalar are given for each node point.  The nodal scalar
        // vector should contain values for all nodes, even those not
        // on the face See above for return values.

        TVector<Vec3D> IntegrateSurfaceVector(
                             int face_num,
                             const TVector<Vec3D> &coords,
                             const Vec3D &value,
                             int iint_pts = 0) const ;

        // integrates a vector over one surface of an element.  The vector
        // is assumed to be uniform over the surface.  The returned vector
        // has a length equal to the number of shape functions, some of the
        // values being zero.

        TVector<Vec3D> IntegrateSurfaceVector(
                             int face_num,
                             const TVector<Vec3D> &coords,
                             const TVector<Vec3D> &nodal_values,
                             int iint_pts = 0) const ;

        // integrates a vector over one surface of an element.  A value for
        // the vector is given at each of the node point.  The nodal
        // values should contain values for all nodes, even those not
        // on the face. See above for the returned value

        TVector<double> IntegrateVolume(
                             const TVector<Vec3D> &coords,
                             double scalar,
                             int iint_pts = 0) const ;

        // integrates a scalar over the volume of an element.  The scalar
        // is assumed to be uniform over the element.

        TVector<double> IntegrateVolume(
                             const TVector<Vec3D> &coords,
                             double *scalars,
                             int iint_pts = 0) const ;

        // integrates a scalar over the volume of an element.  A value
        // of the scalar is given for each of the integration points.

        TVector<Vec3D> IntegrateVolume(
                             const TVector<Vec3D> &coords,
                             const StressVec &stress,
                             int int_pts = 0) const ;

        // integrates a stress vector over the volume of an element.
        // The stress is assumed to be uniform over the element.

        TVector<Vec3D> IntegrateVolume(
                             const TVector<Vec3D> &coords,
                             const TVector<StressVec> &stress,
                             int int_pts = 0) const ;

        // integrates a stress vector over the volume of an element.
        // It is assumed that a stress vector is given for each of
        // of the Gauss points.

        double InterpolateSurfaceScalar(
                             int face_num,
                             const Vec3D &natural_coords,
                             const TVector<double> &node_values) const ;

        double InterpolateSurfaceScalar(
                             int face_num,
                             const Vec3D &natural_coords,
                             const double *node_values) const ;

        // interpolates scalar nodal values over one surface of an element.  It
        // is assumed that nodal values are given for all nodes even though
        // some will not be used.

        double InterpolateVolume(
                             const Vec3D &natural_coords,
                             const TVector<double> &node_values) const ;

        double InterpolateVolume(
                             const Vec3D &natural_coords,
                             const double *node_values) const ;

        // interpolates scalar nodal values over the volume of an element

        Vec3D InterpolateVolume(
                             const Vec3D &natural_coords,
                             const TVector<Vec3D> &node_values) const ;

        // interpolates vector nodal values over the volume of an element

        TVector<Vec3D> InternalPointForces(
                             const TVector<Vec3D> &nat_coords,
                             const TVector<Vec3D> &values) const ;

        // Computes the nodal forces due to an internal point force
        // at the given natural coordinate location

        int NumVolumeShapeFunc() { return(shape->NumVolumeShape()) ; } ;

        // returns the number of volume shape functions for an element

        int NumFaceNodes(int face) { return(shape->GetFaceMap(face)[0]) ; } ;

        // returns the number of nodes for the given face of an element

        TMatrix<double> ScalarFieldDerivMtx(
                             const TVector<Vec3D> &coords,
                             const TMatrix<double> &material_mat,
                             int iint_pts = 0) const ;

        // formulates the derivative matrix for a scalar field (e.g., the
        // conductivity matrix for thermal analysis.  This version assumes
        // that the material matrix is uniform throughout the element.

        TMatrix<double> ScalarFieldDerivMtx(
                             const TVector<Vec3D> &coords,
                             const TMatrix<double> *material_mats,
                             int iint_pts = 0) const ;

        // formulates the derivative matrix for a scalar field (e.g., the
        // conductivity matrix for thermal analysis.  This version assumes
        // that that a value for the material matrix is given for each
        // of the Gauss points

        TMatrix<double> ScalarFieldShapeMtx(
                             const TVector<Vec3D> &coords,
                             double material_prop,
                             int iint_pts = 0) const ;

        // formulates the shape function matrix for a scalar field (e.g.,
        // the storage matrix for thermal analysis or the mass matrix
        // for elasticity).  This version assumes that the material
        // properties are uniform over the element.

        TMatrix<double> ScalarFieldShapeMtx(
                             const TVector<Vec3D> &coords,
                             double *material_props,
                             int iint_pts = 0) const ;

        // formulates the shape function matrix for a scalar field (e.g.,
        // the storage matrix for thermal analysis or the mass matrix
        // for elasticity).  This version assumes that material
        // properties are give for each of the Gauss points.

        const char *ShapeFuncName() const { return(shape_name) ; } ;

        // returns the name of the shape function set for which the
        // element is currently configured.

        TMatrix<double> StiffMtxCheck(
                             const TVector<Vec3D> &coords,
                             const TMatrix<double> &material_mat,
                             int iint_pts = 0) const ;

        // formulates an elasticity stiffness matrix assuming isotropic
        // material properties 

        TMatrix<double> StiffMtxCheck(
                             const TVector<Vec3D> &coords,
                             const TMatrix<double> *material_mats,
                             int iint_pts = 0) const ;

        // formulates an elasticity stiffness matrix assuming isotropic
        // material properties 

        TVector<StrainVec> Strains(
                             const TVector<Vec3D> &coords,
                             const TVector<Vec3D> &nodal_disp,
                             int iint_pts = 0) const ;

        // computes the strains at the integration points

        double Volume(const TVector<Vec3D> &coords,
                      int iint_pts = 0) const ;

        // computes the volume of an element

    private:
        const ShapeFunc *shape ;

        char *shape_name, *intpts_name ;

        TMatrix<double> DeformationStiffMtxHelp(
                             const TVector<Vec3D> &coords,
                             const TMatrix<double> *material_mats,
                             bool uniform_material_mat,
                             int iint_pts) const ;

        TVector<Vec3D> IntegrateNormalSurfaceScalarHelp1(
                             int face_num,
                             const TVector<Vec3D> &coords,
                             double *scalar,
                             bool uniform_scalar,
                             int int_pts) const ;

        TVector<Vec3D> IntegrateNormalSurfaceScalarHelp2(
                             const ShapeFunc::SurfaceShape surface_shape,
                             const TVector<Vec3D> &coords,
                             const TVector<double> &scalar,
                             int int_pts) const ;

        TVector<double> IntegrateSurfaceScalarHelp1(
                             int face_num,
                             const TVector<Vec3D> &coords,
                             double *scalar,
                             bool uniform_scalar,
                             int int_pts) const ;

        TVector<double> IntegrateSurfaceScalarHelp2(
                             const ShapeFunc::SurfaceShape surface_shape,
                             const TVector<Vec3D> &coords,
                             const TVector<double> &scalar,
                             int int_pts) const ;

        TVector<double> IntegrateVolumeScalarHelp(
                             const TVector<Vec3D> &coords,
                             double *scalar,
                             bool uniform_scalar,
                             int int_pts) const ;

        TVector<Vec3D> IntegrateVolumeStressHelp(
                             const TVector<Vec3D> &coords,
                             const TVector<StressVec> &stress,
                             bool uniform_scalar,
                             int int_pts) const ;

        TVector<Vec3D> IntegrateSurfaceVectorHelp1(
                             int face_num,
                             const TVector<Vec3D> &coords,
                             const TVector<Vec3D> &values,
                             bool uniform_scalar,
                             int iint_pts) const ;

        TVector<Vec3D> IntegrateSurfaceVectorHelp2(
                             const ShapeFunc::SurfaceShape surface_shape,
                             const TVector<Vec3D> &coords,
                             const TVector<Vec3D> &values,
                             int int_pts) const ;

        double InterpolateSurfaceScalarHelp(
                             const ShapeFunc::SurfaceShape surface_shape,
                             const Vec3D &natural_coords,
                             const TVector<double> &node_values) const ;

        TMatrix<double> ScalarFieldDerivMtxHelp(
                             const TVector<Vec3D> &coords,
                             const TMatrix<double> *material_mats,
                             bool uniform_scalar,
                             int int_pts = 0) const ;

        TMatrix<double> ScalarFieldShapeMtxHelp(
                             const TVector<Vec3D> &coords,
                             double *material_props,
                             bool uniform_material_props,
                             int iint_pts) const ;

        TMatrix<double> StiffMtxCheckHelp(
                             const TVector<Vec3D> &coords,
                             const TMatrix<double> *material_mats,
                             bool uniform_material_mat,
                             int iint_pts) const ;

        void StiffUpd(int num_shape,double factor,
                       const TMatrix<double> &d,
                       const TMatrix<double> &C,
                       TMatrix<double> &k) const ;

        void StiffUpdChk(int num_shape,double factor,
                       const TMatrix<double> &d,
                       const TMatrix<double> &C,
                       TMatrix<double> &k) const ;
} ;


inline std::ostream &operator << (std::ostream &out,const Element &element)
{
    out << "Element( " << element.ShapeFuncName() << " ) " ;
    return(out) ;
}


inline std::istream &operator >> (std::istream &in, Element &element)
{
    char buffer0[10],buffer1[128],buffer2[5] ;

    in >> buffer0 >> buffer1 >> buffer2 ;
   
    element.Configure(buffer1) ;
    return(in) ;
}


inline std::ostream &operator << (std::ostream &out,const Element *element)
{
    out << "Element( " << element->ShapeFuncName() << " ) " ;
    return(out) ;
}


inline std::istream &operator >> (std::istream &in, Element *element)
{
    char buffer0[10],buffer1[128],buffer2[5] ;

    in >> buffer0 >> buffer1 >> buffer2 ;
   
    element->Configure(buffer1) ;
    return(in) ;
}


//------------------------------------------------------------------
// This is a factory for element classes.  It insures that only
// one instance is created for each element type. This is a
// singleton class so things are set up so that only one instance
// is created.

class ElementFactory {
    public:
        static ElementFactory *Instance() ;

        Element *CreateElement(const char *shape_name) ;

    protected:
        ElementFactory() : num_in_cache(0), num_allocated(0) {} ;
        ~ElementFactory()
            { if (num_allocated > 0) delete [] element_cache ; } ;

    private:
        static ElementFactory *_instance ;
        static int alloc_quantum ;

        struct ElementCacheItem {
            char shape_name[40] ;
            Element *element ;
        } ;

        ElementCacheItem *element_cache ;
        int num_in_cache ;
        int num_allocated ;
} ;

} // namespace

#endif
