/*
// FemLibModule - python extension for the FemLib
//
*/

#include <cstdio>
#include <cstdlib>
#include "Python.h"
#include "Numeric/arrayobject.h"
#include "FemLib.hh"

using FemLib::Element ;
using FemLib::ElementFactory ;
using FemLib::ColTensor ;
using FemLib::TMatrix ;
using FemLib::TVector ;
using FemLib::Vec3D ;

#if BAD_STATIC_FORWARD
staticforward PyTypeObject element_type ;
#else
PyTypeObject element_type ;
#endif

typedef struct {
    PyObject_HEAD
    Element *element ;
    PyObject *vec3d_class ;
    PyObject *coltensor_class ;
    }
    ElementObject ;

extern "C" {
    PyObject *Area(ElementObject *self,PyObject *args) ;
    PyObject *CheckForValidElement(ElementObject *self,PyObject *args) ;
    PyObject *DeformationStiffMtx(ElementObject *self,PyObject *args) ;
    PyObject *FindNaturalCoords(ElementObject *self,PyObject *args) ;
    PyObject *GetDerivDefaultPts(ElementObject *self,PyObject *args) ;
    PyObject *GetShapeDefaultPts(ElementObject *self,PyObject *args) ;
    PyObject *GetSurfDefaultPts(ElementObject *self,PyObject *args) ;
    PyObject *GetVolumeIntegratonPts(ElementObject *self,PyObject *args) ;
    PyObject *InitialStressStiffMtx(ElementObject *self,PyObject *args) ;
    PyObject *IntegrateNormalSurfaceScalar(ElementObject *self,PyObject *args) ;
    PyObject *IntegrateSurfaceScalar(ElementObject *self,PyObject *args) ;
    PyObject *IntegrateSurfaceVector(ElementObject *self,PyObject *args) ;
    PyObject *IntegrateVolume(ElementObject *self,PyObject *args) ;
    PyObject *InterpolateSurfaceScalar(ElementObject *self,PyObject *args) ;
    PyObject *InterpolateVolume(ElementObject *self,PyObject *args) ;
    PyObject *InternalPointForces(ElementObject *self,PyObject *args) ;
    PyObject *NumVolumeShapeFunc(ElementObject *self,PyObject *args) ;
    PyObject *NumFaceNodes(ElementObject *self,PyObject *args) ;
    PyObject *ScalarFieldDerivMtx(ElementObject *self,PyObject *args) ;
    PyObject *ScalarFieldShapeMtx(ElementObject *self,PyObject *args) ;
    PyObject *Strains(ElementObject *self,PyObject *args) ;
    PyObject *Volume(ElementObject *self,PyObject *args) ;

    void initFemLib() ;
}

#define is_ElementObject(v) ((v)->ob_type == &element_type)

/* ----------------------------------------------------------
**  These methods setup and check for known python classes
*/

static bool FindExternalClasses(ElementObject *self)
{
    PyObject *dict = PyImport_GetModuleDict() ;

    // Vec3D

    PyObject *vec3d_mod = PyDict_GetItemString(dict,"Vec3D") ;
    if (vec3d_mod == 0) {
        vec3d_mod = PyImport_ImportModule("Vec3D") ;
        if (vec3d_mod == 0) return(false) ;
        Py_DECREF(vec3d_mod) ;
    }
    PyObject *mod_dict = PyModule_GetDict(vec3d_mod) ;
    self->vec3d_class = PyDict_GetItemString(mod_dict,"Vec3D") ;

    // ColTensor

    PyObject *coltensor_mod = PyDict_GetItemString(dict,"ColTensor") ;
    if (coltensor_mod == 0) {
        coltensor_mod = PyImport_ImportModule("ColTensor") ;
        if (coltensor_mod == 0) return(false) ;
        Py_DECREF(coltensor_mod) ;
    }
    mod_dict = PyModule_GetDict(coltensor_mod) ;
    self->coltensor_class = PyDict_GetItemString(mod_dict,"ColTensor") ;

    return(true) ;
}

static bool Vec3D_Check(ElementObject *self,PyObject *obj)
{
    return(PyObject_IsInstance(obj,self->vec3d_class) != 0) ;
}

static bool ColTensor_Check(ElementObject *self,PyObject *obj)
{
    return(PyObject_IsInstance(obj,self->coltensor_class) != 0) ;
}

/* ----------------------------------------------------------
**  These methods convert input arguments
*/

static TVector<Vec3D> *ConvertVec3DSeq(ElementObject *self,
                                        PyObject *obj,int num_shape)
{
    if ((PySequence_Check(obj) == 0) ||
        (PySequence_Size(obj) < num_shape)) {
        PyErr_SetString(PyExc_TypeError,
            "argument should a sequence of Vec3D's") ;
        return(0) ;
    }

    TVector<Vec3D> *vect = new TVector<Vec3D>(num_shape) ; 
    for (int i=0 ; i<num_shape ; ++i) {
        PyObject *item = PySequence_GetItem(obj,i) ;
        if (!Vec3D_Check(self,item)) {
            Py_DECREF(item) ;
            PyErr_SetString(PyExc_TypeError,
                "argument should a sequence of Vec3D's") ;
            return(0) ;
        }
        PyObject *data = PyObject_GetAttrString(item,"data") ;
        for (int j=0 ; j<3 ; ++j) {
            PyObject *v = PySequence_GetItem(data,j) ;
            if (PyFloat_Check(v) != 0) {
                (*vect)[i][j] = PyFloat_AsDouble(v) ;
            } else {
                (*vect)[i][j] = PyInt_AsLong(v) ;
            }
            Py_DECREF(v) ;
        }
        Py_DECREF(item) ;
        Py_DECREF(data) ;
    }
    return(vect) ;
}

static bool ConvertVec3D(ElementObject *self,
                         PyObject *obj,Vec3D &value)
{
    if (!Vec3D_Check(self,obj)) {
        PyErr_SetString(PyExc_TypeError,
                "argument should a sequence of Vec3D's") ;
        return(false) ;
    }

    PyObject *data = PyObject_GetAttrString(obj,"data") ;
    for (int j=0 ; j<3 ; ++j) {
        PyObject *v = PySequence_GetItem(data,j) ;
        if (PyFloat_Check(v) != 0) {
            value[j] = PyFloat_AsDouble(v) ;
        } else {
            value[j] = PyInt_AsLong(v) ;
        }
        Py_DECREF(v) ;
    }
    Py_DECREF(data) ;
    return(true) ;
}


static TVector<ColTensor> *ConvertColTensorSeq(ElementObject *self,
                                        PyObject *obj,int num_shape)
{
    if ((PySequence_Check(obj) == 0) ||
        (PySequence_Size(obj) < num_shape)) {
        PyErr_SetString(PyExc_TypeError,
            "argument should be a sequence of ColTensors") ;
        return(0) ;
    }

    TVector<ColTensor> *vect = new TVector<ColTensor>(num_shape) ; 
    for (int i=0 ; i<num_shape ; ++i) {
        PyObject *item = PySequence_GetItem(obj,i) ;
        if (!ColTensor_Check(self,item)) {
            Py_DECREF(item) ;
            PyErr_SetString(PyExc_TypeError,
                "argument should a sequence of ColTensors") ;
            return(0) ;
        }
        PyObject *data = PyObject_GetAttrString(item,"data") ;
        for (int j=0 ; j<6 ; ++j) {
            PyObject *v = PySequence_GetItem(data,j) ;
            if (PyFloat_Check(v)) {
                 (*vect)[i][j] = PyFloat_AsDouble(v) ;
            } else {
                 (*vect)[i][j] = PyInt_AsLong(v) ;
            }
            Py_DECREF(v) ;
        }
        Py_DECREF(item) ;
        Py_DECREF(data) ;
    }
    return(vect) ;
}

static bool ConvertColTensor(ElementObject *self,
                             PyObject *obj,ColTensor &value)
{
    if (!ColTensor_Check(self,obj)) {
        PyErr_SetString(PyExc_TypeError,
            "argument should be a ColTensor") ;
        return(false) ;
    }
    PyObject *data = PyObject_GetAttrString(obj,"data") ;
    for (int j=0 ; j<6 ; ++j) {
        PyObject *v = PySequence_GetItem(data,j) ;
        if (PyFloat_Check(v)) {
             value[j] = PyFloat_AsDouble(v) ;
        } else {
             value[j] = PyInt_AsLong(v) ;
        }
        Py_DECREF(v) ;
    }
    return(true) ;
}

static double *ConvertFloatSeq(PyObject *obj,int num)
{
    if ((PySequence_Check(obj) == 0) ||
        (PySequence_Size(obj) < num)) {
        PyErr_SetString(PyExc_TypeError,
            "argument should a sequence of floats") ;
        return(0) ;
    }

    double *vect = new double[num] ; 
    for (int i=0 ; i<num ; ++i) {
        PyObject *item = PySequence_GetItem(obj,i) ;
        if (PyFloat_Check(item) != 0) {
            vect[i] = PyFloat_AsDouble(item) ;
        } else if (PyInt_Check(item) != 0) {
            vect[i] = PyInt_AsLong(item) ;
        } else {
            Py_DECREF(item) ;
            PyErr_SetString(PyExc_TypeError,
                "sequence member should be a float") ;
            return(0) ;
        }
        Py_DECREF(item) ;
    }
    return(vect) ;
}

/* ----------------------------------------------------------
**  These methods convert outputs
*/

static PyObject *TMatrixToPyArray(TMatrix<double> &mtx)
{
    int dims[2] ;
    dims[0] = mtx.Rows() ;  dims[1] = mtx.Cols() ;
    PyArrayObject *out = 
        (PyArrayObject *)PyArray_FromDims(2,dims,PyArray_DOUBLE) ;

    double *adata = (double *)out->data ;
    int cur = 0 ;
    for (int i=0 ; i<mtx.Rows() ; ++i) {
        for (int j=0 ; j<mtx.Cols() ; ++j) {
            adata[cur] = mtx(i,j) ;
            ++cur ;
        }
    }
    return((PyObject *)out) ;
}


static PyObject *TVectorToPyArray(TVector<Vec3D> &vec)
{
    int dims[2] ;
    dims[0] = vec.Rows() ;  dims[1] = 3 ;
    PyArrayObject *out = 
        (PyArrayObject *)PyArray_FromDims(2,dims,PyArray_DOUBLE) ;

    double *adata = (double *)out->data ;
    int cur = 0 ;
    for (int i=0 ; i<vec.Rows() ; ++i) {
        for (int j=0 ; j<3 ; ++j) {
            adata[cur] = vec[i][j] ;
            ++cur ;
        }
    }
    return((PyObject *)out) ;
}


static PyObject *TDVectorToPyArray(TVector<double> &vec)
{
    int dims[1] ;
    dims[0] = vec.Rows() ;
    PyArrayObject *out = 
        (PyArrayObject *)PyArray_FromDims(1,dims,PyArray_DOUBLE) ;

    double *adata = (double *)out->data ;
    int cur = 0 ;
    for (int i=0 ; i<vec.Rows() ; ++i) {
        adata[cur] = vec[i] ;
        ++cur ;
    }
    return((PyObject *)out) ;
}

static PyObject *Vec3DToPyVec3D(ElementObject *self,Vec3D &cd)
{
    PyObject *new_vec = PyInstance_New(self->vec3d_class,
                  Py_BuildValue("ddd",cd.x(),cd.y(),cd.z()),0) ;
    return(new_vec) ;
}

static PyObject *Vec3DListToPyVec3DTuple(ElementObject *self,
                                         const TVector<Vec3D> &list)
{
    // create a tuple object

    PyObject *tuple = PyTuple_New(list.Rows()) ;

    // now fill it

    for (int i=0 ; i<list.Rows() ; ++i) {
        PyObject *new_vec = PyInstance_New(self->vec3d_class,
                  Py_BuildValue("ddd",list[i].x(),list[i].y(),
                                      list[i].z()),0) ;
        PyTuple_SetItem(tuple,i,new_vec) ;
    }
    return(tuple) ;
}

static PyObject *ColTensorListToPyColTensorList(ElementObject *self,
                                     const TVector<ColTensor> &list)
{
    // create a list object

    PyObject *py_list = PyList_New(list.Rows()) ;

    // now fill it

    for (int i=0 ; i<list.Rows() ; ++i) {
        PyObject *new_vec = PyInstance_New(self->coltensor_class,
                  Py_BuildValue("dddddd",list[i].xx(),list[i].yy(),
                                         list[i].zz(),list[i].xy(),
                                         list[i].yz(),list[i].zx()),0) ;
        PyList_SetItem(py_list,i,new_vec) ;
    }
    return(py_list) ;
}

/* ----------------------------------------------------------
**  These routines implement the class methods
*/

static PyObject *Area(ElementObject *self,PyObject *args)
{
    int face ;
    int order = 0 ;
    PyObject *coord_obj ;
    if (!PyArg_ParseTuple(args,"iO|i",&face,&coord_obj,&order))
        return(NULL) ;

    TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                     self->element->NumVolumeShapeFunc()) ;
    if (coords == 0) return(NULL) ;

    return(Py_BuildValue("d",self->element->Area(face,*coords,order))) ;
}

static PyObject *CheckForValidElement(ElementObject *self,PyObject *args)
{
    PyObject *coord_obj ;
    int order = 5 ;

    if (!PyArg_ParseTuple(args,"O|i",&coord_obj,&order)) return(NULL) ;

    TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                     self->element->NumVolumeShapeFunc()) ;
    if (coords == 0) return(NULL) ;

    bool valid = self->element->CheckForValidElement(*coords,order) ;
    delete coords ;

    return(Py_BuildValue("i",valid ? 1 : 0)) ;
}


static PyObject *DeformationStiffMtx(ElementObject *self,PyObject *args)
{
    PyObject *coord_obj,*mat_obj ;
    int order = 0 ;

    if (!PyArg_ParseTuple(args,"OO|i",&coord_obj,&mat_obj,&order))
        return(NULL) ;

    if (order == 0) order =  self->element->GetDerivDefaultPts() ;

    if (PyArray_Check(mat_obj)) {

        PyArrayObject *matarray = (PyArrayObject *)mat_obj ;
        if ((matarray->dimensions[0] != 6) ||
            (matarray->dimensions[1] != 6) ||
            (matarray->descr->type_num != PyArray_DOUBLE)) {
            PyErr_SetString(PyExc_TypeError,
            "2nd argument should be a 6x6 PyArray or a sequence of these") ;
            return(NULL) ;
        }

        TMatrix<double> mat(6,6) ;
        for (int i=0 ; i<6 ; ++i) {
            for (int j=0 ; j<6 ; ++j) {
                char *eptr = matarray->data + i*matarray->strides[0] +
                             j*matarray->strides[1] ;
                mat(i,j) = *((double *)eptr) ;
            }
        }

        TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                         self->element->NumVolumeShapeFunc()) ;
        if (coords == 0) return(NULL) ;

        TMatrix<double> k = self->element->DeformationStiffMtx(
                                    *coords,mat,order) ;
        delete coords ;
        return(TMatrixToPyArray(k)) ;

    } else if (PySequence_Check(mat_obj)) {

        if (PySequence_Size(mat_obj) != order) {
            PyErr_SetString(PyExc_TypeError,
            "Number of material prop values differs from the # of integration points") ;
            return(NULL) ;
        }

        TMatrix<double> *mat_props = new TMatrix<double>[order] ;
        for (int i=0 ; i < order ; ++i) {
            PyObject *item = PySequence_GetItem(mat_obj,i) ;
            PyArrayObject *matarray = (PyArrayObject *)item ;
            if ((matarray->dimensions[0] != 6) ||
                (matarray->dimensions[1] != 6) ||
                (matarray->descr->type_num != PyArray_DOUBLE)) {
                PyErr_SetString(PyExc_TypeError,
                "2nd argument should be a 6x6 PyArray or a sequence of these") ;
                Py_DECREF(item) ;
                delete [] mat_props ;
                return(NULL) ;
            }
            mat_props[i].Configure(6,6) ;
            for (int k=0 ; k<6 ; ++k) {
                for (int j=0 ; j<6 ; ++j) {
                    char *eptr = matarray->data + 
                                 k*matarray->strides[0] +
                                 j*matarray->strides[1] ;
                    mat_props[i](k,j) = *((double *)eptr) ;
                }
            }
            Py_DECREF(item) ;
        }

        TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                         self->element->NumVolumeShapeFunc()) ;
        if (coords == 0) return(NULL) ;

        TMatrix<double> k = self->element->DeformationStiffMtx(
                                *coords,mat_props,order) ;
        delete coords ;
        return(TMatrixToPyArray(k)) ;

    }

    PyErr_SetString(PyExc_TypeError,
        "2nd argument should be a 6x6 PyArray or a sequence of these") ;
    return(NULL) ;
}


static PyObject *FindNaturalCoords(ElementObject *self,PyObject *args)
{
    PyObject *coord_obj,*cart_obj ;

    if (!PyArg_ParseTuple(args,"OO",&cart_obj,&coord_obj)) return(NULL) ;

    Vec3D cartesian ;
    if (!ConvertVec3D(self,cart_obj,cartesian)) return(NULL) ;

    TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                     self->element->NumVolumeShapeFunc()) ;
    if (coords == 0) return(NULL) ;

    Vec3D natural ;
    bool valid = self->element->FindNaturalCoords(cartesian,*coords,
                                                  natural) ;
    delete coords ;

    if (valid) {
        return(Vec3DToPyVec3D(self,natural)) ;
    }
    return(Py_BuildValue("i",0)) ;
}

static PyObject *GetDerivDefaultPts(ElementObject *self,PyObject *args)
{
    if (!PyArg_ParseTuple(args,"")) return(NULL) ;

    return(Py_BuildValue("i",self->element->GetDerivDefaultPts())) ;
}


static PyObject *GetShapeDefaultPts(ElementObject *self,PyObject *args)
{
    if (!PyArg_ParseTuple(args,"")) return(NULL) ;

    return(Py_BuildValue("i",self->element->GetShapeDefaultPts())) ;
}

static PyObject *GetSurfDefaultPts(ElementObject *self,PyObject *args)
{
    int face ;
    if (!PyArg_ParseTuple(args,"i",&face)) return(NULL) ;

    return(Py_BuildValue("i",self->element->GetSurfDefaultPts(face))) ;
}

static PyObject *GetVolumeIntegrationPts(ElementObject *self,PyObject *args)
{
    int num_pts = 0 ;

    if (!PyArg_ParseTuple(args,"|i",&num_pts)) return(NULL) ;

    const TVector<Vec3D> gp =
             self->element->GetVolumeIntegrationPts(num_pts) ;

    return(Vec3DListToPyVec3DTuple(self,gp)) ;
}


static PyObject *InitialStressStiffMtx(ElementObject *self,PyObject *args)
{
    PyObject *coord_obj,*str_obj ;
    int order = 0 ;

    if (!PyArg_ParseTuple(args,"OO|i",&coord_obj,&str_obj,&order))
        return(NULL) ;

    int nums = self->element->NumVolumeShapeFunc() ;
    if (order == 0) order =  self->element->GetDerivDefaultPts() ;

    TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,nums) ;
    if (coords == 0) return(NULL) ;

    TVector<ColTensor> *stress = ConvertColTensorSeq(
                                     self,str_obj,order) ;
    if (stress == 0) {
        delete stress ;
        return(NULL) ;
    }

    TMatrix<double> k(nums*3,nums*3) ;
    self->element->InitialStressStiffUpd(k,*coords,*stress,order) ;
    return(TMatrixToPyArray(k)) ;
}


static PyObject *IntegrateNormalSurfaceScalar(ElementObject *self,PyObject *args)
{
    PyObject *coord_obj,*value_obj ;
    int face,order = 0 ;

    if (!PyArg_ParseTuple(args,"iOO|i",&face,&coord_obj,
                          &value_obj,&order))
        return(NULL) ;

    if (order == 0) order = self->element->GetSurfDefaultPts(face) ;

//    int num_face_nodes = self->element->NumFaceNodes(face) ;

    if (PyInt_Check(value_obj) || PyFloat_Check(value_obj)) {

        TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                     self->element->NumVolumeShapeFunc()) ;
        if (coords == 0) return(NULL) ;

        double value ;
        if (PyInt_Check(value_obj) == 0) {
            value = PyInt_AsLong(value_obj) ;
        } else {
            value = PyFloat_AsDouble(value_obj) ;
        }
        TVector<Vec3D> f = self->element->IntegrateNormalSurfaceScalar(
                            face,*coords,value,order) ;
        delete coords ;
        return(TVectorToPyArray(f)) ;

    } else if (PySequence_Check(value_obj)) {

        TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                     self->element->NumVolumeShapeFunc()) ;
        if (coords == 0) return(NULL) ;

        double *values = ConvertFloatSeq(value_obj,
                     self->element->NumVolumeShapeFunc()) ;
        if (values == 0) return(NULL) ;

        TVector<Vec3D> f = self->element->IntegrateNormalSurfaceScalar(
                       face,*coords,values,order) ;
        delete coords ;
        delete values ;
        return(TVectorToPyArray(f)) ;
    }

    PyErr_SetString(PyExc_TypeError,
        "3rd argument should be a scalar or a sequence of scalars") ;
    return(NULL) ;
}



static PyObject *IntegrateSurfaceScalar(ElementObject *self,PyObject *args)
{
    PyObject *coord_obj,*value_obj ;
    int face,order = 0 ;

    if (!PyArg_ParseTuple(args,"iOO|i",&face,&coord_obj,
                          &value_obj,&order))
        return(NULL) ;

    if (order == 0) order = self->element->GetSurfDefaultPts(face) ;

//    int num_face_nodes = self->element->NumFaceNodes(face) ;

    if (PyInt_Check(value_obj) || PyFloat_Check(value_obj)) {

        TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                     self->element->NumVolumeShapeFunc()) ;
        if (coords == 0) return(NULL) ;

        double value ;
        if (PyInt_Check(value_obj) == 0) {
            value = PyInt_AsLong(value_obj) ;
        } else {
            value = PyFloat_AsDouble(value_obj) ;
        }
        TVector<double> f = self->element->IntegrateSurfaceScalar(
                            face,*coords,value,order) ;
        delete coords ;
        return(TDVectorToPyArray(f)) ;

    } else if (PySequence_Check(value_obj)) {

        TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                     self->element->NumVolumeShapeFunc()) ;
        if (coords == 0) return(NULL) ;

        double *values = ConvertFloatSeq(value_obj,
                     self->element->NumVolumeShapeFunc()) ;
        if (values == 0) return(NULL) ;

        TVector<double> f = self->element->IntegrateSurfaceScalar(
                       face,*coords,values,order) ;
        delete coords ;
        delete values ;
        return(TDVectorToPyArray(f)) ;
    }

    PyErr_SetString(PyExc_TypeError,
        "3rd argument should be a scalar or a sequence of scalars") ;
    return(NULL) ;
}


static PyObject *IntegrateSurfaceVector(ElementObject *self,PyObject *args)
{
    PyObject *coord_obj,*value_obj ;
    int face,order = 0 ;

    if (!PyArg_ParseTuple(args,"iOO|i",&face,&coord_obj,
                          &value_obj,&order))
        return(NULL) ;

    if (order == 0) order = self->element->GetSurfDefaultPts(face) ;

//    int num_face_nodes = self->element->NumFaceNodes(face) ;

    if (Vec3D_Check(self,value_obj)) {

        TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                     self->element->NumVolumeShapeFunc()) ;
        if (coords == 0) return(NULL) ;

        Vec3D value ;
        if (!ConvertVec3D(self,value_obj,value)) return(NULL) ;

        TVector<Vec3D> f = self->element->IntegrateSurfaceVector(
                            face,*coords,value,order) ;
        delete coords ;
        return(TVectorToPyArray(f)) ;

    } else if (PySequence_Check(value_obj)) {

        TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                     self->element->NumVolumeShapeFunc()) ;
        if (coords == 0) return(NULL) ;

        TVector<Vec3D> *values = ConvertVec3DSeq(self,value_obj,
                     self->element->NumVolumeShapeFunc()) ;
        if (values == 0) return(NULL) ;

        TVector<Vec3D> f = self->element->IntegrateSurfaceVector(
                       face,*coords,*values,order) ;
        delete coords ;
        delete values ;
        return(TVectorToPyArray(f)) ;

    }

    PyErr_SetString(PyExc_TypeError,
        "3rd argument should be a Vec3D object or a sequence of these") ;
    return(NULL) ;
}


static PyObject *IntegrateVolume(ElementObject *self,PyObject *args)
{
    PyObject *coord_obj,*value_obj ;
    int order = 0 ;

    if (!PyArg_ParseTuple(args,"OO|i",&coord_obj,
                          &value_obj,&order))
        return(NULL) ;

    TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
               self->element->NumVolumeShapeFunc()) ;
    if (coords == 0) return(NULL) ;

    // check to see if we have scalars or stresses

    if (PyInt_Check(value_obj) || PyFloat_Check(value_obj)) {

        double value ;
        if (PyInt_Check(value_obj) == 0) {
            value = PyInt_AsLong(value_obj) ;
        } else {
            value = PyFloat_AsDouble(value_obj) ;
        }
        TVector<double> nvals = self->element->IntegrateVolume(
                                 *coords,value,order) ;
        return(TDVectorToPyArray(nvals)) ;

    } else if (ColTensor_Check(self,value_obj)) {

        ColTensor value ;
        if (!ConvertColTensor(self,value_obj,value)) return(0) ; 

        TVector<Vec3D> nvals = self->element->IntegrateVolume(
                                 *coords,value,order) ;

        return(TVectorToPyArray(nvals)) ;

    } else if (PySequence_Check(value_obj)) {

        PyObject *item = PySequence_GetItem(value_obj,0) ;

        if (PyInt_Check(item) || PyFloat_Check(item)) {

            // scalar sequence

            double *scalars = ConvertFloatSeq(value_obj,
                               PySequence_Size(value_obj)) ;
            if (scalars == 0) return(NULL) ;
            TVector<double> nvals = self->element->IntegrateVolume(
                                     *coords,scalars,order) ;
            delete scalars ;
            Py_DECREF(item) ;
            return(TDVectorToPyArray(nvals)) ;

        } else {

            // stress sequence

            TVector<ColTensor> *stress = ConvertColTensorSeq(
                               self,value_obj,
                               PySequence_Size(value_obj)) ;
            if (stress == 0) return(NULL) ;
            TVector<Vec3D> nvals = self->element->IntegrateVolume(
                                     *coords,*stress,order) ;
            delete stress ;
            Py_DECREF(item) ;
            return(TVectorToPyArray(nvals)) ;
        }

    }

    PyErr_SetString(PyExc_TypeError,
            "2nd argument should be a scalar, a scalar sequence, ColTensor, or sequence of these") ;
    return(0) ;
}

static PyObject *InterpolateSurfaceScalar(ElementObject *self,PyObject *args)
{
    PyObject *natural_obj,*node_vals_obj ;
    int face_num ;

    if (!PyArg_ParseTuple(args,"iOO",&face_num,&natural_obj,
                          &node_vals_obj))
        return(NULL) ;

    Vec3D natural ;
    if (!ConvertVec3D(self,natural_obj,natural)) return(NULL) ;

    if (PySequence_Check(node_vals_obj) == 0) {
        PyErr_SetString(PyExc_TypeError,
            "3nd argument should a sequence") ;
        return(0) ;
    }

    double *nvalues = ConvertFloatSeq(node_vals_obj,
                           self->element->NumVolumeShapeFunc()) ;
    if (nvalues == 0) return(NULL) ;

    double value = self->element->
             InterpolateSurfaceScalar(face_num,natural,nvalues) ;

    delete nvalues ;
    return(Py_BuildValue("d",value)) ;
}

static PyObject *InterpolateVolume(ElementObject *self,PyObject *args)
{
    PyObject *natural_obj,*node_vals_obj ;

    if (!PyArg_ParseTuple(args,"OO",&natural_obj,&node_vals_obj))
        return(NULL) ;

    Vec3D natural ;
    if (!ConvertVec3D(self,natural_obj,natural)) return(NULL) ;

    // check to see if we have scalars or stresses

    if (PySequence_Check(node_vals_obj) == 0) {
        PyErr_SetString(PyExc_TypeError,
            "2nd argument should a sequence") ;
        return(0) ;
    }

    PyObject *item = PySequence_GetItem(node_vals_obj,0) ;
    if (PyInt_Check(item) || PyFloat_Check(item)) {
         double *node_vals = ConvertFloatSeq(node_vals_obj,
                    self->element->NumVolumeShapeFunc()) ;
         if (node_vals == 0) return(NULL) ;

         double value = self->element->
             InterpolateVolume(natural,node_vals) ;

         delete node_vals ;
         Py_DECREF(item) ;
         return(Py_BuildValue("d",value)) ;
    } else {
         TVector<Vec3D> *node_vals = ConvertVec3DSeq(self,node_vals_obj,
                    self->element->NumVolumeShapeFunc()) ;
         if (node_vals == 0) return(NULL) ;

         Vec3D value = self->element->
             InterpolateVolume(natural,*node_vals) ;

         delete node_vals ;
         Py_DECREF(item) ;
         return(Vec3DToPyVec3D(self,value)) ;
    }
    return(0) ;
}


static PyObject *InternalPointForces(ElementObject *self,PyObject *args)
{
    PyObject *nat_coord_obj,*values_obj ;

    if (!PyArg_ParseTuple(args,"OO",&nat_coord_obj,&values_obj))
        return(NULL) ;

    if ((PySequence_Check(nat_coord_obj) == 0) ||
        (PySequence_Check(values_obj) == 0)) {
        PyErr_SetString(PyExc_TypeError,
            "arguments be sequences") ;
        return(0) ;
    }
    int num = PySequence_Size(nat_coord_obj) ;
    if (num != PySequence_Size(values_obj)) {
        PyErr_SetString(PyExc_TypeError,
            "argument sequences should be the same size") ;
        return(0) ;
    }
        
    TVector<Vec3D> *nat_coords = ConvertVec3DSeq(self,nat_coord_obj,num) ;
    if (nat_coords == 0) return(NULL) ;

    TVector<Vec3D> *values = ConvertVec3DSeq(self,values_obj,num) ;
    if (values == 0) {
        delete nat_coords ;
        return(NULL) ;
    }

    TVector<Vec3D> Fvect = self->element->
         InternalPointForces(*nat_coords,*values) ;
    return(TVectorToPyArray(Fvect)) ;
}

static PyObject *NumVolumeShapeFunc(ElementObject *self,PyObject *args)
{
    if (!PyArg_ParseTuple(args,"")) return(NULL) ;

    return(Py_BuildValue("i",self->element->NumVolumeShapeFunc())) ;
}

static PyObject *NumFaceNodes(ElementObject *self,PyObject *args)
{
    int face ;
    if (!PyArg_ParseTuple(args,"i",&face)) return(NULL) ;

    return(Py_BuildValue("i",self->element->NumFaceNodes(face))) ;
}

static PyObject *ScalarFieldDerivMtx(ElementObject *self,PyObject *args)
{
    PyObject *coord_obj,*mat_obj ;
    int order = 0 ;

    if (!PyArg_ParseTuple(args,"OO|i",&coord_obj,&mat_obj,&order))
        return(NULL) ;

    if (order == 0) order =  self->element->GetDerivDefaultPts() ;

    if (PyArray_Check(mat_obj)) {

        PyArrayObject *matarray = (PyArrayObject *)mat_obj ;
        if ((matarray->dimensions[0] != 3) ||
            (matarray->dimensions[1] != 3) ||
            (matarray->descr->type_num != PyArray_DOUBLE)) {
            PyErr_SetString(PyExc_TypeError,
            "2nd argument should be a 3x3 PyArray or a sequence of these") ;
            return(NULL) ;
        }

        TMatrix<double> mat(3,3) ;
        for (int i=0 ; i<3 ; ++i) {
            for (int j=0 ; j<3 ; ++j) {
                char *eptr = matarray->data + i*matarray->strides[0] +
                             j*matarray->strides[1] ;
                mat(i,j) = *((double *)eptr) ;
            }
        }

        TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                         self->element->NumVolumeShapeFunc()) ;
        if (coords == 0) return(NULL) ;

        TMatrix<double> k = self->element->ScalarFieldDerivMtx(
                                    *coords,mat,order) ;
        delete coords ;
        return(TMatrixToPyArray(k)) ;

    } else if (PySequence_Check(mat_obj)) {

        if (PySequence_Size(mat_obj) != order) {
            PyErr_SetString(PyExc_TypeError,
            "Number of material prop values differs from the # of integration points") ;
            return(NULL) ;
        }

        TMatrix<double> *mat_props = new TMatrix<double>[order] ;
        for (int i=0 ; i < order ; ++i) {
            PyObject *item = PySequence_GetItem(mat_obj,i) ;
            PyArrayObject *matarray = (PyArrayObject *)item ;
            if ((matarray->dimensions[0] != 3) ||
                (matarray->dimensions[1] != 3) ||
                (matarray->descr->type_num != PyArray_DOUBLE)) {
                PyErr_SetString(PyExc_TypeError,
                "2nd argument should be a 3x3 PyArray or a sequence of these") ;
                delete [] mat_props ;
                Py_DECREF(item) ;
                return(NULL) ;
            }
            mat_props[i].Configure(3,3) ;
            for (int k=0 ; k<3 ; ++k) {
                for (int j=0 ; j<3 ; ++j) {
                    char *eptr = matarray->data + 
                                 k*matarray->strides[0] +
                                 j*matarray->strides[1] ;
                    mat_props[i](k,j) = *((double *)eptr) ;
                }
            }
            Py_DECREF(item) ;
        }

        TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                         self->element->NumVolumeShapeFunc()) ;
        if (coords == 0) return(NULL) ;

        TMatrix<double> k = self->element->ScalarFieldDerivMtx(
                                *coords,mat_props,order) ;
        delete coords ;
        return(TMatrixToPyArray(k)) ;

    }

    PyErr_SetString(PyExc_TypeError,
        "2nd argument should be a 3x3 PyArray or a sequence of these") ;
    return(NULL) ;
}

static PyObject *ScalarFieldShapeMtx(ElementObject *self,PyObject *args)
{
    PyObject *coord_obj,*mat_obj ;
    int order = 0 ;

    if (!PyArg_ParseTuple(args,"OO|i",&coord_obj,&mat_obj,&order))
        return(NULL) ;

    if (order == 0) order =  self->element->GetShapeDefaultPts() ;

    // check to see if we have scalar material value for all gauss
    // points or if we have a sequence of points

    if (PyInt_Check(mat_obj) || PyFloat_Check(mat_obj)) {

        PyObject *mat_float = PyNumber_Float(mat_obj) ;
        double matprop = PyFloat_AsDouble(mat_float) ;
        Py_DECREF(mat_float) ;

        TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                     self->element->NumVolumeShapeFunc()) ;
        if (coords == 0) return(NULL) ;

        TMatrix<double> c = self->element->ScalarFieldShapeMtx(
                                *coords,matprop,order) ;
        delete coords ;
        return(TMatrixToPyArray(c)) ;

    } else if (PySequence_Check(mat_obj)) {

        if (PySequence_Size(mat_obj) != order) {
            PyErr_SetString(PyExc_TypeError,
            "Number of material prop values differs from the # of integration points") ;
            return(NULL) ;
        }

        double *mat_props = new double[order] ;
        for (int i=0 ; i < order ; ++i) {
            PyObject *item = PySequence_GetItem(mat_obj,i) ;
            if (PyInt_Check(item)) {
                mat_props[i] = PyInt_AsLong(item) ;
            } else if (PyFloat_Check(item)) {
                mat_props[i] = PyFloat_AsDouble(item) ;
            } else {
                PyErr_SetString(PyExc_TypeError,
                    "Material prop must be a number") ;
                delete [] mat_props ;
                Py_DECREF(item) ;
                return(NULL) ;
            }
            Py_DECREF(item) ;
        }

        TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                     self->element->NumVolumeShapeFunc()) ;
        if (coords == 0) return(NULL) ;

        TMatrix<double> c = self->element->ScalarFieldShapeMtx(
                                *coords,mat_props,order) ;
        delete coords ;
        return(TMatrixToPyArray(c)) ;

    }

    PyErr_SetString(PyExc_TypeError,
          "2nd argument should be a material prop or a sequence of props") ;
    return(NULL) ;
}


static PyObject *Strains(ElementObject *self,PyObject *args)
{
    PyObject *coord_obj,*disp_obj ;
    int order = 0 ;

    if (!PyArg_ParseTuple(args,"OO|i",&coord_obj,&disp_obj,&order))
        return(NULL) ;

    if (order == 0) order =  self->element->GetDerivDefaultPts() ;

    TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                         self->element->NumVolumeShapeFunc()) ;
    if (coords == 0) return(NULL) ;

    TVector<Vec3D> *disp = ConvertVec3DSeq(self,disp_obj,
                         self->element->NumVolumeShapeFunc()) ;
    if (disp == 0) {
        delete coords ;
        return(NULL) ;
    }

    TVector<ColTensor> s = self->element->Strains(
                                    *coords,*disp,order) ;
    delete coords ;
    delete disp ;
    return(ColTensorListToPyColTensorList(self,s)) ;
}



static PyObject *Volume(ElementObject *self,PyObject *args)
{
    PyObject *coord_obj ;
    int order = 0 ;

    if (!PyArg_ParseTuple(args,"O|i",&coord_obj,&order))
        return(NULL) ;

    if (order == 0) order =  self->element->GetDerivDefaultPts() ;

    TVector<Vec3D> *coords = ConvertVec3DSeq(self,coord_obj,
                     self->element->NumVolumeShapeFunc()) ;
    if (coords == 0) return(NULL) ;

    double volume = self->element->Volume(*coords,order) ;
    delete coords ;
    return(Py_BuildValue("d",volume)) ;
}

static PyMethodDef ElementMethods[] = {
    { "Area",                        (PyCFunction)Area,                    METH_VARARGS },
    { "CheckForValidElement",        (PyCFunction)CheckForValidElement,    METH_VARARGS },
    { "DeformationStiffMtx",         (PyCFunction)DeformationStiffMtx,     METH_VARARGS },
    { "FindNaturalCoords",           (PyCFunction)FindNaturalCoords,       METH_VARARGS },
    { "GetDerivDefaultPts",          (PyCFunction)GetDerivDefaultPts,      METH_VARARGS },
    { "GetShapeDefaultPts",          (PyCFunction)GetShapeDefaultPts,      METH_VARARGS },
    { "GetSurfDefaultPts",           (PyCFunction)GetSurfDefaultPts,       METH_VARARGS },
    { "GetVolumeIntegrationPts",     (PyCFunction)GetVolumeIntegrationPts, METH_VARARGS },
    { "InitialStressStiffMtx",       (PyCFunction)InitialStressStiffMtx,   METH_VARARGS },
    { "IntegrateNormalSurfaceScalar",(PyCFunction)IntegrateNormalSurfaceScalar,  METH_VARARGS },
    { "IntegrateSurfaceScalar",      (PyCFunction)IntegrateSurfaceScalar,  METH_VARARGS },
    { "IntegrateSurfaceVector",      (PyCFunction)IntegrateSurfaceVector,  METH_VARARGS },
    { "IntegrateVolume",             (PyCFunction)IntegrateVolume,         METH_VARARGS },
    { "InterpolateSurfaceScalar",    (PyCFunction)InterpolateSurfaceScalar,METH_VARARGS },
    { "InterpolateVolume",           (PyCFunction)InterpolateVolume,       METH_VARARGS },
    { "InternalPointForces",         (PyCFunction)InternalPointForces,     METH_VARARGS },
    { "NumVolumeShapeFunc",          (PyCFunction)NumVolumeShapeFunc,      METH_VARARGS },
    { "NumFaceNodes",                (PyCFunction)NumFaceNodes,            METH_VARARGS },
    { "ScalarFieldDerivMtx",         (PyCFunction)ScalarFieldDerivMtx,     METH_VARARGS },
    { "ScalarFieldShapeMtx",         (PyCFunction)ScalarFieldShapeMtx,     METH_VARARGS },
    { "Strains",                     (PyCFunction)Strains,                 METH_VARARGS },
    { "Volume",                      (PyCFunction)Volume,                  METH_VARARGS },
    { (char *)NULL, (PyCFunction)NULL }
} ;


static ElementObject *newElementObject(PyObject *args)
{

    PyObject *shape_obj ;
    char *shape_name ;

    if (!PyArg_ParseTuple(args,"O",&shape_obj)) return(NULL) ;

    ElementObject *self = PyObject_NEW(ElementObject,&element_type) ;
    if (self == NULL) return NULL ;

    if (PyString_Check(shape_obj)) {
        shape_name = PyString_AsString(shape_obj) ;
    } else {
        PyErr_SetString(PyExc_TypeError,"Argument should be a string") ;
        return NULL ;
    }

    if (!FindExternalClasses(self)) return(NULL) ;

    self->element = ElementFactory::Instance()->
        CreateElement(shape_name) ;

    return(self) ;
}


static void ElementDealloc(PyObject *self)
{
    PyObject_Del(self) ;
}

static int ElementPrint(PyObject *self,FILE *fp,int flags)
{
    ElementObject *obj = (ElementObject *)self ;
    fprintf(fp,"[ Element( %s ) ]",
                obj->element->ShapeFuncName()) ;
    return 0 ;
}

static PyObject *ElementGetattr(PyObject *self,char *name)
{
    return(Py_FindMethod(ElementMethods,self,name)) ;
}

static PyObject *ElementRepr(PyObject *self)
{
    ElementObject *obj = (ElementObject *)self ;
    char buff[128] ;

    sprintf(buff,"[ Element ( %s ) ]",
                obj->element->ShapeFuncName()) ;
    return(PyString_FromString(buff)) ;
}


#if BAD_STATIC_FORWARD
PyTypeObject element_type = {
#else
statichere PyTypeObject element_type = {
#endif

    /* type header */

    PyObject_HEAD_INIT(&PyType_Type)
    0,
    "Element",                    /* tp_name */
    sizeof(ElementObject),        /* tp_basicsize */
    0,                            /* tp_itemsize */

    /* standard methods */

    (destructor)  ElementDealloc,    /* tp_dealloc ref-count==0 */
    (printfunc)   ElementPrint,      /* tp_print   "print x" */
    (getattrfunc) ElementGetattr,    /* tp_getattr "x.attr" */
    (setattrfunc) 0,                 /* tp_setattr "x.attr=v" */
    (cmpfunc)     0,                 /* tp_compare "x > y" */
    (reprfunc)    ElementRepr,       /* tp_repr    'x', print x */

    /* type categories */

    0,                   /* tp_as_number   +,-,*,/,%,&,>>,pow... */
    0,                   /* tp_as_sequence +,[i],[i:j],len,... */
    0,                   /* ty_as_mapping  [key],len,... */

    /* more methods */

    (hashfunc)     0,                   /* tp_hash   "dict[x]" */
    (ternaryfunc)  0,                   /* tp_call   "x()" */
    (reprfunc)     0,                   /* tp_str    "str(x)" */
} ;

/* ==========================================================
** ==========================================================
** ==========================================================
**  Module methods and lookup table
*/

static PyObject *ElementType_new(PyObject *self, PyObject *args)
{
    return((PyObject *)newElementObject(args)) ;
}

static PyMethodDef FemLibTypeMethods[] = {
    { "Element",     ElementType_new,     METH_VARARGS },
    { NULL, NULL }
} ;

#include <stdio.h>

void initFemLib()
{
    Py_InitModule("FemLib",FemLibTypeMethods) ;
    import_array() ;
}

