//
// Integration points header file
//
// Revision -
//   $Revision: 1.2 $  $Date: 2001/12/19 02:47:16 $  $Author: wash $
//
 
#ifndef IntegrationPts_h
#define IntegrationPts_h

#include "Vec3D.hh"
#include "TVector.hh"

namespace FemLib {

// This is the abstract base class.  This should be specialized
// for each element shape (e.g., brick, tetrahedral, wedge, pyramid).
// The data for surfaces (triangles and quadrilaterals) are done
// in this class since these are common to all shapes.

class IntegrationPts {
    public:
        enum SurfaceShape {TRIANGLE, QUADRILATERAL} ;
 
        virtual const TVector<Vec3D>  &GetVolumeCoords(int number) const = 0 ;

        // returns the Gauss point coordinates for volume integrations

        virtual const TVector<double> &GetVolumeWeights(int number) const = 0 ;

        // returns the Gauss weights for volume integrations

        const TVector<Vec3D>  &GetSurfaceCoords(SurfaceShape shape,
                                                int number) const ;

        // returns the Gauss point coordinates for surface integrations

        const TVector<double> &GetSurfaceWeights(SurfaceShape shape,
                                                 int number) const ;

        // returns the Gauss weights for surface integrations


    protected:
        IntegrationPts() ;
        virtual ~IntegrationPts() {} ;

        TVector<Vec3D>  NullPt ;
        TVector<double> NullWt ;

    private:
        TVector<Vec3D> Tri3Pt ;
        TVector<Vec3D> Tri4Pt ;
        TVector<Vec3D> Tri7Pt ;
        TVector<Vec3D> Quad1Pt ;
        TVector<Vec3D> Quad4Pt ;
        TVector<Vec3D> Quad9Pt ;

        TVector<double> Tri3Wt ;
        TVector<double> Tri4Wt ;
        TVector<double> Tri7Wt ;
        TVector<double> Quad1Wt ;
        TVector<double> Quad4Wt ;
        TVector<double> Quad9Wt ;
} ;
                                                 

// subclass for tetrahedrals,  This is a singleton class so set
// things up so that only one instance is created

class TetIntegrationPts : public IntegrationPts {
    public:
        static TetIntegrationPts *Instance() ;
        virtual ~TetIntegrationPts() {} ;

        const TVector<Vec3D>  &GetVolumeCoords(int number) const ;
        const TVector<double> &GetVolumeWeights(int number) const ;

    protected:
        TetIntegrationPts() ;

    private:
        static TetIntegrationPts *_instance ;

        TVector<Vec3D> Tet1Pt ;
        TVector<Vec3D> Tet4Pt ;
        TVector<Vec3D> Tet5Pt ;
        TVector<Vec3D> Tet11Pt ;
        TVector<Vec3D> Tet15Pt ;
        TVector<Vec3D> Tet24Pt ;

        TVector<double> Tet1Wt ;
        TVector<double> Tet4Wt ;
        TVector<double> Tet5Wt ;
        TVector<double> Tet11Wt ;
        TVector<double> Tet15Wt ;
        TVector<double> Tet24Wt ;
} ;

// subclass for hexahedral,  This is a singleton class so set
// things up so that only one instance is created

class BrickIntegrationPts : public IntegrationPts {
    public:
        static BrickIntegrationPts *Instance() ;
        virtual ~BrickIntegrationPts() {} ;

        const TVector<Vec3D>  &GetVolumeCoords(int number) const ;
        const TVector<double> &GetVolumeWeights(int number) const ;

    protected:
        BrickIntegrationPts() ;

    private:
        static BrickIntegrationPts *_instance ;

        TVector<Vec3D> Brick1Pt ;
        TVector<Vec3D> Brick8Pt ;
        TVector<Vec3D> Brick27Pt ;
        TVector<Vec3D> Brick64Pt ;

        TVector<double> Brick1Wt ;
        TVector<double> Brick8Wt ;
        TVector<double> Brick27Wt ;
        TVector<double> Brick64Wt ;
} ;

// subclass for wedges,  This is a singleton class so set
// things up so that only one instance is created

class WedgeIntegrationPts : public IntegrationPts {
    public:
        static WedgeIntegrationPts *Instance() ;
        virtual ~WedgeIntegrationPts() {} ;

        const TVector<Vec3D>  &GetVolumeCoords(int number) const ;
        const TVector<double> &GetVolumeWeights(int number) const ;

    protected:
        WedgeIntegrationPts() ;

    private:
        static WedgeIntegrationPts *_instance ;

        TVector<Vec3D> Wedge6Pt ;
        TVector<Vec3D> Wedge12Pt ;
        TVector<Vec3D> Wedge28Pt ;
        TVector<Vec3D> Wedge48Pt ;

        TVector<double> Wedge6Wt ;
        TVector<double> Wedge12Wt ;
        TVector<double> Wedge28Wt ;
        TVector<double> Wedge48Wt ;
} ;

} // namespace

#endif
