//
// Material mats implementation file
//
// Revision -
//   $Revision: 1.1.1.1 $  $Date: 2001/12/02 16:53:08 $  $Author: wash $
//
 
#include <cmath>
#include "MaterialMat.hh"

namespace FemLib {

MaterialMat *MaterialMat::_instance = 0 ;

MaterialMat *MaterialMat::Instance()
{
    if (_instance == 0) {
        _instance = new MaterialMat ;
    }
    return _instance ;
}

TMatrix<double> MaterialMat::RotateOrthoMatrix(TMatrix<double> Dmtx,
                                               double angles[3])
{
    double ca[3],sa[3] ;
    for (int i=0 ; i<3 ; ++i) {
        ca[i] = cos(angles[i]) ;
        sa[i] = sin(angles[i]) ;
    }
          
    // build a 3x3 transformation matrix

    double t3[3][3] ;
    t3[0][0] = ca[0]*ca[2] - sa[0]*sa[1]*sa[2] ;
    t3[0][1] = -sa[0]*ca[1] ;
    t3[0][2] = ca[0]*sa[2] + sa[0]*sa[1]*ca[2] ;
    t3[1][0] = sa[0]*ca[2] + ca[0]*sa[1]*sa[2] ;
    t3[1][1] = ca[0]*ca[1] ;
    t3[1][2] = sa[0]*sa[2] - ca[0]*sa[1]*ca[2] ;
    t3[2][0] = -ca[1]*sa[2] ;
    t3[2][1] = sa[1] ;
    t3[2][2] = ca[1]*ca[2] ;

    // now the 6x6 transformation matrix

    TMatrix<double> tran(6,6) ;
    for (int i=0 ; i<3 ; ++i) {
        int l = (i+1) % 3 ;
        for (int j=0 ; j<3 ; ++j) {
            int k = (j+1) % 3 ;
            tran(i,j) = t3[i][j]*t3[i][j] ;
            tran(i,j+3) = t3[i][j]*t3[i][k] ; 
            tran(i+3,j) = 2.0*t3[i][j]*t3[l][j] ; 
            tran(i+3,j+3) = t3[i][j]*t3[l][k] + t3[l][j]*t3[i][k] ;
        }
    }

    TMatrix<double> TDmtx(6,6) ;
    TDmtx.TripleProd(tran,Dmtx,1.0) ;
    return(TDmtx) ;
}

TMatrix<double> MaterialMat::IsotropicMatMatrix(double E,double nu)
{
    TMatrix<double> Dmtx(6,6) ;
    double D11 = (E * (1.0 - nu))/ ((1.0 + nu) * (1.0 - 2.0*nu)) ;
    double D12 = D11 * (nu/(1.0-nu)) ;
    double D33 = E / (2 * (1 + nu)) ;

    Dmtx(0,0) = D11 ;
    Dmtx(1,1) = D11 ;
    Dmtx(2,2) = D11 ;
    Dmtx(0,1) = D12 ;
    Dmtx(0,2) = D12 ;
    Dmtx(1,0) = D12 ;
    Dmtx(1,2) = D12 ;
    Dmtx(2,0) = D12 ;
    Dmtx(2,1) = D12 ;
    Dmtx(3,3) = D33 ;
    Dmtx(4,4) = D33 ;
    Dmtx(5,5) = D33 ;
    return(Dmtx) ;
}

TMatrix<double> MaterialMat::OrthotropicMatMatrixFromEngngConstants(
                    double E1,double E2,double E3,
                    double nu12,double nu23,double nu13,
                    double G12,double G23,double G31,
                    double angles[3])
{
    double nu21 = nu12 * (E2/E1) ;
    double nu31 = nu13 * (E3/E1) ;
    double nu32 = nu23 * (E3/E2) ;

    double delta = (1.0 - nu12*nu21 - nu23*nu32 - nu13*nu31 -
                   2.0*nu21*nu32*nu13) / (E1*E2*E3) ;

    double c11 = (1.0 - nu23*nu32) / (E2*E3*delta) ;
    double c22 = (1.0 - nu13*nu31) / (E1*E3*delta) ;
    double c33 = (1.0 - nu12*nu21) / (E1*E2*delta) ;

    double c12 = (nu12 + nu32*nu13) / (E1*E3*delta) ;
    double c23 = (nu23 + nu21*nu13) / (E1*E2*delta) ;
    double c13 = (nu13 + nu12*nu23) / (E1*E2*delta) ;

    double c44 = G23 ;
    double c55 = G31 ;
    double c66 = G12 ;

    TMatrix<double> Dmtx(6,6) ;
    Dmtx(0,0) = c11 ;
    Dmtx(1,1) = c22 ;
    Dmtx(2,2) = c33 ;
    Dmtx(0,1) = c12 ;
    Dmtx(0,2) = c13 ;
    Dmtx(1,0) = c12 ;
    Dmtx(1,2) = c23 ;
    Dmtx(2,0) = c13 ;
    Dmtx(2,1) = c23 ;
    Dmtx(3,3) = c44 ;
    Dmtx(4,4) = c55 ;
    Dmtx(5,5) = c66 ;
    return (RotateOrthoMatrix(Dmtx,angles)) ;
}


TMatrix<double> MaterialMat::OrthotropicMatMatrixFromStiffnessComps(
                    double c[9],double angles[3])
{
    TMatrix<double> Dmtx(6,6) ;
    Dmtx(0,0) = c[0] ;
    Dmtx(1,1) = c[1] ;
    Dmtx(2,2) = c[2] ;
    Dmtx(0,1) = c[3] ;
    Dmtx(0,2) = c[5] ;
    Dmtx(1,0) = c[3] ;
    Dmtx(1,2) = c[4] ;
    Dmtx(2,0) = c[5] ;
    Dmtx(2,1) = c[4] ;
    Dmtx(3,3) = c[6] ;
    Dmtx(4,4) = c[7] ;
    Dmtx(5,5) = c[8] ;
    return (RotateOrthoMatrix(Dmtx,angles)) ;
}


TMatrix<double> MaterialMat::OrthotropicMatMatrixFromComplianceComps(
                    double s[9],double angles[3])
{
    TMatrix<double> Dmtx(6,6) ;
    double Sd = s[0]*s[1]*s[2] - s[0]*s[4]*s[4] - s[1]*s[5]*s[5] -
                s[2]*s[3]*s[3] + 2.0*s[3]*s[4]*s[5] ;
    Dmtx(0,0) = (s[1]*s[2] - s[4]*s[4]) / Sd ;
    Dmtx(1,1) = (s[2]*s[0] - s[5]*s[5]) / Sd ;
    Dmtx(2,2) = (s[0]*s[1] - s[3]*s[3]) / Sd ;
    Dmtx(0,1) = (s[5]*s[4] - s[3]*s[2]) / Sd ;
    Dmtx(0,2) = (s[3]*s[4] - s[5]*s[1]) / Sd ;
    Dmtx(1,0) = Dmtx(0,1) ;
    Dmtx(1,2) = (s[3]*s[5] - s[4]*s[0]) / Sd ;
    Dmtx(2,0) = Dmtx(0,2) ;
    Dmtx(2,1) = Dmtx(1,2) ;
    Dmtx(3,3) = 1.0/s[6] ;
    Dmtx(4,4) = 1.0/s[7] ;
    Dmtx(5,5) = 1.0/s[8] ;
    return (RotateOrthoMatrix(Dmtx,angles)) ;
}

// -----------------------------------------------------------------

ScalarMaterialMat *ScalarMaterialMat::_instance = 0 ;

ScalarMaterialMat *ScalarMaterialMat::Instance()
{
    if (_instance == 0) {
        _instance = new ScalarMaterialMat ;
    }
    return _instance ;
}

TMatrix<double> ScalarMaterialMat::RotateOrthoMatrix(
                             TMatrix<double> Dmtx,
                             double angles[3])
{
    double ca[3],sa[3] ;
    for (int i=0 ; i<3 ; ++i) {
        ca[i] = cos(angles[i]) ;
        sa[i] = sin(angles[i]) ;
    }
          
    // build a 3x3 transformation matrix

    TMatrix<double> tran(3,3) ;
    tran(0,0) = ca[0]*ca[2] - sa[0]*sa[1]*sa[2] ;
    tran(0,1) = -sa[0]*ca[1] ;
    tran(0,2) = ca[0]*sa[2] + sa[0]*sa[1]*ca[2] ;
    tran(1,0) = sa[0]*ca[2] + ca[0]*sa[1]*sa[2] ;
    tran(1,1) = ca[0]*ca[1] ;
    tran(1,2) = sa[0]*sa[2] - ca[0]*sa[1]*ca[2] ;
    tran(2,0) = -ca[1]*sa[2] ;
    tran(2,1) = sa[1] ;
    tran(2,2) = ca[1]*ca[2] ;

    TMatrix<double> TDmtx(3,3) ;
    TDmtx.TripleProd(tran,Dmtx,1.0) ;
    return(TDmtx) ;
}

TMatrix<double> ScalarMaterialMat::IsotropicMatMatrix(double k)
{
    TMatrix<double> Dmtx(3,3) ;

    Dmtx(0,0) = k ;
    Dmtx(1,1) = k ;
    Dmtx(2,2) = k ;
    return(Dmtx) ;
}

TMatrix<double> ScalarMaterialMat::OrthotropicMatMatrix(
                    double k1,double k2,double k3,
                    double angles[3])
{
    TMatrix<double> Dmtx(3,3) ;
    Dmtx(0,0) = k1 ;
    Dmtx(1,1) = k2 ;
    Dmtx(2,2) = k3 ;
    return (RotateOrthoMatrix(Dmtx,angles)) ;
}

} // namespace
