
import Numeric
import math

# This module provides routines to formulate finite element
# constitutive matricies

def _RotateOrthoMatrix(mat,props):
    """Rotate a constitutive matrix

    If 'angle_x', 'angle_y', or 'angle_z' are defined in the
    properties matrix then the constitutive matrix is rotated
    by these angles in the order defined in the documentation
    for the constitutive routines below.

    returns a rotated constitutive matrix"""

    angle_x = props.get('angle_x')
    angle_y = props.get('angle_y')
    angle_z = props.get('angle_z')

    if angle_x == None and angle_y == None and angle_z == None:
        return mat

    if angle_x == None: angle_x = 0.0
    if angle_y == None: angle_y = 0.0
    if angle_z == None: angle_z = 0.0

    # compute the sines and cosines of the angles

    ca = [math.cos(angle_z*math.pi/180.0),
          math.cos(angle_x*math.pi/180.0),
          math.cos(angle_y*math.pi/180.0)]
    sa = [math.sin(angle_z*math.pi/180.0),
          math.sin(angle_x*math.pi/180.0),
          math.sin(angle_y*math.pi/180.0)]
          
    # build a 3x3 transformation matrix

    t3 = [[ca[0]*ca[2] - sa[0]*sa[1]*sa[2],-sa[0]*ca[1],
           ca[0]*sa[2] + sa[0]*sa[1]*ca[2]],
          [sa[0]*ca[2] + ca[0]*sa[1]*sa[2],ca[0]*ca[1],
           sa[0]*sa[2] - ca[0]*sa[1]*ca[2]],
          [-ca[1]*sa[2],sa[1],ca[1]*ca[2]]]

    #print t3

    # now the 6x6 transformation matrix

    tran = Numeric.zeros([6,6],Numeric.Float64)

    for i in range(3):
        l = (i+1) % 3
        for j in range(3):
            k = (j+1) % 3
            tran[i,j] = t3[i][j]**2
            tran[i,j+3] = t3[i][j]*t3[i][k] 
            tran[i+3,j] = 2.0*t3[i][j]*t3[l][j] 
            tran[i+3,j+3] = t3[i][j]*t3[l][k] + t3[l][j]*t3[i][k]

    #print tran

    trant = Numeric.transpose(tran)
    return Numeric.dot(trant,Numeric.dot(mat,tran))

def _RotateOrthoMatrix3x3(mat,props):
    """Rotate a scalar (3x3) material matrix

    If 'angle_x', 'angle_y', or 'angle_z' are defined in the
    properties matrix then the constitutive matrix is rotated
    by these angles in the order defined in the documentation
    for the constitutive routines below.

    returns a rotated constitutive matrix"""

    angle_x = props.get('angle_x')
    angle_y = props.get('angle_y')
    angle_z = props.get('angle_z')

    if angle_x == None and angle_y == None and angle_z == None:
        return mat

    if angle_x == None: angle_x = 0.0
    if angle_y == None: angle_y = 0.0
    if angle_z == None: angle_z = 0.0

    # compute the sines and cosines of the angles

    ca = [math.cos(angle_z*math.pi/180.0),
          math.cos(angle_x*math.pi/180.0),
          math.cos(angle_y*math.pi/180.0)]
    sa = [math.sin(angle_z*math.pi/180.0),
          math.sin(angle_x*math.pi/180.0),
          math.sin(angle_y*math.pi/180.0)]
          
    # build a 3x3 transformation matrix

    tran = Numeric.zeros([3,3],Numeric.Float64)
    tran[0,0] = ca[0]*ca[2] - sa[0]*sa[1]*sa[2]
    tran[0,1] = -sa[0]*ca[1]
    tran[0,2] = ca[0]*sa[2] + sa[0]*sa[1]*ca[2]
    tran[1,0] = sa[0]*ca[2] + ca[0]*sa[1]*sa[2]
    tran[1,1] = ca[0]*ca[1]
    tran[1,2] = sa[0]*sa[2] - ca[0]*sa[1]*ca[2]
    tran[2,0] = -ca[1]*sa[2]
    tran[2,1] = sa[1]
    tran[2,2] = ca[1]*ca[2]

    print tran

    trant = Numeric.transpose(tran)
    return Numeric.dot(trant,Numeric.dot(mat,tran))


def IsotropicMatMatrix(props):
    """Formulate an isotropic constitutive matrix

    props is a dictionary of material properties.  This routine
    assumes that values are defined for 'E' and 'nu', the
    Youngs modulus and Poisson ratio, respectively.

    returns a 6x6 Numpy constitutive matrix"""
   
    E = props.get('E')
    if E == None:
        raise KeyError, "'E' must be defined for isotropic properties"
    nu = props.get('nu')
    if E == None:
        raise KeyError, "'nu' must be defined for isotropic properties"

    D11 = (E * (1.0 - nu))/ ((1.0 + nu) * (1.0 - 2.0*nu))
    D12 = D11 * (nu/(1.0-nu))
    D33 = E / (2 * (1 + nu))

    mat = Numeric.zeros([6,6],Numeric.Float64)
    mat[0,0] = D11
    mat[1,1] = D11
    mat[2,2] = D11

    mat[0,1] = D12
    mat[0,2] = D12
    mat[1,0] = D12
    mat[1,2] = D12
    mat[2,0] = D12
    mat[2,1] = D12

    mat[3,3] = D33
    mat[4,4] = D33
    mat[5,5] = D33
    return mat


def ScalarIsotropicMatMatrix(props):
    """Formulate an isotropic scalar (3x3) material matrix

    props is a dictionary of material properties.  This routine
    assumes that a value is defined for 'k', the scalar material
    property (e.g. conductivity).

    returns a 3x3 Numpy constitutive matrix"""

    k = props.get('k')
    if k == None:
        raise KeyError, "'k' must be defined for isotropic properties"
   
    mat = Numeric.zeros([3,3],Numeric.Float64)
    mat[0,0] = k
    mat[1,1] = k
    mat[2,2] = k
    return mat


def OrthotropicMatMatrixFromEngngConstants(props):
    """Formulate an orthotropic constitutive matrix from engineering constants

    props is a dictionary of material properties.  This routine
    assumes that values are defined for 'E1', 'E2', 'E3',
    'nu12', 'nu23', 'nu13', 'G12', 'G23', and 'G31'.  Optionally,
    three rotations can be specified, 'angle_z', 'angle_x', and
    'angle_y'.  These angles define the rotation of the material
    axes from the Cartesion axes.  The rotation is done in this order:

    angle_z - first rotation about the z axis (positive x axis to y)
    angle_x - second rotation about the (intermediate local) y axis
              (positive z axis to x)
    angle_y - third rotation about (intermediate local) y axis
              (positive z axis to x)

    Angles are given in DEGREES.

    returns a 6x6 Numpy constitutive matrix"""

    E1 = props.get('E1')
    E2 = props.get('E2')
    E3 = props.get('E3')
    if E1 == None or E2 == None or E3 == None:
        raise KeyError, "'E1', 'E2', and 'E3' must be defined for orthotropic properties"

    nu12 = props.get('nu12')
    nu23 = props.get('nu23')
    nu13 = props.get('nu13')
    if nu12 == None or nu23 == None or nu13 ==None:
        raise KeyError, "'nu12', 'nu23', and 'nu13' must be defined for orthotropic properties"

    G12 = props.get('G12')
    G23 = props.get('G23')
    G31 = props.get('G31')
    if G12 == None or G23 == None or G31 ==None:
        raise KeyError, "'G12', 'G23', and 'G31' must be defined for orthotropic properties"

    # compute the complimentary nu's

    nu21 = nu12 * (E2/E1)
    nu31 = nu13 * (E3/E1)
    nu32 = nu23 * (E3/E2)

    # compute the 'stiffness' values

    delta = (1.0 - nu12*nu21 - nu23*nu32 - nu13*nu31 -
             2.0*nu21*nu32*nu13) / (E1*E2*E3)

    c11 = (1.0 - nu23*nu32) / (E2*E3*delta)
    c22 = (1.0 - nu13*nu31) / (E1*E3*delta)
    c33 = (1.0 - nu12*nu21) / (E1*E2*delta)

    c12 = (nu12 + nu32*nu13) / (E1*E3*delta)
    c23 = (nu23 + nu21*nu13) / (E1*E2*delta)
    c13 = (nu13 + nu12*nu23) / (E1*E2*delta)

    c44 = G23
    c55 = G31
    c66 = G12

    # now the material matrix

    mat = Numeric.zeros([6,6],Numeric.Float64)
    mat[0,0] = c11
    mat[1,1] = c22
    mat[2,2] = c33

    mat[0,1] = c12
    mat[0,2] = c13
    mat[1,0] = mat[0,1]
    mat[1,2] = c23
    mat[2,0] = mat[0,2]
    mat[2,1] = mat[1,2]

    mat[3,3] = c44
    mat[4,4] = c55
    mat[5,5] = c66
    return _RotateOrthoMatrix(mat,props)


def OrthotropicMatMatrixFromStiffnessComps(props):
    """Formulate an orthotropic constitutive matrix from 'stiffness' components

    props is a dictionary of material properties.  This routine
    assumes that values are defined for 'c11', 'c22', 'c33',
    'c12', 'c23', 'c13', 'c44', 'c55', and 'c66'.  Note that
    following conventions is observed:

    c44 => xy,  c55 => yz,  c66 => zx

    Optionally, three rotations can be specified, 'angle_z', 'angle_x', and
    'angle_y'.  These angles define the rotation of the material
    axes from the Cartesion axes.  The rotation is done in this order:

    angle_z - first rotation about the z axis (positive x axis to y)
    angle_x - second rotation about the (intermediate local) y axis
              (positive z axis to x)
    angle_y - third rotation about (intermediate local) y axis
              (positive z axis to x)

    Angles are given in DEGREES.

    returns a 6x6 Numpy constitutive matrix"""

    c11 = props.get('c11')
    c22 = props.get('c22')
    c33 = props.get('c33')
    c12 = props.get('c12')
    c23 = props.get('c23')
    c13 = props.get('c13')
    c44 = props.get('c44')
    c55 = props.get('c55')
    c66 = props.get('c66')
    if (c11 == None or c22 == None or c33 == None or
        c12 == None or c23 == None or c13 == None or
        c44 == None or c55 == None or c66 == None):
        raise KeyError, "'c11, c22, c33, c12, c23, c13, c44, c55, and c66 must all be defined"

    mat = Numeric.zeros([6,6],Numeric.Float64)
    mat[0,0] = c11
    mat[1,1] = c22
    mat[2,2] = c33

    mat[0,1] = c12
    mat[0,2] = c13
    mat[1,0] = mat[0,1]
    mat[1,2] = c23
    mat[2,0] = mat[0,2]
    mat[2,1] = mat[1,2]

    mat[3,3] = c44
    mat[4,4] = c55
    mat[5,5] = c66
    return _RotateOrthoMatrix(mat,props)


def OrthotropicMatMatrixFromComplianceComps(props):
    """Formulate an orthotropic constitutive matrix from 'compliance' components

    props is a dictionary of material properties.  This routine
    assumes that values are defined for 's11', 's22', 's33',
    's12', 's23', 's13', 's44', 's55', and 's66'.  The value of
    these are defined as follows:

    s11 = 1.0/E1
    s22 = 1.0/E2
    s33 = 1.0/E3
    s21 = s12 = -nu21/E2 = -nu12/E1
    s32 = s23 = -nu32/E3 = -nu23/E2
    s31 = s13 = -nu31/E1 = -nu13/E3
    s44 = 1.0/G12
    s55 = 1.0/G23
    s66 = 1.0/G31

    Optionally, three rotations can be specified, 'angle_z', 'angle_x', and
    'angle_y'.  These angles define the rotation of the material
    axes from the Cartesion axes.  The rotation is done in this order:

    angle_z - first rotation about the z axis (positive x axis to y)
    angle_x - second rotation about the (intermediate local) y axis
              (positive z axis to x)
    angle_y - third rotation about (intermediate local) y axis
              (positive z axis to x)

    Angles are given in DEGREES.

    returns a 6x6 Numpy constitutive matrix"""

    s11 = props.get('s11')
    s22 = props.get('s22')
    s33 = props.get('s33')
    s12 = props.get('s12')
    s23 = props.get('s23')
    s13 = props.get('s13')
    s44 = props.get('s44')
    s55 = props.get('s55')
    s66 = props.get('s66')
    if (s11 == None or s22 == None or s33 == None or
        s12 == None or s23 == None or s13 == None or
        s44 == None or s55 == None or s66 == None):
        raise KeyError, "'s11, s22, s33, s12, s23, s13, s44, s55, and s66 must all be defined"

    mat = Numeric.zeros([6,6],Numeric.Float64)

    Sd = s11*s22*s33 - s11*s23**2 - s22*s13**2 - s33*s12**2 + 2.0*s12*s23*s13

    mat[0,0] = (s22*s33 - s23**2) / Sd
    mat[1,1] = (s33*s11 - s13**2) / Sd
    mat[2,2] = (s11*s22 - s12**2) / Sd

    mat[0,1] = (s13*s23 - s12*s33) / Sd
    mat[1,0] = mat[0,1]
    mat[1,2] = (s12*s13 - s23*s11) / Sd
    mat[2,1] = mat[1,2]
    mat[0,2] = (s12*s23 - s13*s22) / Sd
    mat[2,0] = mat[0,2]

    mat[3,3] = 1.0/s44
    mat[4,4] = 1.0/s55
    mat[5,5] = 1.0/s66
    return _RotateOrthoMatrix(mat,props)


def ScalarOrthotropicMatMatrix(props):
    """Formulate a scalar (3x3) orthotropic material matrix

    props is a dictionary of material properties.  This routine
    assumes that values are defined for 'k1', 'k2', and 'k3'.
    Optionally, three rotations can be specified, 'angle_z', 'angle_x', and
    'angle_y'.  These angles define the rotation of the material
    axes from the Cartesion axes.  The rotation is done in this order:

    angle_z - first rotation about the z axis (positive x axis to y)
    angle_x - second rotation about the (intermediate local) y axis
              (positive z axis to x)
    angle_y - third rotation about (intermediate local) y axis
              (positive z axis to x)

    Angles are given in DEGREES.

    returns a 3x3 Numpy constitutive matrix"""

    k1 = props.get('k1')
    k2 = props.get('k2')
    k3 = props.get('k3')
    if k1 == None or k2 == None or k3 == None:
        raise KeyError, "'k1', 'k2', and 'k3' must be defined for orthotropic properties"

    # now the material matrix

    mat = Numeric.zeros([3,3],Numeric.Float64)
    mat[0,0] = k1
    mat[1,1] = k2
    mat[2,2] = k3
    return _RotateOrthoMatrix3x3(mat,props)


if __name__ == '__main__':

    # test isotropic

    iprops = {'E':10.0,'nu':0.3,'k':1.0}
    mat = IsotropicMatMatrix(iprops)
    print 'Isotropic:'
    print mat
    print

    mat = ScalarIsotropicMatMatrix(iprops)
    print 'Isotropic Scalar:'
    print mat

    # test orthotropic
    oprops = {'E1':10.0,'E2':8.0,'E3':6.0,
              'nu12':0.1,'nu23':0.2,'nu13':0.3,
              'G23':4.0,'G31':5.0,'G12':6.0,
              'k1':1.0,'k2':2.0,'k3':3.0}
    mat = OrthotropicMatMatrixFromEngngConstants(oprops)
    print
    print 'Ortho from conts:'
    print mat

    cprops = {}
    cprops['c11'] = mat[0,0]
    cprops['c22'] = mat[1,1]
    cprops['c33'] = mat[2,2]
    cprops['c12'] = mat[0,1]
    cprops['c23'] = mat[1,2]
    cprops['c13'] = mat[0,2]
    cprops['c44'] = mat[3,3]
    cprops['c55'] = mat[4,4]
    cprops['c66'] = mat[5,5]
    mat = OrthotropicMatMatrixFromStiffnessComps(cprops)
    print
    print "Ortho from c's:"
    print mat

    sprops = {}
    sprops['s11'] = 1.0/oprops['E1']
    sprops['s22'] = 1.0/oprops['E2']
    sprops['s33'] = 1.0/oprops['E3']
    sprops['s12'] = -oprops['nu12']/oprops['E1']
    sprops['s23'] = -oprops['nu23']/oprops['E2']
    sprops['s13'] = -oprops['nu13']/oprops['E1']
    sprops['s44'] = 1.0/oprops['G12']
    sprops['s55'] = 1.0/oprops['G23']
    sprops['s66'] = 1.0/oprops['G31']
    mat = OrthotropicMatMatrixFromComplianceComps(sprops)
    print
    print "Ortho from s's:"
    print
    print sprops
    print
    print mat

    sprops['angle_x'] = 90.0
    mat = OrthotropicMatMatrixFromComplianceComps(sprops)
    print
    print "rotated:"
    print mat
    print

    oprops['angle_z'] = 90.0
    #oprops['angle_x'] = 90.0
    oprops['angle_y'] = 90.0
    mat = ScalarOrthotropicMatMatrix(oprops)
    print
    print "Scalar Orthotropic:"
    print mat

