
#include <cstdio>
//#include <string.h>
#include <cstdlib>

#include "SparseCG.hh"

#define DEFAULT_TOL 1e-10

// superclass code

SparseCG::SparseCG(int num_nodes,int start) : 
    NumNodes(num_nodes),print_start(start)
{
    x = 0 ;
    r = 0 ;
    z = 0 ;
    p = 0 ;
    Ap = 0 ;
    Binv = 0 ;
    b = 0 ;
    NumEqns = NumNodes ;

    Tolerance = DEFAULT_TOL ;
    CoeffsAlloced = false ;
    PrescribedApplied = false ;

    RecCacheList = 0 ;
    RecCachePtr = 0 ;

    FixedCacheList = 0 ;
    FixedCachePtr = 0 ;

    CData = new CoefData ;
    CData->num_rows = NumNodes ;
    CData->rows = new CoefRec*[NumNodes] ;
    CData->A = 0 ;
    for (int i=0 ; i<NumNodes ; ++i) CData->rows[i] = 0 ;
}


SparseCG::~SparseCG()
{
    if (x != 0) delete x ;
    if (b != 0) delete b ;
    if (CData->A != 0) delete CData->A ;
    delete CData->rows ;

    CoefRecCache *cptr = RecCacheList ;
    while (cptr != 0) {
        CoefRecCache *tmp = cptr->next ;
        delete cptr ;
        cptr = tmp ;
    }
}

double *SparseCG::DoSolution()
{
    AllocateSolveData() ;
    if (x == 0) x = new double[NumEqns] ;
    SolveSetup() ;
    if (!PrescribedApplied) ApplyPrescribedValue() ;
    if (CGSolve()) {
        FreeSolveData() ;
        return(x) ;
    }
    FreeSolveData() ;
    return(0) ;
}

double *SparseCG::DoGaussSeidel()
{
    AllocateSolveData() ;
    if (x == 0) x = new double[NumEqns] ;
    SolveSetup() ;
    if (!PrescribedApplied) ApplyPrescribedValue() ;
    if (GaussSeidelSolve()) {
        FreeSolveData() ;
        return(x) ;
    }
    FreeSolveData() ;
    return(0) ;
}

double *SparseCG::MatrixVectorMultiply()
{
    if (!PrescribedApplied) ApplyPrescribedValue() ;
    if (x == 0) x = new double[NumEqns] ;
    MatrixVectMult(b,x) ;
    return(x) ;
}

void SparseCG::AllocateSolveData()
{
    r = new double[NumEqns] ;
    z = new double[NumEqns] ;
    p = new double[NumEqns] ;
    Ap = new double[NumEqns] ;
    Binv = new double[NumEqns] ;
}


void SparseCG::FreeSolveData()
{
    if (r != 0) delete r ;
    if (z != 0) delete z ;
    if (p != 0) delete p ;
    if (Ap != 0) delete Ap ;
    if (Binv != 0) delete Binv ;

    r = 0 ;
    z = 0 ;
    p = 0 ;
    Ap = 0 ;
    Binv = 0 ;
}

SparseCG::CoefRec *SparseCG::NewRec()
{
    if ((RecCacheList == 0) ||
        (RecCachePtr >= REC_CACHE_BLOCK_SIZE)) {
        CoefRecCache *ctmp = new CoefRecCache ;
        ctmp->next = RecCacheList ;
        RecCacheList = ctmp ;
        RecCachePtr = 0 ;
    }
    CoefRec *tmp = &(RecCacheList->entries[RecCachePtr]) ;
    RecCachePtr++ ;
    return(tmp) ;
}

SparseCG::FixedRec *SparseCG::NewFixed()
{
    if ((FixedCacheList == 0) ||
        (FixedCachePtr >= FIXED_CACHE_BLOCK_SIZE)) {
        FixedRecCache *ctmp = new FixedRecCache ;
        ctmp->next = FixedCacheList ;
        FixedCacheList = ctmp ;
        FixedCachePtr = 0 ;
    }
    FixedRec *tmp = &(FixedCacheList->entries[FixedCachePtr]) ;
    FixedCachePtr++ ;
    return(tmp) ;
}

void SparseCG::ElementEquations(int num_nodes,int *nodes)
{
    for (int l=0 ; l<num_nodes ; ++l) {
        for (int j=0 ; j<=l ; ++j) {
            int ri, ci ;
            if (nodes[l] >= nodes[j]) {
                ri = nodes[l] ;
                ci = nodes[j] ;
            } else {
                ri = nodes[j] ;
                ci = nodes[l] ;
            }
            bool insert = true ;
            CoefRec *next, *prev = 0 ;
            for (next = CData->rows[ri] ; next != 0 ; next = next->next ) {
                if (ci == next->column) {
                    insert = false ;
                    break ;
                } else if (ci < next->column) {
                    CoefRec *rec = NewRec() ;
                    rec->column = ci ;
                    rec->offset = -1 ;
                    rec->next = next ;
                    if (prev != 0)
                        prev->next = rec ;
                    else
                        CData->rows[ri] = rec ;
                     insert = false ;
                    break ;
                } else {
                    prev = next ;
                }
            }

            if (insert) {
                CoefRec *rec = NewRec() ;
                rec->column = ci ;
                rec->offset = -1 ;
                rec->next = 0 ;
                if (prev != 0)
                    prev->next = rec ;
                else
                    CData->rows[ri] = rec ;
            }
        }
    }
}

void SparseCG::AllocateCoeffs()
{
    // count the number of 3x3 blocks and allocate space
    // for the coefficients

    int cur = 0 ;
    CoefRec *next ;
    for (int i=0 ; i<CData->num_rows ; ++i) {
        for (next = CData->rows[i] ; next != 0 ; next = next->next ) ++cur ;
    }
    CData->A = new double[cur] ;
    for (int i=0 ; i<cur ; ++i) CData->A[i] = 0.0 ;
    CoeffsAlloced = true ;
    cur_offset = 0 ;

    b = new double[NumNodes] ;
    for (int i=0 ; i<NumNodes ; ++i) b[i] = 0.0 ;
}

void SparseCG::ElementValues(int num_nodes,int *nodes,
                             const TMatrix<double> &stiff)
{
    if (!CoeffsAlloced) AllocateCoeffs() ;

    // go through the equations associated with
    // the element and build up equations

    for (int l=0 ; l<num_nodes ; ++l) {
        for (int k=0 ; k<num_nodes ; ++k) {
            if (nodes[k] > nodes[l]) continue ;

            int ri = nodes[l] ;
            int ci = nodes[k] ;

            CoefRec *next ;
            for (next = CData->rows[ri] ; next != 0 ; next = next->next ) {
                if (ci == next->column) {
                    if (next->offset == -1) {
                        next->offset = cur_offset ;
                        cur_offset ++ ;
                    }
                    CData->A[next->offset] += stiff(l,k) ;
                    break ;
                }
            }
        }
    }
}

void SparseCG::RhsValues(int num_nodes,int *nodes,
                         const double *contrib_rhs)
{
    if (!CoeffsAlloced) AllocateCoeffs() ;
    for (int l=0 ; l<num_nodes ; ++l) {
        b[nodes[l]] += contrib_rhs[l] ;
    }
}

#ifdef PY_NUMERIC
void SparseCG::ElementValues(int num_nodes,int *nodes,
                             const PyArrayObject *stiff)
{
    if (!CoeffsAlloced) AllocateCoeffs() ;

    // go through the equations associated with
    // the element and build up equations

    for (int l=0 ; l<num_nodes ; ++l) {
        for (int k=0 ; k<num_nodes ; ++k) {
            if (nodes[k] > nodes[l]) continue ;

            int ri = nodes[l] ;
            int ci = nodes[k] ;

            CoefRec *next ;
            for (next = CData->rows[ri] ; next != 0 ; next = next->next ) {
                if (ci == next->column) {
                    if (next->offset == -1) {
                        next->offset = cur_offset ;
                        cur_offset++ ;
                    }
                    char *eptr = stiff->data + 
                                 l*stiff->strides[0] +
                                 k*stiff->strides[1] ;
                    CData->A[next->offset] += *((double *)eptr) ;
                    break ;
                }
            }
        }
    }
}

void SparseCG::RhsValues(int num_nodes,int *nodes,
                         const PyArrayObject *contrib_rhs)
{
    if (!CoeffsAlloced) AllocateCoeffs() ;

    for (int l=0 ; l<num_nodes ; ++l) {
        char *eptr = contrib_rhs->data +
                         l*contrib_rhs->strides[0] ;
        b[nodes[l]] += *((double *)eptr) ;
    }
}
#endif

void SparseCG::AddPrescribedValue(int node,double value)
{
    FixedRec *tmp = NewFixed() ;
    tmp->node = node ;
    tmp->flags = 1 ;
    tmp->values[0] = value ;
    PrescribedApplied = false ;
}

void SparseCG::ApplyPrescribedValue()
{
    bool first = true ;
    for (FixedRecCache *ptr = FixedCacheList ; ptr != 0 ; ptr = ptr->next) {
        int stop = (first) ? FixedCachePtr : FIXED_CACHE_BLOCK_SIZE ;
        if (first) first = false ;
        for (int j=0 ; j<stop ; ++j) {
            int node = ptr->entries[j].node ;
            double *values = ptr->entries[j].values ;

            // zero out this row and column

            for (int i=0 ; i<CData->num_rows ; ++i) {

                if (i == node) {
                    CoefRec *cur, *prev = 0 ;
                    bool update = true ;
                    for (cur = CData->rows[i] ; cur != 0 ; cur = cur->next) {
                        if (cur->column == i) {
                            CData->A[cur->offset]   = 1.0 ;
                            b[i] = values[0] ;
                        } else {
                            CData->A[cur->offset]   = 0.0 ;
                            if (prev == 0) {
                                CData->rows[i] = cur->next ;
                                update = false ;
                            } else {
                                prev->next = cur->next ;
                                update = false ;
                            }
                        }
                        if (update) prev = cur ;
                    }
                } else if (i > node) {
                    CoefRec *cur, *prev = 0 ;
                    bool update = true ;
                    for (cur = CData->rows[i] ; cur != 0 ; cur = cur->next) {
                        if ((cur->column == node) && (cur->column != i)) {
                            if (prev == 0) {
                                CData->rows[i] = cur->next ;
                                update = false ;
                            } else {
                                update = false ;
                                prev->next = cur->next ;
                            }
                            CData->A[cur->offset] = 0.0 ;
                        }
                        if (update) prev = cur ;
                    }
                }
            }
        }
    }
}

void SparseCG::SolveSetup()
{
    // extract the diagonal for a preconditioner

    for (int i=0 ; i<NumNodes ; ++i) {
        CoefRec *next ;
        for (next = CData->rows[i] ; next != 0 ; next = next->next ) {
            if (i == next->column) {
                Binv[i] = 1.0 / CData->A[next->offset] ;
            }
        }
    }
}

bool SparseCG::CGSolve()
{
    int i ;
    double rz, rz_old = 0.0, beta, pAp, alpha, norm, target = 0.0 ;

    double tol_sqr = Tolerance * Tolerance ;

    // PrintMatrix() ;

    // zero out the unknown and residual vector

    // for (i=0 ; i<NumEqns ; ++i) printf("%d %g\n",i,b[i]) ;

    for (i=0 ; i<NumEqns ; ++i) {
        x[i] = 0 ;
        r[i] = b[i] ;
    }

    // iterate until convergance

    int iter = 0 ;
    int max_iter = int(1.5 * NumEqns) ;

    while (iter < max_iter) {

        rz = 0 ;
        for (i=0 ; i<NumEqns ; ++i) {
            z[i] = Binv[i] * r[i] ;
            rz += z[i] * r[i] ;
        }
        iter += 1 ;

        if (iter == 1) {
            for (i=0 ; i<NumEqns ; ++i) {
                p[i] = z[i] ;
            }
        } else {
            beta = rz / rz_old ;
            rz_old = rz ;
            for (i=0 ; i<NumEqns ; ++i) {
                p[i] = z[i] + beta*p[i] ;
            }
        }

        // do the matrix/vector multiply

        MatrixVectMult(p,Ap) ;

        pAp = 0 ;
        for (i=0 ; i<NumEqns ; ++i) {
            pAp += p[i] * Ap[i] ;
        }

        alpha = rz / pAp ;

        for (i=0 ; i<NumEqns ; ++i) {
            x[i] += alpha * p[i] ;
            r[i] -= alpha * Ap[i] ;
        }

        // find the norm of the residual

        norm = 0.0 ;
        for (i=0 ; i<NumEqns ; ++i) {
            norm += r[i]*r[i] ;
        }

//  cout << iter << ' ' << norm << endl ;
        if (iter == 1) {
            target = tol_sqr*norm ;
        } else if (norm < target) {
            return(true) ;
        }
        rz_old = rz ;
    }

    return(false) ;
}

bool SparseCG::GaussSeidelSolve()
{
    double   *A = CData->A ;
    CoefRec  *cur ;
    int i, j, k ;

    double tol_sqr = Tolerance * Tolerance ;
    double *x_tmp = new double[CData->num_rows] ;
    double target, norm ;

    // PrintMatrix() ;

    // iterate until convergance

    int iter = 0 ;
    int max_iter = int(4.0 * NumEqns) ;
    for (i=0 ; i<CData->num_rows ; ++i) x[i] = b[i] ;

    while (iter < max_iter) {

        norm = 0.0 ;

        // save the current vector and set equal to the rhs

        for (i=0 ; i<CData->num_rows ; ++i) {
            x_tmp[i] = x[i] ;
            x[i] = b[i] ;
        }

        // loop through rows, first pass

        for (i=0 ; i<CData->num_rows ; ++i) {
            for (cur = CData->rows[i] ; cur != 0 ; cur = cur->next) {

                k = cur->offset ;
                j = cur->column ;

                if (j != i) {
                    x[j] -= A[k]*x_tmp[i] ;
                }
            }
        }

        // loop through rows, second pass

        for (i=0 ; i<CData->num_rows ; ++i) {
            double aii ;
            for (cur = CData->rows[i] ; cur != 0 ; cur = cur->next) {

                k = cur->offset ;
                j = cur->column ;

                if (j != i) {
                    x[i] -= A[k]*x[j] ;
                } else {
                    aii = A[k] ;
                }
            }
            x[i] /= aii ;
        }

        for (i=0 ; i<CData->num_rows ; ++i) {
            double r = x[i] - x_tmp[i] ; 
            norm += r * r ;
        }


//  cout << iter << ' ' << norm << endl ;

        printf("%d %g %g\n",iter,norm,target) ;

        if (iter == 0) {
            target = tol_sqr*norm ;
//        } else if (norm < target) {
        } else if (norm < 0.001) {
            delete x_tmp ;
            return(true) ;
        }

        ++iter ;
    }

    delete x_tmp ;
    return(false) ;
}

void SparseCG::ZeroRhs()
{
    if (b == 0) return ;
    for (int l=0 ; l<NumEqns ; ++l) {
        b[l] = 0.0 ;
    }
}

/* ===========================================================
**
**  SparseMult - matrix/vector multiply for a sparce symmetric
**               matrix stored as 3x3 blocks in compresses
**               column format.
**
*/

void SparseCG::MatrixVectMult(double *b,double *x)
{
    double   *A = CData->A ;
    CoefRec  *cur ;
    int i, j, k ;

    // zero the output vector

    for (i=0 ; i<CData->num_rows ; ++i) x[i] = 0.0 ;

    // loop through rows

    for (i=0 ; i<CData->num_rows ; ++i) {
        for (cur = CData->rows[i] ; cur != 0 ; cur = cur->next) {

            k = cur->offset ;
            j = cur->column ;

            x[i] += A[k]*b[j] ;

            if (i != cur->column) {
                x[j] += A[k]*b[i] ;
            }
        }
    }
}

double SparseCG::GetEntry(int row,int col)
{
    CoefRec  *cur ;

    if (row >= CData->num_rows) return(0.0) ;

    for (cur = CData->rows[row] ; cur != 0 ; cur = cur->next) {
        if (cur->column == col) {
            return(CData->A[cur->offset]) ;
        } else if (cur->column > col) {
            return(0.0) ;
        }
    }
    return(0.0) ;
}

#if 1
void SparseCG::PrintMatrix()
{
    double   *A = CData->A ;
    CoefRec  *cur ;
    int start = print_start ;
    int stop = print_start+4 ;

    for (int i=start ; i<CData->num_rows ; ++i) {
        printf("%3d ",i) ;

        cur = CData->rows[i] ;
        for (int k=0 ; k<stop ; ++k) {

            if (cur == 0) break ;

            if (k == i) {
                int kk = cur->offset ;
                printf(" %5.2f",A[kk]) ;
            } else if (cur->column == k) {
                int kk = cur->offset ;
                if (k >= start) {
                    printf(" %5.2f",A[kk]) ;
                }
                cur = cur->next ;
            } else {
                if (k >= start) {
//                    printf("                  ") ;
                    printf("  0.00") ;
                }
            }
        }
        printf("\n") ;
    }
    printf("\n") ;

//    for (int i=0 ; i<NumEqns ; ++i) {
//        printf("%d %g\n",i,b[i]) ;
//    }
}
#endif

// -------------------------------------------------------------
// Elasticity solver


void SparseCGElasticity::AllocateCoeffs()
{
    // count the number of 3x3 blocks and allocate space
    // for the coefficients

    int cur = 0 ;
    CoefRec *next ;
    for (int i=0 ; i<CData->num_rows ; ++i) {
        for (next = CData->rows[i] ; next != 0 ; next = next->next ) ++cur ;
    }
    CData->A = new double[9*cur] ;
    for (int i=0 ; i<9*cur ; ++i) CData->A[i] = 0.0 ;
    CoeffsAlloced = true ;
    cur_offset = 0 ;

    b = new double[3*NumNodes] ;
    for (int i=0 ; i<3*NumNodes ; ++i) b[i] = 0.0 ;
}


void SparseCGElasticity::ElementValues(int num_nodes,int *nodes,
                             const TMatrix<double> &stiff)
{
    if (!CoeffsAlloced) AllocateCoeffs() ;

    // go through the equations associated with
    // the element and build up equations

    for (int l=0 ; l<num_nodes ; ++l) {
        for (int k=0 ; k<num_nodes ; ++k) {
            if (nodes[k] > nodes[l]) continue ;

            int ri = nodes[l] ;
            int ci = nodes[k] ;

            CoefRec *next ;
            for (next = CData->rows[ri] ; next != 0 ; next = next->next ) {
                if (ci == next->column) {
                    if (next->offset == -1) {
                        next->offset = cur_offset ;
                        cur_offset += 9 ;
                    }
                    int cur = 0 ;
                    for (int i=0 ; i<3 ; ++i) {
                        for (int j=0 ; j<3 ; ++j) {
                            cur = i*3 + j ;
                            CData->A[next->offset+cur] += 
                                 stiff(l*3+i,k*3+j) ;
                            ++cur ;
                        }
                    }
                    break ;
                }
            }
        }
    }
}

void SparseCGElasticity::RhsValues(int num_nodes,int *nodes,
                         const TVector<Vec3D> &contrib_rhs)
{
    if (!CoeffsAlloced) AllocateCoeffs() ;
    for (int l=0 ; l<num_nodes ; ++l) {
        int jj = 3*nodes[l] ;
        b[jj] += contrib_rhs[l].x() ;
        b[jj+1] += contrib_rhs[l].y() ;
        b[jj+2] += contrib_rhs[l].z() ;
    }
}


#ifdef PY_NUMERIC
void SparseCGElasticity::ElementValues(int num_nodes,int *nodes,
                                       const PyArrayObject *stiff)
{
    if (!CoeffsAlloced) AllocateCoeffs() ;

    // go through the equations associated with
    // the element and build up equations

    for (int l=0 ; l<num_nodes ; ++l) {
        for (int k=0 ; k<num_nodes ; ++k) {
            if (nodes[k] > nodes[l]) continue ;

            int ri = nodes[l] ;
            int ci = nodes[k] ;

            CoefRec *next ;
            for (next = CData->rows[ri] ; next != 0 ; next = next->next ) {
                if (ci == next->column) {
                    if (next->offset == -1) {
                        next->offset = cur_offset ;
                        cur_offset += 9 ;
                    }
                    int cur = 0 ;
                    for (int i=0 ; i<3 ; ++i) {
                        for (int j=0 ; j<3 ; ++j) {
                            cur = i*3 + j ;
                            char *eptr = stiff->data + 
                                         (l*3+i)*stiff->strides[0] +
                                         (k*3+j)*stiff->strides[1] ;
                            cur = i*3 + j ;
                            CData->A[next->offset+cur] +=
                                 *((double *)eptr) ;
                            ++cur ;
                        }
                    }
                    break ;
                }
            }
        }
    }
}

void SparseCGElasticity::RhsValues(int num_nodes,int *nodes,
                                   const PyArrayObject *contrib_rhs)
{
    if (!CoeffsAlloced) AllocateCoeffs() ;

    if (contrib_rhs->nd == 2) {
        for (int l=0 ; l<num_nodes ; ++l) {
            int jj = 3*nodes[l] ;
            for (int k=0 ; k<3 ; ++k) {
                char *eptr = contrib_rhs->data +
                                 l*contrib_rhs->strides[0] +
                                 k*contrib_rhs->strides[1] ;
                b[jj+k] += *((double *)eptr) ;
            }
        }
    } else {
        for (int l=0 ; l<num_nodes ; ++l) {
            int jj = 3*nodes[l] ;
            for (int k=0 ; k<3 ; ++k) {
                char *eptr = contrib_rhs->data +
                             (l*3+k)*contrib_rhs->strides[0] ;
                b[jj+k] += *((double *)eptr) ;
            }
        }
    }
}
#endif

void SparseCGElasticity::AddPrescribedValue(
                             int node,bool x,bool y,bool z,
                             double values[3])
{
    FixedRec *tmp = NewFixed() ;
    tmp->node = node ;
    tmp->flags = 0 ;
    if (x) {
        tmp->flags |= 1 ;
        tmp->values[0] = values[0] ;
    }
    if (y) {
        tmp->flags |= 1<<1 ;
        tmp->values[1] = values[1] ;
    }
    if (z) {
        tmp->flags |= 1<<2 ;
        tmp->values[2] = values[2] ;
    }
}


void SparseCGElasticity::ApplyPrescribedValue()
{
//    PrintMatrix() ;

    bool first = true ;
    for (FixedRecCache *ptr = FixedCacheList ; ptr != 0 ; ptr = ptr->next) {
        int stop = (first) ? FixedCachePtr : FIXED_CACHE_BLOCK_SIZE ;
        if (first) first = false ;
        for (int j=0 ; j<stop ; ++j) {
            int node = ptr->entries[j].node ;
            int flags = ptr->entries[j].flags ;
            double *values = ptr->entries[j].values ;

            // zero out this row and column

            for (int i=0 ; i<CData->num_rows ; ++i) {

                if (i == node) {
                    CoefRec *cur, *prev = 0 ;
                    bool update = true ;
                    for (cur = CData->rows[i] ; cur != 0 ; cur = cur->next) {
                        if (cur->column == i) {
                            if ((flags & 1) != 0) {
                                CData->A[cur->offset]   = 1.0 ;
                                CData->A[cur->offset+1] = 0.0 ;
                                CData->A[cur->offset+2] = 0.0 ;
                                CData->A[cur->offset+3] = 0.0 ;
                                CData->A[cur->offset+6] = 0.0 ;
                                b[i*3] = values[0] ;
                            }
                            if ((flags & (1<<1)) != 0) {
                                CData->A[cur->offset+1] = 0.0 ;
                                CData->A[cur->offset+3] = 0.0 ;
                                CData->A[cur->offset+4] = 1.0 ;
                                CData->A[cur->offset+5] = 0.0 ;
                                CData->A[cur->offset+7] = 0.0 ;
                                b[i*3+1] = values[1] ;
                            }
                            if ((flags & (1<<2)) != 0) {
                                CData->A[cur->offset+2] = 0.0 ;
                                CData->A[cur->offset+5] = 0.0 ;
                                CData->A[cur->offset+6] = 0.0 ;
                                CData->A[cur->offset+7] = 0.0 ;
                                CData->A[cur->offset+8] = 1.0 ;
                                b[i*3+2] = values[2] ;
                            }
                        } else {
                            if ((flags & 0x07) == 0x07) {
                                if (prev == 0) {
                                    CData->rows[i] = cur->next ;
                                    update = false ;
                                } else {
                                    prev->next = cur->next ;
                                    update = false ;
                                }
                            } else {
                                if ((flags & 1) != 0) {
                                    CData->A[cur->offset]   = 0.0 ;
                                    CData->A[cur->offset+1] = 0.0 ;
                                    CData->A[cur->offset+2] = 0.0 ;
                                }
                                if ((flags & (1<<1)) != 0) {
                                    CData->A[cur->offset+3] = 0.0 ;
                                    CData->A[cur->offset+4] = 0.0 ;
                                    CData->A[cur->offset+5] = 0.0 ;
                                }
                                if ((flags & (1<<2)) != 0) {
                                    CData->A[cur->offset+6] = 0.0 ;
                                    CData->A[cur->offset+7] = 0.0 ;
                                    CData->A[cur->offset+8] = 0.0 ;
                                }
                            }
                        }
                        if (update) prev = cur ;
                    }
                } else if (i > node) {
                    CoefRec *cur, *prev = 0 ;
                    bool update = true ;
                    for (cur = CData->rows[i] ; cur != 0 ; cur = cur->next) {
                        if ((cur->column == node) && (cur->column != i)) {
                            if ((flags & 0x07) == 0x07) {
                                if (prev == 0) {
                                    CData->rows[i] = cur->next ;
                                    update = false ;
                                } else {
                                    update = false ;
                                    prev->next = cur->next ;
                                }
                            } else {
                                if ((flags & 1) != 0) {
                                    CData->A[cur->offset]   = 0.0 ;
                                    CData->A[cur->offset+3] = 0.0 ;
                                    CData->A[cur->offset+6] = 0.0 ;
                                }
                                if ((flags & (1<<1)) != 0) {
                                    CData->A[cur->offset+1] = 0.0 ;
                                    CData->A[cur->offset+4] = 0.0 ;
                                    CData->A[cur->offset+7] = 0.0 ;
                                }
                                if ((flags & (1<<2)) != 0) {
                                    CData->A[cur->offset+2] = 0.0 ;
                                    CData->A[cur->offset+5] = 0.0 ;
                                    CData->A[cur->offset+8] = 0.0 ;
                                }
                            }
                        }
                        if (update) prev = cur ;
                    }
                }
            }
        }
    }
}


void SparseCGElasticity::SolveSetup()
{
    // extract the diagonal for a preconditioner

    for (int i=0 ; i<NumNodes ; ++i) {
        CoefRec *next ;
        for (next = CData->rows[i] ; next != 0 ; next = next->next ) {
            if (i == next->column) {
                Binv[i*3]   = 1.0 / CData->A[next->offset] ;
                Binv[i*3+1] = 1.0 / CData->A[next->offset+4] ;
                Binv[i*3+2] = 1.0 / CData->A[next->offset+8] ;
            }
        }
    }
}

bool SparseCGElasticity::GaussSeidelSolve()
{
    double   *A = CData->A ;
    CoefRec  *cur ;
    int i, ii, jj, kk ;

    double tol_sqr = Tolerance * Tolerance ;
    double *x_tmp = new double[CData->num_rows*3] ;
    double target, norm ;

    // PrintMatrix() ;

    // iterate until convergance

    int iter = 0 ;
    int max_iter = int(4.0 * NumEqns) ;
    for (i=0 ; i<CData->num_rows*3 ; ++i) x[i] = b[i] ;

    while (iter < max_iter) {

        norm = 0.0 ;

        // save the current vector and set equal to the rhs

        for (i=0 ; i<CData->num_rows*3 ; ++i) {
            x_tmp[i] = x[i] ;
            x[i] = b[i] ;
        }

        // loop through rows, first pass

        for (i=0 ; i<CData->num_rows ; ++i) {

            ii = 3 * i ;

            for (cur = CData->rows[i] ; cur != 0 ; cur = cur->next) {

                kk = cur->offset ;
                jj = 3 * cur->column ;

                if (cur->column != i) {
                    x[jj] -= (A[kk]*x_tmp[ii] + 
                                  A[kk+1]*x_tmp[ii+1] +
                                  A[kk+2]*x_tmp[ii+2]) ;
                    x[jj+1] -= (A[kk+3]*x_tmp[ii] +
                                    A[kk+4]*x_tmp[ii+1] +
                                    A[kk+5]*x_tmp[ii+2]) ;
                    x[jj+2] -= (A[kk+6]*x_tmp[ii] +
                                    A[kk+7]*x_tmp[ii+1] +
                                    A[kk+8]*x_tmp[ii+2]) ;
                } else {
                    x[ii+1] -= A[kk+3]*x_tmp[ii] ;
                    x[ii+2] -= (A[kk+6]*x_tmp[ii] +
                                    A[kk+7]*x_tmp[ii+1]) ;
                }
            }
       }

        // loop through rows, second pass

        for (i=0 ; i<CData->num_rows ; ++i) {

            int diag ;
            ii = 3 * i ;

            for (cur = CData->rows[i] ; cur != 0 ; cur = cur->next) {

                kk = cur->offset ;
                jj = 3 * cur->column ;

                if (cur->column != i) {
                    x[ii] -= (A[kk]*x[jj] + 
                                  A[kk+1]*x[jj+1] +
                                  A[kk+2]*x[jj+2]) ;
                    x[ii+1] -= (A[kk+3]*x[jj] +
                                    A[kk+4]*x[jj+1] +
                                    A[kk+5]*x[jj+2]) ;
                    x[ii+2] -= (A[kk+6]*x[jj] +
                                    A[kk+7]*x[jj+1] +
                                    A[kk+8]*x[jj+2]) ;
                } else {
                    diag = kk ;
                    x[ii+1] -= A[kk+3]*x[jj] ;
                    x[ii+2] -= (A[kk+6]*x[jj] +
                                    A[kk+7]*x[jj+1]) ;
                }
                x[ii] /= A[kk] ;
                x[ii+1] /= A[kk+4] ;
                x[ii+2] /= A[kk+8] ;
            }
       }

       for (i=0 ; i<CData->num_rows*3 ; ++i) {
//            fprintf(stdout,"i%d %g %g %g\n",i,x_cur[i]*inv_aii[i],x_cur[i],inv_aii[i]) ;
            double r = x[i] - x_tmp[i] ; 
            norm += r * r ;
       }

//  cout << iter << ' ' << norm << endl ;
//fprintf(stdout,"inter, norm : %d %g\n",iter,norm) ;

//       for (i=0 ; i<CData->num_rows*3 ; ++i)
//            fprintf(stdout,"%d %g %g %g\n",i,x_cur[i],x_old[i],b[i]) ;

        if (iter == 1) {
            target = tol_sqr*norm ;
        } else if (norm < target) {
            delete x_tmp ;
            return(true) ;
        }
    }

    delete x_tmp ;
    return(false) ;
}


/* ===========================================================
**
**  SparseMult - matrix/vector multiply for a sparce symmetric
**               matrix stored as 3x3 blocks in compresses
**               column format.
**
*/

void SparseCGElasticity::MatrixVectMult(double *b,double *x)
{
    double   *A = CData->A ;
    CoefRec  *cur ;
    int i, ii, jj, kk ;

    // zero the output vector

    for (i=0 ; i<CData->num_rows*3 ; ++i) x[i] = 0.0 ;

    // loop through rows

    for (i=0 ; i<CData->num_rows ; ++i) {

        ii = 3 * i ;

        for (cur = CData->rows[i] ; cur != 0 ; cur = cur->next) {

            kk = cur->offset ;
            jj = 3 * cur->column ;

            x[ii] += A[kk]*b[jj] + A[kk+1]*b[jj+1] + A[kk+2]*b[jj+2] ;
            x[ii+1] += A[kk+3]*b[jj] + A[kk+4]*b[jj+1] + A[kk+5]*b[jj+2] ;
            x[ii+2] += A[kk+6]*b[jj] + A[kk+7]*b[jj+1] + A[kk+8]*b[jj+2] ;

            if (i != cur->column) {
                x[jj] += A[kk]*b[ii] + A[kk+3]*b[ii+1] + A[kk+6]*b[ii+2] ;
                x[jj+1] += A[kk+1]*b[ii] + A[kk+4]*b[ii+1] + A[kk+7]*b[ii+2] ;
                x[jj+2] += A[kk+2]*b[ii] + A[kk+5]*b[ii+1] + A[kk+8]*b[ii+2] ;
            }
        }
    }
}

double SparseCGElasticity::GetEntry(int row,int col)
{
    CoefRec  *cur ;
    int r3, rr, c3, cr ;
    if (row > col) {
        r3 = row/3 ;
        rr = row % 3 ;
        c3 = col/3 ;
        cr = col % 3 ;
    } else {
        r3 = col/3 ;
        rr = col % 3 ;
        c3 = row/3 ;
        cr = row % 3 ;
    }

    if (r3 >= CData->num_rows) return(0.0) ;

    for (cur = CData->rows[r3] ; cur != 0 ; cur = cur->next) {
        if (cur->column == c3) {
            return(CData->A[cur->offset + (rr*3) + cr]) ;
        } else if (cur->column > c3) {
            return(0.0) ;
        }
    }
    return(0.0) ;
}

#if 1
void SparseCGElasticity::PrintMatrix()
{
    double   *A = CData->A ;
    CoefRec  *cur ;
    int start = print_start ;
    int stop = print_start+4 ;

    for (int i=start ; i<CData->num_rows ; ++i) {
        // int ii = 3 * i ;
        for (int j=0 ; j<3 ; ++j) {

//            if (j == 0)
//                printf("%3d ",i) ;
//            else
//                printf("    ") ;
            printf("%3d ",i*3+j) ;

            cur = CData->rows[i] ;
            for (int k=0 ; k<stop ; ++k) {

                if (cur == 0) break ;

                if (k == i) {
                    int kk = cur->offset ;
                    if (j == 0) {
                        printf(" %5.2f",A[kk+(j*3)]) ;
                    } else if (j == 1) {
                        printf(" %5.2f %5.2f",A[kk+(j*3)],
                           A[kk+(j*3)+1]) ;
                    } else {
                        printf(" %5.2f %5.2f %5.2f",A[kk+(j*3)],
                           A[kk+(j*3)+1],A[kk+(j*3)+2]) ;
                    }
                    cur = cur->next ;
                } else if (cur->column == k) {
                    int kk = cur->offset ;
                    if (k >= start) {
                        printf(" %5.2f %5.2f %5.2f",A[kk+(j*3)],
                           A[kk+(j*3)+1],A[kk+(j*3)+2]) ;
                    }
                    cur = cur->next ;
                } else {
                    if (k >= start) {
//                        printf("                  ") ;
                        printf("  0.00  0.00  0.00") ;
                    }
                }
            }
            printf("\n") ;
        }
    }
    printf("\n") ;

//    for (int i=0 ; i<NumEqns ; ++i) {
//        printf("%d %g\n",i,b[i]) ;
//    }
}
#endif
