/*
// SparseCGModule - python extension for the Sparse CG solver
//
*/

#include <cstdio>
#include <cstdlib>
#include "Python.h"
#include "Numeric/arrayobject.h"
#include "SparseCG.hh"

staticforward PyTypeObject scalar_solver_type ;
staticforward PyTypeObject elasticity_solver_type ;

typedef struct {
    PyObject_HEAD
    SparseCG *solver ;
    }
    SparseCGObject ;

typedef struct {
    PyObject_HEAD
    SparseCGElasticity *solver ;
    }
    SparseCGEObject ;

extern "C" {
    PyObject *AddElementEquations(SparseCGObject *self,PyObject *args) ;
    PyObject *AddElementValues(SparseCGObject *self,PyObject *args) ;
    PyObject *AddRhsValues(SparseCGObject *self,PyObject *args) ;
    PyObject *ScalarAddPrescribedValue(SparseCGObject *self,PyObject *args) ;
    PyObject *ElasticityAddPrescribedValue(SparseCGEObject *self,PyObject *args) ;
    PyObject *DoSolution(SparseCGObject *self,PyObject *args) ;
    PyObject *ElasticityDoSolution(SparseCGObject *self,PyObject *args) ;
    PyObject *DoGaussSeidel(SparseCGObject *self,PyObject *args) ;
    PyObject *ElasticityDoGaussSeidel(SparseCGObject *self,PyObject *args) ;
    PyObject *MatrixVectorMultiply(SparseCGObject *self,PyObject *args) ;
    PyObject *ElasticityMatrixVectorMultiply(SparseCGObject *self,PyObject *args) ;
    PyObject *SetTolerance(SparseCGObject *self,PyObject *args) ;
    PyObject *GetEntry(SparseCGObject *self,PyObject *args) ;
    PyObject *ZeroRhs(SparseCGObject *self,PyObject *args) ;

    void initSparseCG() ;
}

// forward declaration of module API's

/* ==========================================================
** ==========================================================
** ==========================================================
*/

#define is_SparseCGObject(v) ((v)->ob_type == &solver_type)


/* ----------------------------------------------------------
**  Object methods and lookup table
*/

static int *ConvertIntList(int num,PyObject *obj)
{
    if ((PySequence_Check(obj) == 0) ||
        (PySequence_Size(obj) < num)) {
        PyErr_SetString(PyExc_TypeError,
            "argument should a sequence of ints") ;
        return(0) ;
    }

    int *list = new int[num] ; 
    for (int i=0 ; i<num ; ++i) {
        PyObject *item = PySequence_GetItem(obj,i) ;
        if (PyInt_Check(item) == 0) {
            PyErr_SetString(PyExc_TypeError,
                "sequence member should be an int") ;
            Py_DECREF(item) ;
            return(0) ;
        }
        list[i] = PyInt_AsLong(item) ;
        Py_DECREF(item) ;
    }
    return(list) ;
}


static PyObject *AddElementEquations(SparseCGObject *self,PyObject *args)
{
    int num_nodes ;
    PyObject *nodes_obj ;
    if (!PyArg_ParseTuple(args,"iO",&num_nodes,&nodes_obj))
        return(NULL) ;

    int *nodes = ConvertIntList(num_nodes,nodes_obj) ;
    if (nodes == 0) return(NULL) ;

    self->solver->ElementEquations(num_nodes,nodes) ;
    delete nodes ;
    Py_INCREF(Py_None) ;
    return(Py_None) ;
}


static PyObject *AddElementValues(SparseCGObject *self,PyObject *args)
{
    int num_nodes ;
    PyObject *nodes_obj, *stiff_obj ;
    if (!PyArg_ParseTuple(args,"iOO",&num_nodes,&nodes_obj,
              &stiff_obj)) return(NULL) ;

    int *nodes = ConvertIntList(num_nodes,nodes_obj) ;
    if (nodes == 0) return(NULL) ;

    if (PyArray_Check(stiff_obj) == 0) {
        PyErr_SetString(PyExc_TypeError,
        "Third argument should be a NumPy array") ;
        return(NULL) ;
    }

    self->solver->ElementValues(num_nodes,nodes,
                               (PyArrayObject *)stiff_obj) ;
    delete nodes ;
    Py_INCREF(Py_None) ;
    return(Py_None) ;
}


static PyObject *AddRhsValues(SparseCGObject *self,PyObject *args)
{
    int num_nodes ;
    PyObject *nodes_obj, *rhs_obj ;
    if (!PyArg_ParseTuple(args,"iOO",&num_nodes,&nodes_obj,
              &rhs_obj)) return(NULL) ;

    int *nodes = ConvertIntList(num_nodes,nodes_obj) ;
    if (nodes == 0) return(NULL) ;

    if (PyArray_Check(rhs_obj) == 0) {
        PyErr_SetString(PyExc_TypeError,
        "Third argument should be a NumPy array") ;
        return(NULL) ;
    }

    self->solver->RhsValues(num_nodes,nodes,
                           (PyArrayObject *)rhs_obj) ;
    delete nodes ;
    Py_INCREF(Py_None) ;
    return(Py_None) ;
}

static PyObject *ScalarAddPrescribedValue(SparseCGObject *self,PyObject *args)
{
    int node ;
    double value ;
    if (!PyArg_ParseTuple(args,"id",&node,&value)) return(NULL) ;
    self->solver->AddPrescribedValue(node,value) ;
    Py_INCREF(Py_None) ;
    return(Py_None) ;
}

static PyObject *ElasticityAddPrescribedValue(SparseCGEObject *self,PyObject *args)
{
    int node,x,y,z ;
    PyObject *val_obj ;
    if (!PyArg_ParseTuple(args,"iiiiO",&node,&x,&y,&z,&val_obj))
        return(NULL) ;

    // convert the values object

    if ((PySequence_Check(val_obj) == 0) ||
        (PySequence_Size(val_obj) < 3)) {
        PyErr_SetString(PyExc_TypeError,
            "last argument should a sequence of floats") ;
        return(0) ;
    }
    double values[3] ;
    for (int i=0 ; i<3 ; ++i) {
        PyObject *item = PySequence_GetItem(val_obj,i) ;
        if (PyNumber_Check(item) == 0) {
            PyErr_SetString(PyExc_TypeError,
                "sequence member should be a number") ;
            Py_DECREF(item) ;
            return(0) ;
        }
        PyObject *flt = PyNumber_Float(item) ;
        values[i] = PyFloat_AsDouble(flt) ;
        Py_DECREF(flt) ;
        Py_DECREF(item) ;
    }

    self->solver->AddPrescribedValue(node,(x!=0),(y!=0),(z!=0),values) ;
    Py_INCREF(Py_None) ;
    return(Py_None) ;
}


static PyObject *DoSolution(SparseCGObject *self,PyObject *args)
{
    if (!PyArg_ParseTuple(args,"")) return(NULL) ;
    double *x = self->solver->DoSolution() ;
    if (x == 0) {
        Py_INCREF(Py_None) ;
        return(Py_None) ;
    }

    // store the results in a NumPy Array

    int dims[1] ;
    dims[0] = self->solver->GetNumEqns() ;
    PyArrayObject *out = 
        (PyArrayObject *)PyArray_FromDims(1,dims,PyArray_DOUBLE) ;

    double *adata = (double *)out->data ;
    for (int i=0 ; i<self->solver->GetNumEqns() ; ++i) {
        adata[i] = x[i] ;
    }
    return((PyObject *)out) ;

}

static PyObject *ElasticityDoSolution(SparseCGObject *self,PyObject *args)
{
    if (!PyArg_ParseTuple(args,"")) return(NULL) ;
    double *x = self->solver->DoSolution() ;
    if (x == 0) {
        Py_INCREF(Py_None) ;
        return(Py_None) ;
    }

    // store the results in a NumPy Array

    int dims[2] ;
    dims[0] = self->solver->GetNumNodes() ;
    dims[1] = 3 ;
    PyArrayObject *out = 
        (PyArrayObject *)PyArray_FromDims(2,dims,PyArray_DOUBLE) ;

    double *adata = (double *)out->data ;
    int cur = 0 ;
    for (int i=0 ; i<self->solver->GetNumNodes() ; ++i) {
        for (int j=0 ; j<3 ; ++j) {
            adata[cur] = x[cur] ;
            ++cur ;
        }
    }
    return((PyObject *)out) ;

}

static PyObject *DoGaussSeidel(SparseCGObject *self,PyObject *args)
{
    if (!PyArg_ParseTuple(args,"")) return(NULL) ;
    double *x = self->solver->DoGaussSeidel() ;
    if (x == 0) {
        Py_INCREF(Py_None) ;
        return(Py_None) ;
    }

    // store the results in a NumPy Array

    int dims[1] ;
    dims[0] = self->solver->GetNumEqns() ;
    PyArrayObject *out = 
        (PyArrayObject *)PyArray_FromDims(1,dims,PyArray_DOUBLE) ;

    double *adata = (double *)out->data ;
    for (int i=0 ; i<self->solver->GetNumEqns() ; ++i) {
        adata[i] = x[i] ;
    }
    return((PyObject *)out) ;

}

static PyObject *ElasticityDoGaussSeidel(SparseCGObject *self,PyObject *args)
{
    if (!PyArg_ParseTuple(args,"")) return(NULL) ;
    double *x = self->solver->DoGaussSeidel() ;
    if (x == 0) {
        Py_INCREF(Py_None) ;
        return(Py_None) ;
    }

    // store the results in a NumPy Array

    int dims[2] ;
    dims[0] = self->solver->GetNumNodes() ;
    dims[1] = 3 ;
    PyArrayObject *out = 
        (PyArrayObject *)PyArray_FromDims(2,dims,PyArray_DOUBLE) ;

    double *adata = (double *)out->data ;
    int cur = 0 ;
    for (int i=0 ; i<self->solver->GetNumNodes() ; ++i) {
        for (int j=0 ; j<3 ; ++j) {
            adata[cur] = x[cur] ;
            ++cur ;
        }
    }
    return((PyObject *)out) ;

}

static PyObject *MatrixVectorMultiply(SparseCGObject *self,PyObject *args)
{
    if (!PyArg_ParseTuple(args,"")) return(NULL) ;
    double *x = self->solver->MatrixVectorMultiply() ;
    if (x == 0) {
        Py_INCREF(Py_None) ;
        return(Py_None) ;
    }

    // store the results in a NumPy Array

    int dims[1] ;
    dims[0] = self->solver->GetNumEqns() ;
    PyArrayObject *out = 
        (PyArrayObject *)PyArray_FromDims(1,dims,PyArray_DOUBLE) ;

    double *adata = (double *)out->data ;
    for (int i=0 ; i<self->solver->GetNumEqns() ; ++i) {
        adata[i] = x[i] ;
    }
    return((PyObject *)out) ;

}

static PyObject *ElasticityMatrixVectorMultiply(SparseCGObject *self,PyObject *args)
{
    if (!PyArg_ParseTuple(args,"")) return(NULL) ;
    double *x = self->solver->MatrixVectorMultiply() ;
    if (x == 0) {
        Py_INCREF(Py_None) ;
        return(Py_None) ;
    }

    // store the results in a NumPy Array

    int dims[2] ;
    dims[0] = self->solver->GetNumNodes() ;
    dims[1] = 3 ;
    PyArrayObject *out = 
        (PyArrayObject *)PyArray_FromDims(2,dims,PyArray_DOUBLE) ;

    double *adata = (double *)out->data ;
    int cur = 0 ;
    for (int i=0 ; i<self->solver->GetNumNodes() ; ++i) {
        for (int j=0 ; j<3 ; ++j) {
            adata[cur] = x[cur] ;
            ++cur ;
        }
    }
    return((PyObject *)out) ;

}


static PyObject *SetTolerance(SparseCGObject *self,PyObject *args)
{
    double tol ;
    if (!PyArg_ParseTuple(args,"d",&tol)) return(NULL) ;
    self->solver->SetTolerance(tol) ;
    Py_INCREF(Py_None) ;
    return(Py_None) ;
}


static PyObject *GetEntry(SparseCGObject *self,PyObject *args)
{
    int i,j ;
    if (!PyArg_ParseTuple(args,"ii",&i,&j)) return(NULL) ;
    double val = self->solver->GetEntry(i,j) ;
    return(Py_BuildValue("d",val)) ;
}

static PyObject *ZeroRhs(SparseCGObject *self,PyObject *args)
{
    if (!PyArg_ParseTuple(args,"")) return(NULL) ;
    self->solver->ZeroRhs() ;
    Py_INCREF(Py_None) ;
    return(Py_None) ;
}

static PyMethodDef SparseCGScalarMethods[] = {
    { "AddElementEquations", (PyCFunction)AddElementEquations,      METH_VARARGS },
    { "AddElementValues",    (PyCFunction)AddElementValues,         METH_VARARGS },
    { "AddRhsValues",        (PyCFunction)AddRhsValues,             METH_VARARGS },
    { "AddPrescribedValue",  (PyCFunction)ScalarAddPrescribedValue, METH_VARARGS },
    { "DoSolution",          (PyCFunction)DoSolution,               METH_VARARGS },
    { "DoGaussSeidel",       (PyCFunction)DoGaussSeidel,            METH_VARARGS },
    { "MatrixVectorMultiply",(PyCFunction)MatrixVectorMultiply,     METH_VARARGS },
    { "SetTolerance",        (PyCFunction)SetTolerance,             METH_VARARGS },
    { "GetEntry",            (PyCFunction)GetEntry,                 METH_VARARGS },
    { "ZeroRhs",             (PyCFunction)ZeroRhs,                  METH_VARARGS },
    { (char *)NULL, (PyCFunction)NULL }
} ;

static PyMethodDef SparseCGElasticityMethods[] = {
    { "AddElementEquations", (PyCFunction)AddElementEquations,           METH_VARARGS },
    { "AddElementValues",    (PyCFunction)AddElementValues,              METH_VARARGS },
    { "AddRhsValues",        (PyCFunction)AddRhsValues,                  METH_VARARGS },
    { "AddPrescribedValue",  (PyCFunction)ElasticityAddPrescribedValue,  METH_VARARGS },
    { "DoSolution",          (PyCFunction)ElasticityDoSolution,          METH_VARARGS },
    { "DoGaussSeidel",       (PyCFunction)ElasticityDoGaussSeidel,       METH_VARARGS },
    { "MatrixVectorMultiply",(PyCFunction)ElasticityMatrixVectorMultiply,METH_VARARGS },
    { "SetTolerance",        (PyCFunction)SetTolerance,                  METH_VARARGS },
    { "GetEntry",            (PyCFunction)GetEntry,                      METH_VARARGS },
    { "ZeroRhs",             (PyCFunction)ZeroRhs,                       METH_VARARGS },
    { (char *)NULL, (PyCFunction)NULL }
} ;


static SparseCGObject *newSparseCGScalarObject(PyObject *args)
{
    int num_nodes ;

    if (!PyArg_ParseTuple(args,"i",&num_nodes)) return(NULL) ;

    SparseCGObject *self = PyObject_NEW(SparseCGObject,
                                        &scalar_solver_type) ;
    if (self == NULL) return NULL ;

    self->solver = new SparseCG(num_nodes) ;
    return(self) ;
}

static SparseCGEObject *newSparseCGElasticityObject(PyObject *args)
{
    int num_nodes ;

    if (!PyArg_ParseTuple(args,"i",&num_nodes)) return(NULL) ;

    SparseCGEObject *self = PyObject_NEW(SparseCGEObject,
                                         &elasticity_solver_type) ;
    if (self == NULL) return NULL ;

    self->solver = new SparseCGElasticity(num_nodes) ;
    return(self) ;
}

static void SparseCGDealloc(PyObject *self)
{
    SparseCGObject *obj = (SparseCGObject *)self ;
    delete obj->solver ;
}

static PyObject *SparseCGScalarGetattr(PyObject *self,char *name)
{
    return(Py_FindMethod(SparseCGScalarMethods,self,name)) ;
}

static PyObject *SparseCGElasticityGetattr(PyObject *self,char *name)
{
    return(Py_FindMethod(SparseCGElasticityMethods,self,name)) ;
}

PyTypeObject scalar_solver_type = {

    /* type header */

    PyObject_HEAD_INIT(&PyType_Type)
    0,
    "SparseCGScalar",             /* tp_name */
    sizeof(SparseCGObject),       /* tp_basicsize */
    0,                            /* tp_itemsize */

    /* standard methods */

    (destructor)  SparseCGDealloc,        /* tp_dealloc ref-count==0 */
    (printfunc)   0,                      /* tp_print   "print x" */
    (getattrfunc) SparseCGScalarGetattr,  /* tp_getattr "x.attr" */
    (setattrfunc) 0,                      /* tp_setattr "x.attr=v" */
    (cmpfunc)     0,                      /* tp_compare "x > y" */
    (reprfunc)    0,                     /* tp_repr    'x', print x */
 
    /* type categories */

    0,                   /* tp_as_number   +,-,*,/,%,&,>>,pow... */
    0,                   /* tp_as_sequence +,[i],[i:j],len,... */
    0,                   /* ty_as_mapping  [key],len,... */

    /* more methods */

    (hashfunc)     0,                   /* tp_hash   "dict[x]" */
    (ternaryfunc)  0,                   /* tp_call   "x()" */
    (reprfunc)     0,                   /* tp_str    "str(x)" */
} ;

PyTypeObject elasticity_solver_type = {

    /* type header */

    PyObject_HEAD_INIT(&PyType_Type)
    0,
    "SparseCGElasticity",         /* tp_name */
    sizeof(SparseCGObject),       /* tp_basicsize */
    0,                            /* tp_itemsize */

    /* standard methods */

    (destructor)  SparseCGDealloc,           /* tp_dealloc ref-count==0 */
    (printfunc)   0,                         /* tp_print   "print x" */
    (getattrfunc) SparseCGElasticityGetattr, /* tp_getattr "x.attr" */
    (setattrfunc) 0,                         /* tp_setattr "x.attr=v" */
    (cmpfunc)     0,                         /* tp_compare "x > y" */
    (reprfunc)    0,                         /* tp_repr    'x', print x */

    /* type categories */

    0,                   /* tp_as_number   +,-,*,/,%,&,>>,pow... */
    0,                   /* tp_as_sequence +,[i],[i:j],len,... */
    0,                   /* ty_as_mapping  [key],len,... */

    /* more methods */

    (hashfunc)     0,                   /* tp_hash   "dict[x]" */
    (ternaryfunc)  0,                   /* tp_call   "x()" */
    (reprfunc)     0,                   /* tp_str    "str(x)" */
} ;

// ==========================================================
//  Module methods and lookup table
//

static PyObject *SparseCGScalarType_new(PyObject *self, PyObject *args)
{
    return((PyObject *)newSparseCGScalarObject(args)) ;
}

static PyObject *SparseCGElasticityType_new(PyObject *self, PyObject *args)
{
    return((PyObject *)newSparseCGElasticityObject(args)) ;
}

static PyMethodDef SparseCGTypeMethods[] = {
    { "SparseCGScalar",     SparseCGScalarType_new,     METH_VARARGS },
    { "SparseCGElasticity", SparseCGElasticityType_new, METH_VARARGS },
    { NULL, NULL }
} ;

#include <stdio.h>

void initSparseCG()
{
    Py_InitModule("SparseCG",SparseCGTypeMethods) ;
    import_array() ;
}

