//
// Shape function implementation file
//
// Revision -
//   $Revision: 1.2 $  $Date: 2002/01/31 15:08:30 $  $Author: wash $
//

/*  Here's how these elements are numbered


   10-noded numbering

                            
                         2 o +++
                          / \   +++   7     
                         /   \     ++o    
                        /     \       +++ o 3 
                       /       \5     ++++
                    6 o         o ++++  +
                     /     ++o++ \     o 8
                    / +++++  9    \   +
                   /++             \ +
                  o--------o--------o  - - - - - - - - t = -1
                   0       4        1

   face    nodes
    0   0 1 2 4 5 6       r is zero on face 1, 1 at node 0
    1   1 3 2 8 7 5       s is zero on face 2, 1 at node 1
    2   2 3 0 7 9 6       t is zero on face 3, 1 at node 2
    3   0 3 1 9 8 4



                       2 o-------o-------o 6
  20-noded numbering    +|      18      +|
                       + |             + |          t
                    9 o  |         11 o  |          ^
                     +   o 12        +   o 14       |
                    +    | 19       +    |          |
                 3 o-------o-------o 7   |          -----> s
                   |   0 o-------o-|-----o 4       /
                   |    +       16 |    +         /
                   |   +           |   +         r
                13 o  o 8       15 o  o 10
                   | +             | +
                   |+              |+
                 1 o-------o-------o 5
                          17        

    face         nodes
     0   0  1  3  2  8 13  9 12
     1   4  6  7  5 14 11 15 10
     2   0  4  5  1 16 10 17  8
     3   2  3  7  6  9 19 11 18
     4   0  2  6  4 12 18 14 16
     5   1  5  7  3 17 15 19 13


                                    5
                                  o
   15-noded numbering           +/ \
                            14 o/   \
                             + /     \ 
                          2+  o 11    o 10
                          /\ /         \
                         /  \     9     \
                        /  o-\----o------o 4 - - - - - t = +1
                       /  +   \         +
                    8 o +      o 7     +
               r     / o 12     \     o 13
               ^    / +          \   +
               |   /+             \ +
              --- o--------o-------o  - - - - - - - - t = -1
                 / 0       6        1
                /--> s


    face        nodes
     0   0  3  4  1 12  9 13  6
     1   2  5  3  0 14 11 12  8
     2   1  4  5  2 13 10 14  7
     3   0  1  2  6  7  8
     4   3  5  4 11 10  9
*/

#include <cstring>

#include "ShapeFunc.hh"

namespace FemLib {

TVector<double> QIShapeFunc::SurfaceShapeFunc(
            SurfaceShape shape,Vec3D loc) const
{
    if (shape == TRIANGLE) {
        TVector<double> shape(6) ;
        double u = 1.0 - loc.x() - loc.y() ;
        shape[0] = (2 * loc.x() - 1) * loc.x() ;
        shape[1] = (2 * loc.y() - 1) * loc.y() ;
        shape[2] = (2 * u - 1) * u ;
        shape[3] = 4.0 * loc.x() * loc.y() ;
        shape[4] = 4.0 * u * loc.y() ;
        shape[5] = 4.0 * u * loc.x() ;
        return(shape) ;
    } else if (shape == QUADRILATERAL) {
        TVector<double> shape(8) ;
        double rr  = loc.x() * loc.x() ;
        double rs  = loc.x() * loc.y() ;
        double ss  = loc.y() * loc.y() ;
        double rrs = rr * loc.y() ;
        double rss = loc.x() * ss ;

        shape[0] = 0.25 * (rr - rrs + rs - rss + ss - 1.0) ;
        shape[1] = 0.25 * (rr - rrs - rs + rss + ss - 1.0) ;
        shape[2] = 0.25 * (rr + rrs + rs + rss + ss - 1.0) ;
        shape[3] = 0.25 * (rr + rrs - rs - rss + ss - 1.0) ;
        shape[4] = 0.5 * (-rr + rrs - loc.y() + 1.0) ;
        shape[5] = 0.5 * (loc.x() - rss - ss + 1.0) ;
        shape[6] = 0.5 * (-rr - rrs + loc.y() + 1.0) ;
        shape[7] = 0.5 * (-loc.x() + rss - ss + 1.0) ;
        return(shape) ;
    }
    return(NullShp) ;
}

TMatrix<double> QIShapeFunc::SurfaceShapeDeriv(
            SurfaceShape shape,Vec3D loc) const
{
    if (shape == TRIANGLE) {
        TMatrix<double> deriv(2,6) ;
        deriv(0,0) =  4.0 * loc.x() - 1.0 ;
        deriv(0,1) =  0.0 ;
        deriv(0,2) = -3.0 + 4.0 * loc.x() + 4.0 * loc.y() ;
        deriv(0,3) =  4.0 * loc.y() ;
        deriv(0,4) = -4.0 * loc.y() ;
        deriv(0,5) =  4.0 - 4.0 * loc.y() - 8.0 * loc.x() ;

        deriv(1,0) =  0.0 ;
        deriv(1,1) =  4.0 * loc.y() - 1.0 ;
        deriv(1,2) = -3.0 + 4.0 * loc.y() + 4.0 * loc.x() ;
        deriv(1,3) =  4.0 * loc.x() ;
        deriv(1,4) =  4.0 - 4.0 * loc.x() - 8.0 * loc.y() ;
        deriv(1,5) = -4.0 * loc.x() ;
        return(deriv) ;
    } else if (shape == QUADRILATERAL) {
        TMatrix<double> deriv(2,8) ;
        double r2  = 2.0 * loc.x() ;
        double rr  = loc.x() * loc.x() ;
        double rs2 = 2.0 * loc.x() * loc.y() ;
        double ss  = loc.y() * loc.y() ;
        double s2  = 2.0 * loc.y() ;

        deriv(0,0) = 0.25 * (r2 - rs2 + loc.y() - ss) ;
        deriv(0,1) = 0.25 * (r2 - rs2 - loc.y() + ss) ;
        deriv(0,2) = 0.25 * (r2 + rs2 + loc.y() + ss) ;
        deriv(0,3) = 0.25 * (r2 + rs2 - loc.y() - ss) ;
        deriv(0,4) = 0.5 * (-r2 + rs2) ;
        deriv(0,5) = 0.5 * (1.0 - ss) ;
        deriv(0,6) = 0.5 * (-r2 - rs2) ;
        deriv(0,7) = 0.5 * (-1.0 + ss) ;

        deriv(1,0) = 0.25 * (-rr + loc.x() - rs2 + s2) ;
        deriv(1,1) = 0.25 * (-rr - loc.x() + rs2 + s2) ;
        deriv(1,2) = 0.25 * (rr + loc.x() + rs2 + s2) ;
        deriv(1,3) = 0.25 * (rr - loc.x() - rs2 + s2) ;
        deriv(1,4) = 0.5 * (-1.0 + rr) ;
        deriv(1,5) = 0.5 * (-rs2 - s2) ;
        deriv(1,6) = 0.5 * (1.0 - rr) ;
        deriv(1,7) = 0.5 * (rs2 - s2) ;
        return(deriv) ;
    }
    return(NullDrv) ;
}

// ------------------------------------------------------------

TVector<double> LIShapeFunc::SurfaceShapeFunc(
            SurfaceShape shape,Vec3D loc) const
{
    if (shape == TRIANGLE) {
        TVector<double> shape(3) ;
        shape[0] = loc.x() ;
        shape[1] = loc.y() ;
        shape[2] = 1.0 - loc.x() - loc.y() ;
        return(shape) ;
    } else if (shape == QUADRILATERAL) {
//         TVector<double> shape(8) ;
//         double rr  = loc.x() * loc.x() ;
//         double rs  = loc.x() * loc.y() ;
//         double ss  = loc.y() * loc.y() ;
//         double rrs = rr * loc.y() ;
//         double rss = loc.x() * ss ;
// 
//         shape[0] = 0.25 * (rr - rrs + rs - rss + ss - 1.0) ;
//         shape[1] = 0.25 * (rr - rrs - rs + rss + ss - 1.0) ;
//         shape[2] = 0.25 * (rr + rrs + rs + rss + ss - 1.0) ;
//         shape[3] = 0.25 * (rr + rrs - rs - rss + ss - 1.0) ;
//         shape[4] = 0.5 * (-rr + rrs - loc.y() + 1.0) ;
//         shape[5] = 0.5 * (loc.x() - rss - ss + 1.0) ;
//         shape[6] = 0.5 * (-rr - rrs + loc.y() + 1.0) ;
//         shape[7] = 0.5 * (-loc.x() + rss - ss + 1.0) ;
//         return(shape) ;
    }
    return(NullShp) ;
}

TMatrix<double> LIShapeFunc::SurfaceShapeDeriv(
            SurfaceShape shape,Vec3D loc) const
{
    if (shape == TRIANGLE) {
        TMatrix<double> deriv(2,3) ;
        deriv(0,0) =  1.0 ;
        deriv(0,1) =  0.0 ;
        deriv(0,2) = -1.0 ;

        deriv(1,0) =  0.0 ;
        deriv(1,1) =  1.0 ;
        deriv(1,2) = -1.0 ;
        return(deriv) ;
    } else if (shape == QUADRILATERAL) {
//         TMatrix<double> deriv(2,8) ;
//         double r2  = 2.0 * loc.x() ;
//         double rr  = loc.x() * loc.x() ;
//         double rs2 = 2.0 * loc.x() * loc.y() ;
//         double ss  = loc.y() * loc.y() ;
//         double s2  = 2.0 * loc.y() ;
// 
//         deriv(0,0) = 0.25 * (r2 - rs2 + loc.y() - ss) ;
//         deriv(0,1) = 0.25 * (r2 - rs2 - loc.y() + ss) ;
//         deriv(0,2) = 0.25 * (r2 + rs2 + loc.y() + ss) ;
//         deriv(0,3) = 0.25 * (r2 + rs2 - loc.y() - ss) ;
//         deriv(0,4) = 0.5 * (-r2 + rs2) ;
//         deriv(0,5) = 0.5 * (1.0 - ss) ;
//         deriv(0,6) = 0.5 * (-r2 - rs2) ;
//         deriv(0,7) = 0.5 * (-1.0 + ss) ;
// 
//         deriv(1,0) = 0.25 * (-rr + loc.x() - rs2 + s2) ;
//         deriv(1,1) = 0.25 * (-rr - loc.x() + rs2 + s2) ;
//         deriv(1,2) = 0.25 * (rr + loc.x() + rs2 + s2) ;
//         deriv(1,3) = 0.25 * (rr - loc.x() - rs2 + s2) ;
//         deriv(1,4) = 0.5 * (-1.0 + rr) ;
//         deriv(1,5) = 0.5 * (-rs2 - s2) ;
//         deriv(1,6) = 0.5 * (1.0 - rr) ;
//         deriv(1,7) = 0.5 * (rs2 - s2) ;
//         return(deriv) ;
    }
    return(NullDrv) ;
}

// ----------------------------------------------------------------
//  Tet2IShapeFunc implementation

Tet2IShapeFunc *Tet2IShapeFunc::Instance()
{
    if (_instance == 0) {
        _instance = new Tet2IShapeFunc ;
    }
    return _instance ;
}

Tet2IShapeFunc *Tet2IShapeFunc::_instance = 0 ;

int Tet2IShapeFunc::face_map[4][7] = {
    { 6, 0, 1, 2, 4, 5, 6 },
    { 6, 1, 3, 2, 8, 7, 5 },
    { 6, 2, 3, 0, 7, 9, 6 },
    { 6, 0, 3, 1, 9, 8, 4 }} ;


TVector<double> Tet2IShapeFunc::VolumeShapeFunc(Vec3D loc) const
{
    TVector<double> shape(10) ;
    double u = 1.0 - loc.x() - loc.y() - loc.z() ;
    shape[0] = (2 * loc.x() - 1) * loc.x() ;
    shape[1] = (2 * loc.y() - 1) * loc.y() ;
    shape[2] = (2 * loc.z() - 1) * loc.z() ;
    shape[3] = (2 * u - 1) * u ;
    shape[4] = 4.0 * loc.x() * loc.y() ;
    shape[5] = 4.0 * loc.y() * loc.z() ;
    shape[6] = 4.0 * loc.x() * loc.z() ;
    shape[7] = 4.0 * loc.z() * u ;
    shape[8] = 4.0 * loc.y() * u ;
    shape[9] = 4.0 * loc.x() * u ;
    return(shape) ;
}

TMatrix<double> Tet2IShapeFunc::VolumeShapeDeriv(Vec3D loc) const
{
    TMatrix<double> deriv(3,10) ;
    double u = 1.0 - loc.x() - loc.y() - loc.z() ;

    deriv(0,0) =  4.0*loc.x() - 1.0 ;
    deriv(0,1) =  0.0 ;
    deriv(0,2) =  0.0 ;
    deriv(0,3) =  1.0 - 4.0*u ;
    deriv(0,4) =  4.0*loc.y() ;
    deriv(0,5) =  0.0 ;
    deriv(0,6) =  4.0*loc.z() ;
    deriv(0,7) =  -4.0*loc.z() ;
    deriv(0,8) =  -4.0*loc.y() ;
    deriv(0,9) =  4.0 * (u - loc.x()) ;

    deriv(1,0) =  0.0 ;
    deriv(1,1) =  4.0*loc.y() - 1.0 ;
    deriv(1,2) =  0.0 ;
    deriv(1,3) =  1.0 - 4.0*u ;
    deriv(1,4) =  4.0*loc.x() ;
    deriv(1,5) =  4.0*loc.z() ;
    deriv(1,6) =  0.0 ;
    deriv(1,7) =  -4.0*loc.z() ;
    deriv(1,8) =  4.0 * (u - loc.y()) ;
    deriv(1,9) =  -4.0*loc.x() ;

    deriv(2,0) =  0.0 ;
    deriv(2,1) =  0.0 ;
    deriv(2,2) =  4.0*loc.z() - 1.0;
    deriv(2,3) =  1.0 - 4.0*u;
    deriv(2,4) =  0.0 ;
    deriv(2,5) =  4.0*loc.y() ;
    deriv(2,6) =  4.0*loc.x() ;
    deriv(2,7) =  4.0 * (u - loc.z()) ;
    deriv(2,8) = -4.0*loc.y() ;
    deriv(2,9) = -4.0*loc.x() ;

    return(deriv) ;
}


bool Tet2IShapeFunc::InsideElement(const Vec3D &natural) const
{
    double u = 1.0 - natural[0] - natural[1] - natural[2] ;
    if ((natural[0] < 0.0) || (natural[0] > 1.0) ||
        (natural[1] < 0.0) || (natural[1] > 1.0) ||
        (natural[2] < 0.0) || (natural[2] > 1.0) ||
        (u < 0.0) || (u > 1.0)) return(false) ;
    return(true) ;
}


// ----------------------------------------------------------------
//  Brick2IShapeFunc implementation

Brick2IShapeFunc *Brick2IShapeFunc::Instance()
{
    if (_instance == 0) {
        _instance = new Brick2IShapeFunc ;
    }
    return _instance ;
}

Brick2IShapeFunc *Brick2IShapeFunc::_instance = 0 ;

int Brick2IShapeFunc::face_map[6][9] = {
    { 8, 0, 1, 3, 2,  8, 13,  9, 12 },
    { 8, 4, 6, 7, 5, 14, 11, 15, 10 },
    { 8, 0, 4, 5, 1, 16, 10, 17,  8 },
    { 8, 2, 3, 7, 6,  9, 19, 11, 18 },
    { 8, 0, 2, 6, 4, 12, 18, 14, 16 },
    { 8, 1, 5, 7, 3, 17, 15, 19, 13 }} ;


TVector<double> Brick2IShapeFunc::VolumeShapeFunc(Vec3D loc) const
{
    TVector<double> shape(20) ;
    double r = loc.x() ;
    double s = loc.y() ;
    double t = loc.z() ;
    double r2 = r*r ;
    double s2 = s*s ;
    double t2 = t*t ;
    shape[ 0] = 0.125*(1.0-r)*(1.0-s)*(1.0-t)*(-r-s-t-2.0) ;
    shape[ 1] = 0.125*(1.0-r)*(1.0-s)*(1.0+t)*(-r-s+t-2.0) ;
    shape[ 2] = 0.125*(1.0-r)*(1.0+s)*(1.0-t)*(-r+s-t-2.0) ;
    shape[ 3] = 0.125*(1.0-r)*(1.0+s)*(1.0+t)*(-r+s+t-2.0) ;
    shape[ 4] = 0.125*(1.0+r)*(1.0-s)*(1.0-t)*(r-s-t-2.0) ;
    shape[ 5] = 0.125*(1.0+r)*(1.0-s)*(1.0+t)*(r-s+t-2.0) ;
    shape[ 6] = 0.125*(1.0+r)*(1.0+s)*(1.0-t)*(r+s-t-2.0) ;
    shape[ 7] = 0.125*(1.0+r)*(1.0+s)*(1.0+t)*(r+s+t-2.0) ;
    shape[ 8] = 0.250*(1.0-r)*(1.0-s)*(1.0-t2) ;
    shape[ 9] = 0.250*(1.0-r)*(1.0+s)*(1.0-t2) ;
    shape[10] = 0.250*(1.0+r)*(1.0-s)*(1.0-t2) ;
    shape[11] = 0.250*(1.0+r)*(1.0+s)*(1.0-t2) ;
    shape[12] = 0.250*(1.0-r)*(1.0-s2)*(1.0-t) ;
    shape[13] = 0.250*(1.0-r)*(1.0-s2)*(1.0+t) ;
    shape[14] = 0.250*(1.0+r)*(1.0-s2)*(1.0-t) ;
    shape[15] = 0.250*(1.0+r)*(1.0-s2)*(1.0+t) ;
    shape[16] = 0.250*(1.0-r2)*(1.0-s)*(1.0-t) ;
    shape[17] = 0.250*(1.0-r2)*(1.0-s)*(1.0+t) ;
    shape[18] = 0.250*(1.0-r2)*(1.0+s)*(1.0-t) ;
    shape[19] = 0.250*(1.0-r2)*(1.0+s)*(1.0+t) ;
    return(shape) ;
}

TMatrix<double> Brick2IShapeFunc::VolumeShapeDeriv(Vec3D loc) const
{
    TMatrix<double> deriv(3,20) ;
    double r = loc.x() ;
    double s = loc.y() ;
    double t = loc.z() ;
    double r2 = r*r ;
    double s2 = s*s ;
    double t2 = t*t ;

    deriv(0, 0) = -.125 * (1.0-s) * (1.0-t) * (-2.0*r-s-t-1.0) ;
    deriv(0, 1) = -.125 * (1.0-s) * (1.0+t) * (-2.0*r-s+t-1.0) ;
    deriv(0, 2) = -.125 * (1.0+s) * (1.0-t) * (-2.0*r+s-t-1.0) ; 
    deriv(0, 3) = -.125 * (1.0+s) * (1.0+t) * (-2.0*r+s+t-1.0) ;
    deriv(0, 4) = 0.125 * (1.0-s) * (1.0-t) * (2.0*r-s-t-1.0) ;
    deriv(0, 5) = 0.125 * (1.0-s) * (1.0+t) * (2.0*r-s+t-1.0) ;
    deriv(0, 6) = 0.125 * (1.0+s) * (1.0-t) * (2.0*r+s-t-1.0) ;
    deriv(0, 7) = 0.125 * (1.0+s) * (1.0+t) * (2.0*r+s+t-1.0) ;
    deriv(0, 8) = -.250 * (1.0-s) * (1.0-t2) ;
    deriv(0, 9) = -.250 * (1.0+s) * (1.0-t2) ;
    deriv(0,10) = 0.250 * (1.0-s) * (1.0-t2) ;
    deriv(0,11) = 0.250 * (1.0+s) * (1.0-t2) ;
    deriv(0,12) = -.250 * (1.0-s2) * (1.0-t) ;
    deriv(0,13) = -.250 * (1.0-s2) * (1.0+t) ;
    deriv(0,14) = 0.250 * (1.0-s2) * (1.0-t) ;
    deriv(0,15) = 0.250 * (1.0-s2) * (1.0+t) ;
    deriv(0,16) = -0.50 * r * (1.0-s) * (1.0-t) ;
    deriv(0,17) = -0.50 * r * (1.0-s) * (1.0+t) ;
    deriv(0,18) = -0.50 * r * (1.0+s) * (1.0-t) ;
    deriv(0,19) = -0.50 * r * (1.0+s) * (1.0+t) ;

    deriv(1, 0) = -.125 * (1.0-r) * (1.0-t) * (-r-2.0*s-t-1.0) ;
    deriv(1, 1) = -.125 * (1.0-r) * (1.0+t) * (-r-2.0*s+t-1.0) ;
    deriv(1, 2) = 0.125 * (1.0-r) * (1.0-t) * (-r+2.0*s-t-1.0) ; 
    deriv(1, 3) = 0.125 * (1.0-r) * (1.0+t) * (-r+2.0*s+t-1.0) ;
    deriv(1, 4) = -.125 * (1.0+r) * (1.0-t) * (r-2.0*s-t-1.0) ;
    deriv(1, 5) = -.125 * (1.0+r) * (1.0+t) * (r-2.0*s+t-1.0) ;
    deriv(1, 6) = 0.125 * (1.0+r) * (1.0-t) * (r+2.0*s-t-1.0) ;
    deriv(1, 7) = 0.125 * (1.0+r) * (1.0+t) * (r+2.0*s+t-1.0) ;
    deriv(1, 8) = -.250 * (1.0-r) * (1-t2) ;
    deriv(1, 9) = 0.250 * (1.0-r) * (1.0-t2) ;
    deriv(1,10) = -.250 * (1.0+r) * (1.0-t2) ;
    deriv(1,11) = 0.250 * (1.0+r) * (1.0-t2) ;
    deriv(1,12) = -.500 * (1.0-r) * s * (1.0-t) ;
    deriv(1,13) = -.500 * (1.0-r) * s * (1.0+t) ;
    deriv(1,14) = -.500 * (1.0+r) * s * (1.0-t) ;
    deriv(1,15) = -.500 * (1.0+r) * s * (1.0+t) ;
    deriv(1,16) = -.250 * (1.0-r2) * (1.0-t) ;
    deriv(1,17) = -.250 * (1.0-r2) * (1.0+t) ;
    deriv(1,18) = 0.250 * (1.0-r2) * (1.0-t) ;
    deriv(1,19) = 0.250 * (1.0-r2) * (1.0+t) ;

    deriv(2, 0) = -.125 * (1.0-r) * (1.0-s) * (-r-s-2.0*t-1.0) ;
    deriv(2, 1) = 0.125 * (1.0-r) * (1.0-s) * (-r-s+2.0*t-1.0) ;
    deriv(2, 2) = -.125 * (1.0-r) * (1.0+s) * (-r+s-2.0*t-1.0) ;
    deriv(2, 3) = 0.125 * (1.0-r) * (1.0+s) * (-r+s+2.0*t-1.0) ;
    deriv(2, 4) = -.125 * (1.0+r) * (1.0-s) * (r-s-2.0*t-1.0) ;
    deriv(2, 5) = 0.125 * (1.0+r) * (1.0-s) * (r-s+2.0*t-1.0) ;
    deriv(2, 6) = -.125 * (1.0+r) * (1.0+s) * (r+s-2.0*t-1.0) ;
    deriv(2, 7) = 0.125 * (1.0+r) * (1.0+s) * (r+s+2.0*t-1.0) ;
    deriv(2, 8) = -.500 * (1.0-r) * (1.0-s) * t ;
    deriv(2, 9) = -.500 * (1.0-r) * (1.0+s) * t ;
    deriv(2,10) = -.500 * (1.0+r) * (1.0-s) * t ;
    deriv(2,11) = -.500 * (1.0+r) * (1.0+s) * t ;
    deriv(2,12) = -.250 * (1.0-r) * (1.0-s2) ;
    deriv(2,13) = 0.250 * (1.0-r) * (1.0-s2) ;
    deriv(2,14) = -.250 * (1.0+r) * (1.0-s2) ;
    deriv(2,15) = 0.250 * (1.0+r) * (1.0-s2) ;
    deriv(2,16) = -.250 * (1.0-r2) * (1.0-s) ;
    deriv(2,17) = 0.250 * (1.0-r2) * (1.0-s) ;
    deriv(2,18) = -.250 * (1.0-r2) * (1.0+s) ;
    deriv(2,19) = 0.250 * (1.0-r2) * (1.0+s) ;
    return(deriv) ;
}


bool Brick2IShapeFunc::InsideElement(const Vec3D &natural) const
{
    if ((natural[0] < -1.0) || (natural[0] > 1.0) ||
        (natural[1] < -1.0) || (natural[1] > 1.0) ||
        (natural[2] < -1.0) || (natural[2] > 1.0)) return(false) ;
    return(true) ;
}


// ----------------------------------------------------------------
//  Wedge2IShapeFunc implementation

Wedge2IShapeFunc *Wedge2IShapeFunc::Instance()
{
    if (_instance == 0) {
        _instance = new Wedge2IShapeFunc ;
    }
    return _instance ;
}

Wedge2IShapeFunc *Wedge2IShapeFunc::_instance = 0 ;

int Wedge2IShapeFunc::face_map[5][9] = {
    { 8, 0, 3, 4,  1, 12,  9, 13,  6 },
    { 8, 2, 5, 3,  0, 14, 11, 12,  8 },
    { 8, 1, 4, 5,  2, 13, 10, 14,  7 },
    { 6, 0, 1, 2,  6,  7,  8, -1, -1 },
    { 6, 3, 5, 4, 11, 10,  9, -1, -1 }} ;

ShapeFunc::SurfaceShape Wedge2IShapeFunc::face_type_map[5] = {
    QUADRILATERAL,QUADRILATERAL,QUADRILATERAL,
    TRIANGLE,TRIANGLE } ;

TVector<double> Wedge2IShapeFunc::VolumeShapeFunc(Vec3D loc) const
{
    TVector<double> shape(15) ;
    double r = loc.x() ;
    double s = loc.y() ;
    double t = loc.z() ;
    double u = 1.0 - r - s ;
    shape[ 0] = 0.5*u * ((2.0*u-1.0)*(1.0-t) - (1.0-t*t)) ;
    shape[ 1] = 0.5*s * ((2.0*s-1.0)*(1.0-t) - (1.0-t*t)) ; 
    shape[ 2] = 0.5*r * ((2.0*r-1.0)*(1.0-t) - (1.0-t*t)) ; 
    shape[ 3] = 0.5*u * ((2.0*u-1.0)*(1.0+t) - (1.0-t*t)) ; 
    shape[ 4] = 0.5*s * ((2.0*s-1.0)*(1.0+t) - (1.0-t*t)) ;
    shape[ 5] = 0.5*r * ((2.0*r-1.0)*(1.0+t) - (1.0-t*t)) ;   
    shape[ 6] = 2.0 * u * s * (1.0 - t) ;
    shape[ 7] = 2.0 * s * r * (1.0 - t) ;
    shape[ 8] = 2.0 * r * u * (1.0 - t) ; 
    shape[ 9] = 2.0 * u * s * (1.0 + t) ; 
    shape[10] = 2.0 * s * r * (1.0 + t) ;
    shape[11] = 2.0 * r * u * (1.0 + t) ;
    shape[12] = u * (1.0 - t * t) ;
    shape[13] = s * (1.0 - t * t) ; 
    shape[14] = r * (1.0 - t * t) ;
    return(shape) ;
}

TMatrix<double> Wedge2IShapeFunc::VolumeShapeDeriv(Vec3D loc) const
{
    TMatrix<double> deriv(3,15) ;
    double r = loc.x() ;
    double s = loc.y() ;
    double t = loc.z() ;
    double u = 1.0 - r - s ;

    deriv(0, 0) = -0.5*(t-2.0+4.0*r+4.0*s)*(t-1.0) ;
    deriv(0, 1) =  0.0 ;
    deriv(0, 2) =  0.5*(t-4.0*r+2.0)*(t-1.0) ;
    deriv(0, 3) = -0.5*(t+2.0-4.0*r-4.0*s)*(t+1.0) ;
    deriv(0, 4) =  0.0 ;
    deriv(0, 5) =  0.5*(t+4.0*r-2.0)*(t+1.0) ;
    deriv(0, 6) =  2.0*s*(t-1.0) ;
    deriv(0, 7) = -2.0*s*(t-1.0) ;
    deriv(0, 8) =  2.0*(-1.0+2.0*r+s)*(t-1.0) ;
    deriv(0, 9) = -2.0*s*(t+1.0) ;
    deriv(0,10) =  2.0*s*(t+1.0) ;
    deriv(0,11) = -2.0*(-1.0+2.0*r+s)*(t+1.0) ;
    deriv(0,12) =  (t-1.0)*(t+1.0) ;
    deriv(0,13) =  0.0 ;
    deriv(0,14) = -(t-1.0)*(t+1.0);

    deriv(1, 0) = -0.5*(t-2.0+4.0*r+4.0*s)*(t-1.0) ;
    deriv(1, 1) =  0.5*(t-4.0*s+2)*(t-1.0) ;
    deriv(1, 2) =  0.0 ;
    deriv(1, 3) = -0.5*(t+2.0-4.0*r-4.0*s)*(t+1.0) ;
    deriv(1, 4) =  0.5*(t+4.0*s-2.0)*(t+1.0) ;
    deriv(1, 5) =  0.0 ;
    deriv(1, 6) =  2.0*(-1.0+2.0*s+r)*(t-1.0) ;
    deriv(1, 7) = -2.0*r*(t-1.0) ;
    deriv(1, 8) =  2.0*r*(t-1.0) ;
    deriv(1, 9) = -2.0*(-1.0+2.0*s+r)*(t+1.0) ;
    deriv(1,10) =  2.0*r*(t+1.0) ;
    deriv(1,11) = -2.0*r*(t+1.0) ;
    deriv(1,12) =  (t-1.0)*(t+1.0) ;
    deriv(1,13) = -(t-1.0)*(t+1.0) ;
    deriv(1,14) =  0.0 ;

    deriv(2, 0) = -0.5*u*(2.0*u-1.0-2.0*t) ;
    deriv(2, 1) = -0.5*s*(2.0*s-1.0-2.0*t) ;
    deriv(2, 2) = -0.5*r*(2.0*r-1.0-2.0*t) ;
    deriv(2, 3) =  0.5*u*(2.0*u-1.0+2.0*t) ;
    deriv(2, 4) =  0.5*s*(2.0*s-1.0+2.0*t) ;
    deriv(2, 5) =  0.5*r*(2.0*r-1.0+2.0*t) ;
    deriv(2, 6) = -2.0*u*s ;
    deriv(2, 7) = -2.0*s*r ;
    deriv(2, 8) = -2.0*r*u ;
    deriv(2, 9) =  2.0*u*s ;
    deriv(2,10) =  2.0*s*r ;
    deriv(2,11) =  2.0*r*u ;
    deriv(2,12) = -2.0*u*t ;
    deriv(2,13) = -2.0*s*t ;
    deriv(2,14) = -2.0*r*t ;
    return(deriv) ;
}


bool Wedge2IShapeFunc::InsideElement(const Vec3D &natural) const
{
    double u = 1.0 - natural[0] - natural[1] ;
    if ((natural[0] < 0.0) || (natural[0] > 1.0) ||
        (natural[1] < 0.0) || (natural[1] > 1.0) ||
        (natural[2] < -1.0) || (natural[2] > 1.0) ||
        (u < 0.0) || (u > 1.0)) return(false) ;
    return(true) ;
}

// ----------------------------------------------------------------
//  Pyramid2IShapeFunc implementation

Pyramid2IShapeFunc *Pyramid2IShapeFunc::Instance()
{
    if (_instance == 0) {
        _instance = new Pyramid2IShapeFunc ;
    }
    return _instance ;
}

Pyramid2IShapeFunc *Pyramid2IShapeFunc::_instance = 0 ;

int Pyramid2IShapeFunc::face_map[5][9] = {
    { 8, 0, 1, 2,  3,  5,  6,  7,  8 },
    { 6, 0, 4, 1,  9, 10,  5,  0,  0 },
    { 6, 1, 4, 2, 10, 11,  6,  0,  0 },
    { 6, 2, 4, 3, 11, 12,  7,  0,  0 },
    { 6, 3, 4, 0, 12,  9,  8,  0,  0 }} ;

ShapeFunc::SurfaceShape Pyramid2IShapeFunc::face_type_map[5] = {
    QUADRILATERAL,TRIANGLE,TRIANGLE,TRIANGLE,TRIANGLE } ;

TVector<double> Pyramid2IShapeFunc::VolumeShapeFunc(Vec3D loc) const
{
    TVector<double> shape(13) ;
    double r = loc.x() ;
    double s = loc.y() ;
    double t = loc.z() ;
    double r2 = r*r ;
    double s2 = s*s ;
    double t2 = t*t ;
    shape[ 0] = 0.125*(1.0-r)*(1.0-s)*(1.0-t)*(-r-s-t-2.0) ;
    shape[ 1] = 0.125*(1.0+r)*(1.0-s)*(1.0-t)*(r-s-t-2.0) ;
    shape[ 2] = 0.125*(1.0+r)*(1.0-s)*(1.0+t)*(r-s+t-2.0) ;
    shape[ 3] = 0.125*(1.0-r)*(1.0-s)*(1.0+t)*(-r-s+t-2.0) ;
    shape[ 4] = 0.125*(1.0-r)*(1.0+s)*(1.0+t)*(-r+s+t-2.0) +
                0.125*(1.0-r)*(1.0+s)*(1.0-t)*(-r+s-t-2.0) +
                0.125*(1.0+r)*(1.0+s)*(1.0-t)*(r+s-t-2.0) +
                0.125*(1.0+r)*(1.0+s)*(1.0+t)*(r+s+t-2.0) +
                0.250*(1.0-r)*(1.0+s)*(1.0-t2) +
                0.250*(1.0+r)*(1.0+s)*(1.0-t2) +
                0.250*(1.0-r2)*(1.0+s)*(1.0-t) +
                0.250*(1.0-r2)*(1.0+s)*(1.0+t) ;
    shape[ 5] = 0.250*(1.0-r2)*(1.0-s)*(1.0-t) ;
    shape[ 6] = 0.250*(1.0+r)*(1.0-s)*(1.0-t2) ;
    shape[ 7] = 0.250*(1.0-r2)*(1.0-s)*(1.0+t) ;
    shape[ 8] = 0.250*(1.0-r)*(1.0-s)*(1.0-t2) ;
    shape[ 9] = 0.250*(1.0-r)*(1.0-s2)*(1.0-t) ;
    shape[10] = 0.250*(1.0+r)*(1.0-s2)*(1.0-t) ;
    shape[11] = 0.250*(1.0+r)*(1.0-s2)*(1.0+t) ;
    shape[12] = 0.250*(1.0-r)*(1.0-s2)*(1.0+t) ;
    return(shape) ;
}

TMatrix<double> Pyramid2IShapeFunc::VolumeShapeDeriv(Vec3D loc) const
{
    TMatrix<double> deriv(3,13) ;
    double r = loc.x() ;
    double s = loc.y() ;
    double t = loc.z() ;
    double r2 = r*r ;
    double s2 = s*s ;
    double t2 = t*t ;

    deriv(0, 0) = -.125 * (1.0-s) * (1.0-t) * (-2.0*r-s-t-1.0) ;
    deriv(0, 1) = 0.125 * (1.0-s) * (1.0-t) * (2.0*r-s-t-1.0) ;
    deriv(0, 2) = 0.125 * (1.0-s) * (1.0+t) * (2.0*r-s+t-1.0) ;
    deriv(0, 3) = -.125 * (1.0-s) * (1.0+t) * (-2.0*r-s+t-1.0) ;
    deriv(0, 4) = -.125 * (1.0+s) * (1.0-t) * (-2.0*r+s-t-1.0) +
                  -.125 * (1.0+s) * (1.0+t) * (-2.0*r+s+t-1.0) +
                  0.125 * (1.0+s) * (1.0-t) * (2.0*r+s-t-1.0) +
                  0.125 * (1.0+s) * (1.0+t) * (2.0*r+s+t-1.0) +
                  -.250 * (1.0+s) * (1.0-t2) +
                  0.250 * (1.0+s) * (1.0-t2) +
                  -0.50 * r * (1.0+s) * (1.0-t) +
                  -0.50 * r * (1.0+s) * (1.0+t) ;
    deriv(0, 5) = -0.50 * r * (1.0-s) * (1.0-t) ;
    deriv(0, 6) = 0.250 * (1.0-s) * (1.0-t2) ;
    deriv(0, 7) = -0.50 * r * (1.0-s) * (1.0+t) ;
    deriv(0, 8) = -.250 * (1.0-s) * (1.0-t2) ;
    deriv(0, 9) = -.250 * (1.0-s2) * (1.0-t) ;
    deriv(0,10) = 0.250 * (1.0-s2) * (1.0-t) ;
    deriv(0,11) = 0.250 * (1.0-s2) * (1.0+t) ;
    deriv(0,12) = -.250 * (1.0-s2) * (1.0+t) ;

    deriv(1, 0) = -.125 * (1.0-r) * (1.0-t) * (-r-2.0*s-t-1.0) ;
    deriv(1, 1) = -.125 * (1.0+r) * (1.0-t) * (r-2.0*s-t-1.0) ;
    deriv(1, 2) = -.125 * (1.0+r) * (1.0+t) * (r-2.0*s+t-1.0) ;
    deriv(1, 3) = -.125 * (1.0-r) * (1.0+t) * (-r-2.0*s+t-1.0) ;
    deriv(1, 4) = 0.125 * (1.0-r) * (1.0-t) * (-r+2.0*s-t-1.0) +
                  0.125 * (1.0-r) * (1.0+t) * (-r+2.0*s+t-1.0) +
                  0.125 * (1.0+r) * (1.0-t) * (r+2.0*s-t-1.0) +
                  0.125 * (1.0+r) * (1.0+t) * (r+2.0*s+t-1.0) +
                  0.250 * (1.0-r) * (1.0-t2) +
                  0.250 * (1.0+r) * (1.0-t2) +
                  0.250 * (1.0-r2) * (1.0-t) +
                  0.250 * (1.0-r2) * (1.0+t) ;
    deriv(1, 5) = -.250 * (1.0-r2) * (1.0-t) ;
    deriv(1, 6) = -.250 * (1.0+r) * (1.0-t2) ;
    deriv(1, 7) = -.250 * (1.0-r2) * (1.0+t) ;
    deriv(1, 8) = -.250 * (1.0-r) * (1-t2) ;
    deriv(1, 9) = -.500 * (1.0-r) * s * (1.0-t) ;
    deriv(1,10) = -.500 * (1.0+r) * s * (1.0-t) ;
    deriv(1,11) = -.500 * (1.0+r) * s * (1.0+t) ;
    deriv(1,12) = -.500 * (1.0-r) * s * (1.0+t) ;

    deriv(2, 0) = -.125 * (1.0-r) * (1.0-s) * (-r-s-2.0*t-1.0) ;
    deriv(2, 1) = -.125 * (1.0+r) * (1.0-s) * (r-s-2.0*t-1.0) ;
    deriv(2, 2) = 0.125 * (1.0+r) * (1.0-s) * (r-s+2.0*t-1.0) ;
    deriv(2, 3) = 0.125 * (1.0-r) * (1.0-s) * (-r-s+2.0*t-1.0) ;
    deriv(2, 4) = -.125 * (1.0-r) * (1.0+s) * (-r+s-2.0*t-1.0) +
                  0.125 * (1.0-r) * (1.0+s) * (-r+s+2.0*t-1.0) +
                  -.125 * (1.0+r) * (1.0+s) * (r+s-2.0*t-1.0) +
                  0.125 * (1.0+r) * (1.0+s) * (r+s+2.0*t-1.0) +
                  -.500 * (1.0-r) * (1.0+s) * t +
                  -.500 * (1.0+r) * (1.0+s) * t +
                  -.250 * (1.0-r2) * (1.0+s) +
                  0.250 * (1.0-r2) * (1.0+s) ;
    deriv(2, 5) = -.250 * (1.0-r2) * (1.0-s) ;
    deriv(2, 6) = -.500 * (1.0+r) * (1.0-s) * t ;
    deriv(2, 7) = 0.250 * (1.0-r2) * (1.0-s) ;
    deriv(2, 8) = -.500 * (1.0-r) * (1.0-s) * t ;
    deriv(2, 9) = -.250 * (1.0-r) * (1.0-s2) ;
    deriv(2,10) = -.250 * (1.0+r) * (1.0-s2) ;
    deriv(2,11) = 0.250 * (1.0+r) * (1.0-s2) ;
    deriv(2,12) = 0.250 * (1.0-r) * (1.0-s2) ;

    return(deriv) ;
}


bool Pyramid2IShapeFunc::InsideElement(const Vec3D &natural) const
{
    if ((natural[0] < -1.0) || (natural[0] > 1.0) ||
        (natural[1] < -1.0) || (natural[1] > 1.0) ||
        (natural[2] < -1.0) || (natural[2] > 1.0)) return(false) ;
    return(true) ;
}


// ----------------------------------------------------------------
//  Tet1IShapeFunc implementation

Tet1IShapeFunc *Tet1IShapeFunc::Instance()
{
    if (_instance == 0) {
        _instance = new Tet1IShapeFunc ;
    }
    return _instance ;
}

Tet1IShapeFunc *Tet1IShapeFunc::_instance = 0 ;

int Tet1IShapeFunc::face_map[4][4] = {
    { 3, 0, 1, 2 },
    { 3, 1, 3, 2 },
    { 3, 2, 3, 0 },
    { 3, 0, 3, 1 }} ;


TVector<double> Tet1IShapeFunc::VolumeShapeFunc(Vec3D loc) const
{
    TVector<double> shape(10) ;
    shape[0] = loc.x() ;
    shape[1] = loc.y() ;
    shape[2] = loc.z() ;
    shape[3] = 1.0 - loc.x() - loc.y() - loc.z() ;
    return(shape) ;
}

TMatrix<double> Tet1IShapeFunc::VolumeShapeDeriv(Vec3D loc) const
{
    TMatrix<double> deriv(3,10) ;

    deriv(0,0) =  1.0 ;
    deriv(0,1) =  0.0 ;
    deriv(0,2) =  0.0 ;
    deriv(0,3) = -1.0 ;

    deriv(1,0) =  0.0 ;
    deriv(1,1) =  1.0 ;
    deriv(1,2) =  0.0 ;
    deriv(1,3) = -1.0 ;

    deriv(2,0) =  0.0 ;
    deriv(2,1) =  0.0 ;
    deriv(2,2) =  1.0 ;
    deriv(2,3) = -1.0 ;

    return(deriv) ;
}


bool Tet1IShapeFunc::InsideElement(const Vec3D &natural) const
{
    double u = 1.0 - natural[0] - natural[1] - natural[2] ;
    if ((natural[0] < 0.0) || (natural[0] > 1.0) ||
        (natural[1] < 0.0) || (natural[1] > 1.0) ||
        (natural[2] < 0.0) || (natural[2] > 1.0) ||
        (u < 0.0) || (u > 1.0)) return(false) ;
    return(true) ;
}


// ----------------------------------------------------------------
//  ShapeFuncFactory implementation

ShapeFuncFactory *ShapeFuncFactory::Instance()
{
    if (_instance == 0) {
        _instance = new ShapeFuncFactory ;
    }
    return _instance ;
}

ShapeFuncFactory *ShapeFuncFactory::_instance = 0 ;

ShapeFunc *ShapeFuncFactory::CreateShapeFunc(const char *type_name)
{
    if (strcmp(type_name,"Tet2IShapeFunc") == 0) {
        return(Tet2IShapeFunc::Instance()) ;
    } else if (strcmp(type_name,"Brick2IShapeFunc") == 0) {
        return(Brick2IShapeFunc::Instance()) ;
    } else if (strcmp(type_name,"Wedge2IShapeFunc") == 0) {
        return(Wedge2IShapeFunc::Instance()) ;
    } else if (strcmp(type_name,"Pyramid2IShapeFunc") == 0) {
        return(Pyramid2IShapeFunc::Instance()) ;
    } else if (strcmp(type_name,"Tet1IShapeFunc") == 0) {
        return(Tet1IShapeFunc::Instance()) ;
    }

    return(0) ;
}

} // namespace
