//
// Shape function header file
//
// Revision -
//   $Revision: 1.3 $  $Date: 2002/01/31 15:08:30 $  $Author: wash $
//
 
#ifndef ShapeFunc_h
#define ShapeFunc_h

#include "Vec3D.hh"
#include "TVector.hh"
#include "TMatrix.hh"
#include "IntegrationPts.hh"

namespace FemLib {

// This is the abstract base class.  This should be specialized
// for each element family, (e.g. 2nd order isoparametric), and
// for each element shape (e.g., brick, tetrahedral, wedge, pyramid).

class ShapeFunc {
    public:
        enum SurfaceShape {TRIANGLE, QUADRILATERAL} ;
 
        virtual int NumVolumeShape() const = 0 ;

        // returns the number of volumetric shape functions

        virtual int NumSurfaceShape(SurfaceShape shape) const = 0 ;

        // returns the number of shape functions for the given face

        virtual IntegrationPts *GetIntegrationPtsObject() const
                    { return(int_pts) ; } ;

        // returns a pointer to the object that is used for retrieving
        // Gauss integration points and weights.

        virtual TVector<double> VolumeShapeFunc(Vec3D loc) const = 0 ;

        // returns the values of the volumetric shape functions evaluated
        // at the given natural coordinate location.

        virtual TMatrix<double> VolumeShapeDeriv(Vec3D loc) const = 0 ;

        // returns the values of the derivatives of the shape functions
        // (with respect to the natural coordinates) evaluated at the
        // given natural coordinate location.

        virtual TVector<double> SurfaceShapeFunc(SurfaceShape shape,
                                                 Vec3D loc) const = 0 ;

        // returns the values of the surface shape functions for the
        // given face evaluated at the given natural coordinate location
        // (in the face, the 3rd value being ignored).

        virtual TMatrix<double> SurfaceShapeDeriv(SurfaceShape shape,
                                                  Vec3D loc) const = 0 ;

        // returns the values of the derivatives of the surface shape
        // functions for the given face evaluated at the given natural
        // coordinate location (in the face, the 3rd value being ignored).

        virtual const int *GetFaceMap(int face_num) const = 0 ;

        // returns a map (list) of the nodes that make up the given face

        virtual SurfaceShape GetFaceShape(int face_num) const = 0 ;

        // returns the shape of the given face

        virtual int GetDerivDefaultPts() const = 0 ;

        // returns the default number of Gauss points used for derivative
        // matrix integrations

        virtual int GetShapeDefaultPts() const = 0 ;

        // returns the default number of Gauss points used for shape
        // function matrix integrations

        virtual int GetSurfaceDerivDefaultPts(int face_num) const = 0 ;

        // returns the default number of Gauss points used for surface
        // derivative matrix integrations

        virtual int GetSurfaceShapeDefaultPts(int face_num) const = 0 ;

        // returns the default number of Gauss points used for surface
        // shape function matrix integrations.

        virtual bool InsideElement(const Vec3D &natural) const = 0 ;

        // returns true if these natural coordinates are in the element

    protected:
        ShapeFunc() {} ;
        virtual ~ShapeFunc() {} ;

        TVector<double> NullShp ;
        TMatrix<double> NullDrv ;
        IntegrationPts *int_pts ;
} ;
                                                 

// subclass for 2nd order isoparametric elements

class QIShapeFunc : public ShapeFunc {
    public:
        TVector<double> SurfaceShapeFunc(SurfaceShape shape,
                                         Vec3D loc) const ;
        TMatrix<double> SurfaceShapeDeriv(SurfaceShape shape,
                                          Vec3D loc) const ;

        int GetSurfaceDerivDefaultPts(int face_num) const
            { return(GetFaceShape(face_num)==TRIANGLE ? 3 : 4) ; } ;
        int GetSurfaceShapeDefaultPts(int face_num) const
            { return(GetFaceShape(face_num)==TRIANGLE ? 7 : 9) ; } ;

    protected:
        QIShapeFunc() {} ;
        virtual ~QIShapeFunc() {} ;

} ;

// subclass for 1st order isoparametric elements

class LIShapeFunc : public ShapeFunc {
    public:
        TVector<double> SurfaceShapeFunc(SurfaceShape shape,
                                         Vec3D loc) const ;
        TMatrix<double> SurfaceShapeDeriv(SurfaceShape shape,
                                          Vec3D loc) const ;

        int GetSurfaceDerivDefaultPts(int face_num) const
            { return(GetFaceShape(face_num)==TRIANGLE ? 1 : 3) ; } ;
        int GetSurfaceShapeDefaultPts(int face_num) const
            { return(GetFaceShape(face_num)==TRIANGLE ? 1 : 4) ; } ;

    protected:
        LIShapeFunc() {} ;
        virtual ~LIShapeFunc() {} ;

} ;

// subclass for 2nd order isoparametric tetrahedrals.  This is a
// singleton class so things are set up so that only one instance
// is created.

class Tet2IShapeFunc : public QIShapeFunc {
    public:
        static Tet2IShapeFunc *Instance() ;

        int NumVolumeShape() const { return(10) ; } ;
        int NumSurfaceShape(SurfaceShape shape) const
            { return(shape==TRIANGLE ? 6 : 8) ; } ;

        TVector<double> VolumeShapeFunc(Vec3D loc) const ;
        TMatrix<double> VolumeShapeDeriv(Vec3D loc) const ;

        const int *GetFaceMap(int face_num) const
            { return(&face_map[face_num][0]) ; } ;

        SurfaceShape GetFaceShape(int face_num) const
            { return(TRIANGLE) ; } ;

        int GetDerivDefaultPts() const { return(5) ; } ;
        int GetShapeDefaultPts() const { return(15) ; } ;
        int GetSurfDefaultPts(int face) const { return(7) ; } ;

        bool InsideElement(const Vec3D &natural) const ;

    protected:
        Tet2IShapeFunc()
            {int_pts = TetIntegrationPts::Instance() ; } ;
        virtual ~Tet2IShapeFunc() {} ;

    private:
        static Tet2IShapeFunc *_instance ;
        static int face_map[4][7] ;
} ;

// subclass for 2nd order hexahedrals (20 noded bricks).  This is a
// singleton class so things are set up so that only one instance
// is created.

class Brick2IShapeFunc : public QIShapeFunc {
    public:
        static Brick2IShapeFunc *Instance() ;

        int NumVolumeShape() const { return(20) ; } ;
        int NumSurfaceShape(SurfaceShape shape) const
            { return(8) ; } ;

        TVector<double> VolumeShapeFunc(Vec3D loc) const ;
        TMatrix<double> VolumeShapeDeriv(Vec3D loc) const ;

        const int *GetFaceMap(int face_num) const
            { return(&face_map[face_num][0]) ; } ;

        SurfaceShape GetFaceShape(int face_num) const
            { return(QUADRILATERAL) ; } ;

        int GetDerivDefaultPts() const { return(27) ; } ;
        int GetShapeDefaultPts() const { return(27) ; } ;
        int GetSurfDefaultPts(int face) const { return(9) ; } ;

        bool InsideElement(const Vec3D &natural) const ;

    protected:
        Brick2IShapeFunc()
            {int_pts = BrickIntegrationPts::Instance() ; } ;
        virtual ~Brick2IShapeFunc() {} ;

    private:
        static Brick2IShapeFunc *_instance ;
        static int face_map[6][9] ;
} ;

// subclass for 2nd order wedges (15 noded triangular prism). This is a
// singleton class so things are set up so that only one instance
// is created.

class Wedge2IShapeFunc : public QIShapeFunc {
    public:
        static Wedge2IShapeFunc *Instance() ;

        int NumVolumeShape() const { return(15) ; } ;
        int NumSurfaceShape(SurfaceShape shape) const
            { return(shape==TRIANGLE ? 6 : 8) ; } ;

        TVector<double> VolumeShapeFunc(Vec3D loc) const ;
        TMatrix<double> VolumeShapeDeriv(Vec3D loc) const ;

        const int *GetFaceMap(int face_num) const
            { return(&face_map[face_num][0]) ; } ;

        SurfaceShape GetFaceShape(int face_num) const
            { return(face_type_map[face_num]) ; } ;

        int GetDerivDefaultPts() const { return(28) ; } ;
        int GetShapeDefaultPts() const { return(28) ; } ;
        int GetSurfDefaultPts(int face) const
            { return(face_type_map[face] == QUADRILATERAL ? 9 : 7) ; } ;

        bool InsideElement(const Vec3D &natural) const ;

    protected:
        Wedge2IShapeFunc()
            {int_pts = WedgeIntegrationPts::Instance() ; } ;
        virtual ~Wedge2IShapeFunc() {} ;

    private:
        static Wedge2IShapeFunc *_instance ;
        static int face_map[5][9] ;
        static SurfaceShape face_type_map[5] ;
} ;

// subclass for 2nd order pyramids.  This is a
// singleton class so things are set up so that only one instance
// is created.

class Pyramid2IShapeFunc : public QIShapeFunc {
    public:
        static Pyramid2IShapeFunc *Instance() ;

        int NumVolumeShape() const { return(13) ; } ;
        int NumSurfaceShape(SurfaceShape shape) const
            { return(shape==TRIANGLE ? 6 : 8) ; } ;

        TVector<double> VolumeShapeFunc(Vec3D loc) const ;
        TMatrix<double> VolumeShapeDeriv(Vec3D loc) const ;

        const int *GetFaceMap(int face_num) const
            { return(&face_map[face_num][0]) ; } ;

        SurfaceShape GetFaceShape(int face_num) const
            { return(face_type_map[face_num]) ; } ;

        int GetDerivDefaultPts() const { return(27) ; } ;
        int GetShapeDefaultPts() const { return(27) ; } ;
        int GetSurfDefaultPts(int face) const
            { return(face_type_map[face] == QUADRILATERAL ? 9 : 7) ; } ;

        bool InsideElement(const Vec3D &natural) const ;

    protected:
        Pyramid2IShapeFunc()
            {int_pts = BrickIntegrationPts::Instance() ; } ;
        virtual ~Pyramid2IShapeFunc() {} ;

    private:
        static Pyramid2IShapeFunc *_instance ;
        static int face_map[5][9] ;
        static SurfaceShape face_type_map[5] ;
} ;

// subclass for 1nd order isoparametric tetrahedrals.  This is a
// singleton class so things are set up so that only one instance
// is created.

class Tet1IShapeFunc : public LIShapeFunc {
    public:
        static Tet1IShapeFunc *Instance() ;

        int NumVolumeShape() const { return(10) ; } ;
        int NumSurfaceShape(SurfaceShape shape) const
            { return(3) ; } ;

        TVector<double> VolumeShapeFunc(Vec3D loc) const ;
        TMatrix<double> VolumeShapeDeriv(Vec3D loc) const ;

        const int *GetFaceMap(int face_num) const
            { return(&face_map[face_num][0]) ; } ;

        SurfaceShape GetFaceShape(int face_num) const
            { return(TRIANGLE) ; } ;

        int GetDerivDefaultPts() const { return(3) ; } ;
        int GetShapeDefaultPts() const { return(4) ; } ;
        int GetSurfDefaultPts(int face) const { return(5) ; } ;

        bool InsideElement(const Vec3D &natural) const ;

    protected:
        Tet1IShapeFunc()
            {int_pts = TetIntegrationPts::Instance() ; } ;
        virtual ~Tet1IShapeFunc() {} ;

    private:
        static Tet1IShapeFunc *_instance ;
        static int face_map[4][4] ;
} ;

//------------------------------------------------------------------
// This is a factory for shape function classes.  This is a
// singleton class so things are set up so that only one instance
// is created.

class ShapeFuncFactory {
    public:
        static ShapeFuncFactory *Instance() ;

        ShapeFunc *CreateShapeFunc(const char *type_name) ;

    protected:
        ShapeFuncFactory() {} ;

    private:
        static ShapeFuncFactory *_instance ;
} ;

} // namespace

#endif
