
#ifndef SparseCG_h
#define SparseCG_h

#define PY_NUMERIC 1

#include "TMatrix.hh"
#include "TVector.hh"
#include "Vec3D.hh"

#ifdef PY_NUMERIC
#include "Python.h"
#include "Numeric/arrayobject.h"
#endif

using FemLib::Vec3D ;
using FemLib::TMatrix ;
using FemLib::TVector ;

#define REC_CACHE_BLOCK_SIZE 1000
#define FIXED_CACHE_BLOCK_SIZE 500

class SparseCG {
    public:

        SparseCG(int num_nodes,int start=0) ;
        virtual ~SparseCG() ;

        virtual void ElementEquations(int num_nodes,int *nodes) ;
        virtual void ElementValues(int num_nodes,int *nodes,
                                   const TMatrix<double> &stiff) ;
        virtual void RhsValues(int num_nodes,int *nodes,
                               const double *contrib_rhs) ;
#ifdef PY_NUMERIC
        virtual void ElementValues(int num_nodes,int *nodes,
                                   const PyArrayObject *stiff) ;
        virtual void RhsValues(int num_nodes,int *nodes,
                               const PyArrayObject *contrib_rhs) ;
#endif
        virtual void AddPrescribedValue(int node,double value) ;
        virtual double *DoSolution() ;
        virtual double *DoGaussSeidel() ;
        virtual double *MatrixVectorMultiply() ;
        void SetTolerance(double tol) { Tolerance = tol ; } ;
        int GetNumEqns() { return(NumEqns) ; } ;
        int GetNumNodes() { return(NumNodes) ; } ;
        virtual double GetEntry(int row,int col) ;
        virtual void ZeroRhs() ;

	virtual void MatrixVectMult(double *b,double *x) ;
    protected:

        struct CoefRec {
            int    column ;         // start column for this block
            int    offset ;         // offset into coeff data for the
                                    // start of this block information
            struct CoefRec *next ;  // linked list pointer
        } ;

        struct CoefData {
            double  *A ;            // A matrix coefficients
            CoefRec **rows ;        // heads of the row linked lists
            int     num_rows ;
        } ;

        struct CoefRecCache {
            struct CoefRecCache *next ;
            struct CoefRec entries[REC_CACHE_BLOCK_SIZE] ;
        } ;

        struct FixedRec {           // here is a bit of questionable
            int    node ;           // OO design, but we have set
            int    flags ;          // this structure up for use by
            double values[3] ;      // the elasticity subclass
        } ;

        struct FixedRecCache {
            struct FixedRecCache *next ;
            struct FixedRec entries[FIXED_CACHE_BLOCK_SIZE] ;
        } ;

        int NumNodes ;
        int NumEqns ;
        CoefData *CData ;
        double *x ;
        double *b ;
        double *r ;
        double *z ;
        double *p ;
        double *Ap ;
        double *Binv ;
        double Tolerance ;
        int cur_offset ;
        int print_start ;

        bool CoeffsAlloced ;
        bool PrescribedApplied ;

        CoefRecCache *RecCacheList ;
        int RecCachePtr ;

        FixedRecCache *FixedCacheList ;
        int FixedCachePtr ;

        SparseCG() {} ;
        SparseCG(const SparseCG &other) {} ;
        SparseCG operator = (const SparseCG &other) { return(*this) ;} ;

        virtual bool CGSolve() ;
        virtual bool GaussSeidelSolve() ;

        virtual void AllocateSolveData() ;
        virtual void AllocateCoeffs() ;
        virtual void FreeSolveData() ;
        virtual void SolveSetup() ;
        virtual void ApplyPrescribedValue() ;
        virtual CoefRec *NewRec() ;
        virtual FixedRec *NewFixed() ;

        virtual void PrintMatrix() ;
} ;

// elasticity version of the solver

class SparseCGElasticity : public SparseCG {
    public:

        SparseCGElasticity(int num_nodes,int start=0) :
            SparseCG(num_nodes,start) {NumEqns = 3*NumNodes ; } ;

        void ElementValues(int num_nodes,int *nodes,
                           const TMatrix<double> &stiff) ;
        void RhsValues(int num_nodes,int *nodes,
                       const TVector<Vec3D> &contrib_rhs) ;
#ifdef PY_NUMERIC
        void ElementValues(int num_nodes,int *nodes,
                           const PyArrayObject *stiff) ;
        void RhsValues(int num_nodes,int *nodes,
                       const PyArrayObject *contrib_rhs) ;
#endif
        void AddPrescribedValue(int node,bool x,bool y,
                                bool z,double values[3]) ;
        double GetEntry(int row,int col) ;
        void MatrixVectMult(double *b,double *x) ;
    private:

        SparseCGElasticity(const SparseCGElasticity &other) {} ;
        SparseCGElasticity operator = (const SparseCGElasticity &other) { return(*this) ;} ;


        void AllocateCoeffs() ;
        void SolveSetup() ;
        void ApplyPrescribedValue() ;
        bool GaussSeidelSolve() ;

        void PrintMatrix() ;
} ;

#endif
