#!/usr/local/bin/python

"""Unit test for the SparseCG solver"""

import SparseCG
import unittest
import zlib
import base64
import string
import StringIO
import Numeric

# Here we strore the stiffness matrix for a 20 noded brick element
# in zipped and base64 format.

brick_stiffness = """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==
"""

# -------------------------------------------------------------------

# define an class that makes a zlib compressed string look like
# a file.  This is based on Fredrik Lundh's class in the Python
# Standar Library book

class CompressedStream:
    """A class that makes a zlib compressed string look like a file

    This is based on Fredrik Lundh's class in the Python standard
    library book"""

    def __init__(self,data):
        self.file = StringIO.StringIO(data)
        self._rewind()

    def _rewind(self):
        self.zip = zlib.decompressobj()
        self.pos = 0
        self.offset = 0
        self.data = ""

    def _fill(self,bytes):
        if self.zip:
            # read until we have enough bytes in the buffer
            while not bytes or len(self.data) < bytes:
                self.file.seek(self.pos)
                data = self.file.read(16384)
                if not data:
                    self.data = self.data + self.zip.flush()
                    self.zip = None # no more data
                    break
                self.pos = self.pos + len(data)
                self.data = self.data + self.zip.decompress(data)

    def read(self,bytes=0):
        self._fill(bytes)
        if bytes:
            data = self.data[:bytes]
            self.data = self.data[bytes:]
        else:
            data = self.data
            self.data = ""
        self.offset = self.offset + len(data)
        return data

    def readline(self):
        # make sure we have an entire line
        while self.zip and "\n" not in self.data:
            self._fill(len(self.data) + 512)
        i = string.find(self.data,"\n") + 1
        if i <= 0:
            return self.read()
        return self.read(i)

    def readlines(self):
        lines = []
        while 1:
            s = self.readline()
            if not s:
                break
            lines.append(s)
        return lines

def GetStoredMat(mat,rows,cols):
    """This routine returns the zipped matrix as a NumPy matrix"""

    smat = {}
    zdata =  base64.decodestring(mat)
    fd = CompressedStream(zdata) 
    svals = string.splitfields(fd.readline())
    smat = Numeric.zeros([rows,cols],Numeric.Float64) 
    for i in xrange(int(svals[0])):
        svals = string.splitfields(fd.readline())
        row = int(svals[0])
        col = int(svals[1])
        value = float(svals[2])
        smat[row,col] = value
        if row != col: smat[col,row] = value
    return smat

# -------------------------------------------------------------------

class TwoElemUniaxialTension(unittest.TestCase):
    def assertEqualFloat(self,x,y,tol):
        self.failUnless(abs(x-y) < tol)

    def testTwoElemUniaxialTension(self):
        """two element, tension load"""

        # element connectivity

        conn0 = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
                 11, 12, 13, 14, 15, 16, 17, 18, 19)
        conn1 = (4, 5, 6, 7, 20, 21, 22, 23, 10, 11,
                24, 25, 14, 15, 26, 27, 28, 29, 30, 31)

        # loads (index, value)

        load = ((4,-0.333333),(5,-0.333333),(6,-0.333333),
                (7,-0.333333),(10,1.333333),(11,1.333333),
                (14,1.333333),(15,1.333333))
 
        # prescribed nodal displacements

        pdisp = ((0,(0.0,0.0,0.0)),(1,(0.0,0.0,None)),
                 (2,(0.0,None,0.0)),(3,(0.0,None,None)),
                 (4,(None,0.0,0.0)),(5,(None,0.0,None)),
                 (6,(None,None,0.0)),(8,(0.0,0.0,None)),
                 (9,(0.0,None,None)),(10,(None,0.0,None)),
                 (12,(0.0,None,0.0)),(13,(0.0,None,None)),
                 (14,(None,None,0.0)),(16,(None,0.0,0.0)),
                 (17,(None,0.0,None)),(18,(None,None,0.0)),
                 (20,(None,0.0,0.0)),(21,(None,0.0,None)),
                 (22,(None,None,0.0)),(24,(None,0.0,None)),
                 (26,(None,None,0.0)),(28,(None,0.0,0.0)),
                 (29,(None,0.0,None)),(30,(None,None,0.0))) 

        # expected results

        sdisp = ((0.00,0.00,0.00),(0.00,0.00,-0.06),(0.00,-0.06,0.00),
                 (0.00,-0.06,-0.06),(0.20,0.00,0.00),(0.20,0.00,-0.06),
                 (0.20,-0.06,0.00),(0.20,-0.06,-0.06),(0.00,0.00,-0.03),
                 (0.00,-0.06,-0.03),(0.20,0.00,-0.03),(0.20,-0.06,-0.03),
                 (0.00,-0.03,0.00),(0.00,-0.03,-0.06),(0.20,-0.03,0.00),
                 (0.20,-0.03,-0.06),(0.10,0.00,0.00),(0.10,0.00,-0.06),
                 (0.10,-0.06,0.00),(0.10,-0.06,-0.06),(0.40,0.00,0.00),
                 (0.40,0.00,-0.06),(0.40,-0.06,0.00),(0.40,-0.06,-0.06),
                 (0.40,0.00,-0.03),(0.40,-0.06,-0.03),(0.40,-0.03,0.00),
                 (0.40,-0.03,-0.06),(0.30,0.00,0.00),(0.30,0.00,-0.06),
                 (0.30,-0.06,0.00),(0.30,-0.06,-0.06))

        # initialize the sovler class

        solver = SparseCG.SparseCGElasticity(32)

        # get the stored stiffness matrix and add to the solver

        k = GetStoredMat(brick_stiffness,60,60)
        solver.AddElementEquations(len(conn0),conn0)
        solver.AddElementEquations(len(conn1),conn1)
        solver.AddElementValues(len(conn0),conn0,k)
        solver.AddElementValues(len(conn1),conn1,k)

        # build a loads matrix and add it

        lmat = Numeric.zeros([20,3],Numeric.Float64) 
        for indx,value in load: lmat[indx,0] = value
        solver.AddRhsValues(20,conn1,lmat)

        # do the prescribed displacements

        values = [0.0,0.0,0.0]
        for node,svalues in pdisp:
            if svalues[0] == None:
                have_x = 0
                values[0] = 0.0
            else:
                have_x = 1
                values[0] = svalues[0]
            if svalues[1] == None:
                have_y = 0
                values[1] = 0.0
            else:
                have_y = 1
                values[1] = svalues[1]
            if svalues[2] == None:
                have_z = 0
                values[2] = 0.0
            else:
                have_z = 1
                values[2] = svalues[2]
            solver.AddPrescribedValue(node,have_x,have_y,have_z,values)

        # solve for the displacements and check to see that
        # they are what we expect

        tol = 0.001
        disp = solver.DoSolution()
        for i in xrange(32):
            for j in xrange(3):
                self.assertEqualFloat(disp[i,j],sdisp[i][j],tol)

        # zero out the right hand side and add back the negative
        # of the loads

        solver.ZeroRhs()
        lmat = Numeric.zeros([20,3],Numeric.Float64) 
        for indx,value in load: lmat[indx,0] = -value
        solver.AddRhsValues(20,conn1,lmat)

        # solve and check again

        disp = solver.DoSolution()
        for i in xrange(32):
            for j in xrange(3):
                self.assertEqualFloat(disp[i,j],-sdisp[i][j],tol)

    def testTwoElemUniaxialTensionGS(self):
        """two element, tension load"""

        # element connectivity

        conn0 = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
                 11, 12, 13, 14, 15, 16, 17, 18, 19)
        conn1 = (4, 5, 6, 7, 20, 21, 22, 23, 10, 11,
                24, 25, 14, 15, 26, 27, 28, 29, 30, 31)

        # loads (index, value)

        load = ((4,-0.333333),(5,-0.333333),(6,-0.333333),
                (7,-0.333333),(10,1.333333),(11,1.333333),
                (14,1.333333),(15,1.333333))
 
        # prescribed nodal displacements

        pdisp = ((0,(0.0,0.0,0.0)),(1,(0.0,0.0,None)),
                 (2,(0.0,None,0.0)),(3,(0.0,None,None)),
                 (4,(None,0.0,0.0)),(5,(None,0.0,None)),
                 (6,(None,None,0.0)),(8,(0.0,0.0,None)),
                 (9,(0.0,None,None)),(10,(None,0.0,None)),
                 (12,(0.0,None,0.0)),(13,(0.0,None,None)),
                 (14,(None,None,0.0)),(16,(None,0.0,0.0)),
                 (17,(None,0.0,None)),(18,(None,None,0.0)),
                 (20,(None,0.0,0.0)),(21,(None,0.0,None)),
                 (22,(None,None,0.0)),(24,(None,0.0,None)),
                 (26,(None,None,0.0)),(28,(None,0.0,0.0)),
                 (29,(None,0.0,None)),(30,(None,None,0.0))) 

        # expected results

        sdisp = ((0.00,0.00,0.00),(0.00,0.00,-0.06),(0.00,-0.06,0.00),
                 (0.00,-0.06,-0.06),(0.20,0.00,0.00),(0.20,0.00,-0.06),
                 (0.20,-0.06,0.00),(0.20,-0.06,-0.06),(0.00,0.00,-0.03),
                 (0.00,-0.06,-0.03),(0.20,0.00,-0.03),(0.20,-0.06,-0.03),
                 (0.00,-0.03,0.00),(0.00,-0.03,-0.06),(0.20,-0.03,0.00),
                 (0.20,-0.03,-0.06),(0.10,0.00,0.00),(0.10,0.00,-0.06),
                 (0.10,-0.06,0.00),(0.10,-0.06,-0.06),(0.40,0.00,0.00),
                 (0.40,0.00,-0.06),(0.40,-0.06,0.00),(0.40,-0.06,-0.06),
                 (0.40,0.00,-0.03),(0.40,-0.06,-0.03),(0.40,-0.03,0.00),
                 (0.40,-0.03,-0.06),(0.30,0.00,0.00),(0.30,0.00,-0.06),
                 (0.30,-0.06,0.00),(0.30,-0.06,-0.06))

        # initialize the sovler class

#        solver = SparseCG.SparseCGElasticity(32)

        # get the stored stiffness matrix and add to the solver

#        k = GetStoredMat(brick_stiffness,60,60)
#        solver.AddElementEquations(len(conn0),conn0)
#        solver.AddElementEquations(len(conn1),conn1)
#        solver.AddElementValues(len(conn0),conn0,k)
#        solver.AddElementValues(len(conn1),conn1,k)

        # build a loads matrix and add it

#        lmat = Numeric.zeros([20,3],Numeric.Float64) 
#        for indx,value in load: lmat[indx,0] = value
#        solver.AddRhsValues(20,conn1,lmat)

        # do the prescribed displacements

#        values = [0.0,0.0,0.0]
#        for node,svalues in pdisp:
#            if svalues[0] == None:
#                have_x = 0
#                values[0] = 0.0
#            else:
#                have_x = 1
#                values[0] = svalues[0]
#            if svalues[1] == None:
#                have_y = 0
#                values[1] = 0.0
#            else:
#                have_y = 1
#                values[1] = svalues[1]
#            if svalues[2] == None:
#                have_z = 0
#                values[2] = 0.0
#            else:
#                have_z = 1
#                values[2] = svalues[2]
#            solver.AddPrescribedValue(node,have_x,have_y,have_z,values)

        # solve for the displacements and check to see that
        # they are what we expect

#        tol = 0.001
#        disp = solver.DoGaussSeidel()
#        print disp
#        for i in xrange(32):
#            for j in xrange(3):
#                self.assertEqualFloat(disp[i,j],sdisp[i][j],tol)

        # zero out the right hand side and add back the negative
        # of the loads

#        solver.ZeroRhs()
#        lmat = Numeric.zeros([20,3],Numeric.Float64) 
#        for indx,value in load: lmat[indx,0] = -value
#        solver.AddRhsValues(20,conn1,lmat)

        # solve and check again

#        disp = solver.DoGaussSeidel()
#        for i in xrange(32):
#            for j in xrange(3):
#                self.assertEqualFloat(disp[i,j],-sdisp[i][j],tol)

    def testScalarUniaxialTensionGS(self):
        """two element, tension load"""

        # element connectivity

        conn = ((0, 1),(1, 2),(2,3))

        # loads (index, value)

        load = (0, 0, 0, 3.0)
 
        # expected results

        sdisp = (0.0, 0.5, 3.0, 4.5)

        # initialize the sovler class

        solver = SparseCG.SparseCGScalar(4)

        # add the stiffness matrix to the solver

        k = Numeric.zeros([2,2],Numeric.Float64) 
        k[0,0] = 2.0
        k[0,1] = -2.0
        k[1,0] = -2.0
        k[1,1] = 2.0

        for elem in conn:
            solver.AddElementEquations(len(elem),elem)
        
        for elem in conn:
            solver.AddElementValues(len(elem),elem,k)

        # add the load vector

        loads = Numeric.zeros([2],Numeric.Float64) 
        loads[1] = 3.0
        solver.AddRhsValues(2,conn[2],loads)

        # do the prescribed displacements

        solver.AddPrescribedValue(0,0.0)

        # solve for the displacements and check to see that
        # they are what we expect

#        tol = 0.001
#        disp = solver.DoSolution()
        disp = solver.DoGaussSeidel()
        print disp
#        for i in xrange(32):
#            for j in xrange(3):
#                self.assertEqualFloat(disp[i,j],sdisp[i][j],tol)

        # zero out the right hand side and add back the negative
        # of the loads

#        solver.ZeroRhs()
#        lmat = Numeric.zeros([20,3],Numeric.Float64) 
#        for indx,value in load: lmat[indx,0] = -value
#        solver.AddRhsValues(20,conn1,lmat)

        # solve and check again

#        disp = solver.DoGaussSeidel()
#        for i in xrange(32):
#            for j in xrange(3):
#                self.assertEqualFloat(disp[i,j],-sdisp[i][j],tol)


# -------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()
