#!/usr/local/bin/python

import sys
import string
import FemLib
import DMatrix
import SparseCG

from Vec3D import Vec3D
from Numeric import *
from LinearAlgebra import *

class SimpleFemModel:
    """The PyFemModel class acts as a container to hold a description
       of a mesh and associated elements and boundary conditions."""

    # constructor that reads a model description from a file

    def __init__(self,file):
        """Read the model description file:

           file - input file descriptor object
        """
    
        # get number of nodes, mats, and elems

        vals = string.splitfields(self.GetLine(file))
        self.num_nodes = int(vals[0])
        self.num_mats  = int(vals[1])
        self.num_elems = int(vals[2])

        # read the node coordinates

        self.nodes = {}
        for i in xrange(self.num_nodes):
            vals = string.splitfields(self.GetLine(file))
            self.nodes[int(vals[0])] = (
                float(vals[1]),float(vals[2]),float(vals[3]))

        # read the material properties

        self.mat_props = []
        for i in xrange(self.num_mats):
            tmp = {}
            vals = string.splitfields(self.GetLine(file))
            if len(vals) == 3:
                tmp['E']    = float(vals[1])
                tmp['nu']   = float(vals[2])
                self.mat_props.append((int(vals[0]),'Iso',tmp))
            elif len(vals) == 13:
                tmp['c11'] = float(vals[1])
                tmp['c22'] = float(vals[2])
                tmp['c33'] = float(vals[3])
                tmp['c12'] = float(vals[4])
                tmp['c23'] = float(vals[5])
                tmp['c13'] = float(vals[6])
                tmp['c44'] = float(vals[7])
                tmp['c55'] = float(vals[8])
                tmp['c66'] = float(vals[9])
                tmp['angle_z']  = float(vals[10])
                tmp['angle_x']  = float(vals[11])
                tmp['angle_y']  = float(vals[12])
                self.mat_props.append((int(vals[0]),'Ortho',tmp))

        # the element data

        self.elem = []
        for i in xrange(self.num_elems):
            vals = string.splitfields(self.GetLine(file))
            id   = int(vals[0])
            type = vals[1]
            mat  = int(vals[2])
            conn = []
            for node in vals[3:]: conn.append(int(node))
            self.elem.append((id,type,mat,conn))      
 
        # now read the boundary conditions

        self.disp_bc = {}
        self.tract_bc = {}
        self.iforce_bc = []

        buff = self.GetLine(file)
        while buff != None:
            vals = string.splitfields(buff)

            bc_type = vals[0]
            bc_num = int(vals[1])

            if bc_type == 'disp':
                for i in xrange(bc_num):
                    vals = string.splitfields(self.GetLine(file))
                    bc = self.disp_bc.get(int(vals[0]))
                    if bc == None:
                        bc = [None,None,None]
                        self.disp_bc[int(vals[0])] = bc
                    if vals[1] == 'x':
                        bc[0] = float(vals[2])
                    elif vals[1] == 'y':
                        bc[1] = float(vals[2])
                    elif vals[1] == 'z':
                        bc[2] = float(vals[2])

            elif bc_type == 'tract':
                for i in xrange(bc_num):
                    vals = string.splitfields(self.GetLine(file))
                    bc_list = self.tract_bc.get(int(vals[0]))
                    if bc_list == None:
                        bc_list = []
                        self.tract_bc[int(vals[0])] = bc_list
                    bc_list.append((int(vals[1]),int(vals[2]),
                        float(vals[3]),float(vals[4]),float(vals[5])))

            elif bc_type == 'iforce':
                for i in xrange(bc_num):
                    vals = string.splitfields(self.GetLine(file))
                    self.iforce_bc.append((
                      Vec3D(float(vals[0]),float(vals[1]),float(vals[2])),
                      Vec3D(float(vals[3]),float(vals[4]),float(vals[5]))))

            buff = self.GetLine(file)

    def __repr__(self):
        list = ["[ FEM Model:",""]
        list.append("  Number of Nodes:     %d" % self.num_nodes)
        list.append("  Number of Materials: %d" % self.num_mats)
        list.append("  Number of Elements:  %d" % self.num_elems)
        list.append("")

        list.append("  Nodes:")
        items = self.nodes.items()
        items.sort()
        for item in items:
            list.append("  %4d  %11.4g  %11.4g  %11.4g" %
                        (item[0],item[1][0],item[1][1],item[1][2]))
        list.append("")

        list.append("  Elements:")
        for item in self.elem:
            tmp = ["  %4d  %s %4d  " % (item[0],item[1],item[2])]
            for nd in item[3]:
                tmp.append(" %4d" % nd)
            list.append(string.joinfields(tmp))
        list.append("")

        list.append("  Materials:")
        for item in self.mat_props:
            tmp = "    %2d %s" % (item[0],item[1])
            tmp = tmp + str(item[1])
            list.append(tmp)
        list.append("")

        list.append("  Displacement Boundary Conditions:")
        if len(self.disp_bc) == 0:
            list.append("     None")
        else:
            items = self.disp_bc.items()
            items.sort()
            for item in items:
                tmp = ["    %4d" % item[0]]
                for i in range(3):
                    if item[1][i] == None:
                        tmp.append("        None")
                    else:
                        tmp.append(" %11.4g" % item[1][i])
                list.append(string.joinfields(tmp,""))

        list.append("")

        list.append("  Traction Boundary Conditions:")
        if len(self.tract_bc) == 0:
            list.append("     None")
        else:
            items = self.tract_bc.items()
            items.sort()
            for item in items:
                for i in item[1]:
                    list.append("  %4d  %1d  %1d  %11.4g %11.4g %11.4g" %
                            (item[0],i[0],i[1],i[2],i[3],i[4]))
        list.append("")

        list.append("]")
        return string.joinfields(list,"\n")

    def GetLine(self,file):
        """this is a little helper function that scans the input
           file and ignores lines starting with a '#'"""
        buff = file.readline()
        if len(buff) == 0: return None
        while buff[0] == '#':
            buff = file.readline()
            if len(buff) == 0: return None
        return buff

    def GetElemCoords(self,i):
        """Return a list of nodal coordinates for element i"""
        coords = []
        for node in self.elem[i][3]:
            vals = self.nodes[node]
            coords.append(Vec3D(vals[0],vals[1],vals[2]))
        return coords

    def NumNodes(self):
        return self.num_nodes

    def NumElems(self):
        return self.num_elems

    def NumMats(self):
        return self.num_mats

    def GetElemData(self,i):
        return self.elem[i]

    def GetMatData(self,i):
        return self.mat_props[i]

    def GetNodeIds(self):
        return self.nodes.keys()

    def GetDispBCs(self):
        return self.disp_bc

    def GetTractBCs(self):
        return self.tract_bc

def BuildEqnVect(conn,eqn_nums):
    """This routine builds an equation vector for an element"""
    eqns = []
    for node in conn:
        node_eqns = eqn_nums[node]
        eqns.append(node_eqns[0])
        eqns.append(node_eqns[1])
        eqns.append(node_eqns[2])
    return eqns

if __name__ == '__main__':

    verbose = None
    if "-v" in sys.argv: verbose = 1

    # read the input file

    fd = open(sys.argv[-1], 'r')
    model = SimpleFemModel(fd)
    if verbose: print model

    # set up the linear system

    solver = SparseCG.SparseCGElasticity(model.NumNodes())

    for i in xrange(model.NumElems()):
        id,type,mat,conn = model.GetElemData(i)
        solver.AddElementEquations(len(conn),conn)

    # assign equation numbers to all the nodes

    node_ids = model.GetNodeIds()
    eqn_nums = {}
    cur = 0
    for node in node_ids:
        eqn_nums[node] = (cur,cur+1,cur+2)
        cur += 3
    #if verbose == 1:
    #    items = eqn_nums.items()
    #    items.sort()
    #    print
    #    print "Equation Numbers:"
    #    for item in items:
    #        print " %d   %d %d %d" % (item[0],item[1][0],
    #                                  item[1][1],item[1][2])
        
    # build the material matricies

    mats = {}
    for i in xrange(model.NumMats()):
        id,type,props = model.GetMatData(i)
        if type == 'Ortho':
            mats[id] = DMatrix.OrthotropicMatMatrixFromStiffnessComps(props)
        else:
            mats[id] = DMatrix.IsotropicMatMatrix(props)

    # loop through all the elements

    tract_bc = model.GetTractBCs()
    disp_bc = model.GetDispBCs()

    elem_hash = {}
    for i in xrange(model.NumElems()):
        id,type,mat,conn = model.GetElemData(i)

        # get the formulator object for this element type

        ef = elem_hash.get(type)
        if ef == None:
            ef = FemLib.Element(type+"ShapeFunc")
            elem_hash[type] = ef

        # get the element's stiffness matrix

        coords = model.GetElemCoords(i)
        k = ef.DeformationStiffMtx(coords,mats[mat])
        solver.AddElementValues(len(conn),conn,k)

        # check for traction boundary conditions and add the
        # associated contributions into the rhs

        tract = None
        bc_list = tract_bc.get(id)
        if bc_list != None:
            for bc in bc_list:
                tract = ef.IntegrateSurfaceVector(bc[0],
                              coords,Vec3D(bc[2],bc[3],bc[4]))

            if tract:
                solver.AddRhsValues(len(conn),conn,tract)

        # check for associated displacement boundary conditions
        # and add a contribution to the rhs

        found = 0
        d = zeros((len(conn)*3),Float64)
        for i in xrange(len(conn)):
            disp = disp_bc.get(conn[i])
            if disp != None:
                found = 1
                for j in range(3):
                    if disp[j] != None:
                        d[i*3+j] = disp[j]

        if found == 1:
            e = -dot(k,d)
            solver.AddRhsValues(len(conn),conn,e)

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    #
    # Here we compute contributions to the RHS due to internal
    # point forces

    for (pt_coord,value) in model.iforce_bc:

        # first determine which element contains this point

        nat_coord = None
        for i in xrange(model.NumElems()):
            id,type,mat,conn = model.GetElemData(i)
            ef = elem_hash.get(type)
            coords = model.GetElemCoords(i)
            nat_coord = ef.FindNaturalCoords(pt_coord,coords)
            if nat_coord: break

        if not nat_coord:
            sys.stderr.write(
                "Cannot find element that contains point : %s\n" %
                str(pt_coord))
        else:
            forces = ef.InternalPointForces([nat_coord],[value])
            AddElementTract(forces,conn,b,eqn_nums)

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # Now we enforce deformation boundary conditions.  To do this
    # we zero out the associated rows and colums (with a 1 on the
    # diagonal, and place the displacement value in the rhs

    disp_bc = model.GetDispBCs()
    for item in disp_bc.items():
        eqns = eqn_nums[item[0]]
        vals = item[1]
        dvals = []
        flags = []
        for i in vals:
            if i == None:
                dvals.append(0.0)
                flags.append(0)
            else:
                dvals.append(i)
                flags.append(1)

        solver.AddPrescribedValue(item[0],flags[0],flags[1],flags[2],dvals)

    # solve the the linear system of equations

    x = solver.DoSolution()

#    for i in xrange(96):
#        for j in xrange(96):
#            val = solver.GetEntry(i,j)
#            print "%2d %2d %6.3f" % (i,j,val)

    # print the nodal displacements

    print
    print "Nodal Displacements: "
    print
    node_ids.sort()
    for node in node_ids:
        eqs = eqn_nums[node]
        print " %d  %11.4g %11.4g %11.4g" % (node,x[eqs[0]],
                                             x[eqs[1]],x[eqs[2]])

    #!!!!!!!!!!!!!!!!!!!!!!!!!!

    # Find the displacements at the force points

    for (pt_coord,value) in model.iforce_bc:

        # first determine which element contains this point

        nat_coord = None
        for i in xrange(model.NumElems()):
            id,type,mat,conn = model.GetElemData(i)
            ef = elem_hash.get(type)
            coords = model.GetElemCoords(i)
            nat_coord = ef.FindNaturalCoords(pt_coord,coords)
            if nat_coord: break

        if not nat_coord:
            sys.stderr.write(
                "Cannot find element that contains point : %s\n" %
                str(pt_coord))
        else:
            disp = []
            for i in conn:
                eqs = eqn_nums[i]
                disp.append(Vec3D(x[eqs[0]],x[eqs[1]],x[eqs[2]]))

            locdisp = ef.InterpolateVolumeVector(nat_coord,disp)
            print
            print "Local disp : %g %g %g" % (locdisp[0],locdisp[1],locdisp[2])


