
//
// TMatrix Templated Matrix class
//
// Revision -
//   $Revision: 1.1.1.1 $  $Date: 2001/12/02 16:53:08 $  $Author: wash $
//

#ifndef TMatrix_hh
#define TMatrix_hh

#include <cassert>
#include <iostream>

namespace FemLib {

// Class TMatrix

template<class Dtype> class TMatrix {
    public:

        // constructors and destructors

        TMatrix() { rows = cols = 0 ; data = 0 ; } ;

        TMatrix(const int irows, const int icols)
        {
            rows = irows ;
            cols = icols ;
            data = new Dtype[rows*cols] ;
            for (int i=0 ; i<rows*cols ; ++i) data[i] = 0 ;
        } ;

        TMatrix(const TMatrix &other)
        {
            rows = other.rows ;
            cols = other.cols ;
            data = new Dtype[rows*cols] ;
            for (int i=0 ; i<rows*cols ; ++i) data[i] = other.data[i] ;
        } ;

        TMatrix operator = (const TMatrix &other)
        {
            if (data != 0) delete [] data ;
            rows = other.rows ;
            cols = other.cols ;
            data = new double[rows*cols] ;
            for (int i=0 ; i<rows*cols ; i++) data[i] = other.data[i] ;
            return(*this) ;
        } ;

        ~TMatrix() { if (data != 0) delete [] data ; } ;

        void Configure(const int irows, const int icols)
        {
            if (data != 0) delete [] data ;
            rows = irows ;
            cols = icols ;
            data = new Dtype[rows*cols] ;
            for (int i=0 ; i<rows*cols ; ++i) data[i] = 0 ;
        } ;

        // access functions and operators

        int Rows() const { return(rows) ; } ;
        int Cols() const { return(cols) ; } ;

        Dtype &operator() (int i,int j)
        {
            assert((i<rows) && (j<cols)) ;
            return(data[i*cols+j]) ;
        } ;

        Dtype operator() (int i,int j) const
        {
            assert((i<rows) && (j<cols)) ;
            return(data[i*cols+j]) ;
        } ;

        Dtype &at(int i,int j)
        {
            assert((i<rows) && (j<cols)) ;
            return(data[i*cols+j]) ;
        } ;

        Dtype at(int i,int j) const
        {
            assert((i<rows) && (j<cols)) ;
            return(data[i*cols+j]) ;
        } ;

        // computational operators

        TMatrix<Dtype> &operator += (const Dtype &oper)
        {
            for (int i=0 ; i<rows*cols ; ++i) data[i] += oper ;
            return(*this) ;
        } ;

        TMatrix<Dtype> &operator += (const TMatrix<Dtype> &oper)
        {
            assert((oper.Rows() == rows) &&
                   (oper.Cols() == cols)) ;
            for (int i=0 ; i<rows*cols ; ++i) data[i] += oper.data[i] ;
            return(*this) ;
        } ;

        TMatrix<Dtype> &operator *= (Dtype oper) 
        {
            for (int i=0 ; i<rows*cols ; ++i) data[i] *= oper ;
            return(*this) ;
        } ;

        TMatrix<Dtype> &operator *= (const TMatrix<Dtype> &oper) 
        {
            assert((oper.Rows() == oper.Cols()) &&
                   ((oper.Rows() == cols) || (oper.Rows() == rows))) ;

            if (oper.Rows() == cols) {
                Dtype *tmp = new Dtype[cols] ;
                for (int i=0 ; i<rows ; ++i) {
                    for (int j=0 ; j<cols ; ++j) {
                        tmp[j] = 0 ;
                        for (int k=0 ; k<cols ; ++k) {
                            tmp[j] += data[i*cols+k] *
                                      oper.data[k*cols+j] ;
                        }
                    }
                    for (int l=0 ; l<cols ; ++l) data[i*cols+l] = tmp[l] ;
                }
                delete [] tmp ;
            } else {
                Dtype *tmp = new Dtype[rows] ;
                for (int j=0 ; j<cols ; ++j) {
                    for (int i=0 ; i<rows ; ++i) {
                        tmp[i] = 0 ;
                        for (int k=0 ; k<rows ; ++k) {
                            tmp[i] += oper.data[i*rows+k] *
                                          data[k*cols+j] ;
                        }
                    }
                    for (int l=0 ; l<rows ; ++l) data[l*cols+j] = tmp[l] ;
                }
                delete [] tmp ;
            }
            return(*this) ;
        } ;

        void TripleProd(const TMatrix<Dtype> &oper1,
                        const TMatrix<Dtype> &oper2,
                        double scalar)
        {
            assert(rows == cols) ;
            assert(oper2.Rows() == oper2.Cols()) ;
            assert(oper1.Rows() == oper2.Rows()) ;
            assert(oper1.Cols() == rows) ;

            for (int i=0 ; i<oper2.Rows() ; i++) {
                for (int j=0 ; j<oper1.Cols() ; j++) {
                    double tmp = 0.0 ;
                    for (int k=0 ; k<oper1.Rows() ; k++) {
                        tmp += oper2(i,k) * oper1(k,j) ;
                    }
                    for (int k=0 ; k<oper1.Cols() ; k++) {
                        data[k*rows+j] += tmp * oper1(i,k) * scalar ;
                    }
                }
            }
        }

    protected:

        int    rows ;
        int    cols ;
        Dtype  *data ;

        // private constructor

        TMatrix(int irows,int icols,Dtype *idata) :
            rows(irows),cols(icols),data(idata) {} ;

        // friends

#ifdef WIN32
    friend TMatrix<Dtype> operator* (const TMatrix<Dtype> &oper1,
                                     const TMatrix<Dtype> &oper2) ;

    friend TMatrix<Dtype> operator* (const Dtype &oper1,
                                     const TMatrix<Dtype> &oper2) ;

    friend TMatrix<Dtype> operator* (const TMatrix<Dtype> &oper1,
                                     const Dtype &oper2) ;
#else
    friend TMatrix<Dtype> operator* <Dtype> (const TMatrix<Dtype> &oper1,
                                             const TMatrix<Dtype> &oper2) ;

    friend TMatrix<Dtype> operator* <Dtype> (const Dtype &oper1,
                                             const TMatrix<Dtype> &oper2) ;

    friend TMatrix<Dtype> operator* <Dtype> (const TMatrix<Dtype> &oper1,
                                             const Dtype &oper2) ;
#endif
} ;


template<class Dtype>
inline TMatrix<Dtype> operator *(const TMatrix<Dtype> &oper1,
                                 const TMatrix<Dtype> &oper2)
{
    assert(oper1.Cols() == oper2.Rows()) ;
    Dtype *tmp = new Dtype[oper1.Rows()*oper2.Cols()] ;

    for (int i=0 ; i<oper1.Rows() ; ++i) {
        for (int j=0 ; j<oper2.Cols() ; ++j) {
            tmp[i*oper2.Cols()+j] = 0 ;
            for (int k=0 ; k<oper1.Cols() ; ++k) {
                tmp[i*oper2.Cols()+j] += oper1(i,k) * oper2(k,j) ;
            }
        }
    }
    return(TMatrix<Dtype>(oper1.Rows(),oper2.Cols(),tmp)) ;
}


template<class Dtype>
inline TMatrix<Dtype> operator *(const Dtype &oper1,
                                 const TMatrix<Dtype> &oper2)
{
    Dtype *tmp = new Dtype[oper2.Rows()*oper2.Cols()] ;

    for (int i=0 ; i<oper2.Rows() ; ++i) {
        for (int j=0 ; j<oper2.Cols() ; ++j) {
            tmp[i*oper2.Cols()+j] = oper1 * oper2(i,j) ;
        }
    }
    return(TMatrix<Dtype>(oper2.Rows(),oper2.Cols(),tmp)) ;
}


template<class Dtype>
inline TMatrix<Dtype> operator *(const TMatrix<Dtype> &oper1,
                                 const Dtype &oper2)
{
    return(oper2*oper1) ;
}


#if 0
template<class Dtype> inline ostream &operator << (ostream &out,
                                      const TMatrix<Dtype> &mat)
{
    for (int i=0 ; i<mat.Rows() ; ++i) {
        for (int j=0 ; j<mat.Cols() ; ++j) {
            out << mat(i,j) << ' ' ;
        }
        out << endl ;
    } 
    return(out) ;
}
#endif

} // namespace

#endif
