
//
// TableFunction implementation
//
// Revision -
//   $Revision: 1.2 $  $Date: 2001/12/05 03:16:01 $  $Author: wash $
//

#include <cmath>
#include <cstdlib>

#include "TableFunction.hh"

TableFunction::TableFunction()
{
    NumSegments = -1 ;
    Vals = 0 ;
    XCoefs = 0 ;
    YCoefs = 0 ;
    CurrentSeg = 0 ;
}

TableFunction::TableFunction(int num_points,
                             const double *x_values,
                             const double *y_values,
                             InterpolationType interp,
                             ExtrapolationType extrap)
{
    NumSegments = -1 ;
    Vals = 0 ;
    XCoefs = 0 ;
    YCoefs = 0 ;
    CurrentSeg = 0 ;

    DerivTolerance = 0.0001 ;
    StopTolerance = 0.005 ;
    MaxIterations = 10 ;

    Configure(num_points,x_values,y_values,interp,extrap) ;
}

TableFunction::TableFunction(const TableFunction &other)
{
    NumSegments = other.NumSegments ;
    Vals = new Data[NumSegments+1] ;
    for (int i=0 ; i<NumSegments+1 ; ++i) Vals[i] = other.Vals[i] ;
    XCoefs = new HermiteCoef[NumSegments] ;
    YCoefs = new HermiteCoef[NumSegments] ;
    for (int j=0 ; j<NumSegments ; ++j) {
        XCoefs[j] = other.XCoefs[j] ;
        YCoefs[j] = other.YCoefs[j] ;
    }
    CurrentSeg = other.CurrentSeg ;

    DerivTolerance = other.DerivTolerance ;
    StopTolerance = other.StopTolerance ;
    MaxIterations = other.MaxIterations ;
}


TableFunction TableFunction::operator = (const TableFunction &other)
{
    if (Vals != 0) delete Vals ;
    if (XCoefs != 0) delete XCoefs ;
    if (YCoefs != 0) delete YCoefs ;

    NumSegments = other.NumSegments ;
    Vals = new Data[NumSegments+1] ;
    for (int i=0 ; i<NumSegments+1 ; ++i) Vals[i] = other.Vals[i] ;
    XCoefs = new HermiteCoef[NumSegments] ;
    YCoefs = new HermiteCoef[NumSegments] ;
    for (int j=0 ; j<NumSegments ; ++j) {
        XCoefs[j] = other.XCoefs[j] ;
        YCoefs[j] = other.YCoefs[j] ;
    }
    CurrentSeg = other.CurrentSeg ;

    DerivTolerance = other.DerivTolerance ;
    StopTolerance = other.StopTolerance ;
    MaxIterations = other.MaxIterations ;
    return(*this) ;
}


TableFunction::~TableFunction()
{
    if (Vals != 0) delete Vals ;
    if (XCoefs != 0) delete XCoefs ;
    if (YCoefs != 0) delete YCoefs ;
}


static int Compare(const void *op0,const void *op1)
{
    TableFunction::Data *pt0 = (TableFunction::Data *)op0 ;
    TableFunction::Data *pt1 = (TableFunction::Data *)op1 ;
    if (pt0->x < pt1->x) return(-1) ;
    if (pt0->x > pt1->x) return(1) ;
    return(0) ;
}

void TableFunction::Configure(int num_points,
                              const double *x_values,
                              const double *y_values,
                              InterpolationType interp,
                              ExtrapolationType extrap)
{
    if (Vals != 0) delete Vals ;
    if (XCoefs != 0) delete XCoefs ;
    if (YCoefs != 0) delete YCoefs ;

    NumSegments = num_points-1 ;
    InterpType = interp ;
    ExtrapType = extrap ;
    if (num_points == 0) return ;

    // allocate space

    Vals = new Data[num_points] ;
    if (InterpType == I_SPLINE) {
        XCoefs = new HermiteCoef[NumSegments] ;
        YCoefs = new HermiteCoef[NumSegments] ;
    }

    // copy the data, check to see if we need to sort

    bool sort = false ;
    for (int i=0 ; i<num_points ; ++i) {
        Vals[i].x = x_values[i] ;
        Vals[i].y = y_values[i] ;
        if ((i > 0) && (Vals[i].x < Vals[i-1].x)) sort = true ;
    }

    if (sort) {
        qsort((void *)Vals,num_points,sizeof(Data),Compare) ;
    }

    if ((interp == I_LINEAR) || (num_points == 1)) return ;

    // compute the spline coefficients

    double cur_dx = Vals[1].x - Vals[0].x ;
    double cur_dy = Vals[1].x - Vals[0].y ;
    double cur_length = sqrt(cur_dx*cur_dx + cur_dy*cur_dy) ;

    double x_slope_0 = cur_dx ;
    double y_slope_0 = cur_dy ;

    for (int i=0 ; i<NumSegments ; i++) {
        double x_slope_1 = cur_dx ;
        double y_slope_1 = cur_dy ;
        double next_dx, next_dy, next_length ;

        if (i < NumSegments-1) {
            next_dx = Vals[i+2].x - Vals[i+1].x ;
            next_dy = Vals[i+2].y - Vals[i+1].y ;
            next_length = sqrt(next_dx*next_dx + next_dy*next_dy) ;

            x_slope_1 = (x_slope_1 + next_dx)/2.0 ;
            y_slope_1 = (y_slope_1 + next_dy)/2.0 ;
        }

        XCoefs[i].a0 = Vals[i].x ;
        YCoefs[i].a0 = Vals[i].y ;
        XCoefs[i].a1 = x_slope_0 ;
        YCoefs[i].a1 = y_slope_0 ;
        XCoefs[i].a2 = 3*(Vals[i+1].x-Vals[i].x) - 2*x_slope_0 - x_slope_1 ;
        YCoefs[i].a2 = 3*(Vals[i+1].y-Vals[i].y) - 2*y_slope_0 - y_slope_1 ;
        XCoefs[i].a3 = 2*(Vals[i].x-Vals[i+1].x) + x_slope_0 + x_slope_1 ;
        YCoefs[i].a3 = 2*(Vals[i].y-Vals[i+1].y) + y_slope_0 + y_slope_1 ;

        x_slope_0 = x_slope_1 ;
        y_slope_0 = y_slope_1 ;
        cur_dx = next_dx ;
        cur_dy = next_dy ;
        cur_length = next_length ;
    }
}


double TableFunction::HermiteEvaluate(HermiteCoef &coefs,double t) const
{
    return(coefs.a0 + t*(coefs.a1 + t*(coefs.a2 + t*coefs.a3))) ;
}

double TableFunction::HermiteDerivEvaluate(HermiteCoef &coefs,double t) const
{
    return(coefs.a1 + t*(2*coefs.a2 + t*3*coefs.a3)) ;
}

bool TableFunction::ExtrapolateLow(double x,double *y) const
{
    if (ExtrapType == E_NONE) return(false) ;

    if (ExtrapType == E_CONSTANT) {
        *y = Vals[0].y ;
    } else {
        *y = Vals[0].y + (x-Vals[0].x) *
               (Vals[1].y-Vals[0].y) /
               (Vals[1].x-Vals[0].x) ;
    }
    return(true) ;
}

bool TableFunction::ExtrapolateHigh(double x,double *y) const
{
    if (ExtrapType == E_NONE) return(false) ;

    if (ExtrapType == E_CONSTANT) {
        *y = Vals[NumSegments].y ;
    } else {
        *y = Vals[NumSegments-1].y + (x-Vals[NumSegments-1].x) *
               (Vals[NumSegments].y-Vals[NumSegments-1].y) /
               (Vals[NumSegments].x-Vals[NumSegments-1].x) ;
    }
    return(true) ;
}



bool TableFunction::GetValue(double x,double *y)
{
    // check for no data

    if (NumSegments == -1) return(false) ;

    // check for one data point

    if (NumSegments == 0) {
        if (ExtrapType == E_NONE) return(false) ;
        *y = Vals[0].y ;
        return(true) ;
    } 

    // check to see if we need to do extrapolation

    if (x < Vals[0].x) return(ExtrapolateLow(x,y)) ;
    if (x > Vals[NumSegments].x) return(ExtrapolateHigh(x,y)) ;

    // check to see if we are in the cached segment first,
    // then we start from the beginning and find the segment

    double t ;
    int cur_seg = (CurrentSeg < NumSegments) ? CurrentSeg : 0 ;
    if ((x < Vals[cur_seg].x) || (x > Vals[cur_seg+1].x)) {
        cur_seg = 0 ;
        while (cur_seg < NumSegments) {
            if ((x >= Vals[cur_seg].x) && (x <= Vals[cur_seg+1].x)) {
                t = (x - Vals[cur_seg].x) /
                    (Vals[cur_seg+1].x - Vals[cur_seg].x) ;
                CurrentSeg = cur_seg ;
                break ;
            } else {
                cur_seg++ ;
            }
        }
    }

    // do the appropriate type of interpolation

    if (InterpType == I_LINEAR) {

        if (Vals[cur_seg+1].x == Vals[cur_seg].x) {
            *y = 0.5*(Vals[cur_seg+1].y-Vals[cur_seg].y) ;
        } else {
            *y = Vals[cur_seg].y + (x-Vals[cur_seg].x) *
                   (Vals[cur_seg+1].y-Vals[cur_seg].y) /
                   (Vals[cur_seg+1].x-Vals[cur_seg].x) ;
        }
        return(true) ;

    } else {

       // find the corresponding parametric coordinate
       // using a Newton procedure

        bool fallback = false ;
        int num_it = 0 ;
        double t_cur, t=0 ;

        do {
            t_cur = t ;
            double x_trial = HermiteEvaluate(XCoefs[cur_seg],t_cur) ;
            double f = x - x_trial ;
            double f_deriv = HermiteDerivEvaluate(XCoefs[cur_seg],t_cur) ;
            if (fabs(f_deriv) < DerivTolerance) {
                fallback = true ;
                break ;
            } else {
                t = t_cur + f / f_deriv ;
                t = (t > 1.0) ? 1.0 : t ;
                t = (t < 0.0) ? 0.0 : t ;
            }

            if (num_it >= MaxIterations) {
                fallback = true ;
                break ;
            }
            num_it++ ;
        } while (fabs(t - t_cur) >= StopTolerance) ; 

        if (fallback) {
            t = (x - Vals[cur_seg].x) /
                (Vals[cur_seg+1].x - Vals[cur_seg].x) ;
        }

        *y = HermiteEvaluate(YCoefs[cur_seg],t) ; 
    }

    return(true) ;
}


