//
// TableFunctionModule - python extension for a table function
//
// Revision -
//   $Revision$  $Date$  $Author$
//

#include <cstdio>
#include <cstdlib>
#include "Python.h"
#include "TableFunction.hh"

#if BAD_STATIC_FORWARD
staticforward PyTypeObject tf_type ;
#else
PyTypeObject tf_type ;
#endif

typedef struct {
    PyObject_HEAD
    TableFunction *tf ;
    }
    TableFunctionObject ;

extern "C" {
    PyObject *GetValue(TableFunctionObject *self,PyObject *args) ;
    PyObject *Configure(TableFunctionObject *self,PyObject *args) ;

    void initTableFunction() ;
}

#define is_TableFunctionObject(v) ((v)->ob_type == &tf_type)

/* ----------------------------------------------------------
**  Object methods and lookup table
*/

static double *ConvertNumSequence(PyObject *obj,int *num)
{
    if (PySequence_Check(obj) == 0) {
        PyErr_SetString(PyExc_TypeError,
            "argument should a sequence of numbers") ;
        return(0) ;
    }
    *num = PySequence_Size(obj) ;

    double *vect = new double[*num] ;
    for (int i=0 ; i<(*num) ; ++i) {
        PyObject *item = PySequence_GetItem(obj,i) ;
        if (PyInt_Check(item) == 1) {
            vect[i] = PyInt_AsLong(item) ;
        } else if (PyFloat_Check(item) == 1) {
            vect[i] = PyFloat_AsDouble(item) ;
        } else {
            Py_DECREF(item) ;
            PyErr_SetString(PyExc_TypeError,
                "sequence member should be a number") ;
            return(0) ;
        }
        Py_DECREF(item) ;
    }
    return(vect) ;
}

static bool ConvertArgs(PyObject *Xseq_obj,PyObject *Yseq_obj,
                        char *interp_type,char *extrap_type,
                        double **X,double **Y,int *len,
                        TableFunction::InterpolationType *itype,
                        TableFunction::ExtrapolationType *etype)
{
    if (interp_type == 0) {
        *itype = TableFunction::I_SPLINE ;
    } else if ((strcmp(interp_type,"LINEAR") == 0) ||
               (strcmp(interp_type,"Linear") == 0) ||
               (strcmp(interp_type,"linear") == 0)) {
        *itype = TableFunction::I_LINEAR ;
    } else if ((strcmp(interp_type,"SPLINE") == 0) ||
               (strcmp(interp_type,"Spline") == 0) ||
               (strcmp(interp_type,"spline") == 0)) {
        *itype = TableFunction::I_SPLINE ;
    } else {
        PyErr_SetString(PyExc_TypeError,
            "interpolation type should be LINEAR or SPLINE") ;
    }

    if (extrap_type == 0) {
        *etype = TableFunction::E_NONE ;
    } else if ((strcmp(extrap_type,"NONE") == 0) ||
               (strcmp(extrap_type,"None") == 0) ||
               (strcmp(extrap_type,"none") == 0)) {
        *etype = TableFunction::E_NONE ;
    } else if ((strcmp(extrap_type,"CONSTANT") == 0) ||
               (strcmp(extrap_type,"Constant") == 0) ||
               (strcmp(extrap_type,"constant") == 0)) {
        *etype = TableFunction::E_CONSTANT ;
    } else if ((strcmp(extrap_type,"LINEAR") == 0) ||
               (strcmp(extrap_type,"Linear") == 0) ||
               (strcmp(extrap_type,"linear") == 0)) {
        *etype = TableFunction::E_LINEAR ;
    } else {
        PyErr_SetString(PyExc_TypeError,
            "enterpolation type should be NONE, CONSTANT, or LINEAR") ;
    }

    int x_len,y_len ;

    *X = ConvertNumSequence(Xseq_obj,&x_len) ;
    if (*X == 0) return(false) ;

    *Y = ConvertNumSequence(Yseq_obj,&y_len) ;
    if (*Y == 0) {
        delete *X ;
        return(false) ;
    }

    if (x_len != y_len) {
        PyErr_SetString(PyExc_TypeError,
            "X and Y sequences should have the same length") ;
        delete *X ;
        delete *Y ;
        return(false) ;
    }
    *len = x_len ;
    return(true) ;
}


static PyObject *GetValue(TableFunctionObject *self,PyObject *args)
{
    double x, y ;
    if (!PyArg_ParseTuple(args,"d",&x)) return(NULL) ;

    if (!self->tf->GetValue(x,&y)) {
        Py_INCREF(Py_None) ;
        return(Py_None) ;
    }
    return(Py_BuildValue("d",y)) ;
}


static PyObject *Configure(TableFunctionObject *self,PyObject *args)
{
    PyObject *Xseq_obj, *Yseq_obj ;
    char *interp_type = 0 ;
    char *extrap_type = 0 ;

    if (!PyArg_ParseTuple(args,"OO|ss",&Xseq_obj,&Yseq_obj,
              &interp_type,&extrap_type)) return(NULL) ;

    double *X,*Y ;
    int len ;
    TableFunction::InterpolationType itype ;
    TableFunction::ExtrapolationType etype ;

    if (!ConvertArgs(Xseq_obj,Yseq_obj,interp_type,extrap_type,
                     &X,&Y,&len,&itype,&etype)) return(NULL) ;

    self->tf->Configure(len,X,Y,itype,etype) ;

    delete X ;
    delete Y ;

    Py_INCREF(Py_None) ;
    return(Py_None) ;
}


static PyMethodDef TableFunctionMethods[] = {
    { "GetValue",  (PyCFunction)GetValue,  METH_VARARGS },
    { "Configure", (PyCFunction)Configure, METH_VARARGS },
    { (char *)NULL, (PyCFunction)NULL }
} ;


static TableFunctionObject *newTableFunctionObject(PyObject *args)
{
    PyObject *Xseq_obj, *Yseq_obj ;
    char *interp_type = 0 ;
    char *extrap_type = 0 ;

    if (!PyArg_ParseTuple(args,"OO|ss",&Xseq_obj,&Yseq_obj,
              &interp_type,&extrap_type)) return(NULL) ;

    double *X,*Y ;
    int len ;
    TableFunction::InterpolationType itype ;
    TableFunction::ExtrapolationType etype ;

    if (!ConvertArgs(Xseq_obj,Yseq_obj,interp_type,extrap_type,
                     &X,&Y,&len,&itype,&etype)) return(NULL) ;

    TableFunctionObject *self =
        PyObject_NEW(TableFunctionObject,&tf_type) ;
    if (self == NULL) return NULL ;

    self->tf = new TableFunction(len,X,Y,itype,etype) ;

    delete X ;
    delete Y ;

    return(self) ;
}


static void TableFunctionDealloc(PyObject *self)
{
    TableFunctionObject *obj = (TableFunctionObject *)self ;
    delete obj->tf ;
}

static PyObject *TableFunctionGetattr(PyObject *self,char *name)
{
    return(Py_FindMethod(TableFunctionMethods,self,name)) ;
}


#if BAD_STATIC_FORWARD
PyTypeObject tf_type = {
#else
statichere PyTypeObject tf_type = {
#endif

    /* type header */

    PyObject_HEAD_INIT(&PyType_Type)
    0,
    "TableFunction",              /* tp_name */
    sizeof(TableFunctionObject),  /* tp_basicsize */
    0,                            /* tp_itemsize */

    /* standard methods */

    (destructor)  TableFunctionDealloc, /* tp_dealloc ref-count==0 */
    (printfunc)   0,                    /* tp_print   "print x" */
    (getattrfunc) TableFunctionGetattr, /* tp_getattr "x.attr" */
    (setattrfunc) 0,                    /* tp_setattr "x.attr=v" */
    (cmpfunc)     0,                    /* tp_compare "x > y" */
    (reprfunc)    0,                    /* tp_repr    'x', print x */

    /* type categories */

    0,                   /* tp_as_number   +,-,*,/,%,&,>>,pow... */
    0,                   /* tp_as_sequence +,[i],[i:j],len,... */
    0,                   /* ty_as_mapping  [key],len,... */

    /* more methods */

    (hashfunc)     0,                   /* tp_hash   "dict[x]" */
    (ternaryfunc)  0,                   /* tp_call   "x()" */
    (reprfunc)     0,                   /* tp_str    "str(x)" */
} ;

/* ==========================================================
** ==========================================================
** ==========================================================
**  Module methods and lookup table
*/

static PyObject *TableFunctionType_new(PyObject *self, PyObject *args)
{
    return((PyObject *)newTableFunctionObject(args)) ;
}

static PyMethodDef TableFunctionTypeMethods[] = {
    { "TableFunction",  TableFunctionType_new,     METH_VARARGS },
    { NULL, NULL }
} ;

void initTableFunction()
{
    Py_InitModule("TableFunction",TableFunctionTypeMethods) ;
}

