//
// Vec3D object header file
//
// Revision -
//   $Revision: 1.2 $  $Date: 2001/12/03 03:21:44 $  $Author: wash $
//
 
#ifndef Vec3D_h
#define Vec3D_h

#include <cassert>
#include <cmath>
#include <iostream>

namespace FemLib {

class Vec3D {

    public:

    // constructors and destructors

        Vec3D() { data[0] = data[1] = data[2] = 0.0 ; } ;
        Vec3D(double x,double y,double z) {
            data[0] = x ; data[1] = y ; data[2] = z ; } ;
        Vec3D(double *x)
        {
            if (x == 0) {
                data[0] = 0.0 ; data[1] = 0.0 ; data[2] = 0.0 ;
            } else {
                data[0] = x[0] ; data[1] = x[1] ; data[2] = x[2] ;
            }
        } ;

    // access
     
        double x() const { return(data[0]) ; } ;
        double y() const { return(data[1]) ; } ;
        double z() const { return(data[2]) ; } ;

        void x(double x) { data[0] = x ; } ;
        void y(double y) { data[1] = y ; } ;
        void z(double z) { data[2] = z ; } ;

    // operators

        double operator [] (const int indx) const {
            assert((indx >= 0) && (indx <= 2)) ;
            return(data[indx]) ;
        } ;

        double &operator[] (const int indx) {
            assert((indx >= 0) && (indx <= 2)) ;
            return(data[indx]) ;
        } ;

        Vec3D operator += (const Vec3D &op2) {
            data[0] += op2.data[0] ;
            data[1] += op2.data[1] ;
	    data[2] += op2.data[2] ;
            return(*this) ;
        } ;

        Vec3D operator -= (const Vec3D &op2) {
            data[0] -= op2.data[0] ;
            data[1] -= op2.data[1] ;
	    data[2] -= op2.data[2] ;
            return(*this) ;
        } ;

        Vec3D operator *= (const double &op2) {
            data[0] *= op2 ;
            data[1] *= op2 ;
	    data[2] *= op2 ;
            return(*this) ;
        } ;

        Vec3D operator /= (const double &op2) {
            data[0] /= op2 ;
            data[1] /= op2 ;
	    data[2] /= op2 ;
            return(*this) ;
        } ;

        Vec3D operator - () {
            return(Vec3D(-data[0],-data[1],-data[2])) ;
        } ;

    // member functions

        double Magnitude() {
            return(sqrt(data[0]*data[0]+
	                data[1]*data[1]+
			data[2]*data[2])) ;
        } ;

        Vec3D Normalize() {
            double len = sqrt(data[0]*data[0]+
	                      data[1]*data[1]+
			      data[2]*data[2]) ;
            return(Vec3D(data[0]/len,data[1]/len,data[2]/len)) ;
        } ;

        Vec3D max(const Vec3D &op2) {
            data[0] = data[0] >= op2.x() ? data[0] : op2.x() ;
            data[1] = data[1] >= op2.y() ? data[1] : op2.y() ;
            data[2] = data[2] >= op2.z() ? data[2] : op2.z() ;
            return(*this) ;
        } ;

        Vec3D min(const Vec3D &op2) {
            data[0] = data[0] <= op2.x() ? data[0] : op2.x() ;
            data[1] = data[1] <= op2.y() ? data[1] : op2.y() ;
            data[2] = data[2] <= op2.z() ? data[2] : op2.z() ;
            return(*this) ;
        } ;

    private:
        double data[3] ;

        friend double operator * (const Vec3D &op1,
                                  const Vec3D &op2) ;
} ;
                                                 
inline int operator == (const Vec3D &c1,const Vec3D &c2)
{
    return((c1[0] == c2[0]) && (c1[1] == c2[1]) && (c1[2] == c2[2])) ;
}

inline int operator < (const Vec3D &c1,const Vec3D &c2)
{
    return((c1[0] <= c2[0]) && (c1[1] <= c2[1]) && (c1[2] <= c2[2])) ;
}

inline int operator > (const Vec3D &c1,const Vec3D &c2)
{
    return((c1[0] >= c2[0]) && (c1[1] >= c2[1]) && (c1[2] >= c2[2])) ;
}

inline Vec3D operator + (const Vec3D &op1,const Vec3D &op2)
{
    Vec3D op = op1 ;
    op += op2 ;
    return(op) ;
}

inline Vec3D operator - (const Vec3D &op1,const Vec3D &op2)
{
    Vec3D op = op1 ;
    op -= op2 ;
    return(op) ;
}

inline Vec3D operator * (const Vec3D &op1,const double &op2)
{
    Vec3D op = op1 ;
    op *= op2 ;
    return(op) ;
}

inline Vec3D operator * (const double &op1,const Vec3D &op2)
{
    Vec3D op = op2 ;
    op *= op1 ;
    return(op) ;
}

inline double operator * (const Vec3D &op1,const Vec3D &op2)
{
    return(op1.x()*op2.x() + op1.y()*op2.y() + op1.z()*op2.z()) ;
}

inline Vec3D operator / (const Vec3D &op1,const double &op2)
{
    Vec3D op = op1 ;
    op /= op2 ;
    return(op) ;
}

inline Vec3D max(const Vec3D &op1,const Vec3D &op2)
{
    return(Vec3D((op1.x() >= op2.x() ? op1.x() : op2.x()),
                 (op1.y() >= op2.y() ? op1.y() : op2.y()),
                 (op1.z() >= op2.z() ? op1.z() : op2.z()))) ;
}

inline Vec3D min(const Vec3D &op1,const Vec3D &op2)
{
    return(Vec3D((op1.x() <= op2.x() ? op1.x() : op2.x()),
                 (op1.y() <= op2.y() ? op1.y() : op2.y()),
                 (op1.z() <= op2.z() ? op1.z() : op2.z()))) ;
}

inline Vec3D CrossProd(const Vec3D &op1,const Vec3D &op2)
{
    return(Vec3D((op1.y()*op2.z()) - (op1.z()*op2.y()), 
                 (op1.z()*op2.x()) - (op1.x()*op2.z()), 
                 (op1.x()*op2.y()) - (op1.y()*op2.x()))) ; 
}

inline double TripleProd(const Vec3D &op1,const Vec3D &op2,
                         const Vec3D &op3)
{
    Vec3D tmp = CrossProd(op1,op2) ;
    return(tmp * op3) ; 
}

inline std::ostream &operator << (std::ostream &out,const Vec3D &data)
{
    out <<data.x() << ' ' << data.y() << ' ' << data.z() ;
    return(out) ;
}

inline std::istream &operator >> (std::istream &in, Vec3D &data)
{
    in >> data[0] >> data[1] >> data[2] ;
    return(in) ;
}

} // namespace
 
#endif
