#!/usr/local/bin/python


from SimpleFemModel import *
import os
import sys
from xml.dom.ext.reader import Sax2
from xml.dom.DOMImplementation import implementation
from xml.dom import ext
from xml.dom.NodeFilter import NodeFilter


class WSFemModel(SimpleFemModel):
    """The WSFemModel class acts as a container to hold a description
       of a mesh and associated elements and boundary conditions."""

    # constructor that reads a model description from a web service

    def __init__(self,soapAction,webService,modelName,lengthScale,randomNumberSeed):
        """Reads the entire mesh from a web service

        """

        # read the node coordinates
        method = "GetVertices"
        self.ConstructSoapEnvelope(soapAction,modelName,method)
        walker = self.CallWebService(soapAction,webService,method)

        self.nodes = {}
        while 1:
            node = walker.currentNode
            if node.tagName == "vertex":
                atts = node.attributes
                for k in node.attributes.keys():
                    if atts[k].localName == "id":
                        VertexID = atts[k].value
                    elif atts[k].localName == "x":
                        x = atts[k].value
                    elif atts[k].localName == "y":
                        y = atts[k].value
                    elif atts[k].localName == "z":
                        z = atts[k].value
                self.nodes[int(VertexID)] = (lengthScale*float(x),lengthScale*float(y),lengthScale*float(z))
            next = walker.nextNode()
            if next is None: break



        #the element data and material numbers
        method = "GetRegionElements"
        self.ConstructSoapEnvelope(soapAction,modelName,method)
        walker = self.CallWebService(soapAction,webService,method)

        self.elem = {}
        self.materials = {}
        orientFile = open("orientations.dat","w")
        while  1:
            node = walker.currentNode
            if node.tagName == "region":
                atts = node.attributes
                for k in node.attributes.keys():
                     if atts[k].localName == "id":
                        materialID = int(atts[k].value)
                # for now, the materials only consist of different grains
                # and we make up random orientations here
                orientation = GetRandomRotationMatrix(randomNumberSeed+materialID)
                orientFile.write(str(orientation)+"\n")
                self.materials[materialID] = orientation
            elif node.tagName == "tet":
                atts = node.attributes
                for k in node.attributes.keys():
                    if atts[k].localName == "id":
                        id = int(atts[k].value)
                    elif atts[k].localName == "v0":
                        v0 = int(atts[k].value)
                    elif atts[k].localName == "v1":
                        v1 = int(atts[k].value)
                    elif atts[k].localName == "v2":
                        v2 = int(atts[k].value)
                    elif atts[k].localName == "v3":
                        v3 = int(atts[k].value)
                self.elem[id] = ("Tet2I",materialID,[v0,v1,v2,v3])
            next = walker.nextNode()
            if next is None:
                next = walker.previousNode()
                next = walker.nextSibling()
                if next is None: break


        # get boundary conditions XXX not working now
        method = "GetFaceBoundaryConditions"
        self.ConstructSoapEnvelope(soapAction,modelName,method)
        walker = self.CallWebService(soapAction,webService,method)

        # bc will be stored as a dictionary of faces
        # and their displacements
        
        self.faceDisplacements = {}
        while  1:
            node = walker.currentNode
            if node.tagName == "face":
                atts = node.attributes
                for k in node.attributes.keys():
                    if atts[k].localName == "id":
                        faceID = int(atts[k].value)
            if node.tagName == "type1":
                atts = node.attributes
                for k in node.attributes.keys():
                    if atts[k].localName == "dim":
                        dim = atts[k].value
                    elif atts[k].localName == "value":
                        value = float(atts[k].value)
                self.faceDisplacements[faceID]=(dim,value)
            next = walker.nextNode()
            if next is None:
                if node.tagName == "face":
                    next = walker.previousNode()
                    next = walker.firstChild()
                    next = walker.firstChild()
                elif node.tagName == "type1":
                    next = walker.parentNode()
                    next = walker.nextSibling()
                if next is None: break
          

        # find triples of vertices forming triangles on each face
        self.faceTriangles = {}
        method = "GetFaceElements"
        self.ConstructSoapEnvelope(soapAction,modelName,method)
        walker = self.CallWebService(soapAction,webService,method)

        while  1:
            node = walker.currentNode
            if node.tagName == "face":
                atts = node.attributes
                for k in node.attributes.keys():
                     if atts[k].localName == "id":
                        faceID = int(atts[k].value)
                triangles = []
            elif node.tagName == "tri":
                atts = node.attributes
                for k in node.attributes.keys():
                    if atts[k].localName == "v0":
                        v0 = int(atts[k].value)
                    elif atts[k].localName == "v1":
                        v1 = int(atts[k].value)
                    elif atts[k].localName == "v2":
                        v2 = int(atts[k].value)
                triangles.append((v0,v1,v2))
            next = walker.nextNode()
            if next is None:
                if faceID in self.faceDisplacements.keys():
                    self.faceTriangles[faceID]=triangles
                next = walker.previousNode()
                next = walker.nextSibling()
                if next is None: break

        

    def ConstructSoapEnvelope(self,soapAction,modelName,method):

        # Construct the request SOAP Envelope
        doc = implementation.createDocument(None,None,None)
        
        envelope = doc.createElement('soap:Envelope')
        envelope.setAttribute('xmlns:soap','http://schemas.xmlsoap.org/soap/envelope/')
        envelope.setAttribute('xmlns:xsi','http://www.w3.org/2001/XMLSchema-instance')
        envelope.setAttribute('xmlns:xsd','http://www.w3.org/2001/XMLSchema')
        doc.appendChild(envelope)
        
        body = doc.createElement('soap:Body')
        envelope.appendChild(body)

        if method=="GetFaceBoundaryConditions":
            GetFaceBoundaryConditions = doc.createElement('GetFaceBoundaryConditions')
            GetFaceBoundaryConditions.setAttribute('xmlns','http://www.tc.cornell.edu/~heber')
            body.appendChild(GetFaceBoundaryConditions)
        
        modelname = doc.createElement('modelname')
        modelname.setAttribute('xmlns',soapAction)
        modelname.appendChild(doc.createTextNode(modelName))
        if method=="GetFaceBoundaryConditions":
            GetFaceBoundaryConditions.appendChild(modelname)
        else:
            body.appendChild(modelname)

        # If you want to see the request, uncomment this
        #print "The response:"
        #ext.PrettyPrint(doc)

        # Write it to a file
        f = open('request.xml','w')
        ext.PrettyPrint(doc,f)
        print >> f, ''
        f.close()        



    def CallWebService(self,soapAction,webService,method):
        #returns xml "walker" to xml structure
        stat = os.system("""curl -s --data-binary @request.xml"""
                 """ -H 'SOAPAction: """ + "\"" + soapAction +  "/" + method + "\"\'"
                 """ -H 'Content-Type: text/xml' """
                 + webService +
                 """ > response.xml""")

        if stat <> 0:
            print "curl exited with status", stat

        # Read the response SOAP Envelope from the file
        doc = Sax2.FromXmlFile("response.xml")
        walker = doc.createTreeWalker(doc.documentElement,NodeFilter.SHOW_ELEMENT, None, 0)
        return walker


# if calling directly from command line or web service
if __name__ == "__main__":

    soapAction = "http://www.tc.cornell.edu/~heber"
    webService = "http://sonofstager.tc.cornell.edu/~heber/Valerie/default.asmx"
    modelName = "2grainPart-3D"
    model = WSFemModel(soapAction,webService,modelName,1,1)
    print "NODES"
    print model.nodes
    print "MATERIALS"
    print model.materials
    print "ELEMENTS"
    print model.elem
    print "BOUNDARY CONDITIONS"
    print model.faceDisplacements
    print "FACE TRIANGLES"
    print model.faceTriangles
