#!/usr/local/bin/python

"""Unit test for the FemLib library"""

import FemLib
import unittest
import zlib
import base64
import string
import StringIO
import Numeric
import DMatrix
import ColTensor
import Vec3D
import math

# -------------------------------------------------------------------

# define an class that makes a zlib compressed string look like
# a file.  This is based on Fredrik Lundh's class in the Python
# Standard Library book

class CompressedStream:
    """A class that makes a zlib compressed string look like a file

    This is based on Fredrik Lundh's class in the Python standard
    library book"""

    def __init__(self,data):
        self.file = StringIO.StringIO(data)
        self._rewind()

    def _rewind(self):
        self.zip = zlib.decompressobj()
        self.pos = 0
        self.offset = 0
        self.data = ""

    def _fill(self,bytes):
        if self.zip:
            # read until we have enough bytes in the buffer
            while not bytes or len(self.data) < bytes:
                self.file.seek(self.pos)
                data = self.file.read(16384)
                if not data:
                    self.data = self.data + self.zip.flush()
                    self.zip = None # no more data
                    break
                self.pos = self.pos + len(data)
                self.data = self.data + self.zip.decompress(data)

    def read(self,bytes=0):
        self._fill(bytes)
        if bytes:
            data = self.data[:bytes]
            self.data = self.data[bytes:]
        else:
            data = self.data
            self.data = ""
        self.offset = self.offset + len(data)
        return data

    def readline(self):
        # make sure we have an entire line
        while self.zip and "\n" not in self.data:
            self._fill(len(self.data) + 512)
        i = string.find(self.data,"\n") + 1
        if i <= 0:
            return self.read()
        return self.read(i)

    def readlines(self):
        lines = []
        while 1:
            s = self.readline()
            if not s:
                break
            lines.append(s)
        return lines

# -------------------------------------------------------------------
# expected resulst stored in a zipped and base64 format

tet_conductivity = """
eNplVEGSwzAIu/cV+cBm4mBj+P/HVjCbFLE+dMYqAlmiHdcxrs91XMd5X8/BfRznpc8xAPcRsOTn
PH7O+/lyA1i1WgEoKhjZybRe6K1whJI6eYSU5X/nwv0uBdMBSAdCnzyU6LkqED30OIc8B/f9vd8x
0zrBiQAz8JybXMLkO7S9Ng7cJctmKbNgr2qgAiC/smRXZEav6l0CjuaSSqQrkcxLwpqiR7oSoeRm
VGjy+nRJR4pk4eTQa0IJrcWEKMphUnRJkQ6AdOq7hzP0vS3j+zBqPTkMALsCC4A1hjNj8YIpgMFh
L8ikty7esOgxaciCzPFdUtxJZhRUmdiwVWWuYDi/Q9PNqlMj47KDmjKre5rrprn+5cWa67YqE71e
5sZ9l+fEbOtivBT4Z+fS7SoIb9osCD5tCOrI5NabxGUXJTEb4qrYTeKS4ESwHrD1gI3/QgYA4cSt
BWx1DxGXccAxtAVskFlXwjhg+fi/gL0F7PkbdvbQAm8BOwe8ACjl5S1g54CjpZPJv0vXHBY=
"""

tet_var_conduct = """
eNqdkksSxCAIRPc5hRcYCn+A97/YtNYkItlNL6xKFzT4TOaLEycq9daV5zfrLYORE7X9XVDAOOFy
vyUwSqIsP+lVV1mNZRVl065HcUPxh2q+BSPDaPbTgFFgcHsEp74ctBE/qX1eJe/N+wot4gpmaHkE
I2Z2ZBLbXqQnxFB7WmSt7qbIvHLeQGWC8ZGCIYyzzVEuWZD8oeGT5LiPLqS6kLo+jQN00j2Ndu6o
a5QjoRgVQjVR3e9mgaVFlvZ6IFswHV2LLA1rkPgCQUff7AxbhCl20B+R/gj0x/rfRiQyXvTHQtJ9
J5AcLZpCh53vNdI43utPfQEdrZ/A
"""

tet_mass = """
eNpllEt2xCAMBPc+xVwgeQb0vf/F0mI8Exo2XtRDUGqE2/1q93W/7tfvfaf2EfUVkFakafjANwdI
J9JBxkHk9QNkQ7Ljq7W1TvTZu9XexqscyGnVXYVxFuZe2N7mi2d7mz9lAoVG5nYHyGoezUDYfNQi
PU8zclIBInMsA4qzMPfCvpl3kLZ59rf5Z29NkNV8oN/+mH9zqo3IXLwQZ24O5LQqyokyl6nA5inX
OMwHmVeegzIvz/E2/+wjAXKYD54WEQfizHttTuY9C1HmfSrwtHheAvNtPgXq28UL3DcvgfyBBP3o
7TmsvgAK0HWkRa/+BOb/YAwAB2jdMqR1bwCxlySX6GO8vALdjDFe+hgvF65sLNKAhE/Taby0oGQs
dbwvwOrw1XgYQHJPtmfsQG2ffmPjmkfbjHGrRsaIwzjjAra0MHdxbtKmMQmuGUu/fMsYPx9n43rV
vmWM5+KP8TKHPo2X05yMrYqMBX0aPy00NOmUsRYgY73inIo4pyJ4jus1Bxlrtlq1TUWQMa44pvHS
U1DGEIxp/L2F2oOM5cpzKvKcitxeXhaijMUa0DYVOY0XweSMO4Bzgskvr85Jfgt/GDFNjA==
"""

tet_var_mass = """
eNpdkktyxDAIRPc6hS+QlPjD/S+WxlMZC29w1TM0jYD22te+fvd20WJEi0UfIhTqiFkg9ORYCS9G
zkakOzL+Um0Np6RcMgXCQegk3jldhI8c3aP2UhT/HIZEgGggLyAeiB1IBooAUqizSXmyKS/7iBN7
pRIHgdBJpHMwHKKc3B3kFLtLDQAz9+Tme/lU9wAZ6nhKv9V9qkeTl1W/1SV3GSMqgD/tfMeK1zAK
Modp0u3ibMe7HKTbfUsN4GxnCeAHYAKIp7+6rHxtCqeTc1MkBjQ2RdXo3JRlKdDp51ay5z2UOsOn
47z9/L8YbiKvfEBfRL0NGtA8pWo0DOruwnlKOMsaBrOBTT91G/xusDNOg6YAeY1jqKvOEWT9AR/Q
pag=
"""

tet_stiffness = """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==
"""

tet_var_stiff = """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"""

tet_init_str_stiff = """
eNpdmF2W7CAIhN9nFdnA7RPKX/a/sat2onydk4dhtLEoUIuk++++7uujNJ8/u6b1fYZhe0BxQJdF
Q3tWGv/+97ltPsMIs9KlaKQxzep42l+OnvP4zXaQ42/ylaKRj4MSHZS4aBkOtrcSHZQrR6Mcb3Wy
cb9/B2c1oqnRWY3O6lWiUa/XcYsw26T2fgcUB1I0cjRKNGo02l6lx1V6xN9nnu53VooDORolGjUa
LRp9L+mBMY8regzM45Iel/S4pI8lR8rKeOowWhzp0fC9vt0o2jtQOwxhKMHKsAqsCqsdUMPqGHMu
fh9g3E0WuRnWSccwEoYyrAKrwmqweoBpAGaX/WCxg1Nr29pzBGB3m8CgAFQAKgDVN5Gaz7Qaxjos
x0wA1WVEM95Pn4//WQK/aZ0dbxQ4cCwBdwLuBNwJBKdv5h9sCbgTCE4Td5wK3OkywhnvDgMnoGXQ
n9cR9gaFQ9AywsgIIyOMDPrzt04epBlhZNKfZxhxLsLIlxHPeHdUZR0LeT7TiEEV5GYezHsn4GC2
gpjKt6T2zIaxDmuV1Fn8xqARjE2K85mcOEpAFhFVZK6uk+fxg+vCcF/YvDB2HLgwrCJv9Vt+e2bH
mHPtO4RcGWNl3uqM8YRcGWP9jdEiJFxj1pDXts60xytuMsNVZvMu22HhLrOGtLZvqe6ZzrV/oFhg
oDHkxrS2GfJhoDHk9huyRUx9nZZ6MOGOtY6s96Vz3s2De9Zw0VpfpX18cr0Oa9x8xyUo6Mx6X4dl
eu6DTgY6k94nA5/72budBPRfAsb7Kc+Z5CgIX4fwGwgUgEECmC8N90YCEWCOkvC1C47PjjHn6vcV
fIIOZ0X4OoNfdpx0OAvCJx+bHScdfhkRjfdlR9AmuiGc73W2P3EJ4kQQJ7qXQn3iEsSJ7lgsuteG
OT6dq/+AsSs4jexMk3Dm0f6QNa3CQQIa9LxkTatzkJDGu8my1Yk8gCCXZLGSZOvKeFaAXBLkkqZc
+mgbDUMdlseJIMpQR9OceTugwZShkKY59pW18fRlVQ4Skc3k1UOAkwGCGu+nfXsbsT3TatweNxBw
goCT1kX08Aj9JqHIpn7b9EC+SaixJd/CTDAn1pi+19BLpMicWGRaW/AlUmROLDJN5g6RInO6RFDj
3URCUgotrNLqdB+nUJSColS6zlkqCEol1OAUlJss6Ekl1uDSk2EqiEwswfS93F5eE4lMrMG0duvL
ayKRiSWYJpGH10Qi0yWCGu/mNQc1KEhcoc1XDmpQULiCwlX+qsHtsmGsw/IjIkSFq8wSzVCDosJV
ZolmbO5MYjNLNJPZTGYzSzSvzX0ciwyJmBQx4WuISpCkguwWvoeoBEkqyG4VFHD5StLtsmPMufYd
eKfsVmEJF0hSUXaLsnua4SwoJLqQ6EKiC4kuLOGyzoLjWKToh2hFTGgFhC9HqkEXC62A8O1INehi
oRVQRYHXry7eLp1r/0CxkAa2Aqos8ApdLLYCYiswzXB0VPJeyXsl75W8V1Z4XUfHcSxy9MO7IqYW
xbnQnggf2tSiOBf6E+Fbm2aDEiY2DHVYzsVvDBrBBHEutidqLH+2J2J7IrYn0+yHvMY0NKahzTSc
A70xDY3132Ye9rXWmIb2m4bx7hsIHyXVY4sgdEzCh0n12CIIHZPwbVKzYwoTO4aci99YHUnp3Bw9
tghix6TOpLBlElsmdW6OvnqEN0edSencG30m5eSoMymdm6PPrOwcdSalXyImnTZOaOOE77hyNCpo
44RPuXI0KmjjhK+5mm1cmOhc/AeLYXnkyLlzHI0K2zg5c8Q+Tuzj5Nw5vjqVN2XOHDk3js8cnZQ5
c+TcOT6TtFPmzJFfIiad3vI/ipDlZw==
"""

tet_surf_normal = """
eNpVk1GOgCAQQ//3FF5gDTMFgftfbKW8NSF+kLZgpXbqFeXST7nWc0dT2SCu4jW9vuj6vYtmaWkY
0FtOy5ExZjMM6C1rvdxrsG6++tjo6zEM6C23JcedWbMaBvSWH5+eY2athgG95Y5ph+/wA37AD/gJ
P+Gn+XAuew3WzQd8wK8sVkRD72UcjOlEXlm8amvKFwlVqDrUilpRK5YNvsGvJOJuJZXL8kF+kB+O
dfgO3/ex3ou/dCAP5HG+dSJP5J2Lr7f+eY+eRnGg9zr/YAeVBJWUxt147zxrjjCKAznMf6hjq46t
OrdW3CpuFbcG3+Ab/Aot7vLUUqphnDBXBT/cvVtd/QnDOKF3f3jgOfAceM7Tc56e8/AUBZQjZkBF
C0W4crjfgIqMRQtFC8VEiv7JUX6DKwZThCiH+A2oyFJkKbIUWYoCigKKAooCigKKURT5yPl8gy5i
Es2T8/kGXRTwD5l0v+8=
"""

tet_surf_scalar = """
eNpdkVGOAzEIQ/97ilygEWBI4P4Xq0Ha1bQaCclvIA6OL5WXLFnvrQGxqiwQ6NqCkjC3zAa2tppm
hbq2xuo5Xzsvv3MMSR1Lt3EGVRrUZ20eaI64uNR3pnJqsSqduyr9BYk8qYgksb5ABDQ1rxLgAY4R
+Ey2Y4jBTkRRn6G36b3C0WqdP13tbbO11eWKQhMCHdyrHjAKKI2snf+1Uvt0xdTDg9OtpOD97y7b
uOXmjJM6p6seXfLCb9x4xl1yCHg5VjxCb/odN+YOmJUx0WIW/XsAp66vB/gAiA5WbA==
"""

tet_surf_vector = """
eNplVEGSwyAMu/cVfGAz2MZg/v+xJUJNIaQHzQiDVQlTkiXJn5zwAwpRiUYsRCdWYiMGsQPHKenv
yvX7NXCyclHAabrE+AkY4wklXfotdjCeLv/WBpi6MLOmcXcQpx6FnqeRgZL0U+MVlJ6U8YAhJ/L3
A+NJnuoGph41jbuDeMsR2i2we7dJYP3CuYKztKoXBiII5GfT3F+50hZzBExsdgnjEsQ18dTzjkgQ
0R6kIK55gr96CPXIpifAxBaxMC5BXBOFqEQjFuKdweqLstvt/G+lgImj9u6GXvflEC8tZ7cKSrik
61KAGlmo5V6y5DaZwmInjntQqkRkHTvBjP/e1VrNxbKDCdb2ZcU+ytF5WnZQUw2ieFo3MHYWFxY7
sS4d5qZ2MMHavijXj3Jwnp4OaqrBtGzWKK7pKkYpRinmjmeEEF5LrqWCuYOK0lxzVAETrO1brTGo
9Xrae4SGaOPrZcfYGKJax8UYmnFsjC+avYZFwPTtahvHxpiPIZ+JRixEJ1ZiIwaxA49Xys5XyjgM
Nt1+L+2vlNF441wYXyTjNOTFiNvude8/40EINQ==
"""

tet_int_scalar = """
eNpVysERACEIBMH/REECWoKHYv6J+bIO9te1o4Mhret8QzM3lmnMyk+6PzmeFKzy7V8WRClPKREu
9hAa1w==
"""    

tet_int_var_scalar = """
eNozNOAyUNDVMzCBAy5DNL6RggGXMZqYiYKegQUMWHKZKugZmkOBBZcZCs8chWeBqtESlYsACiAE
AN0qG2I=
"""

tet_int_str = """
eNpVjlESgCAIRP85BRfQAUzM+1+shcoMHZQ3u4AKNxLGQdbMhqxJNIkmsSSWxJK0JA0kH4vnACuz
tjcAlIt6lScAjMvYFD0sVv2O0QGUv1JQY6pX8hDq5nQIVb7WHq2hfL004NiXGWFApwHhvsIJnf7q
WFoqPjn6aXjShHCtigEzhKvuIbhXIKYVHPcCxGo7eg==
"""

tet_int_var_str = """
eNpdkEEOxSAIRPc9BReoAdS23v9ifyBpGT8LE0bmDWoq/VBRaaZvoTdp2r+C4NK8BiwNdW9puN66
IfguOBznxvTwENPDQkv0yHDu7e/e9yVHRkwWDMIoy4DlbE8JMzNqy5kOetiMECJcmbEJ+AucAV7l
uwM8ubeNcweXvufBvOKMeGcZY4RdGU8DK+NXxhNOjh+W/1MQ
"""

tet_intrnl_force = """
eNpNzcENxDAIRNE7VdDARgZix+6/sR07kWY4faEniOZlzZv/rmwYdKDvrxM9ToeYEBNiUkyKSTEl
psSUmBvmml/i2Vkjcedo6wSdoBMMgkEwCJ4NnoVBbtDf3KB2ToJJMAkWwSJYBOZubn+ZFTsh
"""

tet_strain = """
eNpN0MEJxDAMRNH7VqEKgq2ZmKT/xlY2AX2d9En8ML5j/UaMmNc4UzEZGepQjA4z7o5Z2rX2uPb5
7U/tGdk/iSfMgJWwElaee2lPVognzIAlWCor+wMwERMxEfN5snfPquibmZiJmZiJ/QHsZUH+
"""

wedge_mass = """
eNpVlFuyxCAIRP+zimzgTolgxP1v7CIk2FjzQcujj9TM0Lja3e5fk30u8rivh5vFlPcd7rvff3HP
GvYa3sfiXcMkShZ3iDnrJer9XqLeewXqxevDS27J3gG9A7xG9PqcAb3De7945JzH5vwdQSj6Fjzt
Y4IxIyjGFjH5uZ/7R6SD1jVjcvRPnDxj8if49E+cPGPyJ2xyG6uPPXmmiyK/hktYKroo8mu4hKWi
i4KLbpeMNR0XOi581wrHmLvQceG7VjgG5QLH5Y5fu0K80p1a2L/f0Ib+pnpRfHxMScmNop4bRp53
m1DMLMzYJ6kIl0IUVG8bFSrCvZgSYKRCRUhFsBsTekAIqcipQNFh7MH42JlbOeOw82zVi+JSKSU3
itqMascN5vkf6c6YmYWCGtZtRsjZ75emHb2Ig/j14iB+qbgQcxCnklI5Sg6I2YlT6MFiJOYgPoqw
cBNDjg+/FH4pG5fgfxml8EvZuAR/Vj7HXJBfcOPi/FmG/FI2LoVfnB+U5Gv+AV+5FK0=
"""

wedge_var_mass = """
eNpVlFmSxCAIQP9zilxgumRTvP/FhsUopvLBGxBeSGdAnva299dYUfiBiIlHmxbD/jtmDcwmFkOJ
0WIFHO2hUkNRQ9QFLPYa7kBqMe2enLOihktPLvUc9enDL++zkrOiRvLsiuPsFPQa2j0lz65YtnO3
Pn82QNsUA6iAFSiAtKMBV5AK3Tr3MVCekZ1FGnaD6Dy6CRhE50480SA6MxKqQXT+Gkg9UzuP94eN
YMijOSWLNKfkSK1TNKfkec0pOVLrFC1T1KdAA9Fhse6Jc20sHniujS3ACnTWN9fGFkiF/p4jo8Ra
4rmtoOX4lDeC81xGeJ7FiM42jPjKyVmO0ZEwOM9t4BrSxmSHy8PubycAabVykFbLES4rSKtNXBzh
soJqBbGbJQJu9SlCtYKwOim7tyOunwrY5ZS/lUb2FQAuxy9HF/FVKVfOHWWiTIdRwR1BBQJmzbhj
IaiFdv+A0a4HaH1COZnWN5ST6TKm9a6/Sr5yclExpnjXazblu/4ys0Jsdde5ccn5f5zd3e7tz+EP
rQ0GJ7jI/QHZLie6cnyRXJXFn8N/2KUOWmHWsvDfqdj4IazGt7/dP8be+3j+ARfxEzY=
"""

wedge_stiff = """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"""

wedge_var_stiff = """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=
"""

wedge_init_str_stf = """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"""

wedge_surf_scalar = """
eNpVk1GyZTEERf/fKDKBToVNMP+J9Xa6qp/c+2MVwcbxJf5z1ll/tqWcWwQhoG6hQen79yOAnluO
AMHmG59hd+lGmnunjunJCbVkpxT8dtEzXSIP6ROJmV7sN1I+HVkZqgSZHu3WI9KNQB2ypVzRcU8G
Z1x4VQrhTk9MyIVdFlJJqOmZQqSFYOs9dW+TsvVKWKIJT6Qt2+pZlj899P8u/RaSeuw6oYWEpUkQ
MMN6ISUoLYLPN3eGBTtyT1Yn5PTUU/R05yqGTjcXoq1jEp5I+82PbyG8kGK3oI59hOsBbaUdUXJo
DxWY28A8K1CFbE4y0wjxncQJzhVUQU/cE0mo+WZuA48KPCrwbAPPWdnchs2zsvl52Pd5iOOgwSgQ
dVxp+9qiVO6073wRE3JCrZ3R/67IM+eATe2rzzFWlpo1PA1MDfZo+AvgFZsZ
"""

wedge_surf_normal = """
eNpdlsGV5CAMRO8bRScwPIQkQPkntiAu/po+lfnWFK5q2v4T/+m//nuf95dSfs229BmpxndRyMpl
NWZopAI7yI68+SMBK2G9g2e4Lk0F1sgaDBtZJ+s04YQn4Xngv9F0m7unBr1IL45ehDfhTXgTDsJx
fUjbEuozNTMpAUrncKkZ1hCl8CVHKUHKKH6kZCklTNHyIKUEKiVRseIIoT77bapbjFRn+leCxWbl
h70KK/tEsxUqnuoO/kiwd5Pt7EjOM5As7N9XgjWaMJoou7uVbbf5e6W6gz8S7OTgycGT8CK8CC/C
t7AnsxXS4+nzBL4SdHB0cHSUSHrJpJdQaoTyvGj37fIuXDNfzRvGu+F4nSbvwr3hq3mDFkdaHJU8
s7B/2rqIq7wLJ3Tozw2DZ+z4YcODlX3n5OnS6DY9pXAVcG61LdsmK6VwFbDShhJWwrnHFqIxIqVw
FbDTsxN2wpM2JuFJ+LZWm/s+IaQUrgLenLwJb8JBOAhHCaWnjzHENJ6Wsk5eOF1qjCVHnrLjVZaa
vBY/JUwpaWZlP89RSp5srLKxysZqNvavnV/cOK1THrLKxmo29nw1dmz9p3wtUBZWs7Ctr3vOHKVk
lazSsRE2wmV7TtgJO+FJeBK+fZU2e+xtRy7Ci/BK+LysrDmP3IQ34f0mr9nXPjIIB+GgZykBSkmQ
fdXSVy191dJXLX3V0lctfdVyxGrpq5a+ankn0NJXLX019tXYV+MJa+yrsa/GE9ZYWGNhjSessbHG
xlo2tp2fiK5XGVkja+8rptF9HOmEnbAnLON8yfzISXgSnrS8CC/Ci/AmvAlvwkE4CEd63ut+7lMv
CUqJMAt7XjvHsGF5oaZYY5T8B7FjmF1dgpSSpJQopWQpJUw21kpjrTTWSmP/Awwc9tk=
"""

wedge_surf_vector = """
eNpdlluSrSAMRf97FGcC1zIPBOc/sSsBkRW7+iPFNm7IOrsoPyk/+zt//e/f4fFEKc//eKLUZ1Xj
+RtrKCEWijV6r1WlWGN9lUYbRrFF57XqtOEUO8WFNgrFhTaufgCHxROlcPXZxFH7056ysnOluLJz
o7hR3Ci+aeOm+H7ER+vP3U/9ZGdJI5STvUWSPk9Rkr5X2+fSHEXjUJZXsfC+DknSKMWS3pOdNE3x
3Y5MXi9txe4ohWWfvYlr7WJsVcir/ISdNTqvdzU6byU6G20YbQxeV+m04bSRNlhoo9BGYeeLnS92
vti5UlwprhS3GGNr5ZIWpbB8zuMo4lLqU97sfLPzzc4EVhKwkoCVAez27QCWdSdKSz1rP5MAlrWk
WvG+JT+W/Fjy08e5bT6IZb31Hym4XlcmrP6wWZ0Je/WnRilflikTViexSzyI3VYhNtow2jDacHZ2
2kDC6iR2iQttFIov2rho46KNGie9OtcY5IwODWJX0igTVpmwyoRVEqskVkmsDmI3H0HsZiSIfSNS
U8JqSlhNCasDWNaSauot2emz3D6PhNWX10/u4X6d2+R16m0mrHxlH+c53jbyOiD6FoVaoVah7Zv8
vqKUAlYjrEZYjfFqjFdjvBqvA8brgPE6YAHr0b7qcayv/4unViN01u46qkf5FqFtaNsgbZTetHvT
7p2GcVLNZLWUrJZAtQSqJVAtgWoJVEugWkpWS8lqKVkt3QUs3QUs3QWcyepMViepzruA8y7gRNV5
d3XeXZ2w+rwLTJR93gUmCU5YPWD9Fp3atL8Cbdl+ME5SnbHqjFVnrDovAs6LgPMi4IxVZ6w6Y9UD
15dsD1rfH4wTVh+wyre9gPX9HXhi1QerX+tAdVOn+UkaoKQJShohSfVEqidSPZHqiVRPpP4Hg2cK
8Q==
"""

wedge_int_scalar = """
eNpV0MsRwCAIBNB7qqCBMPLx139jATOJ6554IwdWJalXoUI3j8wMCEIRhnBERTQStpVARwzEREg5
JIc01DMjZYd8S1YPfc+Q1eNH9jDJBAxfHFFxrRFri/SYO3F975FosecJO9FhP0QFgBL7SsIQTt/H
PUcORdE=
"""

wedge_int_vector = """
eNpdVNu5xiAIe+8ULlA/46Xagc7+KxwuvwXEpyiGIGlrwkjtKolXy01CIBI8rLQ0Lj4r6Q4YaVg2
KPc2WDkbHiPdT3445qXUdP7hFqW0KKVFKf2Q0qOUHqWMQ8o4pIxDynPkP5xfm9+gbhye/HLTY6Rl
fJPFL1O7mL97TG/XLX8xvbv/cv5r9C/racUIXm6guo7Aw3M3oOOx0RVp2e/IdI8dWitscBn3MNAh
t7BD2kbYoEuehQcdMQ5WHrbHPO0RMCnx3fK8PwxxNTIkBCLlbqimPDieC+Jpy+SVmyqDvJhl1pDJ
bea6KF5BPrPF8k3KT4oliFI/pT2QdilvZ5TZJQiNkMkOztgFRyiv7t0PDHVvXpuWvbvfEuLc6iHx
PvvmlDlUCYLrIF5CXDYxu9YUvvoNfZ2KZ+lVdiWxrCODDu3LF8cauRrWzQ0Hv7j1Nu1q1jz3u6tV
q3WjTs1rzxC/H5KV4DnaAKB/JC34d6PoHhcxET1ekYFaBf0t/Y7/AXkb60s=
"""

wedge_int_force = """
eNpVzcsNBCEMA9D7VpEGBsX5sNB/YwuBHSW5WXmy4aQfJqanDdC6FUAPWpebhB5pppEQcLqfEHDi
vgKOHj/JUAqUAjVPa5nWMm250UqjlUbP0Av0AvuCaHxefUFpLDcIaeMz/A1mfsNmk24QsqbnM3Lb
yG0jt83MZmYzM3BeBedZcN4FciOQK4HSKSGFmU/a0t60Zb9JQ3b/J6yfvr+9Pu/PirQiLcsfvqhf
lQ==
"""

wedge_var_cond = """
eNpVVMm1xDAIu6cKNxC/gPFW0O+/hS8WL+OZgxWDEMIJ1edLX8pddD2ke6oJC3v88ke6HtaDbxQs
7AlB0xI4ccq1YfWnaIzlJuw1huI5X/uScqtY7REr5s9F44NHLN5ricZHLUmS8mRdT71yq9Xy+Kq5
7PzVcrn3PrBHLpcxxsS+pizdnjfwvFmctIHozW0BTi/lbnF/Lzxo4Hszr2NR4IqbMpI71RJocnMb
usqUYWW7sddiOruyZ14nECpO1Y1XfBwdvG/+Fmg6gsgAQS5ecVgPxRsaakZwDTc+glT8zoAbde2h
PUoMlHhXjYEaMGCHgTlPd3a6a97jtL5C8dSSn1s0rWRbUdrXWKCeOlP7Clum1jwHA4C8r5nmsZXs
zkbHANfYgPhMCUit3VmqoXtDQCoirAK6WgfqKY8NVMdBM11xWm/ZSuS+xCHZKEIImSzyNoHUmupu
AsnphtybkEUmi/ZZv5QQdO1uCLKOflJdZ3iK6YyP2HSOeOfZ/FsV2f0Ly9iE9p2nBsaNAlKlvFna
vkcA/Sdt7LcAYJ6JsOnMNfpjlXnr4mvmxb9NTllM89hHfCmxb8wSWX5mXlxySCk/5had+QnUmc99
BM0nTse+/Sqq+ZZCfD4MCstlvJjxy1xx4+OGiDUR7xGJf2w2kmtEYl2UzdnOmw3Uz6sJNM5QxK7I
uoLi1p9Dtf7koYtLC5rYdwsI/yXtH3SFFAs=
"""

wedge_conductivity = """
eNpVVNu1wzAI++8UXiA5wYAfA939V7iCODb49AMVGQlwS/p7ylPuRnZ+ZDFNwUFM5Z4Np/+qf/+o
akNMm1NL3XfZONxxBmLjaMFBDA6R1+HCmy9e8+WI8fXVEud/3/PWkiLbj/pdWvHRUru7PKjffT2r
3V11tOj20FDnusUPAAG0V6FZpVq8akOl665+ACReQa36xShwT3XBbgbX9R7rdtTd17tZrF8sMaFH
saPw/QycibgfkeHuF2vYGOSLa0xwBBiEvFZGFBkQue7xOh5QCWAcyRklpze2LE8f2BdzTMhpf/rA
3sLT+urvIqYrriYnFK9b2Q7APPLkj3WpAAR9oNA0EG8hgGABSNO1dpSB+vYEEHwAzRJ49tm2KI6F
/JezhgxQU4oTkr0zAE2pFrTJx3PQOCYJtvauTPrJkILN6jbbSlYf37N81uSzus/DlMTUxLQfUdnM
HmZbzegBMxHpSUyiSLXPngy77U+e3XZfLbHbPjlOSBJTU64dPbatH2BbH2vRbFs/KZiOiBKTanhJ
gBxGL+mFSBq9eA/fsCU9EUmjF+/hMFsYofhfw/vI/y567JswfvF3cpC/38yOKxDrJiIOvQHK2c8/
NfUTSA==
"""

wedge_strain = """
eNpV1UsS3CAMRdF5VuENxIUeH4n9byzQg/g+j1C5fBqQ1FI960972qN33WecIJ54+310Aj39TUl5
g/78fdvvOcF4vvX8XsQPi9jjfH+o51WvOnDQClpxLATAREy/nVW1ueoE0M7y+T4ZWMPqxzpfrIi+
TgCr0+rcWT9Y+4L5Xx4/bO5Rc53gHOeN2tnqBMAGNjZ4ysGdTWLzXtloM0adNawJa+KQk9Q61FvR
tNdZf9KitI7Uvi8+auGEyaJIFkUe67zJvdcNYCWvK4EV81goiqJVvPridRWsTWszj5vYxnVtWhtW
NBZFNGrRyEWDF80q1us/mM0I1kaEkUEykIcI7lEGClkNmSdrKjZCiGBHkUSn183rzG10euyFGCyV
GKyVuO1wIrXW60YUrSNikJz2LzL5NzJNnHbmaeLtjO/XlpHLUr3MXMzLMnJZqtOqJ41MI5NkGplG
liW7rHrKyCJp3RJl5DZyM9/bxG13aU0T7Bo1FJAaQDWCaky3rGlkTaNgCSlYQrpN8yozVTcgaZND
Pjp8dgglJFG0rpFMtAGibqKNEHWSnfOom9hNHCwgDRMHxWGbtMYRG0c2SzRZP5oUrXHEgaJJcBnI
maJFj00jDhVxqihZOkktqaXtLslZv/wDyQCC1w==
"""

brick_stiffness = """
eNplXEmubbtt7f9RaAC+B1siqaLhzCWFGwYcBIHtRmYfkTq+m2vp431ciEfcVMFKJKX+/PGUp1QT
+/Ta9p8/6m5rlU99bP/ZzQo/N/y57Z+h2aC37N6j94/Muv/s5v5YfT6y5uxlN1v5qePzLH3q2E0B
ZPWRvb11I6dvKSLrRk4D0aLwLfNv/XY2H0dqtTLnpqN1zrKbUn7Sp2x/CtsGn+44x76/ncbVfUXe
SXQf5myf9diaspv72/2j9WnmqFZSowOZATMYm0oa83Aq6Ud5vzM2Cavzs//f/3bT9vD009SeqrvZ
YXKjDKA6cQ8mTm5usmnq01ful/D0ub2U5ib8O461mz3v19x0f6A9YRzL2e5dt+Xj+KW0nBVSS/Kg
Vgzjd8gr5v/OaO1xQO9BP0/EXigsLj2/lHer5rXe7b0tvyuwm5J2aTc17+FuW6a12x3GsgEjj3W3
J35vwRr6cEi2Kww3ZPsdXt2jTTu024L0qwvVS2/PFYZTnWPzzwOmsxkWplt5uNWHSwBUPtW1T/5G
A0GuDcRgNwW2o+3xv7NvKAm73WH6Lcaff5+AvTlhtI907aPFSEBSHFAZgMqxunbMKySo4aqg7qgC
8rWbmmVit5P62K0Ow99CAYshztm5+yJaPp80QfH5/FCXhju4IajCq9IUFS1AVdQgVYPn3lErKMjd
BCWy2x1/HiW3JqyP0hTVp/hDkAqT1muK6lMkAJqa6rYmr7Rbm3dUhjrBrU3maEMZM9IJFjL2DtBc
X6cZWOgrACxYMTuTzgOoWZd6G7fEfMb0zUqDrGgSq9vE9E23iXknOijs2mPT3++5Vcwj6KSza9hG
aA9Yle6KEX5fMOfui5C3yD9XYLyNKVahMVWlQe9FAFbZEDThNWz4O4wBjDFI/Q60FiPsWBrSIPU7
2FwM4o0R5uJdpbFXJZOvsGeOC5s2zqLkEVUaYUWD5q4EyMrwFfkhCDocdWbpWd4GI+UuxzvIGWv0
TmnyIrnXge0OQ5xkpCaa1JmlZ8ZgYJHmUYp5eCQ9kxZpHr6BIfkqpYWevkpAYxBN9I3qQrZa5IUs
WLLFVn2FxnnXcNGSLTTri5yQxRpnAVutMOppdov0zWK2Wr5ieY8WqZsVkgZf6DThiqZ7+ZJlttsA
dOTaAxqrPeBgNnfk0iRaeHLvnJu7cti2vKbtuHK/U2gPuZnNXbm0qu1BjdXClYPh1SwlDmj8xSpE
syrRqIbf7ExzMJEJ4uuQxZCGPmerBRyJVgv6Ea2WxgABuuF3vuNyvzPtTS2w+y38zvzzROSVmavV
Yw/fZapHpAGAm199ZTOBqtgEveyALEDeBr3sgMkYi3q0Bz/R6FTeCroy7XjG76o3XmR3jtO6NDwm
toaL7M5xbg1YEXCN23GNfwcbnnGeHnvGLTxjIFaFBlMVB1vBLXIA6HUHDO4xmeiiYeEiN1/kzNob
gA58E151Yd4OFz5tg+SjeZN8IGnuv+dVdQ8+TVqQr91/z0IiuOpC55EW/nuer9CqH+c906tKFKrx
JzvRHEwCjs8OWARocKJzAK66+KoTAA8ZLQ4ZffZPf1afwwFbKegnJqbe3BNt2j612/7jAGEERYTt
pUP/cFvss0y1dgdsb03lY3MNDQSf58aItQuMxRgeDQAUlwJAOSq8f+azaj3T8BWUz+hPnwdnOw+j
flqzZ9QA2IXTGWdvCqBsQVi7g7S6ToftE3TZGz9ml1iuh2cTOinjuP5AHKHZBFaC+LEorbGFhU17
YLxLfjQCBGUE3KU4G6WZ2zGy79rE4SgthZ+NEOEcCDOGC03G8BhD2sVzOMobbcfBg4n7oQ+R+oU0
CuFMWGALm5C2wK5dMt8lxAktBzhCZJzXa/vM9lg9KIbbFkc5aePTxaRNB1QGON/2NT/PY/uPQ7ae
gx5KbbswepqMOGDAXvmR7mdu9bT/n9UcsAoiuJnJCG7HEcMFDLsILuE51RGS0Wd7IZRxoUyeTKg9
wPLda31z7zP3n4BUmlDYPcCRC8d13/kvmobNvRFDxmeuOmSvsp8/ge/8BPqTWXPEXiaiowijKKNY
QYyOcjrCcL2C7OdQEHQ/iBJGHLMyikcnAaUia8dh9Cdz/7gEMc6jhNQvpEGzic1MOmGczUx6Y8Rm
JiYZvpeE1NqFJAVxtpnpz2cvQe8H4AqpffYK1CfGFsL4r+3dzYGiOsMdSTROuD4NIwL2L80I2COC
MoLBzGZs7rsxcX7OW+cH6LxRM/Y2I9SHMWpFfohDNGzLOUWnnZvHI8xzd0tIOJ1xBjJEHKWBmef3
QPPy+wy3EJBaZaTWLiRhUs2XNm3n9P3dI9bP9gylW0DyBk/f4J+snTdk4pav8qCq8cN/2q9IjWTl
tIoU7I8eReRHEKGjuvcAALYnKXOPAGCP+vA34pCFSLVdU0FrFYEAUIARCiCcjgt/ggGENGnOdREh
F2hchcbr2piObznhZO28UDuvEtbG5qfPpbYCMrDDxOYCbS5PQW9FImiRxFkeEHd5TugnIygjgLiL
xy2SCEnELZKQSaSg3lUQj1sgQmw9YJyQWcaJg8irEuTELl6tIRG7SCpCnhMDQpzOOAP0ivwrfvEr
p3LiF0mW5SFxd0BlpNYuJOE5ubgn4XaIgQJwSBJ3b/rizvHRrnKm5Pr3lX8HLMCI0SWyHmuBkUWw
JQ++koGXyPIhijIKGHipBY2meMQlb3ql05BEqg8x/HQKKHgaknqY4t3Oengi7Xg96iBtcD1cAUj9
QhqFcCYKQT0GPslJJQMv9Rh4QGrtQsoG3tua7bkDLFt8B2SGqM4Qefur8wNwSEWGqM4QP4gSY00L
65GiNKpWSBobWQzxQNEPYIDllUYWQ1rpWSfLN5GaMSZoZGllgeqXk00FlHASAMfzu69NkXY8QIQo
Whk5kSPq1Pk748aaNCVnEVgUshjSDosAUmu49GQxHJIshjcNm9lixLiyxfDmxOa25k0/LXg9dvKB
30NfpGZDA+PhrSykHt3KUhzBrbRJUcAD/ZX6GzLCyVEnmfEYV5YqD3KlFZOyuL/zCSA4nwBKRd9M
Is6VFZ6QLynyTeUBTmecwXNxPZJVTQS7kjKKWFdWPRHrIhSPtAKKoIqTY1iSDpAwLFnoBfWIHMOS
bJGU8MNfVSSoR8QZJWsmcVb5QYzQJAlDUM84I+fNj8x/Yg4l5vGYHPZX6k/M40G5vCV6SojeTVPy
TTwmhwj1YQz2TSIolxlMv9nwlwcjKJc4LmJyhNIvFIhuSATl0r5HTA44Q5l5IiYHKK1dKMQ8EZQD
VtFzCEnspMg8GsyTd16deTJzKDKPlvD6MoIzT+Y/ReZRZx4clWDwUKJ+L+nSqOCD3xuaByuC2tcK
RgMkSisQped9jNIK6I8RKPHw4Q8g1IdJUAxKIoBIExGkShEosRPjRZzOlMeFNMGqRfwQARE/zGYu
Aog/1KldELnRstUytFrm/JStkKHRMjRaVuKYl4ycodUytFpWwsy/xyiHCHbA0fm/bOU6ZA+6AzAe
IR7gTJLWKSQmnVmrFwj4SoeQWGCgE+rRTaCwmAL5Q/0ExDIJyh7IiW4mLongZlZMneJh0vn0LBHd
zGq5czxMIrgJZNpDZDB7IBzZlIhsIhWPS28ea5EMC8CWP/voeJ5p0d4jRcBgjMk9FgHCWwYcP/1s
bhd79p8AuPeAELmQ9OpjF6SjQhsUVpcIxYJcRSwWu5BOi1hsFvtB53ePxeIXxkVkUg8Iq8sJxSIN
yInIoLC6nFAsoZCEjEupnVhs1sfj8rEiGIuUFlOi4/ugsLqcSCzSwayIRCj2l1N6AFzFbpPnWcsg
7Eqtry0Ej40SgFEQYzLGHiqiePwlMck4ThYghU+vn/U8bcb0nPsISS8kN2yIlU36bg5UgRNSczJZ
Ac7SUH9F3VUS8YlRRJkF05LicWOggPpvlskEFhGA7I58C6+AAqbmZHJ2R07tVdarETXOimdeGjCi
xll7RdCYCEH4UCJkTHQwNSeTVWBEjJGMZ5VfbRXh4sw/ES3OGz8zDx6MiSw3w7gCiqvAzE8zAgaJ
eWaoQEKSC0mR5WYwIWF1cPJmkewH7CbGtmVRtkoWxypWoUDcgsR+oFCIKOLbiRMXpfZlkdu3ymQa
i2lUdM2/JW6ZCGarZF28eWLbmUUW5/Zlsd/3LXRLcrIouy8c25YT20ZClK2SRZGKE9tGOmSgFxno
CG9jezAC2ucIcEObzfNi8+weLQPkwiHrvJw1CdDhQLOQM1cRPC1vwAJW1YfSMBqh+GSYFEPxGqH4
ZHf1hOKTZVaPxWMPCKtpxOKRBKQ19ME0jJ5YPNGolcdBaRg9wfhkMPWhgImeaDyQdk7FwWFiVR+K
qek3Fg90qMRBH0zD6AnF0zrLhQPWXB+y5hqB+Kwo9UFzrg+Zc41IPGKgNXcAWnOHNNCSDkFN6hC9
sKwQUrLm3hxwAHLIxA4LmoHyWn+tlDbWCjlE/dZn/qoHjZTBq97UMwZJ/+kp0Hw1pEbGAClA/ENP
xgBILCJRH1Dkeuo0kQqljbVSBlEjZZDUo56MQVLvWqnAQyNjkE2EnpQBEcICDz0lm0CJ0sZaOYeo
kTIgUpl5LQCGvFrJD9DIGxDOLISyCIW4N/IGwKuV/QCHCCNpIRy7cDqEkxySlK03JwSLHLIgaOX8
9gCKYo5DG2VGtSFHn2LYxI8NI9dKOQ6NHEfe8EaRa23E0ZHjQBKLSFC0SE+OA6lUdLy1neD1KwYn
x5EkpXHwWhtzdPtmRl9WbBi8PkgLBe6kOIASZUa1XRzdvjHITCpikK+J1khzgBaMTAcBxoU0ucsi
gKezAcV5OslFZDywLYyh1MGo3ROzHkbK3NyCmzO7t+DmFHB1XnsAxfVn4nb3cJC9BVJ4KuxrCOV4
VbDoQ6VgHYVyZkYjMwM0BtOA4LoKpfBUjsYGIpjjVaEUnspxi7OHIscvxk5Gs2FnQyi+ric5Q19Z
Fym/1wF9KIWncvkbgllePfmZzIcnP5OVYORnsp6MDA3hzAtnoUKOJE3yUSJHgyiem01qP5I0P4ii
jELWRdDbEPQ2YpHA2xD0NsSZO0V/HVDx94ZNQedEqaJBlQ6BGtXdWQtFdXdWo0rxWfVcUtbFUd+N
RNBZjfpupLGYBkZo9dR3AxGsaNCTSsoq9tR3ZzX8LfBOdkTpEKgnmZRtkR5tDpRWQULtYUJU0aCK
h0A96SSk44XziW0jm5R5NJJJwMfKrB7ppB9CWozkR8GkeCOllNk0MkqEIoyiyNpaxBin54OeA7I2
V2f1dFB0wAJdrajL1ZkdlL2vFzozWlQQBVNfapiFV0p9qae+MicaVXmpZ77SPsb7FWmfDWu81Khc
Qz3xhQQWE6hQHqOR90IalIVXY9fcTonXK4OGJV4HpwMbR94r8blxgZfa5ZjbSaRmOq0ynYZXETSy
XkCInRi7nJhIfDFkXFjoxUT6i5DIjTF2YyIFRgBhFPBjIglGGB0Z1ZD1LRyZzLd2HJnk+xgyv1/9
BWnyqAuIjyHrm7P+D2HgbQTtmAfWk6tLTkbHsgL1ZF3izE4ZE+1Y3KZ8EUHjIgJSwOo27ZgH1s4p
E+1cV6Cd8sDaOWmiHesKNLJ12T35ZuuyC3PSddhn8nwoZ6Kdqtu0cx5Y+3HskVSTqxO5Pv1yfU7e
Lmv7zgahX75PZO4QB12fzq7PydtlZd/Z9+ns+0TWjlCy6+N3KcCV6ej5dPR8eok4WfKUOro+HV2f
XsLXTXGcXlSxg2ETb3fowIobHWQo4pGgpOg9nZgV4aBDbyQTszKNbCIQIEsRLwwAhUUU+MwbyUQi
ghU3OrBCQr/JxGR/IpmYtPNgO3EyiVnFRyaRqKAUn0sdSAYqbnRcZiIyiUSHZSJyiSAB43KSIpvI
WLMQ0mKkJBTRw+1EZujBTlKkExFFC2EYY3S0AwMNxTjxm2RJRhiKrNYHGorhcpHtguuDQgiCCFhc
7RArOCrtiIF3ZfTclckKbdK9Np2Uf9e4LpM154SDhTgA4+BxXeYXQx0wqMPkL1AQfKLMRJdaL0i7
ZwN38DSSnwQwnk7FFMPEu6N67svgfKD4Q/G2THyiVZ5Qa9xFGICq+VyVSYw6Kf+u564M4aBrNdmK
RPITUbx0PVNhK3JSn4RDAj45/66R+iSkQcOXCdZpHqEBnMhC5K9qpQlqIzIqPH7VAmTULjKdR6+D
ZxgzSKZnFYrOr0I+9YIaKl2FsjSLbh3owhoqXRSRWnS7Vz1RmzX6KotJ5EsHEgDIyus6NQQwDeF5
VMWJVHQgF1/u1XWuHOSpQBmLLr5woItqqHRBGUuQyXd7Y+xYQ6ULCwg0MrT5VLCohkoXu2InR0s4
dD5ZfDaPLC3i0PlkXWmvk6UlJLv6oMxHmjbzbiRqkQ5K/HIZ+kEUP6EjGSWhX+GZIRKtrJLEr5Ai
whkXoew+7mZOKssf9pBQ2QPPGthTsODXHrypYfEqzctE9lBVjj0Fi13swXcN7KG4lz0kVBZJ5cR2
9lDcyx4SKouEMsyjKk8ELl3bQ48a2MMxL8N08qGyaC5UlWMPCZU99KyBPVSVY5FMTsxhD9Y92EPZ
OItUcuINi1xy5gR72Jez59Q+AB2wTfaQL2cnl4yEyDTZw/k4i2QyfhZyhxbJ5CSYFulk/Kg+TMaF
CpG00WdVGKBER+36bL8g4yKUhMqbEKBzEXhQyiL7nbwgqwW9K6tQumHnWczka1ncmEuulVUs3bDI
fycfyCqcnIIGFENY5L+To2WVYmyGN+YCpbZrJsJdlD9i12TwPpSd/DdBJs+nLqLUnmsVK0+p0cI2
uXDAEbRKjqBVLt2wyp6gVfIErXIuxSq5glZJ3CrlUqyyJ2j1ErdKpRv2zX4j0uAu8/rKorF5TgXG
ojV5dd5uTEaFZ6xKi6LGZK6l1ZG9VAdMiHE4JAVBXBAebFZwHK0VDAhYowur1qjGz1rBrIm1gkkT
a1jjZ43NXMsXVsXbWM9hDUN61vhlGDvJ+myTGoT0zlSE51KVJ1PhERtreGP1IA2aDpm6Bm/DxPCp
xs8aFk5Z4+uq1rjGzxo6kNbIgbSGNX7WyH+0xv6jcZbeGrmP1sh9NMrSW2Pv0Rp7j9aoxs8iU5/5
PxL1ICORqyccqPGyyNQjilYavGKJl0OE6ZDktbB0RKgzocGEsp1rLnYphOIiAJewHFABwRraQSwq
MKG3BEyoCsykQCWKRVFBYlMpeEPfpEAZigk9JmCC4UOTsgihYjLYhMOHJnQbwb41BTATCiCafKvA
8mTwNQETDCEeJEwE2KkpwBlhFZhRSUEg0WsCJlwFZqeoIHOIsCEU9jvPrc/MisJ2UC6/M2oKkA4m
m0zY7zw3P4GMKJFhrzOKCn6oz2AkUh1RWYBD0RzwtCgtQAxt9FEVBihTUaOPdmoPJpIlUVwSc7TT
5eCBaKVDKqAY5oQdIiidCl6qOgD0uRas2rUogsgumubofnQwxqBskim+9GJRBYEATCUZVkHER/E+
til6qWcmwl14qNV4MrXz6HN98QHMaz4hnzABLPkxRS81ptSupYWSHzu3ajN/KyeDTankx6IOgnBy
8MKUKoztFEEgCtbDm3KFsUUZBOFAQakpl/yYUrjSogyCu0C80pTjlRalEDQW5eFiwNKiFgLpQLzS
23aR4YWlQItDJmTWHJK9VEUv1V/Ey4k+BzT8XbCp6NQamlKjV7jMqPzUrEAthFm++y7exvJTswIJ
LrOC+S2Ly8rJfBmZUuO772bfZ7heO2RsSo3vvtv3sUOYSsVSQjN4hysGXwdPp+KTcGZ4AT7Gh+Wn
ZpcpNX6Hy4zLT804IWenciOrdKPyUzOqqLao2yCUfHnFjAqq7VRtIApeXjHDbJxF1QZhGHWgE5pR
SNQMBPXg0NnXOChqxoIalRtIB0Oidmo3gI7i2dc4JOqQwWSyKbUwpdlwWpjSbFwNTamFKU3ZR4fA
2wIOUMQwNLWdj6SdHkm0ThWH1imbYVFmkjm8Y8Wh9VNx+DJ8x0cSrRd8TcSoysROlUkWvc53Iq3z
kbR/Sw7zXOhOpHXKaFjnVxKt051I6+dOZJ5OXTSf9vCE8Eja+YlE61xxaFFikrkqKkwyD3VOptu5
GZ6Yl2+GW6frEHZuhiMdMrWdr0PY92Y4ECJb2ymdbp3Cr1Fjgm3UM52jr1FmQgBSNJ2Dr1Fqgm0l
Khx7jWoTAoyLTJbeXiLh9opqd+HNZ9SOsuuvRICkdhddEOaOsttddn8IpaM0xz377DpCVUx0aNQW
9E8HlE8aPHca/bH43EYZ1GEiPtWdW1TFZCc0imII0K5ZUKyU7tgfiPFUsHbY6Ir9QZo8n7q4DxaM
2fe9U5wUvRxigwvGbFyO87gc58EFYzbYcx7kOV9X7W1crvNgeear9jbYcx4szYMKxmxc4d0ojqHR
c3x3cHw3ymNo/BjgHVeANypkkA5FmQbHd6NEhqhgeNcf+QdHeKDfPNBv9sfq0M8e6DgPdJxHiaNe
ih6PYoYdOjbx3QKbBWrubeY3psTbmM6bUAxq/NqtTSyNtkkRq0lPTNl57fa1prNAnb3Ra7eBUSGJ
O/G+7cERmkVVmgZet7XJL0zZebUApsLRquupW5tUGm3/euo2T6g1nhCURtu8HOx5OdiTS6Ntsoc9
2cOe9HyLTXax5+ViT36+xSb52JN97Eml0RbFOyDC8wodz0uoT/1Olq7JweOZhTo+q6iB5hU6jvId
oqOogyaHjqN8B8lkMz2Pk52M7AwnO1vViYZ6Hic7+eWzRGz2tfUTDfV0uc6m36+joiM/iw3EmGjI
F8n6wscIbWF1qy06U68C1wtsUXWrLXj+yxY+RWiLrijbIlmP8qIkCAsvKAcGy3rUF+Es8IKyres8
vegtQlt8Qdn4hWM7r0DAbKi61dYl6wsfI7SF1a0HB9/9sHUZ8HU55Otcg0hMu8iCL3bIo8gI6Uhl
OuyQR5kRkoFrELbYHV9XXDpqjBhCgenFgelTZoQAjEwvjkxHjRHR4dD0otD0KTJCwGAyWdajxAgk
e53YdJL+hbK+QtazoC6X9SzLC2V9FYNn6x3QUV0slPVVDJ45dgCWQv1b+f33v//86z/++I//+8df
/l7++fe//Nef63YmR6+zPvqn8u9/+9v//OefbfW2mfxP5R9//e+//Nlkbcr6x/8DkM7Fzw==
"""

brick_var_stiff = """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"""

brick_init_str = """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"""

brick_mass = """
eNpVVkuWIyEM2/cpcoHJK4O/97/YGJKA1C8LBMZIWC56PH/P63m9c5TP15+ssZukjR7LmR8wP9a8
1bCaPR4nZq6Y7/xcMd/4uWLOeJ54hZwKOXXFn3HH2zPj0R7r2WtwlsFZtvd+4m3v/cUoxNjJ45DH
N4fPXgfODnkc8vjKc+b95Iyd85MnQFeArlg5v3li5TzzBvMO4zj5s/P/e5emaDYQBKNBRJ+9wiYC
RWAIvIHN9dcgEHTyd5b3sX+1D/3uqX3oAeMyqH3oAXqz1T70AMcEgSBfd39dAvIgg0ZCaNzkjSYh
vZQaGSGnLEHoMmlQr7eGPqH77AeIySb2O042sYugIo0mISVkhBxoyiZ2UV4usojdlSaGSIDmQNvI
oPsbWMNGk9aUEJSxkdNa0NqyT7u3fwsVIbl9t5AQGkB7Utkn2l0m3e5E7zUC8zUyinRaC1pD2pNo
T6I9ifZctOH09TsilEQoWURJhJJ3lbyrJELJIkoidFnkx0yXRS7Y3v0p0o9HAIIIXSKO0xopSDJy
vZGdjNrRyE5GdjJqRyM7GbneUJKhJGNJxpJsS/qJsC3p5mxJiAwEOvWLk0CnmjnVzKlmTjVzEujU
L07Gc1TorNDJeP7pF1ictKi8aLToIDjIpEGdFlTRoIoGVTSookGdFmTSIMGBgoMFBwmOXVFAk0OV
Fo0X/QXc1u/IT5Kf1KNJ30d6WYWeVkn6Pib1aJL8REMnfWeSvjPJhk6ud1K9k+udVO8k+bnkn8Zo
BK+9FHU3vfdSZH568YWefCkyPz36UtTdhZdRdBlFl1F8GUVeKPZCkReKvVD7Mn7ya1/GZSNJCP4P
+Q8qPvkc
"""

brick_var_mass = """
eNpVVkuSxSAI3L9T5AJjBfnp/S82oEmkrWy6wIZGlPT7d1/31ZvdLrf/aAHiXAHoWHp168vCfrNf
Afol7e7T+PrxcnssXAk43Ljdg2RQAD57pAaV2EPNc40A/bBJ7Ontzv0aQA61BsG3R1fQh00XwbNH
F8HDpovgc9PDZlWCLbaHwKoEW2zTUkMAqQRaCewo9UX9sHktoq9EHzYP6i+OL+rPrSTqQV3c/MQZ
Eefv8xsRqKIOiAHJ9UdtrDwCKSAD5ICCp/nsOvk3Mzg3zWWBCFAElzb3Uc0MXpBkKo/SmcELsvTs
FssDOSAIHgEpti1KyuY+aQYkhB0hI5RSmoCK0BA6wih6YZqAMtKbMNE+q0co0T6sA3uRGpARSilu
QEVopbwBHWFNkSJFbTZizZUEpri+r6p9n/Br7jvlIbkSdrRyQqahNBMKOtceC2jo7LUYvZ50oJpy
z5QrUZauhEmupkE79Ee8a/5aeffoS8xbwJsEbwHHWu9IQEUqqxlzCqhUVQCDAEYBjAI4BdStSf3p
EdQj2OaCemRf+TdF2T30MgvqkX0gx+pIVXtIQg8kUXtIUs9RK6hHUk9lykifPMVbrNhvuh+0V49i
vyn2m+4r8sZV7Dfd8o61ylOQpyhPUZ6mPMiY0VkQalFr8HiT4WEaPN9k+GYZvlmGt8vwzTJ4wwOO
M2gCTUB0IySE9TAND9NQraVasFqxOh61w+gixwfbYXiR4810fLAd5hc5PtgO4h3EO4p3FO941I5H
7SkerIpWQ6sX64BBSgMnw4BRSgMnw8DJMGCa0sDJMGCe0oBajKiFN1prJXGDMcdKtfZyC0bW4lyZ
kbUAq6LVQHyW4ytNwFFH0MSpiX8ZNHFq4n8G4Y8GTZya+KtBE6fmhNpMqM3E2kyszczaHLkza3OK
MbM2YFW01j6Z2CczawPO9R/oHxbwH3k=
"""

brick_surf_normal = """
eNpd1luSm1AUQ9H/jMITMGWdxwXPf2KxD12GraTycY2aAqtW1OsRr0f+ez2+f3vrFe/jmKM+/84/
c4z78Xstt6OObs1RvIpwMBwMB8M5Ye2fbMxRvIpwzVPfjuIR4Wa4GW6GF8OL4cXwzvDO8M7w8Tk8
VVv35ws5PxCvI/6e+Ht77avyPR+I11nMi88i61FepCzvVVqXCn98WZ+yQpX+BrJSZa3KapX1KitW
1qysWlm3snJl7crqlfUrK1jWsA7LW8OyivW2vFUsdHwWhqN4RFgMi2ExHAwHw8FwMpwMJ8PfSp/7
llJlzFm8jPS30Of6/GeUu9acxctIL0svphfT3zKfvb2PPXOfs3gZ6YMveTB8MPxm+M2wtThSo7d9
V+vvA1mAPzDdHlutfpXOD2QB/oDVKetTVqisUVmlsk7HKd7AWpXVOlDxBlasrFlKlUmVSZVJlUmV
SZVJlUmVSZVJlUmVSQ1KjZH6m9ig1KDUGKm/1QxKDUoNSg1KDUoNSg1KDW5qDNTbYxTDzXBP+G+8
g06DmxpkGmQa3NSg0qDSoNIYpbeNCjoNOg06jYd1KCtxmN72LAxq2KaGOQ1zGuY0zGmY0zCnYU7D
nIbtaRjTMKZhexqmNExpmNI4leL7l5Ura3ec8ju1hmUVU2qY1DCpYVLDpCalJjc1KTUpNbmpSalJ
qTlSryFMUk1SzaF6TVvSatJqclSTVpNWk5uatJq0mmP1/hyL6Z3pafM3sUmtSa3JTU1aTVrNsXpN
Tj6sRlmPsiLlTXqV8yvRb6HSrKZZTbOaZjXNaprVNKtpVtOspllNs5pmNc1qmtW0RU2TmiY1bVHT
nKY5zdPpvTBZv7KCByoaQMVFqUWp8wvJNbFFqUWpRalFqcVNrYH6G8Ii1OKmFp0WnRad1ji93bkZ
boabd14ML4bX3Plvj4tMi5taVFpUWlRaVFpUWlzUeliDsgplHRJpGdKyPS0zWma0zGidRm/7VKa0
TGmZ0jKlZUrrVHrbszKnZU7LnJY5LXNap1O+g7Urq1fWr6xgWcMDle9gLctqptSm1KbU5qY2pTal
9ki9VrNJtUm1ualNqk2qPVSvaWtabVptbmrTatNqj9X7rRfTi+nFW+8M7wx/y7wmtqm1qbWptam1
qbUf1qKsRlmP8iK9Sa9SlrcuZWXK2pTVKetzrF6L04a1DWvbpLZRbaPaJ9Vrodqstllt29Q2qW1S
+5SKN7B2ZfXK+pUVLGt4oF5v8B+UoS/D
"""

brick_surf_scalar = """
eNpVlEtuxSAQBPc5BRd4lucH+P4Xy5BIfl3etQSmVdQwh98/97jHp66a/uzdycYnrp27yjo5Upxk
q4N3yt75/3UoDVPD0rCH5VXVPzzpGfZc95oZzzn51pVmSI6NFtyJKoYuhjKGNraRnm86h0swDa4h
NOT4rCvMMrxTjc9srrFsdppI6zB/9opYnbb+BC2aide1lpX9xT59XznrTjsRVQxdmgl2FndOrF1I
qKNUHK44XHHl4srFDxdZV2KRqyuurjhc8YEadsv9O1xxUHFQcZjiMMUPEz2woeCMjbVSJ9SVUFfi
zM9rRJz5eR0InZ/Q+Qm4EnAl1JVoKN87DgxQ0JUAlQCVAJUAlYApAVPiMMH5j5yYcCUVS+JZSXUl
4UoqljxYXnFSXUmlkqpKAkrClOSrkoCSB8r3/hNUkq4ksOTBgp3Sp9SVOgP06lE6QAVXSl0pPCyl
A1RwpZRKDbQw1jAk9Ggo3zsuqFJ8VQqu1IGCnWgDV34B5+kBgg==
"""

brick_surf_vector = """
eNpdmFtygzAQBP9zCl8gLs8+ENz/YglSldG08qXKSKwYmll8vOLzyp/Pa/2tMad6/ea773HMedhc
LhfkuuV6j3ucP/dik8ctt3/77unyhDyxe7m8ZnXbNPZpu7hd3C4+XHy4+HDxcPFw8XDx6eLzXzze
mmNO/2/R533cY/zPL1df9w2vd8wx5y4X3NTHa9HHixHtFPSCHn7qNtQKuB3dziM4quSC29P9SIKp
gquCrYKvgrGCs4K1greCuYK7gr2Cv4LBOqE/oYfFuqC/dr2cV73MYL0+LpaL5WK5OFwcLg4Xp4vT
xeni29Lffuccc76/NTRJ3ecNfUPf9h7QpHXXH/be0OR1Xz+gH9AP7H/6YU8/7OmHvVx8uRhuTmBD
TymT2C9AWsQ+wGghaytkLwUtaG0JnBWsFbwVzBXcFeydyFpJZZRrQbufqrmiuaLtvSBwK3ArcCtw
K3ArcCtwK3ArcCtwK3ArcLsceJ7BcHADObsM8rnJPWfD2Q1nN5zdcHbD2Q1nNxa725XK1YW6G/J2
eaPuw0s5XH14KcPFw8XDxbed3ycunNlAzoZDGw5tvGDlhPZ5NgMxG4Q2kLOBnA3kbIDYALEBYgPE
BogNEBsI2UDIBkI2ELKBkA2EbCxY7QbA2UnrZs+k1RbA3Ymr3fMTJcFh5zXAa4DXAK/pOZuOa3rO
pudsOqzpObuceJDJ1RR/0yjRFOfE1eeu9644vStOBza9K07vitNxTe+Kc+XsdqUDlRyofEA/oB+o
/PRiTi/m9GKuHfCcyD6JkmiNE61xgtlEa5wL2eeBy8XsfgXkbILaBLUJahPUJqhNUJugNkFtgtoE
tQlqE9QmIjYBbSJiExGbYDYRsbmQtRt28pae+2shF7S2gjZPbB+bCjlb+KAtJ7ec3HJyy8ktkFv4
nC3P2fKcLc/Z8pwt5GyhRy7ntpzbcm7LuS1wW+iPy3O2PGfLc7Y8Z8uZLWe2nNnymC2P2fKYLRBb
ILZAbCFkCyFbCNlauH4ftuLHbIHXAq8FXgu81uJ1fzbRFxeILRBbILZAbDFnayG7nwnWCt4K5gru
AtlayNqZYLHgsWCy52x7zrbnbDutja648fNTr7b4GzDtbXF7W9yOayNnG78+9crZbfPyzcs3L9/c
++LGJ22vxnjb/PDND9/88M0HNh+ofGDz0zc/ffPTN79cfLkYbgp2Cn6Khgp6QS/o4ahgqeDpRPbL
Q/P3p17MPoHSCNlGyDaQ7YXsA0TzU7YXtHYJmCu4K9iL5rgXtfupBq8AhwWLBY+Rs72otVNddok/
52tiew==
"""

brick_int_scalar = """
eNpVkMsNAyAMQ++dggWKcPiE7r9YA8qB54v1lAgcW7H2aaWVr1a40i29p4/0mb7SPX0XU+1XQT+Q
GlFEI3biIE7iIjqRofSm0r14VLsKFPHc79WPdmAnDi5P4uKyE3d5dyPTruPqhGgYnqbeqeEANCU2
JTal0xReduLmR2+qP7x0Xt8=
"""

brick_int_vector = """
eNplVdu1wyAM++8ULJCcyDwz0N1/hYtMW+TU/TJBtkAWtWRXyq8r8WftbIzuOVIemluqZefgjktz
JKtn9vDc0qELRkBLKzxHOiAVbf4KNiBPwPFh9Hfg8rUJsrGrZLZRnmWiHqBCMoIpE6OQykZKrfI0
RTfMJkqtOUJO35xX1R0WrqMToffbiVAWnU30hgchPSzwxvQkwzFS5CYm295xT8wQHrcjRCYqqwBQ
umPELVN6ySm98gK1D7xA9ZUXqH6s6voXBXEADtxKxmfA5ObBCaght0dvTkCAUP9w5uKNAogzEOhx
BkrcMRsFcj4FNQvIxwA5LrHXEOXRn9bxWRiXlOYwxDrjOXLgOASQT0NEcR5qXEB6bpmtNj/4S4B8
dsbwHI/cQu47cMLDc/B45uG5pX5WRnthvQPymXMQPluoxgmQz0v+b2+qX87BuGda4mYKr7Wou7am
6IKukUn1sf+2otohhWJbPCQlzrsxBQ6pabqkvTd6Gf37IGH5fM80ls9hZ/HwBbEFls33tGD5/ICo
dgVXYNk8YlxZabKMLryw5Ny5xacUb5sr9Y/N93GXy5WKP/OBir/zSr+4HBtSfiBvlwv9Gv6CsEx+
hdzigVt8f/CxuLJfDteyPXoVb4eHyl0fVfz4G8vf4cgjvN1Y7lb6bm5t4t7+MvsHGYlX/Q==
"""
brick_intrnl_force = """
eNpV0cENxCAMRNH7VuEGghgb46T/xhYiReLfGPlx+t4tft26XWq3PM3Wkl3R5h57uV3ZRtS+CVKQ
gnRIh3TIgAzIgByQA3JAJmRCJuSEnJATsiALsiDvJRes93QvGO197+GW36fnZM/JnpOpn079hOqQ
ghSkIB3SIR0yIAMyIAfkgByQCZmQCTkhJ+SELMiCLEjkEfoIgYRCQiKh0R8SL4aw
"""

brick_conductivity = """
eNpVVkt2xDAI2/cUuUDzgvmZ+1+sMG1jaVbWw8YSiHjW8/Vcz7Vuyf7tL2kgt9iTJg3kRBZGVkfu
lO1SvV5nl/au/4D2pveE9i4Aeo4YJjZIbH3krPW6l82v13ZOO1zoeKHP6b8TPqffTYab/KQKSBXD
4+90IPWAVIGpolOdQJy0OWn/UiXIS5CXnfXNlJ32BBwDgSDPHbvv+JbbvE+tRkJoXd93aD6aDRSB
IXAEcd2P1/LqdcK6E9w5a/2qufT/RM2dBywgUHMnIHvT1dTsXQcmSAT7uveq0KvXdQjIgwwaCaH1
5m6gCAz4NHSGQVmS0GHSoDokv32ayx9gNnP0XjhzBADa0UgJGSEnFEhThhjCDWSkqb08ZZghEuC5
yDiyqISL2thQKWiETicbBIUSQ11B+x2GBoVAppkAheEC2kqNV/K7KFVYyX0Nj/0aOO0MDCWFgLYi
bWXayrR1aL+1b6QgwkiEoUmMJBi619i9RhKMTWIkwj4e+admH/seNCYBJIRAhI2IY7WGBpocfe/k
JseBdPKSk5ecB9LJTc6+d9LkpMlJk5MmH01HhY+oV6KPJkQOCoMnJkhjUN8C+xbct6C+BWkMnJhA
6wUJDBIYbL34dA2QctAo6BwMEJxk0+RZS2xqUlOTmprc1MRZS7JpouAkwUmCkwXnb0sBKu01DjoF
4/S7UYL8TfI3Tunm7yO9rUKPq2z8Pm6e0k3yNzl647dm87dmk6M39Xtzvzf1e3O/N8nfI/9MRkN4
8qVwvunNlyLv86Mv+OpLkffp4Zfi+S6qRmE1iqtRVI1iMxSZodgMRWaoqcbRX1OOtzg11UAE/0Z+
AAZV/1g=
"""

brick_var_cond = """
eNpdVluy6zAI+z+ryAbqiW0e9v43dsG5KVIz/agG8xAgJ+P+u6/7kjY1HvvrAUZbLrdLgF6WgZYR
lt585hNg1LEZx76WGcdGG6vL6gEGglk+gqEFQ0v4AJhXkxXPjv9S/oo5FXNq+L8uGu699UySBsFT
WsEMg1kEe/0N6zcMZhjMIhhYrCJ7RH6D+WH5/5Qfll+Qke09JmhRtBgCrzQr0nyk3Y9pRSJE4/qM
JrqHJprXp7dl8XggIZuSza5Pc7+nXwE8QPd4VoCATZf2pX87U/c2nknuTB0lnrbsJ/OOCHeaJiGp
gDsSN0mfjGdQxM68EH1lA55q97Wrin5TGQGhjkCjkgWa32wBBKoKqAyxmIDOSaCcQDuSjJkNOwXc
UF4/e/A/Zz+LcN8ZJtGAUnt26R1dICGbku3U5hG/H+gMF1TTo7aqtGdtBDuUOmiXesq+VqSn8MmK
6xRQ+LDWCgUyPgsrFSg2s42v477aqgqieWNb/A7K0YJxQPGT12A+xb/jnNTqSQsZEDYyEGohoNX8
Jq/BjNK/9czcgh6Sf3NE7R+w9lPQXfUNmETACWSEycjZ6bdAIS6COy2808JUJMeAVuckq3otZ6Xf
4iW51IwkuaARmUgyqf0LKEBMUQ1KUtVzL77aUFKDkhr0ofUuvJ7rqo46GxcQ0aRVQ9CzX2XMEdUE
NYkVE01ixVOTGEEFnsZSMpaS0cXc7XB9xZLvF3KlGRpLyUhKRlKylNI7UCOqxlKyZIqOwRQ9hTyV
PQ1oO++tswidLmPnC8RpxM63saMInffWUYROe+tHgzVR/xGhJ+8vUU/a1QVP3hRLMU83mH5AhzYs
bsMi+S56N3Z+LfdFbVjweuyL5bu4DQvlu6INxWTxVbRIvuv0ANDko4INXGf61cDFbVjZhlJMQPhc
6Bu1T58LfdOVtp+WvBLe57PovRg2vQo3aX+z9jdpf+ObZZ+WFJH9o/39aB+OT3KWH2f9cTZow86u
VJN2doUgfM38A/a1FXQ=
"""

brick_strain = """
eNpV1VtuIzEQQ9H/WUVvYAyRKr32v7GRB0j6Ml8t2DkQW0XL65l/2tOez2ma8n0Wno3n/vz9tP9/
d1FcjOfnWVeqd/FSAiVSIiVQvtT6tLP73nfxUgblS73/AMmQOuL1C9X7gfEBNtVJ9Uv9LooBi1gB
K2yraBWtwYSD1oA1YI1rvc9vwomE80K/6gQ0AU1A80LtZ7GYb5FaoBaoxXwLm9qMt0ltUBvUxq42
pIN4B5NwrlPvl3CAh5s6fOlqMaIcdzV6iolXjLxakGJSxdgryJx8Ia44+jICy5wMOcDv/DesigaO
VT1SRwvUg+xBfpvAFc2K2FEGVZgVZnGbFeRg8MGZ0QhxxJtkLzTibGYEj3JoBsl+aEbuycNZETs6
ohUia6LFPbIn2gy9OT3forz4jswsi9gWnYjMwugEGJVRdEaHP8UtfovZGTeKjs64YYvmNWEhssXJ
sQKMxpiNsXjONkM7OmMH6SDNgzavDfdIHZ1xD7LzGurcZCdYTF0cHFd4vD8cF4iL4IjQ0RePEHmL
eERmXiSekTnq4hnijNc4Y5OTR/MPOGNwQQ==
"""


# here we have a dictionary that maps a model name to the data

mod_map = {"tet_conductivity"  :tet_conductivity,
           "tet_mass"          :tet_mass,
           "tet_var_mass"      :tet_var_mass,
           "tet_stiffness"     :tet_stiffness,
           "tet_var_stiff"     :tet_var_stiff,
           "tet_init_str_stiff":tet_init_str_stiff,
           "tet_surf_normal"   :tet_surf_normal,
           "tet_surf_scalar"   :tet_surf_scalar,
           "tet_surf_vector"   :tet_surf_vector,
           "tet_int_scalar"    :tet_int_scalar,
           "tet_int_var_scalar":tet_int_var_scalar,
           "tet_int_str"       :tet_int_str,
           "tet_int_var_str"   :tet_int_var_str,
           "tet_intrnl_force"  :tet_intrnl_force,
           "tet_var_conduct"   :tet_var_conduct,
           "tet_strain"        :tet_strain,
           "wedge_mass"        :wedge_mass,
           "wedge_var_mass"    :wedge_var_mass,
           "wedge_stiff"       :wedge_stiff,
           "wedge_var_stiff"   :wedge_var_stiff,
           "wedge_init_str_stf":wedge_init_str_stf,
           "wedge_surf_normal" :wedge_surf_normal,
           "wedge_surf_scalar" :wedge_surf_scalar,
           "wedge_surf_vector" :wedge_surf_vector,
           "wedge_int_scalar"  :wedge_int_scalar,
           "wedge_int_vector"  :wedge_int_vector,
           "wedge_int_force"   :wedge_int_force,
           "wedge_conductivity":wedge_conductivity,
           "wedge_var_cond"    :wedge_var_cond,
           "wedge_strain"      :wedge_strain,
           "brick_stiffness"   :brick_stiffness,
           "brick_var_stiff"   :brick_var_stiff,
           "brick_init_str"    :brick_init_str,
           "brick_mass"        :brick_mass,
           "brick_var_mass"    :brick_var_mass,
           "brick_surf_normal" :brick_surf_normal,
           "brick_surf_scalar" :brick_surf_scalar,
           "brick_surf_vector" :brick_surf_vector,
           "brick_int_scalar"  :brick_int_scalar,
           "brick_int_vector"  :brick_int_vector,
           "brick_intrnl_force":brick_intrnl_force,
           "brick_conductivity":brick_conductivity,
           "brick_var_cond"    :brick_var_cond,
           "brick_strain"      :brick_strain}

def GetModelStream(name):
    indata = mod_map.get(name)
    if indata != None:
        outdata = base64.decodestring(indata)
        outobj = CompressedStream(outdata)
        return outobj
    else:
        raise "Do not have a model (" +name+")"

def GetStoredMat(name):
    smat = {}
    fd = GetModelStream(name)
    svals = string.splitfields(fd.readline())
    if len(svals) == 1:
        for i in xrange(int(svals[0])):
            svals = string.splitfields(fd.readline())
            smat[int(svals[0])] = float(svals[1])
    elif len(svals) == 2:
        for i in xrange(int(svals[0])*int(svals[1])):
            svals = string.splitfields(fd.readline())
            smat[(int(svals[0]),int(svals[1]))] = float(svals[2])
    elif len(svals) == 3:
        for i in xrange(int(svals[0])*int(svals[1])*int(svals[2])):
            svals = string.splitfields(fd.readline())
            smat[(int(svals[0]),int(svals[1]),int(svals[2]))] = float(svals[3])
    return smat

def GetStoredMatSym(name):
    """This routine returns the zipped matrix as a NumPy matrix"""
    smat = {}
    fd = GetModelStream(name)
    svals = string.splitfields(fd.readline())
    num = int(svals[0])
    for i in xrange(num*(num+1)/2):
        svals = string.splitfields(fd.readline())
        row = int(svals[0])
        col = int(svals[1])
        value = float(svals[2])
        smat[(row,col)] = value
        if row != col: smat[(col,row)] = value
    return smat
 
# -------------------------------------------------------------------

class Tetrahedral2ITests(unittest.TestCase):
    def assertEqualFloat(self,x,y,tol):
        self.failUnless(abs(x-y) < tol)

    def GetCoords(self):
        coords = []
        coords.append(Vec3D.Vec3D(0.0,0.0,0.0))
        coords.append(Vec3D.Vec3D(0.0,0.0,2.0))
        coords.append(Vec3D.Vec3D(0.0,2.0,2.0))
        coords.append(Vec3D.Vec3D(2.0,0.0,2.0))
        coords.append(Vec3D.Vec3D(0.0,0.0,1.0))
        coords.append(Vec3D.Vec3D(0.0,1.0,2.0))
        coords.append(Vec3D.Vec3D(0.0,1.0,1.0))
        coords.append(Vec3D.Vec3D(1.0,1.0,2.0))
        coords.append(Vec3D.Vec3D(1.0,0.0,2.0))
        coords.append(Vec3D.Vec3D(1.0,0.0,1.0))
        return coords


    def testTet2IArea(self):
        """Tetrahedral Face Area"""
        tol = 0.00001
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()
        self.assertEqualFloat(2.0,elem.Area(0,coords,3),tol)
        self.assertEqualFloat(2.0,elem.Area(0,coords,4),tol)
        self.assertEqualFloat(2.0,elem.Area(0,coords,7),tol)
        self.assertEqualFloat(2.0,elem.Area(1,coords),tol)
        self.assertEqualFloat(3.46410162207,elem.Area(2,coords),tol)
        self.assertEqualFloat(2.0,elem.Area(3,coords),tol)

    def testTet2IStiffness(self):
        """Tetrahedral Stiffness Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()

        # first time through with a uniform E
        mat = DMatrix.IsotropicMatMatrix({'E':1000.0,'nu':0.3})
        k = elem.DeformationStiffMtx(coords,mat)
        smat = GetStoredMatSym('tet_stiffness')
        for i in xrange(30):
            for j in xrange(30):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

        # second time with E equal to the Y coordinates
        mmats = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            mat = DMatrix.IsotropicMatMatrix({'E':gp_coord[1],'nu':0.3})
            mmats.append(mat)
        k = elem.DeformationStiffMtx(coords,mmats)
        smat = GetStoredMatSym("tet_var_stiff")
        for i in xrange(30):
            for j in xrange(30):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

    def testTet2IInitialStressStiff(self):
        """Tetrahedral Initial Stress Stiffness Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()

        # stresses based on the coordinate locations
        stress = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            stress.append(ColTensor.ColTensor(gp_coord[0],
                gp_coord[1],gp_coord[2],(gp_coord[0]+gp_coord[1])/2,
               (gp_coord[1]+gp_coord[2])/2,(gp_coord[2]+gp_coord[0])/2))
        k = elem.InitialStressStiffMtx(coords,stress)
        smat = GetStoredMatSym("tet_init_str_stiff")
        for i in xrange(30):
            for j in xrange(30):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)


    def testTet2IFindNaturalCoords(self):
        """Tetrahedral Find Natural Coordinates"""
        tol = 0.000001
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()
        nat = elem.FindNaturalCoords(Vec3D.Vec3D(0.5,0.5,1.5),coords)
        self.assertEqualFloat(nat[0],0.25,tol)
        self.assertEqualFloat(nat[1],0.25,tol)
        self.assertEqualFloat(nat[2],0.25,tol)

    def testTet2INormalSurfaceScalarIntegration(self):
        """Tetrahedral Normal Surface Scalar Integration"""
        # test the integration of a surface scalar where the nodal
        # value is the distance of the node from the origin
        tol = 0.000001
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()
        smat = GetStoredMat('tet_surf_normal')

        for  i in xrange(4):
            tract = []
            for coord in coords:
                tract.append(coord.Magnitude())

            f = elem.IntegrateNormalSurfaceScalar(i,coords,tract)
            for j in xrange(10):
                for k in xrange(3):
                    self.assertEqualFloat(smat[(i,j,k)],f[j][k],tol)

    def testTet2ISurfaceScalarIntegration(self):
        """Tetrahedral Surface Scalar Integration"""
        # test the integration of a surface scalar where the nodal
        # value is the distance of the node from the origin
        tol = 0.000001
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()
        smat = GetStoredMat('tet_surf_scalar')

        for  i in xrange(4):
            sum = 0.0
            sumv = 0.0
            tract = []
            for coord in coords:
                tract.append(coord.Magnitude())

            f = elem.IntegrateSurfaceScalar(i,coords,tract)

            for j in xrange(10):
                self.assertEqualFloat(smat[(i,j)],f[j],tol)

    def testTet2ISurfaceVectorIntegration(self):
        """Tetrahedral Surface Vector Integration"""
        # test the integration of a surface scalar where the
        # nodal value is the location vector of the node
        tol = 0.000001
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()
        smat = GetStoredMat('tet_surf_vector')

        for  i in xrange(4):
            tract = []
            for coord in coords:
                tract.append(coord)

            f = elem.IntegrateSurfaceVector(i,coords,tract)
            for j in xrange(10):
                for k in xrange(3):
                    self.assertEqualFloat(smat[(i,k,j)],f[j][k],tol)

    def testTet2IIntegrateVolume(self):
        """Tetrahedral Integrate Volume"""
        tol = 0.000001
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()

        # do a volume integration assuming a scalar value of 2.0
        f = elem.IntegrateVolume(coords,2.0)
        smat = GetStoredMat('tet_int_scalar')
        for i in xrange(len(f)):
            self.assertEqualFloat(smat[i],f[i],tol)

        # do a volume integration where the scalar value varies
        # over the element and is equal to the y coordinate
        values = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            values.append(gp_coord[1])
        f = elem.IntegrateVolume(coords,values)
        smat = GetStoredMat('tet_int_var_scalar')
        for i in xrange(len(f)):
            self.assertEqualFloat(smat[i],f[i],tol)

        # do a volume integration with a stress value of (10,20,30,4,5,6)
        value = ColTensor.ColTensor(10.0,20.0,30.0,4.0,5.0,6.0)
        f = elem.IntegrateVolume(coords,value)
        smat = GetStoredMat('tet_int_str')
        for i in xrange(len(f)):
            for j in xrange(3):
                self.assertEqualFloat(smat[(i,j)],f[i][j],tol)

        # do a volume integration with a stress value being
        # (x_coord,y_coord,z_coord,x/2,y/2,z/2)
        values = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            values.append(ColTensor.ColTensor(gp_coord[0],
                gp_coord[1],gp_coord[2],gp_coord[0]/2,
                gp_coord[1]/2,gp_coord[2]/2))
        f = elem.IntegrateVolume(coords,values)
        smat = GetStoredMat('tet_int_var_str')
        for i in xrange(len(f)):
            for j in xrange(3):
                self.assertEqualFloat(smat[(i,j)],f[i][j],tol)


    def testTet2IInterpolateSurfaceScalar(self):
        """Tetrahedral Interpolate Surface Scalar"""
        tol = 0.000001
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()
        node_vals = []
        for coord in coords: node_vals.append(coord[2])
        natural = Vec3D.Vec3D(0.25,0.25,0.25)
        value = elem.InterpolateSurfaceScalar(0,natural,node_vals)
        self.assertEqualFloat(value,1.5,tol)
        value = elem.InterpolateSurfaceScalar(1,natural,node_vals)
        self.assertEqualFloat(value,2.0,tol)
        value = elem.InterpolateSurfaceScalar(2,natural,node_vals)
        self.assertEqualFloat(value,1.0,tol)
        value = elem.InterpolateSurfaceScalar(3,natural,node_vals)
        self.assertEqualFloat(value,1.5,tol)


    def testTet2IInterpolateVolume(self):
        """Tetrahedral Interpolate Volume"""
        tol = 0.000001
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()
        node_vals = []
        for coord in coords: node_vals.append(coord[2])
        natural = Vec3D.Vec3D(0.25,0.25,0.25)
        value = elem.InterpolateVolume(natural,node_vals)
        self.assertEqualFloat(value,1.5,tol)
        value = elem.InterpolateVolume(natural,coords)
        self.assertEqualFloat(value[0],0.5,tol)
        self.assertEqualFloat(value[1],0.5,tol)
        self.assertEqualFloat(value[2],1.5,tol)

    def testTet2IInternalPointForces(self):
        """Tetrahedral Internal Point Forces"""
        tol = 0.0001
        elem = FemLib.Element("Tet2IShapeFunc")
        # place the forces at the gauss points
        c0 = 0.585410
        c1 = 0.138197
        natural = []
        natural.append(Vec3D.Vec3D(c0,c1,c1))
        natural.append(Vec3D.Vec3D(c1,c0,c1))
        natural.append(Vec3D.Vec3D(c1,c1,c0))
        natural.append(Vec3D.Vec3D(c1,c1,c1))
        forces = []
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        f = elem.InternalPointForces(natural,forces)
        smat = GetStoredMat('tet_intrnl_force')
        for i in xrange(len(f)):
            for j in xrange(3):
                self.assertEqualFloat(smat[(i,j)],f[i][j],tol)

    def testTet2IConductivity(self):
        """Tetrahedral Conductivity Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()

        # first time with constant props
        cond = DMatrix.ScalarIsotropicMatMatrix({'k':1.0})
        k = elem.ScalarFieldDerivMtx(coords,cond)
        smat = GetStoredMat('tet_conductivity')
        for i in xrange(10):
            for j in xrange(10):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

        # second time with conductivities equal to the coordinates
        cmats = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            cond = DMatrix.ScalarOrthotropicMatMatrix(
                     {'k1':gp_coord[0],'k2':gp_coord[1],'k3':gp_coord[2]})
            cmats.append(cond)
        k = elem.ScalarFieldDerivMtx(coords,cmats)
        smat = GetStoredMatSym("tet_var_conduct")
        for i in xrange(10):
            for j in xrange(10):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

    def testTet2IMassStorage(self):
        """Tetrahedral Mass/Storage Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Tet2IShapeFunc")
        spec_heat = 0.5
        density = 1.0
        coords = self.GetCoords()

        # fist time with a constant specific heat
        fact = spec_heat * density
        c = elem.ScalarFieldShapeMtx(coords,fact)
        smat = GetStoredMat('tet_mass')
        for i in xrange(10):
            for j in xrange(10):
                self.assertEqualFloat(smat[(i,j)],c[i,j],tol)

        # now with a specific heat equal to the y coordinate
        facts = []
        gpts = elem.GetVolumeIntegrationPts(elem.GetShapeDefaultPts())
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            facts.append(density*gp_coord[1])
            #facts.append(0.5)
        c = elem.ScalarFieldShapeMtx(coords,facts)
        smat = GetStoredMatSym("tet_var_mass")
        for i in xrange(10):
            for j in xrange(10):
                self.assertEqualFloat(smat[(i,j)],c[i,j],tol)

    def testTet2IStrains(self):
        """Tetrahedral Gauss Point Strains"""
        tol = 0.00001
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()

        disp = []
        for nd in coords:
            disp.append(Vec3D.Vec3D(nd[0]**2,nd[1]**2,nd[2]**2))

        gpts = elem.GetVolumeIntegrationPts()

        gp_strain = elem.Strains(coords,disp)
        smat = GetStoredMat("tet_strain")
        for i in xrange(len(gp_strain)):
            for j in xrange(6):
                self.assertEqualFloat(smat[(i,j)],gp_strain[i][j],tol)

    def testTet2IVolume(self):
        """Tetrahedral Volume"""
        tol = 0.00001
        value = 1.3333333333
        elem = FemLib.Element("Tet2IShapeFunc")
        coords = self.GetCoords()
        vol = elem.Volume(coords)
        self.assertEqualFloat(value,vol,tol)

#--------------------------------------------------------------------

class Wedge2ITests(unittest.TestCase):
    def assertEqualFloat(self,x,y,tol):
        self.failUnless(abs(x-y) < tol)

    def GetCoords(self):
        coords = []
        coords.append(Vec3D.Vec3D(0.0,0.0,2.0))
        coords.append(Vec3D.Vec3D(2.0,0.0,2.0))
        coords.append(Vec3D.Vec3D(2.0,2.0,2.0))
        coords.append(Vec3D.Vec3D(0.0,0.0,0.0))
        coords.append(Vec3D.Vec3D(2.0,0.0,0.0))
        coords.append(Vec3D.Vec3D(2.0,2.0,0.0))
        coords.append(Vec3D.Vec3D(1.0,0.0,2.0))
        coords.append(Vec3D.Vec3D(2.0,1.0,2.0))
        coords.append(Vec3D.Vec3D(1.0,1.0,2.0))
        coords.append(Vec3D.Vec3D(1.0,0.0,0.0))
        coords.append(Vec3D.Vec3D(2.0,1.0,0.0))
        coords.append(Vec3D.Vec3D(1.0,1.0,0.0))
        coords.append(Vec3D.Vec3D(0.0,0.0,1.0))
        coords.append(Vec3D.Vec3D(2.0,0.0,1.0))
        coords.append(Vec3D.Vec3D(2.0,2.0,1.0))
        return coords

    def testWedge2IArea(self):
        """Wedge Face Area"""
        tol = 0.00001
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()
        self.assertEqualFloat(4.0,elem.Area(0,coords,1),tol)
        self.assertEqualFloat(4.0,elem.Area(0,coords,4),tol)
        self.assertEqualFloat(4.0,elem.Area(0,coords,9),tol)
        self.assertEqualFloat(4.0*math.sqrt(2),elem.Area(1,coords),tol)
        self.assertEqualFloat(4.0,elem.Area(2,coords),tol)
        self.assertEqualFloat(2.0,elem.Area(3,coords),tol)
        self.assertEqualFloat(2.0,elem.Area(4,coords),tol)

    def testWedge2IStiffness(self):
        """Wedge Stiffness Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()

        # first time through with a uniform E
        mat = DMatrix.IsotropicMatMatrix({'E':1000.0,'nu':0.3})
        k = elem.DeformationStiffMtx(coords,mat)
        smat = GetStoredMatSym('wedge_stiff')
        for i in xrange(45):
            for j in xrange(45):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

        # second time with E equal to the Y coordinates
        mmats = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            mat = DMatrix.IsotropicMatMatrix({'E':gp_coord[1],'nu':0.3})
            mmats.append(mat)
        k = elem.DeformationStiffMtx(coords,mmats)
        smat = GetStoredMatSym("wedge_var_stiff")
        for i in xrange(45):
            for j in xrange(45):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

    def testWedge2IInitialStressStiff(self):
        """Wedge Initial Stress Stiffness Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()

        # stresses based on the coordinate locations
        stress = []
        gpts = elem.GetVolumeIntegrationPts(28)
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            stress.append(ColTensor.ColTensor(gp_coord[0],
                gp_coord[1],gp_coord[2],(gp_coord[0]+gp_coord[1])/2,
               (gp_coord[1]+gp_coord[2])/2,(gp_coord[2]+gp_coord[0])/2))
        k = elem.InitialStressStiffMtx(coords,stress,28)
        smat = GetStoredMatSym("wedge_init_str_stf")
        for i in xrange(45):
            for j in xrange(45):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)


    def testWedge2IFindNaturalCoords(self):
        """Wedge Find Natural Coordinates"""
        tol = 0.000001
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()
        nat = elem.FindNaturalCoords(Vec3D.Vec3D(1.5,0.5,1.0),coords)
        self.assertEqualFloat(nat[0],0.25,tol)
        self.assertEqualFloat(nat[1],0.5,tol)
        self.assertEqualFloat(nat[2],0.0,tol)

    def testWedge2INormalSurfaceScalarIntegration(self):
        """Wedge Normal Surface Scalar Integration"""
        # test the integration of a surface scalar where the nodal
        # value is the distance of the node from the origin
        tol = 0.000001
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()
        smat = GetStoredMat('wedge_surf_normal')

        for  i in xrange(5):
            tract = []
            for coord in coords:
                tract.append(coord.Magnitude())

            f = elem.IntegrateNormalSurfaceScalar(i,coords,tract)
            for j in xrange(15):
                for k in xrange(3):
                    self.assertEqualFloat(smat[(i,j,k)],f[j][k],tol)

    def testWedge2ISurfaceScalarIntegration(self):
        """Wedge Surface Scalar Integration"""
        # test the integration of a surface scalar where the nodal
        # value is the distance of the node from the origin
        tol = 0.000001
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()
        smat = GetStoredMat('wedge_surf_scalar')

        for  i in xrange(5):
            tract = []
            for coord in coords:
                tract.append(coord.Magnitude())

            f = elem.IntegrateSurfaceScalar(i,coords,tract)
            for j in xrange(10):
                self.assertEqualFloat(smat[(i,j)],f[j],tol)

    def testWedge2ISurfaceVectorIntegration(self):
        """Wedge Surface Vector Integration"""
        # test the integration of a surface scalar where the
        # nodal value is the location vector of the node
        tol = 0.000001
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()
        smat = GetStoredMat('wedge_surf_vector')

        for  i in xrange(5):
            tract = []
            for coord in coords:
                tract.append(coord)

            f = elem.IntegrateSurfaceVector(i,coords,tract)
            for j in xrange(15):
                for k in xrange(3):
                    self.assertEqualFloat(smat[(i,j,k)],f[j][k],tol)

    def testWedge2IIntegrateVolume(self):
        """Wedge Integrate Volume"""
        tol = 0.000001
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()

        # do a volume integration assuming a scalar value of 2.0
        f = elem.IntegrateVolume(coords,2.0,12)
        smat = GetStoredMat('wedge_int_scalar')
        for i in xrange(len(f)):
            self.assertEqualFloat(smat[(0,i)],f[i],tol)

        # do a volume integration where the scalar value varies
        # over the element and is equal to the y coordinate
        values = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            values.append(gp_coord[1])
        f = elem.IntegrateVolume(coords,values)
        for i in xrange(len(f)):
            self.assertEqualFloat(smat[(1,i)],f[i],tol)

        # do a volume integration with a stress value of (10,20,30,4,5,6)
        value = ColTensor.ColTensor(10.0,20.0,30.0,4.0,5.0,6.0)
        f = elem.IntegrateVolume(coords,value)
        smat = GetStoredMat('wedge_int_vector')
        for i in xrange(len(f)):
            for j in xrange(3):
               self.assertEqualFloat(smat[(0,i,j)],f[i][j],tol)

        # do a volume integration with a stress value being
        # (x_coord,y_coord,z_coord,x/2,y/2,z/2)
        values = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            values.append(ColTensor.ColTensor(gp_coord[0],
                gp_coord[1],gp_coord[2],gp_coord[0]/2,
                gp_coord[1]/2,gp_coord[2]/2))
        f = elem.IntegrateVolume(coords,values)
        for i in xrange(len(f)):
            for j in xrange(3):
                self.assertEqualFloat(smat[(1,i,j)],f[i][j],tol)

    def testWedge2IInterpolateSurfaceScalar(self):
        """Wedge Interpolate Surface Scalar"""
        tol = 0.000001
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()
        node_vals = []
        for coord in coords: node_vals.append(coord[2])
        natural = Vec3D.Vec3D(0.25,0.25,0.0)
        value = elem.InterpolateSurfaceScalar(0,natural,node_vals)
        self.assertEqualFloat(value,0.75,tol)
        value = elem.InterpolateSurfaceScalar(1,natural,node_vals)
        self.assertEqualFloat(value,0.75,tol)
        value = elem.InterpolateSurfaceScalar(2,natural,node_vals)
        self.assertEqualFloat(value,0.75,tol)
        node_vals = []
        for coord in coords: node_vals.append(coord[1])
        value = elem.InterpolateSurfaceScalar(3,natural,node_vals)
        self.assertEqualFloat(value,1.0,tol)
        value = elem.InterpolateSurfaceScalar(4,natural,node_vals)
        self.assertEqualFloat(value,0.5,tol)

    def testWedge2IInterpolateVolume(self):
        """Wedge Interpolate Volume"""
        tol = 0.000001
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()
        node_vals = []
        for coord in coords: node_vals.append(coord[2])
        natural = Vec3D.Vec3D(0.25,0.5,0.0)
        value = elem.InterpolateVolume(natural,node_vals)
        self.assertEqualFloat(value,1.0,tol)
        value = elem.InterpolateVolume(natural,coords)
        self.assertEqualFloat(value[0],1.5,tol)
        self.assertEqualFloat(value[1],0.5,tol)
        self.assertEqualFloat(value[2],1.0,tol)

    def testWedge2IInternalPointForces(self):
        """Wedge Internal Point Forces"""
        tol = 0.0001
        elem = FemLib.Element("Wedge2IShapeFunc")
        # place the forces at the gauss points
        natural = []
        natural.append(Vec3D.Vec3D(0.25,0.25,-.5))
        natural.append(Vec3D.Vec3D(0.6 ,0.25,-.5))
        natural.append(Vec3D.Vec3D(0.25,0.6 ,-.5))
        natural.append(Vec3D.Vec3D(0.25,0.25,0.5))
        natural.append(Vec3D.Vec3D(0.6 ,0.25,0.5))
        natural.append(Vec3D.Vec3D(0.25,0.6 ,0.5))
        forces = []
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        f = elem.InternalPointForces(natural,forces)
        smat = GetStoredMat('wedge_int_force')
        for i in xrange(len(f)):
            for j in xrange(3):
                self.assertEqualFloat(smat[(i,j)],f[i][j],tol)

    def testWedge2IConductivity(self):
        """Wedge Conductivity Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()

        # first time with constant props
        cond = DMatrix.ScalarIsotropicMatMatrix({'k':1.0})
        k = elem.ScalarFieldDerivMtx(coords,cond)
        smat = GetStoredMatSym('wedge_conductivity')
        for i in xrange(15):
            for j in xrange(15):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

        # second time with conductivities equal to the coordinates
        cmats = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            cond = DMatrix.ScalarOrthotropicMatMatrix(
                     {'k1':gp_coord[0],'k2':gp_coord[1],'k3':gp_coord[2]})
            cmats.append(cond)
        k = elem.ScalarFieldDerivMtx(coords,cmats)
        smat = GetStoredMatSym("wedge_var_cond")
        for i in xrange(15):
            for j in xrange(15):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

    def testWedge2IMassStorage(self):
        """Wedge Mass/Storage Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Wedge2IShapeFunc")
        spec_heat = 0.5
        density = 1.0
        coords = self.GetCoords()

        # fist time with a constant specific heat
        fact = spec_heat * density
        c = elem.ScalarFieldShapeMtx(coords,fact,28)
        smat = GetStoredMatSym('wedge_mass')
        for i in xrange(15):
            for j in xrange(15):
                self.assertEqualFloat(smat[(i,j)],c[i,j],tol)
   
        # now with a specific heat equal to the y coordinate
        facts = []
        gpts = elem.GetVolumeIntegrationPts(elem.GetShapeDefaultPts())
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            facts.append(density*gp_coord[1])
        c = elem.ScalarFieldShapeMtx(coords,facts)
        smat = GetStoredMatSym("wedge_var_mass")
        for i in xrange(15):
            for j in xrange(15):
                self.assertEqualFloat(smat[(i,j)],c[i,j],tol)


    def testWedge2IStrains(self):
        """Wedge Gauss Point Strains"""
        tol = 0.00001
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()

        disp = []
        for nd in coords:
            disp.append(Vec3D.Vec3D(nd[0]**2,nd[1]**2,nd[2]**2))

        gpts = elem.GetVolumeIntegrationPts()

        gp_strain = elem.Strains(coords,disp)
        smat = GetStoredMat("wedge_strain")
        for i in xrange(len(gp_strain)):
            for j in xrange(6):
                self.assertEqualFloat(smat[(i,j)],gp_strain[i][j],tol)


    def testWedge2IVolume(self):
        """Wedge Volume"""
        tol = 0.00001
        value = 4.0
        elem = FemLib.Element("Wedge2IShapeFunc")
        coords = self.GetCoords()
        vol = elem.Volume(coords)
        self.assertEqualFloat(value,vol,tol)


#--------------------------------------------------------------------

class Brick2ITests(unittest.TestCase):
    def assertEqualFloat(self,x,y,tol):
        self.failUnless(abs(x-y) < tol)

    def GetCoords(self):
        coords = []
        coords.append(Vec3D.Vec3D(0.0,0.0,0.0))
        coords.append(Vec3D.Vec3D(0.0,0.0,4.0))
        coords.append(Vec3D.Vec3D(0.0,4.0,0.0))
        coords.append(Vec3D.Vec3D(0.0,4.0,4.0))
        coords.append(Vec3D.Vec3D(4.0,0.0,0.0))
        coords.append(Vec3D.Vec3D(4.0,0.0,4.0))
        coords.append(Vec3D.Vec3D(4.0,4.0,0.0))
        coords.append(Vec3D.Vec3D(4.0,4.0,4.0))
        coords.append(Vec3D.Vec3D(0.0,0.0,2.0))
        coords.append(Vec3D.Vec3D(0.0,4.0,2.0))
        coords.append(Vec3D.Vec3D(4.0,0.0,2.0))
        coords.append(Vec3D.Vec3D(4.0,4.0,2.0))
        coords.append(Vec3D.Vec3D(0.0,2.0,0.0))
        coords.append(Vec3D.Vec3D(0.0,2.0,4.0))
        coords.append(Vec3D.Vec3D(4.0,2.0,0.0))
        coords.append(Vec3D.Vec3D(4.0,2.0,4.0))
        coords.append(Vec3D.Vec3D(2.0,0.0,0.0))
        coords.append(Vec3D.Vec3D(2.0,0.0,4.0))
        coords.append(Vec3D.Vec3D(2.0,4.0,0.0))
        coords.append(Vec3D.Vec3D(2.0,4.0,4.0))
        return coords

    def testBrick2IArea(self):
        """Brick Face Area"""
        tol = 0.00001
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()
        self.assertEqualFloat(16.0,elem.Area(0,coords,1),tol)
        self.assertEqualFloat(16.0,elem.Area(0,coords,4),tol)
        self.assertEqualFloat(16.0,elem.Area(0,coords,9),tol)
        self.assertEqualFloat(16.0,elem.Area(1,coords),tol)
        self.assertEqualFloat(16.0,elem.Area(2,coords),tol)
        self.assertEqualFloat(16.0,elem.Area(3,coords),tol)
        self.assertEqualFloat(16.0,elem.Area(4,coords),tol)
        self.assertEqualFloat(16.0,elem.Area(5,coords),tol)

    def testBrick2IStiffness(self):
        """Brick Stiffness Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()

        # first time through with a uniform E
        mat = DMatrix.IsotropicMatMatrix({'E':1000.0,'nu':0.3})
        k = elem.DeformationStiffMtx(coords,mat)
        smat = GetStoredMatSym('brick_stiffness')
        for i in xrange(60):
            for j in xrange(60):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

        # second time with E equal to the Y coordinates
        mmats = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            mat = DMatrix.IsotropicMatMatrix({'E':gp_coord[1],'nu':0.3})
            mmats.append(mat)
        k = elem.DeformationStiffMtx(coords,mmats)
        smat = GetStoredMatSym("brick_var_stiff")
        for i in xrange(60):
            for j in xrange(60):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

    def testBrick2IInitialStressStiff(self):
        """Brick Initial Stress Stiffness Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()

        # stresses based on the coordinate locations
        stress = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            stress.append(ColTensor.ColTensor(gp_coord[0],
                gp_coord[1],gp_coord[2],(gp_coord[0]+gp_coord[1])/2,
               (gp_coord[1]+gp_coord[2])/2,(gp_coord[2]+gp_coord[0])/2))
        k = elem.InitialStressStiffMtx(coords,stress)
        smat = GetStoredMatSym("brick_init_str")
        for i in xrange(30):
            for j in xrange(30):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

    def testBrick2IFindNaturalCoords(self):
        """Brick Find Natural Coordinates"""
        tol = 0.000001
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()
        nat = elem.FindNaturalCoords(Vec3D.Vec3D(3.0,1.0,2.0),coords)
        self.assertEqualFloat(nat[0],0.5,tol)
        self.assertEqualFloat(nat[1],-.5,tol)
        self.assertEqualFloat(nat[2],0.0,tol)

    def testBrick2INormalSurfaceScalarIntegration(self):
        """Brick Normal Surface Scalar Integration"""
        # test the integration of a surface scalar where the nodal
        # value is the distance of the node from the origin
        tol = 0.000001
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()
        smat = GetStoredMat('brick_surf_normal')

        for  i in xrange(6):
            tract = []
            for coord in coords:
                tract.append(coord.Magnitude())
            f = elem.IntegrateNormalSurfaceScalar(i,coords,tract)
            for j in xrange(20):
                for k in xrange(3):
                    self.assertEqualFloat(smat[(i,j,k)],f[j][k],tol)

    def testBrick2ISurfaceScalarIntegration(self):
        """Brick Surface Scalar Integration"""
        # test the integration of a surface scalar where the nodal
        # value is the distance of the node from the origin
        tol = 0.000001
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()
        smat = GetStoredMat('brick_surf_scalar')

        for  i in xrange(6):
            tract = []
            for coord in coords:
                tract.append(coord.Magnitude())
            f = elem.IntegrateSurfaceScalar(i,coords,tract)
            for j in xrange(20):
                value = smat[(i,j)]
                self.assertEqualFloat(value,f[j],tol)

    def testBrick2ISurfaceVectorIntegration(self):
        """Brick Surface Vector Integration"""
        # test the integration of a surface scalar where the
        # nodal value is the location vector of the node
        tol = 0.000001
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()
        smat = GetStoredMat('brick_surf_vector')

        for  i in xrange(6):
            tract = []
            for coord in coords:
                tract.append(coord)
            f = elem.IntegrateSurfaceVector(i,coords,tract)
            for j in xrange(20):
                for k in xrange(3):
                    self.assertEqualFloat(smat[(i,j,k)],f[j][k],tol)

    def testBrick2IIntegrateVolume(self):
        """Brick Integrate Volume"""
        tol = 0.000001
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()

        # do a volume integration assuming a scalar value of 2.0
        f = elem.IntegrateVolume(coords,2.0)
        smat = GetStoredMat('brick_int_scalar')
        for i in xrange(len(f)):
            self.assertEqualFloat(smat[(0,i)],f[i],tol)

        # do a volume integration where the scalar value varies
        # over the element and is equal to the y coordinate
        values = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            values.append(gp_coord[1])
        f = elem.IntegrateVolume(coords,values)
        for i in xrange(len(f)):
            self.assertEqualFloat(smat[(1,i)],f[i],tol)

        # do a volume integration with a stress value of (10,20,30,4,5,6)
        value = ColTensor.ColTensor(10.0,20.0,30.0,4.0,5.0,6.0)
        f = elem.IntegrateVolume(coords,value)
        smat = GetStoredMat('brick_int_vector')
        for i in xrange(len(f)):
            for j in xrange(3):
                self.assertEqualFloat(smat[(0,i,j)],f[i][j],tol)

        # do a volume integration with a stress value being
        # (x_coord,y_coord,z_coord,x/2,y/2,z/2)
        values = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            values.append(ColTensor.ColTensor(gp_coord[0],
                gp_coord[1],gp_coord[2],gp_coord[0]/2,
                gp_coord[1]/2,gp_coord[2]/2))
        f = elem.IntegrateVolume(coords,values)
        for i in xrange(len(f)):
            for j in xrange(3):
                self.assertEqualFloat(smat[(1,i,j)],f[i][j],tol)

    def testBrick2IInterpolateSurfaceScalar(self):
        """Brick Interpolate Surface Scalar"""
        tol = 0.000001
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()
        node_vals = []
        for coord in coords: node_vals.append(coord[2])
        natural = Vec3D.Vec3D(0.5,-.5,0.0)
        value = elem.InterpolateSurfaceScalar(0,natural,node_vals)
        self.assertEqualFloat(value,3.0,tol)
        value = elem.InterpolateSurfaceScalar(1,natural,node_vals)
        self.assertEqualFloat(value,1.0,tol)
        value = elem.InterpolateSurfaceScalar(2,natural,node_vals)
        self.assertEqualFloat(value,1.0,tol)
        value = elem.InterpolateSurfaceScalar(3,natural,node_vals)
        self.assertEqualFloat(value,3.0,tol)
        value = elem.InterpolateSurfaceScalar(4,natural,node_vals)
        self.assertEqualFloat(value,0.0,tol)
        value = elem.InterpolateSurfaceScalar(5,natural,node_vals)
        self.assertEqualFloat(value,4.0,tol)

    def testBrick2IInterpolateVolume(self):
        """Brick Interpolate Volume"""
        tol = 0.000001
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()
        node_vals = []
        for coord in coords: node_vals.append(coord[2])
        natural = Vec3D.Vec3D(0.5,-.5,0.0)
        value = elem.InterpolateVolume(natural,node_vals)
        self.assertEqualFloat(value,2.0,tol)
        value = elem.InterpolateVolume(natural,coords)
        self.assertEqualFloat(value[0],3.0,tol)
        self.assertEqualFloat(value[1],1.0,tol)
        self.assertEqualFloat(value[2],2.0,tol)

    def testBrick2IInternalPointForces(self):
        """Brick Internal Point Forces"""
        tol = 0.0001
        elem = FemLib.Element("Brick2IShapeFunc")
        # place the forces at the gauss points
        natural = []
        natural.append(Vec3D.Vec3D(-.25,-.25,-.25))
        natural.append(Vec3D.Vec3D(0.25,-.25,-.25))
        natural.append(Vec3D.Vec3D(-.25,0.25,-.25))
        natural.append(Vec3D.Vec3D(0.25,0.25,-.25))
        natural.append(Vec3D.Vec3D(-.25,-.25,0.25))
        natural.append(Vec3D.Vec3D(0.25,-.25,0.25))
        natural.append(Vec3D.Vec3D(-.25,0.25,0.25))
        natural.append(Vec3D.Vec3D(0.25,0.25,0.25))
        forces = []
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        f = elem.InternalPointForces(natural,forces)
        smat = GetStoredMat('brick_intrnl_force')
        for i in xrange(len(f)):
           for j in xrange(3):
               self.assertEqualFloat(smat[(i,j)],f[i][j],tol)

    def testBrick2IConductivity(self):
        """Brick Conductivity Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()

        # first time with constant props
        cond = DMatrix.ScalarIsotropicMatMatrix({'k':1.0})
        k = elem.ScalarFieldDerivMtx(coords,cond)
        smat = GetStoredMatSym('brick_conductivity')
        for i in xrange(20):
            for j in xrange(20):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

        # second time with conductivities equal to the coordinates
        cmats = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            cond = DMatrix.ScalarOrthotropicMatMatrix(
                     {'k1':gp_coord[0],'k2':gp_coord[1],'k3':gp_coord[2]})
            cmats.append(cond)
        k = elem.ScalarFieldDerivMtx(coords,cmats)
        smat = GetStoredMatSym("brick_var_cond")
        for i in xrange(20):
            for j in xrange(20):
                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

    def testBrick2IMassStorage(self):
        """Brick Mass/Storage Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Brick2IShapeFunc")
        spec_heat = 0.5
        density = 1.0
        coords = self.GetCoords()

        # fist time with a constant specific heat
        fact = spec_heat * density
        c = elem.ScalarFieldShapeMtx(coords,fact)
        smat = GetStoredMatSym('brick_mass')
        for i in xrange(20):
            for j in xrange(20):
                self.assertEqualFloat(smat[(i,j)],c[i,j],tol)
   
        # now with a specific heat equal to the y coordinate
        facts = []
        gpts = elem.GetVolumeIntegrationPts(elem.GetShapeDefaultPts())
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            facts.append(density*gp_coord[1])
        c = elem.ScalarFieldShapeMtx(coords,facts)
        smat = GetStoredMatSym("brick_var_mass")
        for i in xrange(20):
            for j in xrange(20):
                self.assertEqualFloat(smat[(i,j)],c[i,j],tol)

    def testBrick2IStrains(self):
        """Brick Gauss Point Strains"""
        tol = 0.00001
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()

        disp = []
        for nd in coords:
            disp.append(Vec3D.Vec3D(nd[0]**2,nd[1]**2,nd[2]**2))

        gp_strain = elem.Strains(coords,disp)
        smat = GetStoredMat("brick_strain")
        for i in xrange(len(gp_strain)):
            for j in xrange(6):
                self.assertEqualFloat(smat[(i,j)],gp_strain[i][j],tol)


    def testBrick2IVolume(self):
        """Brick Volume"""
        tol = 0.00001
        value = 64.0
        elem = FemLib.Element("Brick2IShapeFunc")
        coords = self.GetCoords()
        self.assertEqualFloat(value,elem.Volume(coords,1),tol)
        self.assertEqualFloat(value,elem.Volume(coords,8),tol)
        self.assertEqualFloat(value,elem.Volume(coords,27),tol)
        self.assertEqualFloat(value,elem.Volume(coords,64),tol)

#--------------------------------------------------------------------

class Pyramid2ITests(unittest.TestCase):
    def assertEqualFloat(self,x,y,tol):
        self.failUnless(abs(x-y) < tol)

    def GetCoords(self):
        coords = []
        coords.append(Vec3D.Vec3D(0.0,0.0,0.0))
        coords.append(Vec3D.Vec3D(4.0,0.0,0.0))
        coords.append(Vec3D.Vec3D(4.0,0.0,4.0))
        coords.append(Vec3D.Vec3D(0.0,0.0,4.0))
        coords.append(Vec3D.Vec3D(2.0,4.0,2.0))
        coords.append(Vec3D.Vec3D(2.0,0.0,0.0))
        coords.append(Vec3D.Vec3D(4.0,0.0,2.0))
        coords.append(Vec3D.Vec3D(2.0,0.0,4.0))
        coords.append(Vec3D.Vec3D(0.0,0.0,2.0))
        coords.append(Vec3D.Vec3D(1.0,2.0,1.0))
        coords.append(Vec3D.Vec3D(3.0,2.0,1.0))
        coords.append(Vec3D.Vec3D(3.0,2.0,3.0))
        coords.append(Vec3D.Vec3D(1.0,2.0,3.0))
        return coords

    def testPyramid2IArea(self):
        """Pyramid Face Area"""
        tol = 0.00001
        elem = FemLib.Element("Pyramid2IShapeFunc")
        coords = self.GetCoords()
        self.assertEqualFloat(16.0,elem.Area(0,coords),tol)
        self.assertEqualFloat(2*math.sqrt(20),elem.Area(1,coords),tol)
        self.assertEqualFloat(2*math.sqrt(20),elem.Area(2,coords),tol)
        self.assertEqualFloat(2*math.sqrt(20),elem.Area(3,coords),tol)
        self.assertEqualFloat(2*math.sqrt(20),elem.Area(4,coords),tol)

    def testPyramid2IStiffness(self):
        """Pyramid Stiffness Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Pyramid2IShapeFunc")
        coords = self.GetCoords()

        # first time through with a uniform E
        mat = DMatrix.IsotropicMatMatrix({'E':1000.0,'nu':0.3})
        k = elem.DeformationStiffMtx(coords,mat)
#        smat = GetStoredMatSym('pyramid_stiffness')
#        for i in xrange(60):
#            for j in xrange(60):
#               self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

        # second time with E equal to the Y coordinates
        mmats = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            mat = DMatrix.IsotropicMatMatrix({'E':gp_coord[1],'nu':0.3})
            mmats.append(mat)
        k = elem.DeformationStiffMtx(coords,mmats)
#        smat = GetStoredMatSym("pyramid_var_stiff")
#        for i in xrange(60):
#            for j in xrange(60):
#                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

    def testPyramid2IInitialStressStiff(self):
        """Pyramid Initial Stress Stiffness Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Pyramid2IShapeFunc")
        coords = self.GetCoords()

        # stresses based on the coordinate locations
        stress = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            stress.append(ColTensor.ColTensor(gp_coord[0],
                gp_coord[1],gp_coord[2],(gp_coord[0]+gp_coord[1])/2,
               (gp_coord[1]+gp_coord[2])/2,(gp_coord[2]+gp_coord[0])/2))
        k = elem.InitialStressStiffMtx(coords,stress)
#        smat = GetStoredMatSym("pyramid_init_str")
#        for i in xrange(30):
#            for j in xrange(30):
#                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

    def testPyramid2IFindNaturalCoords(self):
        """Pyramid Find Natural Coordinates"""
        tol = 0.000001
        elem = FemLib.Element("Pyramid2IShapeFunc")
        coords = self.GetCoords()
        nat = elem.FindNaturalCoords(Vec3D.Vec3D(3.0,1.0,2.0),coords)
#        self.assertEqualFloat(nat[0],0.5,tol)
#        self.assertEqualFloat(nat[1],-.5,tol)
#        self.assertEqualFloat(nat[2],0.0,tol)

    def testPyramid2ISurfaceScalarIntegration(self):
        """Pyramid Surface Scalar Integration"""
        # test the integration of a surface scalar where the nodal
        # value is the distance of the node from the origin
        tol = 0.000001
        elem = FemLib.Element("Pyramid2IShapeFunc")
        coords = self.GetCoords()
#        smat = GetStoredMat('pyramid_surf_scalar')

        for  i in xrange(5):
            tract = []
            for coord in coords:
                tract.append(coord.Magnitude())
            f = elem.IntegrateSurfaceScalar(i,coords,tract)
#            for j in xrange(20):
#                value = smat[(i,j)]
#                self.assertEqualFloat(value,f[j],tol)

    def testPyramid2ISurfaceVectorIntegration(self):
        """Pyramid Surface Vector Integration"""
        # test the integration of a surface scalar where the
        # nodal value is the location vector of the node
        tol = 0.000001
        elem = FemLib.Element("Pyramid2IShapeFunc")
        coords = self.GetCoords()
#        smat = GetStoredMat('pyramid_surf_vector')

        for  i in xrange(5):
            tract = []
            for coord in coords:
                tract.append(coord)
            f = elem.IntegrateSurfaceVector(i,coords,tract)
#            for j in xrange(20):
#                for k in xrange(3):
#                    self.assertEqualFloat(smat[(i,j,k)],f[j][k],tol)

    def testPyramid2IIntegrateVolume(self):
        """Pyramid Integrate Volume"""
        tol = 0.000001
        elem = FemLib.Element("Pyramid2IShapeFunc")
        coords = self.GetCoords()

        # do a volume integration assuming a scalar value of 2.0
        f = elem.IntegrateVolume(coords,2.0)
#        smat = GetStoredMat('pyramid_int_scalar')
#        for i in xrange(len(f)):
#            self.assertEqualFloat(smat[(0,i)],f[i],tol)

        # do a volume integration where the scalar value varies
        # over the element and is equal to the y coordinate
        values = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            values.append(gp_coord[1])
        f = elem.IntegrateVolume(coords,values)
#        for i in xrange(len(f)):
#            self.assertEqualFloat(smat[(1,i)],f[i],tol)

        # do a volume integration with a stress value of (10,20,30,4,5,6)
        value = ColTensor.ColTensor(10.0,20.0,30.0,4.0,5.0,6.0)
        f = elem.IntegrateVolume(coords,value)
#        smat = GetStoredMat('pyramid_int_vector')
#        for i in xrange(len(f)):
#            for j in xrange(3):
#                self.assertEqualFloat(smat[(0,i,j)],f[i][j],tol)

        # do a volume integration with a stress value being
        # (x_coord,y_coord,z_coord,x/2,y/2,z/2)
        values = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            values.append(ColTensor.ColTensor(gp_coord[0],
                gp_coord[1],gp_coord[2],gp_coord[0]/2,
                gp_coord[1]/2,gp_coord[2]/2))
        f = elem.IntegrateVolume(coords,values)
#        for i in xrange(len(f)):
#            for j in xrange(3):
#                self.assertEqualFloat(smat[(1,i,j)],f[i][j],tol)

    def testPyramid2IInterpolateSurfaceScalar(self):
        """Pyramid Interpolate Surface Scalar"""
        tol = 0.000001
        elem = FemLib.Element("Pyramid2IShapeFunc")
        coords = self.GetCoords()
        node_vals = []
        for coord in coords: node_vals.append(coord[2])
        natural = Vec3D.Vec3D(0.0,0.0,0.0)
#        print node_vals
        value = elem.InterpolateSurfaceScalar(0,natural,node_vals)
#        self.assertEqualFloat(value,3.0,tol)
        value = elem.InterpolateSurfaceScalar(1,natural,node_vals)
#        self.assertEqualFloat(value,1.0,tol)
        value = elem.InterpolateSurfaceScalar(2,natural,node_vals)
#        self.assertEqualFloat(value,1.0,tol)
        value = elem.InterpolateSurfaceScalar(3,natural,node_vals)
#        self.assertEqualFloat(value,3.0,tol)
        value = elem.InterpolateSurfaceScalar(4,natural,node_vals)
#        self.assertEqualFloat(value,0.0,tol)

    def testPyramid2IInterpolateVolume(self):
        """Pyramid Interpolate Volume"""
        tol = 0.000001
        elem = FemLib.Element("Pyramid2IShapeFunc")
        coords = self.GetCoords()
        node_vals = []
        for coord in coords: node_vals.append(coord[2])
        natural = Vec3D.Vec3D(0.5,-.5,0.0)
        value = elem.InterpolateVolume(natural,node_vals)
#        self.assertEqualFloat(value,2.0,tol)
        value = elem.InterpolateVolume(natural,coords)
#        self.assertEqualFloat(value[0],3.0,tol)
#        self.assertEqualFloat(value[1],1.0,tol)
#        self.assertEqualFloat(value[2],2.0,tol)

    def testPyramid2IInternalPointForces(self):
        """Pyramid Internal Point Forces"""
        tol = 0.0001
        elem = FemLib.Element("Pyramid2IShapeFunc")
        # place the forces at the gauss points
        natural = []
        natural.append(Vec3D.Vec3D(-.25,-.25,-.25))
        natural.append(Vec3D.Vec3D(0.25,-.25,-.25))
        natural.append(Vec3D.Vec3D(-.25,0.25,-.25))
        natural.append(Vec3D.Vec3D(0.25,0.25,-.25))
        natural.append(Vec3D.Vec3D(-.25,-.25,0.25))
        natural.append(Vec3D.Vec3D(0.25,-.25,0.25))
        natural.append(Vec3D.Vec3D(-.25,0.25,0.25))
        natural.append(Vec3D.Vec3D(0.25,0.25,0.25))
        forces = []
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        forces.append(Vec3D.Vec3D(1.0,2.0,3.0))
        f = elem.InternalPointForces(natural,forces)
#        smat = GetStoredMat('pyramid_intrnl_force')
#        for i in xrange(len(f)):
#           for j in xrange(3):
#               self.assertEqualFloat(smat[(i,j)],f[i][j],tol)

    def testPyramid2IConductivity(self):
        """Pyramid Conductivity Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Pyramid2IShapeFunc")
        coords = self.GetCoords()

        # first time with constant props
        cond = DMatrix.ScalarIsotropicMatMatrix({'k':1.0})
        k = elem.ScalarFieldDerivMtx(coords,cond)
#        smat = GetStoredMatSym('pyramid_conductivity')
#        for i in xrange(20):
#            for j in xrange(20):
#                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

        # second time with conductivities equal to the coordinates
        cmats = []
        gpts = elem.GetVolumeIntegrationPts()
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            cond = DMatrix.ScalarOrthotropicMatMatrix(
                     {'k1':gp_coord[0],'k2':gp_coord[1],'k3':gp_coord[2]})
            cmats.append(cond)
        k = elem.ScalarFieldDerivMtx(coords,cmats)
#        smat = GetStoredMatSym("pyramid_var_cond")
#        for i in xrange(20):
#            for j in xrange(20):
#                self.assertEqualFloat(smat[(i,j)],k[i,j],tol)

    def testPyramid2IMassStorage(self):
        """Pyramid Mass/Storage Matrix"""
        tol = 0.00001
        elem = FemLib.Element("Pyramid2IShapeFunc")
        spec_heat = 0.5
        density = 1.0
        coords = self.GetCoords()

        # fist time with a constant specific heat
        fact = spec_heat * density
        c = elem.ScalarFieldShapeMtx(coords,fact)
#        print c
#        smat = GetStoredMatSym('pyramid_mass')
#        for i in xrange(20):
#            for j in xrange(20):
#                self.assertEqualFloat(smat[(i,j)],c[i,j],tol)
   
        # now with a specific heat equal to the y coordinate
        facts = []
        gpts = elem.GetVolumeIntegrationPts(elem.GetShapeDefaultPts())
        for gp in gpts:
            gp_coord = elem.InterpolateVolume(gp,coords)
            facts.append(density*gp_coord[1])
        c = elem.ScalarFieldShapeMtx(coords,facts)
 #       smat = GetStoredMatSym("pyramid_var_mass")
 #       for i in xrange(20):
 #           for j in xrange(20):
 #               self.assertEqualFloat(smat[(i,j)],c[i,j],tol)


    def testPyramid2IVolume(self):
        """Pyramid Volume"""
        tol = 0.00001
        value = 64.0/3.0
        elem = FemLib.Element("Pyramid2IShapeFunc")
        coords = self.GetCoords()
        self.assertEqualFloat(value,elem.Volume(coords,8),tol)
        self.assertEqualFloat(value,elem.Volume(coords,27),tol)
        self.assertEqualFloat(value,elem.Volume(coords,64),tol)

# -------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()
